import { DirectionalLight } from 'three/src/lights/DirectionalLight';
import { AmbientLight } from 'three/src/lights/AmbientLight';
import { WebGLRenderer } from 'three/src/renderers/WebGLRenderer';
import { PerspectiveCamera } from 'three/src/cameras/PerspectiveCamera';
import { Scene } from 'three/src/scenes/Scene';
import { Vector3 } from 'three/src/math/Vector3';
import { GridLayer } from './layer/GridLayer';
// import { BoxLayer } from './layer/BoxLayer';
import { DemLayer } from './layer/DemLayer';
import { Map } from './core/Map';
import * as domEvent from './core/domEvent';
import { Coordinates } from './controls/Coordinates';
import { NorthArrow } from './controls/NorthArrow';
import { LayerControl } from './controls/LayerControl';
import { BasemapControl } from './controls/BasemapControl';
import { SliderControl } from './controls/SliderControl';
// import { SlicerControl } from './controls/SlicerControl';
import { Mesh } from 'three/src/objects/Mesh';
import { SphereGeometry } from 'three/src/geometries/SphereGeometry';
import { MeshLambertMaterial } from 'three/src/materials/MeshLambertMaterial';
import * as util from './core/utilities';
import * as browser from './core/browser';
import * as domUtil from './core/domUtil';
import { MobileDialog } from "./controls/MobileDialog";
import { Picking } from './clip/Picking';

import { Selection } from './clip/Selection';

import '../css/page.scss'; /* style loader will import it */

class Application {

    constructor(container) {
        this.container = container;
        this.running = false;
        this.wireframeMode = false;
        this.canvas;
        this._canvasImageUrl;
        this.downloadButton;
        this.showCaps = true;

        this.objects = [];

        if (container.clientWidth && container.clientHeight) {
            this.width = container.clientWidth;
            this.height = container.clientHeight;
            this._fullWindow = false;
        } else {
            this.width = window.innerWidth;
            this.height = window.innerHeight;
            this._fullWindow = true;
        }

        // this.canvas = document.querySelector('#imgCanvas');
        // this.$topTextInput = document.querySelector('#topText');
        // this.$bottomTextInput = document.querySelector('#bottomText');
        // this.$imageInput = document.querySelector('#image');
        this.downloadButton = document.querySelector('#btnDownloadCanvasImage');
        this.menuIcon = document.querySelector('#menu-icon');
        this.navigation = document.getElementsByClassName('navigation')[0];
        // this.addEventListeners();

        this.dialog = new MobileDialog("Help", container, { klass: "fm_about" });
        this.aboutIcon = document.querySelector('#menu-about-icon');


        // this.createScene();
        // this.addEventListeners();
    }
    async build() {
        await this.createScene();
        this.addEventListeners();
    }

    async createScene() {

        var dirNode = document.getElementsByTagName("body")[0];
        if (browser.touch == true && browser.mobile == true) {
            //dirNode.setAttribute("dir", "ltr");
            domUtil.addClass(dirNode, "touch");
        } else {
            domUtil.addClass(dirNode, "notouch");
        }

        // let opt = { r: 200, c: 0x38eeff, o: 0.8 };
        var opt = { r: 5, c: 0xffff00, o: 1 };
        this.queryMarker = new Mesh(new SphereGeometry(opt.r),
            new MeshLambertMaterial({ color: opt.c, opacity: opt.o, transparent: (opt.o < 1) }));

        this.queryMarker.visible = true;
        // this.queryMarker.position.set(4282010, 2302070, -13616.3);
        /* Renderer */
        // let bgcolor = 0xfdfdfd;
        this.renderer = new WebGLRenderer({ alpha: true, antialias: true, preserveDrawingBuffer: true });
        this.renderer.setPixelRatio(window.devicePixelRatio);
        this.renderer.setSize(this.width, this.height);
        this.renderer.autoClear = false;
        this.renderer.setClearColor(0x000000, 0.0); // second param is opacity, 0 => transparent 

        // enable clipping
        // let Empty = Object.freeze([]);
        // this.renderer.clippingPlanes = Empty; // GUI sets it to globalPlanes
        this.renderer.localClippingEnabled = true;
        this.container.appendChild(this.renderer.domElement);

        /* Scene: that will hold all our elements such as objects, cameras and lights. */
        this.scene = new Scene();
        this.capsScene = new Scene();
        // this.backStencil = new Scene();
        // this.frontStencil = new Scene();
        this._buildDefaultLights(this.scene);
        //app.scene.autoUpdate = false;
        //// show axes in the screen
        //app.scene.add(new THREE.AxisHelper(100));
        this.scene.add(this.queryMarker);

        /* Camera */
        var angle = 45;
        var aspect = this.width / this.height;
        var near = 0.1; //This is the distance at which the camera will start rendering scene objects
        var far = 2000; //Anything beyond this distance will not be rendered                
        // this.camera = new PerspectiveCamera(angle, aspect, near, far);
        // this.camera.position.set(0, -0.1, 150);
        // this.camera.lookAt(new Vector3(0, 0, 0));

        this.camera = new PerspectiveCamera(30, this.width / this.height, 100, 100000);
        // const dirLight = new DirectionalLight(0xffffff, 1);
        // dirLight.position.set(585000 + 10000, 6135000 + 10000, -500 + 5000);
        // this.camera.add(dirLight);

        // let x = { min: 3955850, max: 4527300, avg: 4282010 };  
        // let y = { min: 2182271.6538, max: 2504028.3462 , avg: 2302070 };
        // let z = { min: -60066, max: 3574.94, avg: -13616.3 };

        let x = { min: 4415940, max: 4508490, avg: 4463200 };
        // let y = { min: 2351480, max: 2475820 , avg: 2412610 };
        let y = { min: 2351229.9140, max: 2476070.0860, avg: 2412610 };
        let z = { min: -8798.15, max: 1401.92, avg: -155.17 };

        const center = new Vector3((x.min + x.max) / 2, (y.min + y.max) / 2, 0);
        // const center = new Vector3(x.avg, y.avg, z.avg);
        const size = Math.max(x.max - x.min, y.max - y.min, z.max - z.min);
        let baseExtent = {
            x: x,
            y: y
        };

        const camDirection = new Vector3(-0.5, -Math.SQRT1_2, 0.5);
        // const camDirection = new Vector3(0, 0, 1);
        const camOffset = camDirection.multiplyScalar(size * 2);
        this.camera.position.copy(center);
        this.camera.position.add(camOffset);
        this.camera.near = size * 0.1;
        this.camera.far = size * 25;
        this.camera.updateProjectionMatrix();


        /* Camera */
        // // const center = new Vector3();
        // var angle = 45;
        // var aspect = this.width / this.height;
        // var near = 0.1; //This is the distance at which the camera will start rendering scene objects
        // var far = 2000; //Anything beyond this distance will not be rendered                
        // this.camera = new PerspectiveCamera(angle, aspect, near, far);        
        // this.camera.position.set(0, -0.1, 150);
        // this.camera.lookAt(new Vector3(0, 0, 0));

        // create map 
        // let map = this.map = new Map(x, y, z, size, center, this.camera, this.scene, this.container, 'https://geusegdi01.geus.dk/meta3d/rpc/model_meta?modelid=20');

        // this.selection = new Selection(
        //     // new Vector3(-7, -14, -14),
        //     // new Vector3(14, 9, 3)
        //     new Vector3(x.min, y.min, z.min),
        //     new Vector3(x.max, y.max, z.max)
        // );
        // new Picking(size, center, this);

        let map = this.map = await Map.build(x, y, z, center, this.camera, this.scene, this.container, 'https://geusegdi01.geus.dk/meta3d/rpc/model_meta_all?modelid=20');
        this.mapTitle = document.querySelector('#map-title');
        this.mapTitle.innerHTML += map.title;
        map.on('ready', () => {
          this.selection.setUniforms();
        }, this);

        this.selection = new Selection(
            // new Vector3(-7, -14, -14),
            // new Vector3(14, 9, 3)
            new Vector3(x.min, y.min, z.min),
            new Vector3(x.max, y.max, z.max),
            map
        );

        new Picking(size, center, this);
       

        // let boxLayer = new BoxLayer({ 
        //     width: 10000, height: 10000, depth: 10000, name: 'center-box', color: 800080 , center: center
        // });
        // this.map.addLayer(boxLayer);

        //add map controls:
        if (util.hasTouch() == false) {
            let coordinates = new Coordinates({ camera: this.camera, crs: "EPSG:3034" }).addTo(this.map);
            // coordinates.addListener('onPoint', (args) => {
            //     let vector = args[0];
            //     this.queryMarker.position.set(vector.x, vector.y, vector.z);
            //     // this.queryMarker.updateMatrixWorld();
            //     this.animate();
            // }, this);
        }
        this.northArrow = new NorthArrow({ headLength: 1, headWidth: 1 }).addTo(this.map);

        let demLayer = new DemLayer({
            q: 0, shading: true, type: 'dem', name: 'DEM Layer', color: 16382457, "baseExtent": baseExtent,
            "materialParameter": [{
                "i": 0,
                "materialtypee": 0,
                "ds": 1,
                "bottomZ": 3000,
            }]
        });
        demLayer.addBlock({
            "width": 173, //267,
            "plane": {
                "width": x.max - x.min, //100.0,
                // "offsetX": 0,
                // "offsetY": 0,
                // "offsetX" : x.avg,
                // "offsetY" : y.avg,
                "offsetX": center.x,
                "offsetY": center.y,
                "height": y.max - y.min, //78.59618717504333
            },
            "dem_values": [1326.9, 1303.69, 1272.49, 1241.16, 1209.7, 1178.11, 1146.39, 1114.54, 1082.56, 1050.44, 1018.2, 985.82, 959.53, 950.84, 942.18, 933.56, 924.96, 916.39, 907.85, 899.35, 890.87, 882.43, 874.01, 864.66, 852.23, 839.86, 827.55, 815.29, 803.09, 790.94, 778.85, 766.82, 754.84, 742.92, 733.63, 734.0, 734.34, 734.65, 734.92, 735.16, 735.37, 735.54, 735.69, 735.8, 735.88, 735.35, 732.21, 729.08, 725.96, 722.84, 719.73, 716.63, 713.53, 710.43, 707.35, 704.27, 701.02, 696.82, 692.63, 688.47, 684.31, 680.18, 676.06, 671.95, 667.87, 663.79, 659.74, 656.93, 663.26, 669.66, 676.14, 682.7, 689.33, 696.03, 702.81, 709.67, 716.6, 723.61, 730.93, 740.78, 750.65, 760.52, 770.4, 780.3, 790.2, 800.11, 810.03, 819.97, 829.91, 839.17, 834.77, 830.44, 826.19, 822.01, 817.89, 813.67, 809.33, 804.88, 800.31, 795.63, 790.77, 780.11, 769.48, 758.89, 748.33, 737.8, 727.31, 716.86, 706.44, 696.05, 685.7, 675.38, 669.94, 664.65, 659.36, 654.06, 648.75, 643.43, 638.11, 632.78, 627.45, 622.11, 616.76, 616.03, 615.58, 615.03, 614.38, 613.64, 612.8, 611.85, 610.81, 609.67, 608.44, 607.1, 607.0, 607.11, 607.24, 607.4, 607.58, 607.79, 608.02, 608.27, 608.55, 608.86, 609.19, 613.97, 619.74, 625.6, 631.55, 637.58, 643.69, 649.89, 656.18, 662.56, 669.02, 675.57, 686.13, 697.88, 709.68, 721.56, 733.5, 745.5, 757.57, 769.71, 781.91, 794.18, 806.51, 816.3, 824.79, 832.91, 840.65, 848.02, 855.01, 1328.88, 1306.0, 1274.0, 1241.86, 1209.59, 1177.19, 1144.66, 1112.0, 1079.21, 1046.28, 1013.23, 980.04, 951.19, 942.7, 934.23, 925.79, 917.38, 909.0, 900.66, 892.34, 884.05, 875.8, 867.57, 858.77, 846.68, 834.65, 822.68, 810.76, 798.9, 787.1, 775.35, 763.66, 752.02, 740.44, 730.41, 730.59, 730.74, 730.86, 730.94, 730.99, 731.01, 731.0, 730.96, 730.88, 730.77, 730.34, 727.24, 724.15, 721.07, 717.99, 714.91, 711.85, 708.79, 705.73, 702.68, 699.64, 696.53, 692.42, 688.33, 684.26, 680.2, 676.16, 672.14, 668.13, 664.13, 660.16, 656.19, 652.6, 659.36, 666.2, 673.11, 680.1, 687.16, 694.3, 701.51, 708.8, 716.17, 723.61, 731.13, 741.03, 750.95, 760.88, 771.02, 781.85, 792.54, 803.11, 813.54, 823.83, 834.0, 844.02, 840.31, 835.91, 831.39, 826.75, 822.01, 817.14, 812.17, 807.07, 801.87, 796.54, 791.1, 781.03, 770.59, 760.19, 749.82, 739.48, 729.18, 718.92, 708.68, 698.49, 688.32, 678.2, 672.29, 666.96, 661.63, 656.29, 650.95, 645.6, 640.24, 634.87, 629.5, 624.12, 618.74, 617.1, 616.12, 615.05, 613.87, 612.6, 611.23, 609.77, 608.2, 606.53, 604.77, 602.91, 602.66, 602.9, 603.17, 603.46, 603.77, 604.11, 604.47, 604.85, 605.27, 605.7, 606.16, 610.81, 617.03, 623.34, 629.74, 636.22, 642.78, 649.43, 656.17, 663.0, 669.91, 676.91, 687.36, 699.44, 711.59, 723.8, 736.07, 748.41, 760.82, 773.29, 785.83, 798.44, 811.11, 820.15, 826.73, 832.93, 838.76, 844.21, 849.29, 1330.93, 1308.45, 1275.63, 1242.68, 1209.6, 1176.4, 1143.06, 1109.59, 1075.98, 1042.25, 1008.39, 974.39, 942.83, 934.52, 926.24, 917.99, 909.77, 901.59, 893.43, 885.3, 877.2, 869.14, 861.1, 852.82, 841.08, 829.39, 817.76, 806.18, 794.66, 783.2, 771.79, 760.44, 749.14, 737.91, 727.22, 727.21, 727.17, 727.1, 726.99, 726.86, 726.69, 726.49, 726.26, 725.99, 725.69, 725.32, 722.26, 719.21, 716.16, 713.12, 710.09, 707.06, 704.04, 701.02, 698.01, 695.0, 692.0, 688.01, 684.02, 680.04, 676.08, 672.13, 668.2, 664.28, 660.38, 656.5, 652.63, 648.78, 655.39, 662.66, 670.0, 677.96, 686.38, 694.99, 703.8, 712.79, 721.98, 731.37, 740.95, 751.46, 761.93, 772.27, 782.48, 792.56, 802.5, 812.31, 821.98, 831.52, 840.93, 850.21, 847.1, 842.05, 836.9, 831.62, 826.24, 820.73, 815.12, 809.39, 803.54, 797.58, 791.5, 781.92, 771.67, 761.46, 751.28, 741.13, 731.02, 720.94, 710.9, 700.89, 690.92, 680.98, 674.64, 669.28, 663.91, 658.54, 653.16, 647.77, 642.37, 636.97, 631.56, 626.15, 620.72, 618.26, 616.76, 615.16, 613.46, 611.67, 609.77, 607.78, 605.68, 603.49, 601.2, 598.81, 598.3, 598.67, 599.07, 599.49, 599.93, 600.4, 600.89, 601.41, 601.95, 602.52, 603.11, 607.56, 614.24, 620.99, 627.84, 634.77, 641.78, 648.89, 656.07, 663.35, 670.71, 678.16, 688.53, 700.94, 713.42, 725.97, 738.58, 751.26, 764.0, 776.81, 789.69, 802.63, 815.64, 824.37, 829.03, 833.32, 837.24, 840.78, 843.94, 1333.06, 1311.02, 1277.39, 1243.64, 1209.75, 1175.73, 1141.58, 1107.3, 1072.89, 1038.35, 1003.68, 968.87, 934.43, 926.31, 918.22, 910.16, 902.14, 894.14, 886.17, 878.23, 870.32, 862.45, 854.6, 846.78, 835.42, 824.07, 812.78, 801.54, 790.36, 779.24, 768.18, 757.16, 746.21, 735.31, 724.47, 723.86, 723.64, 723.37, 723.08, 722.76, 722.4, 722.01, 721.59, 721.13, 720.65, 720.13, 717.28, 714.27, 711.26, 708.25, 705.26, 702.27, 699.28, 696.3, 693.33, 690.36, 687.4, 683.58, 679.69, 675.9, 672.45, 669.08, 665.76, 662.51, 659.33, 656.22, 653.17, 650.18, 657.3, 666.24, 675.37, 684.69, 694.2, 703.9, 713.8, 723.89, 734.17, 744.65, 755.32, 765.32, 775.04, 784.62, 794.08, 803.4, 812.59, 821.64, 830.56, 839.35, 848.0, 856.52, 854.0, 848.32, 842.52, 836.61, 830.58, 824.44, 818.18, 811.81, 805.33, 798.72, 792.01, 782.78, 772.72, 762.69, 752.7, 742.74, 732.82, 722.93, 713.07, 703.26, 693.47, 683.72, 677.01, 671.61, 666.2, 660.79, 655.37, 649.94, 644.51, 639.07, 633.63, 628.17, 622.72, 619.53, 617.5, 615.37, 613.15, 610.83, 608.41, 605.89, 603.27, 600.55, 597.73, 594.82, 593.91, 594.41, 594.94, 595.49, 596.07, 596.67, 597.29, 597.94, 598.61, 599.31, 600.04, 604.23, 611.35, 618.56, 625.85, 633.23, 640.7, 648.25, 655.89, 663.61, 671.43, 679.32, 689.62, 702.38, 715.19, 728.08, 741.03, 754.04, 767.12, 780.27, 793.48, 806.76, 820.1, 828.96, 831.71, 834.09, 836.09, 837.72, 838.97, 1335.26, 1313.28, 1279.29, 1244.72, 1210.03, 1175.2, 1140.24, 1105.15, 1069.93, 1034.58, 999.1, 963.48, 927.74, 918.07, 910.17, 902.31, 894.47, 886.66, 878.88, 871.13, 863.41, 855.73, 848.07, 840.44, 829.7, 818.69, 807.74, 796.85, 786.01, 775.23, 764.5, 753.83, 743.22, 732.66, 722.16, 720.55, 720.13, 719.68, 719.2, 718.68, 718.14, 717.56, 716.95, 716.31, 715.63, 714.92, 712.29, 709.32, 706.61, 703.93, 701.26, 698.61, 695.98, 693.37, 690.78, 688.21, 685.66, 682.54, 679.35, 676.22, 673.15, 670.15, 667.21, 664.34, 661.54, 658.8, 656.12, 653.52, 660.56, 670.59, 680.81, 691.22, 701.82, 712.62, 723.61, 734.79, 746.16, 757.73, 769.49, 779.32, 788.28, 797.11, 805.81, 814.37, 822.81, 831.11, 839.28, 847.31, 855.21, 862.98, 861.02, 854.7, 848.26, 841.71, 835.04, 828.26, 821.37, 814.35, 807.23, 799.99, 792.63, 783.6, 773.72, 763.89, 754.08, 744.32, 734.58, 724.88, 715.22, 705.59, 695.99, 686.43, 679.38, 673.94, 668.5, 663.05, 657.59, 652.13, 646.66, 641.18, 635.7, 630.21, 624.71, 620.89, 618.34, 615.69, 612.94, 610.09, 607.14, 604.09, 600.95, 597.71, 594.37, 590.92, 589.5, 590.13, 590.79, 591.47, 592.18, 592.91, 593.67, 594.45, 595.25, 596.08, 596.94, 600.81, 608.38, 616.04, 623.79, 631.61, 639.53, 647.53, 655.62, 663.79, 672.05, 680.4, 690.66, 703.75, 716.9, 730.12, 743.41, 756.76, 770.17, 783.65, 797.2, 810.82, 824.5, 833.93, 834.77, 835.23, 835.32, 835.04, 834.37, 1337.54, 1315.09, 1281.31, 1245.94, 1210.43, 1174.8, 1139.03, 1103.13, 1067.1, 1030.94, 994.65, 958.23, 921.68, 909.8, 902.09, 894.41, 886.77, 879.15, 871.56, 864.0, 856.47, 848.97, 841.51, 834.07, 823.92, 813.26, 802.65, 792.1, 781.6, 771.16, 760.77, 750.44, 740.17, 729.96, 719.8, 717.26, 716.96, 716.62, 716.24, 715.81, 715.33, 714.8, 714.23, 713.6, 712.93, 712.21, 710.04, 707.42, 704.83, 702.26, 699.71, 697.17, 694.66, 692.16, 689.69, 687.23, 684.79, 682.04, 679.22, 676.47, 673.78, 671.15, 668.6, 666.1, 663.68, 661.31, 659.02, 656.79, 663.63, 674.75, 686.06, 697.57, 709.26, 721.15, 733.23, 745.5, 757.96, 770.62, 783.48, 793.44, 801.65, 809.73, 817.67, 825.48, 833.16, 840.71, 848.12, 855.4, 862.55, 869.56, 868.15, 861.19, 854.11, 846.92, 839.62, 832.2, 824.66, 817.01, 809.25, 801.37, 793.37, 784.38, 774.7, 765.05, 755.44, 745.86, 736.31, 726.8, 717.33, 707.89, 698.48, 689.11, 681.75, 676.28, 670.8, 665.31, 659.82, 654.32, 648.81, 643.3, 637.78, 632.25, 626.72, 622.35, 619.27, 616.1, 612.82, 609.44, 605.97, 602.4, 598.73, 594.96, 591.09, 587.13, 585.06, 585.82, 586.61, 587.43, 588.26, 589.13, 590.01, 590.93, 591.86, 592.82, 593.81, 597.31, 605.33, 613.44, 621.63, 629.91, 638.27, 646.73, 655.26, 663.89, 672.6, 681.39, 691.63, 705.05, 718.54, 732.1, 745.72, 759.41, 773.16, 786.98, 800.86, 814.81, 828.83, 839.27, 838.2, 836.75, 834.92, 832.73, 830.15, 1339.89, 1316.98, 1283.46, 1247.28, 1210.97, 1174.52, 1137.95, 1101.24, 1064.41, 1027.44, 990.34, 953.11, 915.74, 901.5, 893.98, 886.49, 879.03, 871.6, 864.21, 856.84, 849.5, 842.19, 834.91, 827.71, 818.95, 809.34, 799.69, 790.0, 780.28, 770.52, 760.72, 750.89, 741.01, 731.1, 721.16, 717.91, 717.34, 716.72, 716.05, 715.34, 714.57, 713.76, 712.91, 712.0, 711.05, 710.05, 708.01, 705.51, 703.04, 700.58, 698.14, 695.72, 693.32, 690.93, 688.57, 686.23, 683.9, 681.47, 679.03, 676.66, 674.34, 672.1, 669.91, 667.8, 665.75, 663.76, 661.84, 659.99, 666.51, 678.72, 691.13, 703.72, 716.51, 729.48, 742.65, 756.02, 769.57, 783.32, 797.27, 807.7, 815.16, 822.48, 829.67, 836.73, 843.65, 850.45, 857.11, 863.63, 870.03, 876.29, 875.4, 867.79, 860.08, 852.25, 844.31, 836.25, 828.07, 819.79, 811.38, 802.86, 794.23, 785.13, 775.64, 766.18, 756.76, 747.37, 738.01, 728.69, 719.4, 710.15, 700.93, 691.75, 684.13, 678.62, 673.11, 667.58, 662.05, 656.52, 650.97, 645.42, 639.87, 634.3, 628.73, 623.91, 620.31, 616.6, 612.8, 608.9, 604.9, 600.81, 596.61, 592.32, 587.92, 583.43, 580.59, 581.49, 582.41, 583.36, 584.33, 585.32, 586.34, 587.38, 588.45, 589.54, 590.66, 593.72, 602.19, 610.75, 619.39, 628.12, 636.93, 645.83, 654.82, 663.89, 673.05, 682.3, 692.53, 706.29, 720.12, 734.01, 747.97, 761.99, 776.08, 790.23, 804.45, 818.74, 833.09, 844.99, 842.0, 838.64, 834.9, 830.79, 826.31, 1342.32, 1318.94, 1285.75, 1248.76, 1211.63, 1174.38, 1137.0, 1099.48, 1061.84, 1024.06, 986.16, 948.21, 910.58, 894.46, 887.77, 881.16, 874.64, 868.21, 861.86, 855.6, 849.43, 843.34, 837.35, 831.43, 822.96, 813.12, 803.25, 793.34, 783.39, 773.4, 763.38, 753.32, 743.22, 733.09, 722.91, 718.62, 717.76, 716.86, 715.91, 714.91, 713.87, 712.78, 711.64, 710.45, 709.21, 707.93, 705.97, 703.59, 701.22, 698.88, 696.55, 694.24, 691.95, 689.69, 687.44, 685.21, 682.99, 680.84, 678.78, 676.78, 674.84, 672.97, 671.17, 669.43, 667.76, 666.15, 664.61, 663.13, 669.2, 682.5, 696.0, 709.68, 723.56, 737.63, 751.89, 766.34, 780.99, 795.83, 810.86, 822.09, 828.79, 835.36, 841.8, 848.11, 854.28, 860.32, 866.22, 872.0, 877.64, 883.14, 882.76, 874.52, 866.16, 857.69, 849.11, 840.41, 831.6, 822.67, 813.63, 804.47, 795.2, 785.85, 776.55, 767.28, 758.04, 748.84, 739.67, 730.54, 721.44, 712.38, 703.35, 694.36, 686.52, 680.98, 675.42, 669.86, 664.29, 658.72, 653.14, 647.55, 641.96, 636.36, 630.75, 625.57, 621.44, 617.21, 612.88, 608.46, 603.93, 599.31, 594.59, 589.77, 584.85, 579.83, 576.1, 577.13, 578.19, 579.26, 580.36, 581.49, 582.64, 583.81, 585.01, 586.24, 587.49, 590.04, 598.96, 607.97, 617.06, 626.24, 635.5, 644.85, 654.29, 663.81, 673.42, 683.12, 693.37, 707.47, 721.63, 735.86, 750.15, 764.51, 778.94, 793.43, 807.98, 822.6, 837.29, 851.08, 846.18, 840.9, 835.26, 829.24, 822.84, 1341.34, 1316.04, 1282.21, 1243.83, 1205.71, 1167.86, 1130.28, 1092.97, 1055.92, 1019.14, 982.63, 946.39, 910.42, 892.8, 886.64, 880.56, 874.57, 868.67, 862.85, 857.12, 851.48, 845.92, 840.45, 835.07, 827.01, 816.94, 806.85, 796.71, 786.53, 776.32, 766.07, 755.79, 745.47, 735.11, 724.71, 719.37, 718.24, 717.05, 715.82, 714.54, 713.21, 711.84, 710.41, 708.94, 707.42, 705.86, 703.91, 701.64, 699.39, 697.15, 694.94, 692.75, 690.57, 688.42, 686.28, 684.16, 682.07, 680.15, 678.46, 676.84, 675.28, 673.78, 672.36, 670.99, 669.7, 668.47, 667.3, 666.2, 671.7, 686.09, 700.68, 715.45, 730.42, 745.58, 760.93, 776.48, 792.22, 808.15, 824.27, 836.62, 842.57, 848.38, 854.07, 859.62, 865.04, 870.32, 875.47, 880.49, 885.38, 890.13, 890.23, 881.35, 872.36, 863.25, 854.03, 844.69, 835.24, 825.68, 816.0, 806.2, 796.29, 786.53, 777.42, 768.34, 759.29, 750.28, 741.3, 732.36, 723.45, 714.57, 705.73, 696.93, 688.92, 683.34, 677.74, 672.15, 666.54, 660.93, 655.32, 649.69, 644.06, 638.42, 632.78, 627.32, 622.67, 617.91, 613.06, 608.11, 603.06, 597.91, 592.66, 587.32, 581.88, 576.33, 571.59, 572.75, 573.93, 575.14, 576.38, 577.63, 578.91, 580.22, 581.55, 582.91, 584.28, 586.28, 595.65, 605.11, 614.65, 624.28, 633.99, 643.79, 653.67, 663.65, 673.7, 683.85, 694.14, 708.58, 723.08, 737.64, 752.27, 766.97, 781.73, 796.55, 811.45, 826.4, 841.43, 856.52, 850.73, 843.54, 835.99, 828.05, 819.75, 1320.2, 1296.32, 1265.13, 1228.42, 1191.97, 1155.79, 1119.87, 1084.22, 1048.84, 1013.73, 978.88, 944.31, 910.0, 891.06, 885.42, 879.87, 874.41, 869.04, 863.75, 858.55, 853.43, 848.41, 843.47, 838.61, 831.09, 820.81, 810.48, 800.12, 789.72, 779.28, 768.81, 758.3, 747.75, 737.16, 726.54, 720.18, 718.76, 717.29, 715.77, 714.21, 712.6, 710.94, 709.24, 707.48, 705.68, 703.83, 701.83, 699.67, 697.53, 695.41, 693.31, 691.23, 689.17, 687.13, 685.11, 683.1, 681.12, 679.38, 678.07, 676.83, 675.65, 674.53, 673.48, 672.49, 671.57, 670.72, 669.93, 669.21, 674.01, 689.49, 705.17, 721.03, 737.09, 753.34, 769.79, 786.42, 803.25, 820.27, 837.48, 851.28, 856.47, 861.53, 866.46, 871.26, 875.93, 880.46, 884.86, 889.13, 893.26, 897.26, 897.82, 888.3, 878.67, 868.92, 859.06, 849.09, 839.0, 828.8, 818.48, 808.04, 797.49, 787.18, 778.25, 769.36, 760.5, 751.68, 742.89, 734.14, 725.42, 716.73, 708.08, 699.47, 691.32, 685.7, 680.07, 674.44, 668.8, 663.15, 657.5, 651.84, 646.17, 640.49, 634.81, 629.18, 623.99, 618.71, 613.33, 607.86, 602.28, 596.61, 590.84, 584.97, 579.0, 572.93, 567.05, 568.34, 569.66, 571.0, 572.36, 573.75, 575.16, 576.6, 578.06, 579.55, 581.06, 582.6, 592.25, 602.16, 612.15, 622.23, 632.39, 642.64, 652.97, 663.39, 673.9, 684.49, 695.17, 709.62, 724.46, 739.36, 754.32, 769.35, 784.45, 799.61, 814.84, 830.14, 845.5, 860.92, 855.65, 846.56, 837.09, 827.24, 817.03, 1298.84, 1276.38, 1247.79, 1212.74, 1177.96, 1143.44, 1109.19, 1075.2, 1041.49, 1008.04, 974.86, 941.95, 909.31, 889.22, 884.12, 879.1, 874.17, 869.32, 864.56, 859.89, 855.31, 850.81, 846.4, 842.07, 835.22, 824.7, 814.15, 803.56, 792.94, 782.28, 771.58, 760.84, 750.07, 739.26, 728.41, 721.03, 719.32, 717.57, 715.78, 713.93, 712.04, 710.1, 708.11, 706.07, 703.99, 701.86, 699.73, 697.68, 695.66, 693.65, 691.67, 689.7, 687.75, 685.82, 683.91, 682.02, 680.15, 678.56, 677.62, 676.75, 675.95, 675.21, 674.54, 673.93, 673.39, 672.91, 672.5, 672.15, 676.12, 692.7, 709.47, 726.42, 743.57, 760.91, 778.45, 796.17, 814.09, 832.2, 850.5, 866.07, 870.51, 874.82, 879.0, 883.04, 886.95, 890.73, 894.38, 897.89, 901.27, 904.52, 905.52, 895.37, 885.1, 874.71, 864.21, 853.6, 842.87, 832.03, 821.07, 810.0, 798.81, 787.8, 779.06, 770.35, 761.69, 753.05, 744.45, 735.89, 727.35, 718.86, 710.4, 701.97, 693.73, 688.08, 682.41, 676.74, 671.06, 665.38, 659.68, 653.99, 648.28, 642.57, 636.85, 631.13, 625.42, 619.61, 613.71, 607.71, 601.61, 595.41, 589.11, 582.72, 576.22, 569.63, 562.94, 563.91, 565.36, 566.83, 568.32, 569.84, 571.39, 572.96, 574.55, 576.17, 577.81, 579.48, 588.77, 599.12, 609.56, 620.09, 630.7, 641.4, 652.18, 663.05, 674.01, 685.05, 696.18, 710.6, 725.77, 741.01, 756.31, 771.68, 787.11, 802.61, 818.17, 833.8, 848.77, 863.25, 858.17, 847.04, 836.07, 825.24, 814.56, 1277.25, 1256.21, 1230.18, 1196.8, 1163.68, 1130.82, 1098.24, 1065.92, 1033.87, 1002.08, 970.57, 939.32, 908.35, 887.31, 882.73, 878.24, 873.84, 869.52, 865.29, 861.15, 857.09, 853.12, 849.24, 845.44, 839.38, 828.64, 817.86, 807.05, 796.2, 785.31, 774.39, 763.43, 752.43, 741.39, 730.32, 721.93, 719.94, 717.91, 715.83, 713.7, 711.52, 709.3, 707.03, 704.71, 702.35, 699.93, 697.61, 695.68, 693.77, 691.87, 690.0, 688.15, 686.31, 684.5, 682.7, 680.92, 679.16, 677.67, 677.11, 676.62, 676.19, 675.83, 675.53, 675.3, 675.13, 675.03, 675.0, 675.03, 678.05, 695.72, 713.58, 731.62, 749.86, 768.29, 786.91, 805.73, 824.74, 843.94, 863.33, 881.0, 884.68, 888.24, 891.66, 894.95, 898.11, 901.14, 904.03, 906.79, 909.42, 911.92, 913.34, 902.55, 891.64, 880.61, 869.48, 858.23, 846.86, 835.38, 823.78, 812.07, 800.25, 788.38, 779.83, 771.31, 762.83, 754.39, 745.98, 737.6, 729.26, 720.95, 712.68, 704.44, 696.23, 690.45, 684.75, 679.04, 673.33, 667.61, 661.88, 656.14, 650.4, 644.65, 638.9, 633.14, 626.94, 620.61, 614.18, 607.65, 601.03, 594.31, 587.48, 580.56, 573.54, 566.43, 559.21, 559.46, 561.03, 562.64, 564.26, 565.91, 567.59, 569.29, 571.01, 572.76, 574.53, 576.33, 585.2, 596.0, 606.89, 617.87, 628.93, 640.08, 651.31, 662.63, 674.03, 685.51, 696.78, 710.74, 725.61, 740.38, 755.05, 769.61, 784.07, 798.44, 812.69, 826.85, 840.9, 854.85, 852.57, 842.2, 831.98, 821.9, 811.97, 1255.43, 1235.81, 1212.3, 1180.58, 1149.13, 1117.94, 1087.02, 1056.36, 1025.98, 995.86, 966.01, 936.43, 907.11, 885.31, 881.26, 877.3, 873.42, 869.64, 865.93, 862.32, 858.79, 855.35, 852.0, 848.73, 843.57, 832.61, 821.61, 810.57, 799.49, 788.38, 777.23, 766.05, 754.82, 743.56, 732.26, 722.87, 720.6, 718.29, 715.93, 713.52, 711.06, 708.55, 706.0, 703.4, 700.75, 698.05, 695.47, 693.65, 691.85, 690.07, 688.31, 686.57, 684.85, 683.15, 681.47, 679.8, 678.16, 676.71, 676.53, 676.41, 676.36, 676.38, 676.46, 676.6, 676.81, 677.09, 677.43, 677.84, 679.79, 698.55, 717.49, 736.63, 755.96, 775.48, 795.19, 815.1, 835.19, 855.48, 875.97, 896.05, 898.99, 901.79, 904.46, 907.0, 909.41, 911.68, 913.82, 915.83, 917.7, 919.45, 921.05, 909.84, 898.29, 886.63, 874.86, 862.97, 850.96, 838.84, 826.61, 814.26, 801.8, 789.22, 780.56, 772.24, 763.94, 755.69, 747.46, 739.28, 731.12, 723.0, 714.92, 706.87, 698.86, 692.84, 687.1, 681.36, 675.6, 669.85, 664.08, 658.31, 652.53, 646.74, 640.95, 635.15, 628.56, 621.7, 614.75, 607.7, 600.55, 593.3, 585.95, 578.51, 570.96, 563.32, 555.58, 554.98, 556.68, 558.42, 560.18, 561.96, 563.76, 565.59, 567.45, 569.33, 571.23, 573.37, 582.01, 593.41, 604.77, 616.1, 627.4, 638.67, 649.9, 661.1, 672.27, 683.4, 694.5, 707.7, 722.05, 736.29, 750.43, 764.47, 778.41, 792.24, 805.97, 819.6, 833.13, 846.55, 846.83, 837.21, 827.74, 818.41, 809.23, 1233.39, 1215.18, 1194.16, 1164.1, 1134.31, 1104.79, 1075.53, 1046.54, 1017.82, 989.36, 961.18, 933.26, 905.61, 883.22, 879.7, 876.27, 872.92, 869.66, 866.49, 863.4, 860.4, 857.49, 854.66, 851.93, 847.81, 836.62, 825.39, 814.13, 802.83, 791.49, 780.12, 768.7, 757.26, 745.77, 734.24, 723.87, 721.31, 718.72, 716.07, 713.38, 710.64, 707.85, 705.01, 702.13, 699.2, 696.22, 693.31, 691.61, 689.92, 688.26, 686.61, 684.98, 683.37, 681.78, 680.21, 678.66, 677.13, 675.69, 675.89, 676.15, 676.47, 676.86, 677.32, 677.84, 678.43, 679.08, 679.8, 680.58, 681.43, 701.18, 721.22, 741.45, 761.87, 782.48, 803.28, 824.27, 845.46, 866.84, 888.41, 910.18, 913.43, 915.48, 917.39, 919.18, 920.83, 922.35, 923.74, 925.0, 926.12, 927.11, 927.97, 917.25, 905.06, 892.76, 880.35, 867.82, 855.18, 842.42, 829.55, 816.57, 803.46, 790.25, 781.26, 773.12, 765.02, 756.95, 748.92, 740.92, 732.96, 725.03, 717.13, 709.27, 701.44, 695.23, 689.46, 683.68, 677.89, 672.09, 666.29, 660.48, 654.66, 648.84, 643.01, 637.18, 630.28, 622.9, 615.42, 607.84, 600.17, 592.39, 584.52, 576.55, 568.48, 560.32, 552.28, 551.64, 554.29, 556.85, 559.31, 561.69, 563.97, 566.16, 568.25, 570.26, 572.17, 573.99, 581.46, 592.69, 603.88, 615.04, 626.17, 637.27, 648.33, 659.36, 670.36, 681.32, 692.26, 704.77, 718.59, 732.31, 745.92, 759.43, 772.85, 786.15, 799.36, 812.46, 825.46, 838.36, 840.94, 832.07, 823.35, 814.77, 806.34, 1211.12, 1194.33, 1175.75, 1147.36, 1119.23, 1091.37, 1063.78, 1036.45, 1009.39, 982.6, 956.07, 929.82, 903.83, 881.04, 878.05, 875.15, 872.33, 869.6, 866.96, 864.4, 861.93, 859.54, 857.25, 855.04, 852.08, 840.66, 829.21, 817.72, 806.2, 794.64, 783.04, 771.4, 759.73, 748.01, 736.27, 724.91, 722.07, 719.19, 716.26, 713.29, 710.27, 707.2, 704.08, 700.91, 697.7, 694.44, 691.13, 689.54, 687.97, 686.42, 684.88, 683.37, 681.88, 680.4, 678.94, 677.5, 676.09, 674.69, 675.18, 675.81, 676.52, 677.28, 678.12, 679.02, 679.98, 681.01, 682.1, 683.26, 684.49, 703.63, 724.76, 746.08, 767.58, 789.28, 811.17, 833.26, 855.53, 878.0, 900.66, 923.52, 928.0, 929.29, 930.46, 931.49, 932.39, 933.16, 933.8, 934.3, 934.67, 934.91, 935.01, 924.77, 911.95, 899.01, 885.96, 872.79, 859.51, 846.12, 832.61, 818.98, 805.24, 791.39, 781.93, 773.98, 766.07, 758.19, 750.34, 742.53, 734.75, 727.01, 719.31, 711.63, 704.0, 697.64, 691.82, 686.0, 680.18, 674.34, 668.5, 662.66, 656.81, 650.95, 645.08, 639.19, 631.72, 623.71, 615.85, 608.16, 600.62, 593.24, 586.02, 578.96, 572.06, 565.31, 558.73, 557.23, 559.39, 561.46, 563.44, 565.32, 567.12, 568.82, 570.43, 571.94, 573.37, 574.7, 580.95, 592.0, 603.03, 614.02, 624.98, 635.9, 646.8, 657.66, 668.48, 679.28, 690.04, 701.94, 715.23, 728.42, 741.51, 754.5, 767.38, 780.17, 792.85, 805.42, 817.9, 830.27, 834.91, 826.79, 818.82, 810.99, 803.31, 1188.62, 1173.25, 1157.07, 1130.34, 1103.88, 1077.68, 1051.75, 1026.09, 1000.69, 975.56, 950.7, 926.11, 901.79, 878.78, 876.32, 873.95, 871.66, 869.45, 867.34, 865.31, 863.37, 861.51, 859.74, 858.06, 856.39, 844.75, 833.07, 821.36, 809.61, 797.82, 785.99, 774.13, 762.23, 750.3, 738.32, 726.31, 722.88, 719.71, 716.5, 713.25, 709.94, 706.59, 703.19, 699.74, 696.25, 692.7, 689.11, 687.46, 686.0, 684.56, 683.14, 681.74, 680.36, 679.0, 677.65, 676.33, 675.02, 673.74, 674.4, 675.42, 676.5, 677.64, 678.85, 680.12, 681.46, 682.87, 684.34, 685.88, 687.48, 705.89, 728.11, 750.51, 773.11, 795.9, 818.88, 842.05, 865.42, 888.97, 912.72, 936.67, 942.7, 943.25, 943.66, 943.94, 944.09, 944.1, 943.99, 943.74, 943.36, 942.84, 942.19, 932.41, 918.95, 905.37, 891.68, 877.88, 863.96, 849.93, 835.78, 821.52, 807.14, 792.65, 782.56, 774.8, 767.07, 759.38, 751.73, 744.11, 736.52, 728.97, 721.45, 713.96, 706.52, 699.92, 693.88, 687.77, 681.6, 675.36, 669.06, 662.69, 656.26, 649.76, 643.19, 636.56, 629.55, 622.38, 615.38, 608.53, 601.84, 595.3, 588.93, 582.71, 576.65, 570.75, 565.01, 562.92, 564.59, 566.17, 567.66, 569.05, 570.36, 571.57, 572.69, 573.72, 574.66, 575.5, 580.46, 591.35, 602.2, 613.03, 623.81, 634.57, 645.29, 655.98, 666.64, 677.27, 687.86, 699.21, 711.98, 724.64, 737.21, 749.67, 762.02, 774.28, 786.43, 798.49, 810.43, 822.28, 828.73, 821.36, 814.14, 807.06, 800.13, 1165.9, 1151.95, 1138.12, 1113.06, 1088.26, 1063.73, 1039.46, 1015.46, 991.73, 968.26, 945.06, 922.13, 899.47, 877.08, 874.5, 872.66, 870.9, 869.22, 867.63, 866.13, 864.72, 863.39, 862.15, 860.99, 859.93, 848.87, 836.97, 825.03, 813.05, 801.04, 788.99, 776.9, 764.78, 752.62, 740.42, 728.18, 723.73, 720.29, 716.79, 713.25, 709.67, 706.03, 702.35, 698.62, 694.84, 691.01, 687.14, 685.36, 684.01, 682.69, 681.38, 680.09, 678.82, 677.57, 676.34, 675.13, 673.94, 672.76, 673.56, 674.95, 676.41, 677.93, 679.52, 681.17, 682.88, 684.66, 686.51, 688.42, 690.4, 707.96, 731.26, 754.76, 778.44, 802.32, 826.39, 850.65, 875.11, 899.75, 924.59, 949.62, 957.54, 957.33, 956.99, 956.52, 955.92, 955.18, 954.31, 953.31, 952.18, 950.91, 949.51, 940.16, 926.06, 911.85, 897.52, 883.08, 868.52, 853.85, 839.07, 824.17, 809.15, 794.01, 782.71, 774.77, 766.92, 759.17, 751.51, 743.95, 736.48, 729.11, 721.84, 714.66, 707.58, 700.91, 694.51, 688.05, 681.53, 674.93, 668.28, 661.55, 654.76, 647.9, 640.98, 633.99, 627.22, 620.91, 614.74, 608.74, 602.89, 597.2, 591.67, 586.3, 581.09, 576.03, 571.14, 568.69, 569.87, 570.97, 571.97, 572.88, 573.69, 574.42, 575.05, 575.59, 576.04, 576.4, 580.01, 590.73, 601.41, 612.06, 622.68, 633.27, 643.82, 654.34, 664.83, 675.29, 685.71, 696.58, 708.82, 720.96, 733.0, 744.94, 756.77, 768.5, 780.13, 791.65, 803.07, 814.39, 822.4, 815.78, 809.31, 802.98, 796.8, 1142.95, 1130.41, 1118.1, 1095.51, 1072.38, 1049.51, 1026.9, 1004.56, 982.49, 960.69, 939.15, 917.89, 896.89, 876.16, 872.6, 871.28, 870.05, 868.9, 867.84, 866.87, 865.98, 865.18, 864.47, 863.84, 863.31, 853.03, 840.9, 828.74, 816.54, 804.3, 792.02, 779.71, 767.36, 754.98, 742.55, 730.09, 724.63, 720.9, 717.13, 713.31, 709.44, 705.52, 701.55, 697.54, 693.48, 689.37, 685.22, 683.24, 682.0, 680.79, 679.6, 678.42, 677.27, 676.13, 675.01, 673.91, 672.83, 671.77, 672.66, 674.43, 676.26, 678.16, 680.12, 682.15, 684.24, 686.39, 688.62, 690.91, 693.26, 709.84, 734.23, 758.81, 783.59, 808.55, 833.71, 859.06, 884.6, 910.34, 936.27, 962.39, 972.51, 971.55, 970.46, 969.23, 967.88, 966.39, 964.77, 963.02, 961.13, 959.11, 956.64, 946.35, 930.44, 914.59, 898.8, 883.07, 867.4, 851.79, 836.24, 820.76, 805.33, 789.96, 778.41, 770.99, 763.67, 756.44, 749.31, 742.28, 735.34, 728.5, 721.75, 715.1, 708.54, 701.97, 695.22, 688.4, 681.52, 674.57, 667.56, 660.47, 653.33, 646.12, 638.84, 631.49, 624.74, 619.27, 613.95, 608.79, 603.79, 598.95, 594.26, 589.73, 585.37, 581.16, 577.11, 574.56, 575.25, 575.86, 576.37, 576.79, 577.12, 577.36, 577.5, 577.56, 577.52, 577.39, 579.59, 590.14, 600.65, 611.14, 621.59, 632.0, 642.39, 652.74, 663.06, 673.34, 683.59, 694.06, 705.77, 717.39, 728.9, 740.31, 751.61, 762.82, 773.92, 784.92, 795.82, 806.61, 815.93, 810.06, 804.34, 798.76, 793.33, 1119.77, 1108.65, 1097.76, 1077.7, 1056.22, 1035.02, 1014.07, 993.4, 972.99, 952.85, 932.98, 913.37, 894.03, 874.97, 870.61, 869.82, 869.11, 868.5, 867.96, 867.52, 867.16, 866.89, 866.7, 866.61, 866.6, 857.22, 844.87, 832.48, 820.06, 807.59, 795.09, 782.56, 769.98, 757.37, 744.72, 732.04, 725.58, 721.57, 717.51, 713.41, 709.25, 705.06, 700.81, 696.51, 692.17, 687.78, 683.34, 681.1, 679.98, 678.88, 677.8, 676.73, 675.69, 674.67, 673.66, 672.68, 671.71, 670.76, 671.69, 673.84, 676.04, 678.32, 680.65, 683.06, 685.53, 688.06, 690.66, 693.32, 696.05, 711.52, 737.01, 762.68, 788.54, 814.6, 840.84, 867.28, 893.91, 920.73, 947.75, 974.19, 985.33, 982.18, 978.91, 975.52, 972.0, 968.35, 964.59, 960.69, 956.68, 952.54, 948.27, 938.84, 923.27, 907.76, 892.31, 876.92, 861.59, 846.32, 831.11, 815.97, 800.88, 785.85, 774.01, 767.12, 760.33, 753.63, 747.02, 740.51, 734.1, 727.78, 721.56, 715.44, 709.41, 703.1, 695.99, 688.82, 681.58, 674.27, 666.9, 659.46, 651.96, 644.39, 636.76, 629.06, 622.1, 617.47, 613.0, 608.69, 604.53, 600.53, 596.69, 593.01, 589.49, 586.12, 582.91, 580.51, 580.72, 580.84, 580.86, 580.8, 580.64, 580.39, 580.05, 579.62, 579.09, 578.47, 579.21, 589.58, 599.93, 610.24, 620.52, 630.77, 640.98, 651.16, 661.31, 671.43, 681.51, 691.63, 702.82, 713.91, 724.89, 735.78, 746.56, 757.24, 767.82, 778.29, 788.66, 798.93, 809.09, 804.19, 799.22, 794.39, 789.7, 1096.37, 1086.67, 1077.19, 1059.61, 1039.8, 1020.26, 1000.98, 981.96, 963.22, 944.74, 926.53, 908.58, 890.91, 873.5, 868.53, 868.27, 868.09, 868.0, 868.0, 868.08, 868.25, 868.51, 868.85, 869.28, 869.8, 861.45, 848.88, 836.26, 823.61, 810.93, 798.2, 785.44, 772.64, 759.8, 746.93, 734.02, 726.57, 722.28, 717.94, 713.55, 709.12, 704.64, 700.11, 695.53, 690.91, 686.24, 681.52, 678.94, 677.93, 676.94, 675.97, 675.03, 674.1, 673.19, 672.29, 671.42, 670.57, 669.73, 670.66, 673.18, 675.76, 678.41, 681.13, 683.91, 686.75, 689.66, 692.64, 695.78, 700.1, 715.27, 742.19, 768.83, 795.19, 821.25, 847.03, 872.53, 897.73, 922.65, 947.29, 971.63, 984.04, 980.19, 976.23, 972.14, 967.92, 963.58, 959.12, 954.53, 949.82, 944.99, 940.03, 931.27, 916.04, 900.87, 885.76, 870.71, 855.72, 840.79, 825.92, 811.12, 796.37, 781.68, 769.52, 763.15, 756.88, 750.71, 744.63, 738.65, 732.77, 726.98, 721.28, 715.68, 710.18, 704.29, 696.82, 689.29, 681.7, 674.04, 666.31, 658.52, 650.66, 642.74, 634.75, 626.69, 619.31, 615.52, 611.9, 608.43, 605.11, 601.96, 598.96, 596.13, 593.45, 590.93, 588.56, 586.56, 586.28, 585.91, 585.45, 584.9, 584.25, 583.52, 582.69, 581.76, 580.75, 579.64, 578.85, 589.06, 599.24, 609.38, 619.49, 629.56, 639.61, 649.62, 659.6, 669.55, 679.46, 689.34, 699.97, 710.53, 720.99, 731.35, 741.61, 751.76, 761.81, 771.76, 781.61, 791.35, 800.99, 798.17, 793.95, 789.87, 785.94, 1072.74, 1064.45, 1056.39, 1041.27, 1023.12, 1005.23, 987.62, 970.26, 953.18, 936.36, 919.81, 903.53, 887.52, 871.77, 866.37, 866.63, 866.99, 867.42, 867.95, 868.56, 869.26, 870.04, 870.91, 871.87, 872.91, 865.72, 852.92, 840.08, 827.21, 814.3, 801.35, 788.36, 775.34, 762.28, 749.18, 736.04, 727.61, 723.04, 718.42, 713.75, 709.03, 704.27, 699.46, 694.6, 689.69, 684.74, 679.74, 676.76, 675.86, 674.99, 674.13, 673.3, 672.48, 671.68, 670.91, 670.15, 670.63, 671.6, 674.12, 678.51, 682.87, 687.21, 691.53, 695.82, 700.09, 704.33, 708.55, 712.75, 716.92, 729.38, 754.67, 779.66, 804.38, 828.81, 852.95, 876.8, 900.37, 923.65, 946.65, 969.35, 982.87, 978.33, 973.67, 968.88, 963.97, 958.94, 953.78, 948.5, 943.09, 937.56, 931.91, 923.64, 908.75, 893.92, 879.15, 864.44, 849.79, 835.2, 820.67, 806.2, 791.79, 777.45, 764.93, 759.09, 753.35, 747.7, 742.15, 736.69, 731.34, 726.07, 720.9, 715.83, 710.86, 705.54, 697.72, 689.84, 681.89, 673.87, 665.79, 657.64, 649.43, 641.15, 632.8, 624.39, 616.35, 613.41, 610.63, 608.01, 605.54, 603.23, 601.08, 599.08, 597.25, 595.57, 594.06, 592.7, 591.94, 591.08, 590.13, 589.09, 587.96, 586.73, 585.42, 584.01, 582.51, 580.91, 579.23, 588.57, 598.57, 608.55, 618.49, 628.39, 638.27, 648.11, 657.92, 667.7, 677.44, 687.15, 697.23, 707.26, 717.2, 727.03, 736.76, 746.39, 755.91, 765.34, 774.66, 783.87, 792.99, 792.01, 788.53, 785.2, 782.02, 1048.88, 1042.01, 1035.37, 1022.65, 1006.16, 989.94, 973.98, 958.29, 942.87, 927.71, 912.82, 898.2, 883.85, 869.77, 864.12, 864.91, 865.79, 866.76, 867.81, 868.95, 870.17, 871.49, 872.88, 874.37, 875.94, 870.03, 857.0, 843.94, 830.84, 817.7, 804.53, 791.32, 778.07, 764.78, 751.46, 738.1, 728.7, 723.85, 718.94, 713.99, 708.99, 703.95, 698.86, 693.72, 689.26, 685.38, 681.56, 679.64, 680.38, 681.2, 682.07, 683.01, 684.02, 685.1, 686.24, 687.44, 688.71, 690.05, 692.48, 696.71, 700.93, 705.12, 709.28, 713.42, 717.54, 721.63, 725.7, 729.75, 733.77, 743.78, 767.42, 790.78, 813.86, 836.65, 859.15, 881.37, 903.3, 924.94, 946.3, 967.36, 981.82, 976.59, 971.23, 965.75, 960.14, 954.41, 948.56, 942.59, 936.48, 930.26, 923.91, 915.95, 901.4, 886.91, 872.48, 858.11, 843.8, 829.55, 815.36, 801.23, 787.16, 773.15, 760.24, 754.93, 749.72, 744.6, 739.57, 734.64, 729.81, 725.07, 720.43, 715.88, 711.43, 706.87, 698.69, 690.45, 682.14, 673.77, 665.33, 656.83, 648.26, 639.62, 630.92, 622.15, 613.32, 611.15, 609.21, 607.43, 605.81, 604.34, 603.03, 601.88, 600.89, 600.06, 599.39, 598.87, 597.68, 596.34, 594.9, 593.37, 591.75, 590.04, 588.24, 586.34, 584.35, 582.27, 580.1, 588.11, 597.95, 607.75, 617.52, 627.26, 636.96, 646.63, 656.27, 665.88, 675.46, 685.0, 694.58, 704.09, 713.5, 722.81, 732.01, 741.11, 750.12, 759.01, 767.81, 776.5, 785.09, 785.7, 782.97, 780.39, 777.96, 1024.8, 1019.35, 1014.12, 1003.77, 988.94, 974.38, 960.09, 946.06, 932.29, 918.8, 905.57, 892.61, 879.92, 867.49, 861.79, 863.11, 864.52, 866.01, 867.59, 869.25, 871.01, 872.85, 874.77, 876.78, 878.88, 874.37, 861.12, 847.83, 834.51, 821.15, 807.75, 794.31, 781.1, 768.18, 755.43, 742.84, 733.6, 729.58, 725.63, 721.73, 717.9, 714.13, 710.42, 706.77, 703.19, 699.66, 696.2, 694.27, 695.39, 696.58, 697.83, 699.15, 700.53, 701.98, 703.5, 705.08, 706.73, 708.45, 710.86, 714.95, 719.01, 723.05, 727.06, 731.05, 735.02, 738.96, 742.88, 746.77, 750.64, 758.46, 780.46, 802.19, 823.62, 844.77, 865.64, 886.22, 906.51, 926.52, 946.23, 965.66, 980.9, 974.97, 968.91, 962.74, 956.44, 950.02, 943.47, 936.8, 930.0, 923.08, 916.04, 908.2, 893.99, 879.84, 865.74, 851.71, 837.74, 823.83, 809.98, 796.19, 782.46, 768.8, 755.46, 750.68, 745.99, 741.39, 736.9, 732.49, 728.18, 723.97, 719.86, 715.84, 711.91, 708.09, 699.72, 691.13, 682.46, 673.74, 664.94, 656.08, 647.16, 638.17, 629.11, 619.98, 610.8, 608.73, 607.63, 606.7, 605.92, 605.3, 604.83, 604.53, 604.38, 604.39, 604.56, 604.89, 603.52, 601.69, 599.76, 597.75, 595.64, 593.44, 591.15, 588.77, 586.29, 583.72, 581.06, 587.68, 597.35, 606.98, 616.58, 626.15, 635.69, 645.19, 654.66, 664.1, 673.5, 682.87, 692.04, 701.02, 709.91, 718.69, 727.37, 735.94, 744.42, 752.79, 761.06, 769.23, 777.29, 779.24, 777.27, 775.44, 773.75, 1000.49, 996.45, 992.64, 984.62, 971.45, 958.55, 945.92, 933.55, 921.45, 909.61, 898.04, 886.74, 875.71, 864.95, 859.37, 861.22, 863.15, 865.17, 867.28, 869.47, 871.76, 874.02, 876.23, 878.38, 880.48, 877.17, 864.1, 851.18, 838.43, 825.84, 813.41, 801.15, 789.05, 777.11, 765.34, 753.74, 744.6, 740.94, 737.34, 733.81, 730.33, 726.92, 723.57, 720.28, 717.05, 713.88, 710.78, 708.84, 710.33, 711.9, 713.53, 715.22, 716.98, 718.81, 720.7, 722.66, 724.68, 726.78, 729.27, 733.2, 737.11, 741.0, 744.87, 748.71, 752.52, 756.31, 760.08, 763.82, 767.54, 773.42, 793.79, 813.88, 833.68, 853.19, 872.42, 891.36, 910.01, 928.38, 946.46, 964.25, 980.09, 973.47, 966.72, 959.85, 952.86, 945.74, 938.5, 931.13, 923.64, 916.03, 908.29, 900.39, 886.51, 872.7, 858.95, 845.26, 831.63, 818.06, 804.55, 791.1, 777.71, 764.38, 751.11, 746.33, 742.16, 738.1, 734.12, 730.25, 726.47, 722.78, 719.19, 715.7, 712.3, 709.0, 700.82, 691.87, 682.85, 673.77, 664.62, 655.4, 646.12, 636.77, 627.36, 617.88, 608.34, 606.15, 605.9, 605.8, 605.87, 606.09, 606.47, 607.01, 607.71, 608.56, 609.58, 610.75, 609.45, 607.13, 604.72, 602.22, 599.62, 596.94, 594.16, 591.29, 588.32, 585.27, 582.12, 587.29, 596.78, 606.25, 615.68, 625.08, 634.44, 643.78, 653.08, 662.35, 671.58, 680.78, 689.59, 698.05, 706.41, 714.67, 722.82, 730.88, 738.83, 746.67, 754.42, 762.06, 769.6, 772.64, 771.41, 770.33, 769.39, 975.96, 973.33, 970.93, 965.2, 953.7, 942.46, 931.49, 921.02, 911.86, 902.66, 893.43, 884.18, 874.89, 865.57, 859.97, 862.22, 864.42, 866.56, 868.66, 870.7, 872.69, 874.63, 876.51, 878.35, 880.13, 878.1, 866.0, 854.07, 842.3, 830.69, 819.24, 807.96, 796.84, 785.89, 775.09, 764.47, 755.54, 752.24, 749.0, 745.82, 742.7, 739.65, 736.65, 733.72, 730.85, 728.04, 725.3, 723.34, 725.21, 727.15, 729.16, 731.23, 733.36, 735.57, 737.84, 740.17, 742.57, 745.04, 747.7, 751.49, 755.25, 758.98, 762.7, 766.39, 770.05, 773.69, 777.31, 780.9, 784.46, 788.68, 807.41, 825.85, 844.01, 861.89, 879.48, 896.78, 913.8, 930.53, 946.98, 963.13, 979.0, 972.09, 964.65, 957.09, 949.4, 941.59, 933.65, 925.59, 917.41, 909.1, 900.66, 892.11, 878.98, 865.51, 852.1, 838.74, 825.45, 812.22, 799.05, 785.94, 772.89, 759.9, 746.97, 741.88, 738.25, 734.7, 731.26, 727.91, 724.65, 721.49, 718.43, 715.46, 712.59, 709.81, 701.98, 692.67, 683.3, 673.86, 664.36, 654.79, 645.15, 635.45, 625.68, 615.84, 605.94, 603.41, 604.0, 604.75, 605.66, 606.73, 607.95, 609.34, 610.88, 612.57, 614.43, 616.45, 615.47, 612.66, 609.77, 606.78, 603.69, 600.52, 597.26, 593.9, 590.45, 586.91, 583.27, 586.93, 596.25, 605.55, 614.81, 624.04, 633.23, 642.4, 651.53, 660.63, 669.69, 678.73, 687.25, 695.19, 703.02, 710.75, 718.38, 725.91, 733.33, 740.66, 747.88, 754.99, 762.01, 765.89, 765.41, 765.08, 764.89, 964.3, 964.87, 965.51, 963.33, 954.1, 944.83, 935.54, 926.21, 916.86, 907.47, 898.06, 888.61, 879.14, 869.64, 862.87, 864.8, 866.68, 868.51, 870.28, 872.0, 873.67, 875.29, 876.85, 878.36, 879.82, 878.86, 867.75, 856.79, 846.0, 835.37, 824.91, 814.61, 804.47, 794.49, 784.68, 775.04, 766.41, 763.47, 760.59, 757.77, 755.01, 752.32, 749.68, 747.11, 744.59, 742.14, 739.75, 737.78, 740.03, 742.34, 744.72, 747.17, 749.68, 752.26, 754.91, 757.62, 760.39, 763.24, 766.16, 769.79, 773.4, 776.99, 780.55, 784.09, 787.6, 791.09, 794.56, 798.0, 801.42, 804.81, 821.31, 838.12, 854.64, 870.88, 886.83, 902.5, 917.88, 932.97, 947.78, 962.3, 976.53, 970.84, 962.71, 954.45, 946.07, 937.56, 928.93, 920.17, 911.3, 902.29, 893.16, 883.91, 871.39, 858.25, 845.18, 832.17, 819.21, 806.32, 793.49, 780.72, 768.01, 755.36, 742.77, 737.34, 734.23, 731.21, 728.29, 725.47, 722.74, 720.11, 717.57, 715.13, 712.78, 710.53, 703.21, 693.55, 683.82, 674.02, 664.16, 654.24, 644.24, 634.19, 624.06, 613.87, 603.62, 600.51, 601.95, 603.55, 605.3, 607.21, 609.28, 611.5, 613.89, 616.43, 619.13, 621.99, 621.58, 618.29, 614.9, 611.43, 607.86, 604.2, 600.45, 596.6, 592.67, 588.64, 584.52, 586.6, 595.75, 604.88, 613.97, 623.03, 632.06, 641.05, 650.01, 658.94, 667.84, 676.7, 685.01, 692.42, 699.73, 706.94, 714.04, 721.04, 727.94, 734.74, 741.44, 748.03, 754.52, 759.0, 759.27, 759.68, 760.24, 968.2, 969.24, 970.36, 969.3, 959.88, 950.42, 940.94, 931.43, 921.89, 912.32, 902.71, 893.08, 883.42, 873.73, 865.82, 867.43, 868.99, 870.5, 871.95, 873.35, 874.7, 876.0, 877.24, 878.44, 879.57, 879.46, 869.33, 859.35, 849.54, 839.9, 830.41, 821.09, 811.93, 802.94, 794.11, 785.44, 777.23, 774.65, 772.13, 769.66, 767.26, 764.92, 762.65, 760.43, 758.27, 756.18, 754.15, 752.18, 754.78, 757.47, 760.23, 763.05, 765.94, 768.89, 771.91, 775.0, 778.15, 781.37, 784.66, 788.13, 791.59, 795.02, 798.44, 801.82, 805.19, 808.52, 811.84, 815.13, 818.39, 821.63, 835.49, 850.66, 865.55, 880.15, 894.47, 908.5, 922.24, 935.7, 948.87, 961.75, 974.35, 969.72, 960.88, 951.93, 942.85, 933.65, 924.33, 914.88, 905.31, 895.61, 885.79, 875.84, 863.73, 850.94, 838.2, 825.53, 812.92, 800.36, 787.87, 775.44, 763.07, 750.75, 738.5, 732.71, 730.12, 727.63, 725.24, 722.94, 720.73, 718.62, 716.61, 714.7, 712.88, 711.15, 704.5, 694.48, 684.4, 674.25, 664.03, 653.75, 643.4, 632.99, 622.51, 611.97, 601.36, 597.46, 599.74, 602.18, 604.78, 607.53, 610.44, 613.51, 616.74, 620.12, 623.67, 627.37, 627.79, 624.01, 620.13, 616.17, 612.12, 607.97, 603.73, 599.4, 594.98, 590.46, 585.85, 586.3, 595.29, 604.24, 613.16, 622.05, 630.91, 639.74, 648.53, 657.29, 666.02, 674.71, 682.87, 689.76, 696.54, 703.22, 709.8, 716.28, 722.66, 728.93, 735.1, 741.17, 747.13, 751.96, 752.98, 754.14, 755.45, 972.03, 973.54, 975.13, 975.3, 965.69, 956.05, 946.38, 936.68, 926.95, 917.19, 907.4, 897.58, 887.73, 877.85, 868.83, 870.12, 871.36, 872.54, 873.68, 874.76, 875.79, 876.76, 877.69, 878.56, 879.38, 879.9, 870.75, 861.75, 852.92, 844.26, 835.75, 827.41, 819.23, 811.22, 803.37, 795.68, 788.16, 785.76, 783.6, 781.49, 779.45, 777.47, 775.55, 773.69, 771.9, 770.16, 768.49, 766.88, 769.46, 772.53, 775.66, 778.86, 782.13, 785.46, 788.86, 792.32, 795.85, 799.44, 803.1, 806.49, 809.8, 813.08, 816.34, 819.58, 822.79, 825.98, 829.14, 832.28, 835.4, 838.49, 849.96, 863.5, 876.75, 889.71, 902.39, 914.79, 926.9, 938.72, 950.25, 961.5, 972.45, 968.71, 959.18, 949.54, 939.77, 929.87, 919.85, 909.71, 899.44, 889.05, 878.54, 867.9, 856.01, 843.56, 831.16, 818.83, 806.56, 794.34, 782.19, 770.1, 758.06, 746.09, 734.18, 727.97, 725.91, 723.95, 722.08, 720.31, 718.63, 717.05, 715.56, 714.17, 712.87, 711.68, 705.86, 695.49, 685.05, 674.54, 663.97, 653.33, 642.63, 631.86, 621.03, 610.13, 599.16, 594.25, 597.38, 600.66, 604.1, 607.7, 611.45, 615.36, 619.43, 623.66, 628.05, 632.6, 634.08, 629.82, 625.46, 621.01, 616.46, 611.83, 607.11, 602.29, 597.38, 592.38, 587.28, 586.04, 594.85, 603.64, 612.39, 621.11, 629.8, 638.46, 647.08, 655.67, 664.23, 672.75, 680.84, 687.2, 693.45, 699.61, 705.67, 711.87, 718.22, 724.54, 730.84, 737.12, 743.36, 749.09, 751.01, 752.93, 754.87, 975.78, 977.76, 979.82, 981.33, 971.53, 961.7, 951.85, 941.96, 932.04, 922.09, 912.12, 902.11, 892.07, 882.01, 871.91, 872.86, 873.78, 874.64, 875.46, 876.22, 876.92, 877.58, 878.18, 878.74, 879.23, 879.68, 872.01, 863.99, 856.14, 848.45, 840.93, 833.57, 826.37, 819.33, 812.46, 805.76, 799.22, 796.82, 795.01, 793.27, 791.58, 789.96, 788.4, 786.89, 785.46, 784.08, 782.76, 781.51, 784.09, 787.53, 791.04, 794.61, 798.25, 801.96, 805.73, 809.57, 813.48, 817.45, 821.48, 824.88, 828.03, 831.17, 834.28, 837.36, 840.43, 843.46, 846.47, 849.46, 852.43, 855.37, 864.72, 876.62, 888.23, 899.56, 910.61, 921.36, 931.84, 942.02, 951.92, 961.53, 970.85, 967.83, 957.61, 947.27, 936.8, 926.21, 915.5, 904.66, 893.7, 882.62, 871.41, 860.07, 848.24, 836.12, 824.07, 812.07, 800.14, 788.26, 776.45, 764.69, 753.0, 741.37, 729.79, 723.15, 721.61, 720.18, 718.83, 717.58, 716.43, 715.37, 714.41, 713.55, 712.78, 712.1, 707.28, 696.55, 685.76, 674.9, 663.97, 652.98, 641.92, 630.8, 619.61, 608.35, 597.03, 590.89, 594.86, 598.98, 603.27, 607.71, 612.3, 617.06, 621.97, 627.04, 632.27, 637.66, 640.47, 635.72, 630.87, 625.93, 620.91, 615.79, 610.57, 605.27, 599.87, 594.39, 588.81, 585.8, 594.45, 603.07, 611.65, 620.2, 628.72, 637.33, 645.95, 654.58, 663.24, 671.92, 680.37, 686.71, 693.03, 699.33, 705.6, 711.84, 718.06, 724.25, 730.42, 736.56, 742.68, 748.7, 750.67, 752.65, 754.64, 979.46, 981.91, 984.44, 987.04, 977.4, 967.39, 957.34, 947.27, 937.16, 927.03, 916.87, 906.67, 896.45, 886.19, 875.91, 875.65, 876.25, 876.79, 877.29, 877.73, 878.11, 878.45, 878.73, 878.96, 879.14, 879.27, 873.1, 866.07, 859.2, 852.49, 845.95, 839.56, 833.34, 827.29, 821.4, 815.67, 810.11, 807.82, 806.37, 804.98, 803.65, 802.38, 801.18, 800.04, 798.96, 797.94, 796.98, 796.08, 798.64, 802.46, 806.35, 810.3, 814.32, 818.4, 822.55, 826.76, 831.04, 835.39, 839.8, 843.29, 846.3, 849.28, 852.24, 855.17, 858.08, 860.97, 863.83, 866.67, 869.48, 872.27, 879.76, 890.02, 900.0, 909.7, 919.1, 928.23, 937.06, 945.61, 953.87, 961.85, 969.54, 967.07, 956.15, 945.12, 933.96, 922.68, 911.27, 899.74, 888.09, 876.31, 864.4, 852.38, 840.4, 828.62, 816.91, 805.25, 793.65, 782.12, 770.64, 759.23, 747.87, 736.58, 725.35, 718.23, 717.22, 716.3, 715.49, 714.76, 714.14, 713.6, 713.17, 712.83, 712.58, 712.44, 708.77, 697.69, 686.54, 675.32, 664.04, 652.69, 641.28, 629.8, 618.26, 606.65, 594.97, 587.37, 592.18, 597.15, 602.27, 607.56, 613.0, 618.59, 624.35, 630.26, 636.34, 642.57, 646.95, 641.71, 636.38, 630.95, 625.44, 619.83, 613.99, 608.03, 602.06, 596.07, 590.07, 585.39, 594.0, 602.62, 611.27, 619.93, 628.6, 637.3, 646.01, 654.74, 663.49, 672.26, 681.02, 687.24, 693.43, 699.59, 705.73, 711.84, 717.93, 723.99, 730.03, 736.04, 742.02, 747.98, 750.32, 752.36, 754.41, 983.06, 985.98, 988.98, 992.06, 983.31, 973.1, 962.87, 952.61, 942.32, 931.99, 921.64, 911.26, 900.85, 890.41, 879.94, 878.49, 878.77, 879.0, 879.17, 879.29, 879.36, 879.37, 879.34, 879.25, 879.1, 878.91, 874.04, 867.98, 862.09, 856.36, 850.8, 845.39, 840.16, 835.08, 830.17, 825.42, 820.84, 818.75, 817.66, 816.63, 815.66, 814.75, 813.9, 813.12, 812.39, 811.73, 811.13, 810.59, 813.14, 817.33, 821.6, 825.92, 830.31, 834.77, 839.3, 843.89, 848.54, 853.26, 858.05, 861.73, 864.58, 867.42, 870.23, 873.01, 875.77, 878.5, 881.22, 883.9, 886.57, 889.2, 895.09, 903.72, 912.06, 920.12, 927.89, 935.38, 942.58, 949.49, 956.12, 962.46, 968.51, 966.43, 954.82, 943.09, 931.24, 919.26, 907.16, 894.94, 882.59, 870.12, 857.52, 844.8, 832.5, 821.06, 809.69, 798.37, 787.11, 775.92, 764.78, 753.7, 742.69, 731.73, 720.84, 713.21, 712.73, 712.34, 712.05, 711.85, 711.75, 711.74, 711.83, 712.01, 712.29, 712.67, 710.33, 698.89, 687.38, 675.81, 664.17, 652.47, 640.7, 628.87, 616.97, 605.01, 592.97, 583.69, 589.34, 595.15, 601.12, 607.25, 613.53, 619.99, 626.52, 632.98, 639.38, 645.72, 651.04, 645.09, 639.13, 633.16, 627.17, 621.16, 615.14, 609.11, 603.06, 597.0, 590.93, 584.84, 593.49, 602.21, 610.94, 619.7, 628.47, 637.25, 646.06, 654.88, 663.73, 672.59, 681.46, 687.79, 693.85, 699.88, 705.89, 711.87, 717.82, 723.75, 729.66, 735.54, 741.39, 747.22, 749.96, 752.05, 754.16, 986.58, 989.98, 993.45, 996.99, 989.24, 978.85, 968.43, 957.98, 947.5, 936.99, 926.45, 915.88, 905.28, 894.66, 884.0, 881.39, 881.35, 881.25, 881.1, 880.9, 880.65, 880.35, 879.99, 879.58, 879.12, 878.6, 874.81, 869.74, 864.82, 860.07, 855.49, 851.06, 846.8, 842.71, 838.77, 835.01, 831.4, 829.63, 828.89, 828.22, 827.61, 827.06, 826.57, 826.14, 825.77, 825.47, 825.23, 825.05, 827.57, 832.14, 836.78, 841.48, 846.25, 851.08, 855.98, 860.95, 865.98, 871.07, 876.24, 880.19, 882.9, 885.58, 888.24, 890.87, 893.48, 896.07, 898.63, 901.16, 903.67, 906.16, 910.7, 917.69, 924.4, 930.82, 936.96, 942.81, 948.38, 953.66, 958.65, 963.36, 967.77, 965.91, 953.61, 941.19, 928.64, 915.97, 903.18, 890.26, 877.22, 864.06, 850.77, 837.35, 824.54, 813.44, 802.4, 791.43, 780.51, 769.65, 758.85, 748.12, 737.44, 726.82, 716.27, 708.1, 708.14, 708.28, 708.51, 708.84, 709.26, 709.78, 710.39, 711.1, 711.91, 712.81, 711.95, 700.15, 688.29, 676.36, 664.37, 652.32, 640.26, 628.53, 616.88, 605.31, 593.82, 583.61, 590.16, 596.66, 603.1, 609.48, 615.81, 622.07, 628.27, 634.41, 640.5, 646.52, 652.48, 646.63, 640.59, 634.54, 628.48, 622.4, 616.31, 610.2, 604.08, 597.95, 591.8, 585.63, 592.97, 601.77, 610.6, 619.45, 628.31, 637.19, 646.09, 655.01, 663.94, 672.89, 681.86, 688.37, 694.3, 700.2, 706.07, 711.92, 717.74, 723.54, 729.32, 735.06, 740.79, 746.48, 749.58, 751.73, 753.9, 990.03, 993.9, 997.84, 1001.86, 995.2, 984.62, 974.02, 963.38, 952.71, 942.02, 931.29, 920.53, 909.75, 898.93, 888.09, 884.34, 883.97, 883.56, 883.09, 882.57, 882.0, 881.37, 880.7, 879.97, 879.19, 878.35, 875.42, 871.33, 867.39, 863.62, 860.02, 856.57, 853.29, 850.17, 847.22, 844.43, 841.8, 840.44, 840.07, 839.75, 839.5, 839.3, 839.17, 839.1, 839.09, 839.14, 839.26, 839.44, 841.94, 846.88, 851.9, 856.97, 862.12, 867.33, 872.6, 877.94, 883.35, 888.82, 894.36, 898.68, 901.24, 903.77, 906.28, 908.76, 911.22, 913.65, 916.06, 918.45, 920.81, 923.15, 926.6, 931.96, 937.03, 941.82, 946.32, 950.54, 954.47, 958.11, 961.47, 964.54, 967.32, 965.52, 952.53, 939.41, 926.17, 912.81, 899.32, 885.71, 871.98, 858.12, 844.14, 830.03, 816.52, 805.76, 795.06, 784.42, 773.84, 763.33, 752.87, 742.47, 732.13, 721.85, 711.64, 702.89, 703.46, 704.12, 704.88, 705.73, 706.68, 707.81, 709.37, 710.76, 711.99, 713.05, 713.26, 701.48, 689.78, 678.16, 666.62, 655.16, 643.78, 632.48, 621.26, 610.13, 599.07, 588.09, 593.89, 600.07, 606.19, 612.25, 618.26, 624.2, 630.08, 635.91, 641.67, 647.38, 653.02, 648.18, 642.07, 635.95, 629.81, 623.65, 617.49, 611.31, 605.11, 598.9, 592.68, 586.44, 592.42, 601.32, 610.24, 619.18, 628.14, 637.11, 646.1, 655.11, 664.14, 673.18, 682.25, 688.97, 694.77, 700.54, 706.28, 712.0, 717.69, 723.36, 729.0, 734.61, 740.21, 745.77, 749.2, 751.41, 753.63, 993.41, 997.74, 1002.16, 1006.64, 1001.19, 990.43, 979.63, 968.81, 957.95, 947.07, 936.16, 925.21, 914.24, 903.24, 892.21, 887.34, 886.65, 885.92, 885.13, 884.29, 883.4, 882.45, 881.46, 880.41, 879.31, 878.15, 875.87, 872.75, 869.8, 867.01, 864.38, 861.91, 859.61, 857.47, 855.5, 853.69, 852.04, 851.2, 851.18, 851.22, 851.33, 851.49, 851.71, 852.0, 852.35, 852.76, 853.23, 853.77, 856.24, 861.56, 866.95, 872.4, 877.92, 883.51, 889.16, 894.87, 900.66, 906.5, 912.42, 917.2, 919.6, 921.98, 924.34, 926.67, 928.98, 931.26, 933.52, 935.76, 937.97, 940.16, 942.78, 946.5, 949.94, 953.09, 955.96, 958.54, 960.84, 962.85, 964.58, 966.01, 967.16, 965.25, 951.56, 937.75, 923.82, 909.76, 895.58, 881.28, 866.85, 852.3, 837.63, 822.83, 808.44, 798.02, 787.66, 777.36, 767.12, 756.94, 747.13, 738.23, 729.5, 720.93, 712.52, 704.7, 706.33, 707.8, 709.11, 710.26, 711.24, 712.06, 712.72, 713.21, 713.53, 713.7, 713.69, 702.76, 691.49, 680.31, 669.2, 658.17, 647.23, 636.36, 625.57, 614.87, 604.24, 593.69, 597.68, 603.54, 609.34, 615.08, 620.77, 626.39, 631.96, 637.46, 642.91, 648.29, 653.62, 649.75, 643.56, 637.36, 631.15, 624.92, 618.68, 612.43, 606.16, 599.87, 593.58, 587.27, 591.86, 600.86, 609.87, 618.9, 627.95, 637.01, 646.1, 655.2, 664.32, 673.46, 682.61, 689.6, 695.26, 700.9, 706.51, 712.1, 717.66, 723.2, 728.71, 734.19, 739.65, 745.08, 748.81, 751.07, 753.34, 996.71, 1001.51, 1006.4, 1011.35, 1007.21, 996.26, 985.28, 974.27, 963.23, 952.16, 941.06, 929.93, 918.77, 907.58, 896.36, 890.39, 889.38, 888.33, 887.22, 886.06, 884.85, 883.59, 882.27, 880.9, 879.48, 878.0, 876.16, 874.02, 872.05, 870.23, 868.58, 867.1, 865.77, 864.61, 863.61, 862.78, 862.11, 861.9, 862.23, 862.63, 863.09, 863.62, 864.2, 864.84, 865.55, 866.31, 867.14, 868.03, 870.48, 876.18, 881.94, 887.77, 893.66, 899.62, 905.65, 911.74, 917.9, 924.12, 930.41, 935.74, 937.99, 940.22, 942.43, 944.61, 946.77, 948.9, 951.01, 953.1, 955.16, 957.19, 959.25, 961.34, 963.14, 964.66, 965.89, 966.84, 967.5, 967.88, 967.97, 967.78, 967.29, 965.1, 950.72, 936.22, 921.59, 906.84, 891.97, 877.08, 862.34, 847.71, 833.18, 818.77, 804.63, 795.66, 786.85, 778.2, 769.72, 761.4, 753.24, 745.24, 737.41, 729.74, 722.23, 714.89, 715.24, 715.82, 716.22, 716.47, 716.55, 716.47, 716.23, 715.82, 715.25, 714.51, 713.61, 703.96, 693.13, 682.37, 671.7, 661.1, 650.59, 640.15, 629.8, 619.52, 609.33, 599.22, 601.53, 607.07, 612.55, 617.97, 623.34, 628.64, 633.89, 639.07, 644.2, 649.27, 654.28, 651.33, 645.07, 638.79, 632.51, 626.21, 619.89, 613.56, 607.22, 600.86, 594.49, 588.1, 591.28, 600.37, 609.47, 618.6, 627.74, 636.9, 646.07, 655.27, 664.48, 673.71, 682.96, 690.26, 695.79, 701.29, 706.77, 712.23, 717.66, 723.06, 728.44, 733.79, 739.12, 744.42, 748.4, 750.72, 753.05, 999.93, 1005.21, 1010.56, 1015.99, 1013.27, 1002.13, 990.96, 979.76, 968.53, 957.27, 945.99, 934.67, 923.32, 911.94, 900.54, 893.49, 892.17, 890.79, 889.37, 887.89, 886.35, 884.77, 883.13, 881.44, 879.7, 877.91, 876.28, 875.12, 874.13, 873.29, 872.62, 872.11, 871.77, 871.59, 871.57, 871.71, 872.02, 872.53, 873.23, 873.98, 874.8, 875.68, 876.62, 877.62, 878.68, 879.81, 881.0, 882.24, 884.66, 890.73, 896.87, 903.07, 909.34, 915.68, 922.08, 928.54, 935.08, 941.68, 948.34, 954.31, 956.41, 958.49, 960.55, 962.58, 964.58, 966.57, 968.53, 970.46, 972.37, 974.26, 976.0, 976.46, 976.62, 976.51, 976.11, 975.42, 973.42, 971.31, 969.14, 966.9, 964.61, 962.04, 947.26, 932.59, 918.02, 903.57, 889.23, 875.0, 860.87, 846.86, 832.96, 819.17, 805.49, 797.1, 789.19, 781.45, 773.86, 766.44, 759.19, 752.09, 745.16, 738.39, 731.79, 725.35, 724.32, 723.99, 723.5, 722.85, 722.03, 721.05, 719.9, 718.6, 717.12, 715.49, 713.69, 705.09, 694.68, 684.36, 674.12, 663.95, 653.87, 643.87, 633.95, 624.1, 614.34, 604.66, 605.43, 610.66, 615.82, 620.92, 625.97, 630.95, 635.88, 640.75, 645.56, 650.3, 654.99, 652.92, 646.59, 640.24, 633.88, 627.5, 621.11, 614.71, 608.29, 601.86, 595.41, 588.95, 590.68, 599.86, 609.06, 618.28, 627.51, 636.76, 646.03, 655.32, 664.63, 673.95, 683.29, 690.94, 696.33, 701.71, 707.06, 712.38, 717.68, 722.95, 728.2, 733.42, 738.62, 743.79, 747.99, 750.36, 752.75, 1003.08, 1008.83, 1014.65, 1020.55, 1019.35, 1008.02, 996.67, 985.28, 973.87, 962.42, 950.95, 939.44, 927.91, 916.34, 904.75, 896.65, 895.0, 893.31, 891.56, 889.76, 887.91, 886.01, 884.05, 882.04, 879.98, 877.87, 876.25, 876.07, 876.05, 876.19, 876.5, 876.97, 877.6, 878.4, 879.36, 880.48, 881.77, 883.11, 884.16, 885.28, 886.45, 887.69, 888.98, 890.34, 891.76, 893.24, 894.79, 896.39, 898.77, 905.22, 911.73, 918.31, 924.95, 931.67, 938.44, 945.28, 952.19, 959.16, 966.21, 972.9, 974.85, 976.78, 978.69, 980.57, 981.77, 982.24, 982.51, 982.59, 982.46, 982.13, 981.6, 979.51, 977.36, 975.15, 972.88, 970.55, 968.16, 965.71, 963.2, 960.63, 958.0, 955.31, 941.96, 927.91, 913.96, 900.13, 886.41, 872.8, 859.3, 845.9, 832.62, 819.45, 806.39, 798.38, 791.37, 784.53, 777.85, 771.33, 764.97, 758.78, 752.75, 746.88, 741.17, 735.63, 733.56, 732.33, 730.94, 729.39, 727.67, 725.79, 723.75, 721.54, 719.17, 716.63, 713.93, 706.13, 696.16, 686.27, 676.46, 666.73, 657.07, 647.5, 638.01, 628.6, 619.27, 610.02, 609.4, 614.3, 619.15, 623.93, 628.66, 633.33, 637.93, 642.48, 646.97, 651.4, 655.77, 654.53, 648.12, 641.7, 635.26, 628.81, 622.35, 615.87, 609.38, 602.87, 596.35, 589.82, 590.07, 599.34, 608.63, 617.94, 627.27, 636.61, 645.97, 655.35, 664.75, 674.17, 683.6, 691.64, 696.91, 702.15, 707.37, 712.56, 717.73, 722.87, 727.98, 733.08, 738.14, 743.18, 747.56, 749.99, 752.43, 1006.16, 1012.37, 1018.67, 1025.04, 1025.46, 1013.95, 1002.4, 990.83, 979.23, 967.6, 955.93, 944.24, 932.52, 920.77, 908.99, 899.86, 897.89, 895.88, 893.81, 891.69, 889.52, 887.3, 885.02, 882.69, 880.31, 877.88, 876.05, 876.85, 877.81, 878.93, 880.22, 881.66, 883.27, 885.05, 886.99, 889.09, 891.35, 893.63, 895.04, 896.51, 898.04, 899.63, 901.29, 903.0, 904.78, 906.62, 908.52, 910.48, 912.82, 919.64, 926.53, 933.48, 940.33, 946.76, 953.2, 959.67, 966.17, 972.69, 979.23, 985.79, 986.14, 986.25, 986.17, 985.89, 985.41, 984.74, 983.86, 982.79, 981.52, 980.05, 978.38, 976.0, 973.51, 970.96, 968.36, 965.69, 962.96, 960.18, 957.33, 954.42, 951.46, 948.43, 936.55, 923.12, 909.8, 896.58, 883.48, 870.49, 857.61, 844.84, 832.18, 819.63, 807.19, 799.49, 793.39, 787.45, 781.67, 776.05, 770.59, 765.3, 760.17, 755.2, 750.4, 745.76, 742.96, 740.83, 738.54, 736.09, 733.47, 730.69, 727.75, 724.64, 721.37, 717.94, 714.34, 707.09, 697.55, 688.1, 678.72, 669.42, 660.2, 651.06, 642.0, 633.02, 624.12, 615.3, 613.42, 618.01, 622.53, 627.0, 631.41, 635.76, 640.05, 644.28, 648.45, 652.56, 656.61, 656.15, 649.67, 643.17, 636.66, 630.14, 623.6, 617.05, 610.48, 603.9, 597.31, 590.7, 589.44, 598.8, 608.19, 617.59, 627.01, 636.44, 645.9, 655.37, 664.86, 674.37, 683.89, 692.37, 697.51, 702.62, 707.7, 712.77, 717.8, 722.81, 727.8, 732.75, 737.69, 742.6, 747.12, 749.61, 752.11, 1009.16, 1015.84, 1022.61, 1029.45, 1031.6, 1019.9, 1008.17, 996.41, 984.62, 972.8, 960.95, 949.08, 937.17, 925.23, 913.26, 903.12, 900.83, 898.5, 896.11, 893.67, 891.18, 888.64, 886.04, 883.39, 880.69, 877.94, 875.69, 877.47, 879.41, 881.51, 883.77, 886.2, 888.78, 891.53, 894.45, 897.53, 900.77, 904.08, 905.85, 907.68, 909.52, 911.14, 912.76, 914.38, 916.0, 917.61, 919.22, 920.83, 922.44, 928.82, 935.3, 941.81, 948.34, 954.9, 961.47, 968.08, 974.7, 981.35, 988.02, 994.72, 994.56, 993.53, 992.3, 990.88, 989.25, 987.43, 985.41, 983.2, 980.78, 978.16, 975.35, 972.55, 969.72, 966.84, 963.89, 960.89, 957.82, 954.7, 951.51, 948.27, 944.97, 941.6, 931.02, 918.22, 905.52, 892.93, 880.44, 868.07, 855.81, 843.66, 831.62, 819.69, 807.87, 800.45, 795.25, 790.2, 785.32, 780.61, 776.05, 771.66, 767.43, 763.36, 759.46, 755.72, 752.53, 749.5, 746.31, 742.96, 739.44, 735.76, 731.92, 727.91, 723.74, 719.41, 714.91, 707.98, 698.87, 689.84, 680.9, 672.03, 663.24, 654.53, 645.91, 637.36, 628.89, 620.51, 617.51, 621.77, 625.98, 630.13, 634.22, 638.25, 642.22, 646.13, 649.98, 653.77, 657.5, 657.78, 651.23, 644.66, 638.07, 631.47, 624.86, 618.24, 611.6, 604.94, 598.27, 591.59, 588.79, 598.25, 607.72, 617.22, 626.73, 636.26, 645.8, 655.37, 664.95, 674.55, 684.17, 693.13, 698.13, 703.11, 708.07, 713.0, 717.9, 722.78, 727.63, 732.46, 737.26, 742.04, 746.68, 749.22, 751.77, 1012.08, 1019.24, 1026.47, 1033.78, 1037.78, 1025.89, 1013.97, 1002.02, 990.05, 978.04, 966.0, 953.94, 941.84, 929.72, 917.56, 906.43, 903.83, 901.17, 898.46, 895.71, 892.89, 890.03, 887.12, 884.15, 881.13, 878.06, 875.17, 877.93, 880.96, 884.39, 887.85, 891.32, 894.82, 898.35, 901.89, 905.46, 909.05, 912.67, 914.34, 915.95, 917.55, 919.16, 920.76, 922.36, 923.96, 925.56, 927.15, 928.74, 930.33, 936.41, 943.03, 949.67, 956.33, 963.01, 969.72, 976.46, 983.21, 989.99, 996.79, 1003.62, 1003.18, 1001.01, 998.63, 996.06, 993.29, 990.33, 987.16, 983.8, 980.24, 976.48, 972.52, 969.16, 965.99, 962.77, 959.49, 956.15, 952.74, 949.28, 945.76, 942.18, 938.54, 934.83, 925.39, 913.21, 901.13, 889.16, 877.3, 865.55, 853.9, 842.37, 830.95, 819.64, 808.44, 801.24, 796.94, 792.8, 788.82, 785.0, 781.35, 777.86, 774.53, 771.36, 768.36, 765.52, 762.26, 758.33, 754.24, 749.99, 745.57, 741.0, 736.25, 731.35, 726.28, 721.04, 715.65, 708.78, 700.11, 691.51, 683.0, 674.56, 666.21, 657.93, 649.73, 641.62, 633.58, 625.63, 621.65, 625.6, 629.49, 633.32, 637.09, 640.8, 644.45, 648.04, 651.57, 655.05, 658.46, 659.43, 652.8, 646.16, 639.5, 632.83, 626.14, 619.44, 612.73, 606.0, 599.25, 592.5, 588.12, 597.67, 607.24, 616.83, 626.43, 636.05, 645.69, 655.35, 665.02, 674.72, 684.43, 693.91, 698.78, 703.63, 708.45, 713.25, 718.02, 722.77, 727.49, 732.19, 736.86, 741.51, 746.12, 748.81, 751.42, 1014.93, 1022.56, 1030.26, 1038.04, 1043.98, 1031.9, 1019.8, 1007.67, 995.5, 983.31, 971.09, 958.83, 946.55, 934.24, 921.89, 909.79, 906.92, 904.12, 901.36, 898.65, 895.98, 893.35, 890.77, 888.23, 885.73, 883.28, 880.87, 884.15, 887.69, 891.26, 894.84, 898.45, 902.08, 905.74, 909.41, 913.11, 916.84, 920.58, 922.42, 924.01, 925.6, 927.18, 928.77, 930.35, 931.93, 933.5, 935.08, 936.65, 938.22, 943.98, 950.73, 957.5, 964.29, 971.11, 977.95, 984.81, 991.7, 998.61, 1005.54, 1012.5, 1012.0, 1008.68, 1005.16, 1001.44, 997.53, 993.42, 989.11, 984.6, 979.9, 974.99, 969.89, 965.83, 962.32, 958.76, 955.14, 951.46, 947.72, 943.92, 940.06, 936.15, 932.17, 928.13, 919.66, 908.09, 896.63, 885.28, 874.04, 862.91, 851.89, 840.98, 830.17, 819.48, 808.91, 801.88, 798.48, 795.24, 792.16, 789.24, 786.49, 783.89, 781.46, 779.2, 777.1, 775.16, 772.15, 767.32, 762.34, 757.18, 751.87, 746.39, 740.75, 734.95, 728.98, 722.85, 716.55, 709.51, 701.26, 693.1, 685.02, 677.01, 669.09, 661.24, 653.48, 645.8, 638.19, 630.67, 625.85, 629.48, 633.05, 636.56, 640.01, 643.41, 646.74, 650.01, 653.23, 656.38, 659.48, 661.1, 654.39, 647.67, 640.94, 634.19, 627.43, 620.66, 613.87, 607.07, 600.25, 593.42, 587.44, 597.08, 606.74, 616.42, 626.12, 635.83, 645.56, 655.31, 665.08, 674.87, 684.67, 694.49, 699.46, 704.17, 708.86, 713.53, 718.17, 722.79, 727.38, 731.94, 736.48, 741.0, 745.49, 748.4, 751.06, 1017.71, 1025.8, 1033.98, 1042.22, 1049.38, 1036.44, 1023.63, 1010.96, 998.41, 986.0, 973.71, 961.56, 949.53, 937.64, 925.88, 914.25, 911.2, 908.66, 906.17, 903.72, 901.31, 898.95, 896.63, 894.35, 892.12, 889.93, 887.79, 890.73, 894.4, 898.1, 901.82, 905.56, 909.32, 913.1, 916.91, 920.74, 924.6, 928.47, 930.5, 932.07, 933.64, 935.21, 936.77, 938.34, 939.9, 941.45, 943.01, 944.56, 946.11, 951.53, 958.41, 965.31, 972.24, 979.18, 986.15, 993.15, 1000.17, 1007.21, 1014.27, 1021.36, 1021.02, 1016.55, 1011.88, 1007.02, 1001.96, 996.71, 991.25, 985.6, 979.75, 973.7, 967.45, 962.55, 958.71, 954.81, 950.86, 946.84, 942.76, 938.63, 934.43, 930.17, 925.86, 921.48, 913.81, 902.86, 892.02, 881.29, 870.67, 860.16, 849.76, 839.47, 829.29, 819.22, 809.26, 802.35, 799.85, 797.51, 795.33, 793.31, 791.46, 789.77, 788.24, 786.87, 785.67, 784.63, 782.21, 776.48, 770.59, 764.54, 758.33, 751.95, 745.41, 738.71, 731.84, 724.81, 717.62, 710.15, 702.34, 694.61, 686.96, 679.39, 671.89, 664.48, 657.15, 649.9, 642.72, 635.63, 630.12, 633.43, 636.68, 639.87, 643.0, 646.08, 649.09, 652.05, 654.94, 657.78, 660.55, 662.77, 655.99, 649.2, 642.39, 635.57, 628.74, 621.89, 615.03, 608.15, 601.26, 594.36, 587.44, 596.47, 606.22, 615.99, 625.78, 635.59, 645.41, 655.26, 665.12, 675.0, 684.89, 694.81, 700.16, 704.74, 709.3, 713.84, 718.35, 722.83, 727.29, 731.73, 736.13, 740.52, 744.87, 747.97, 750.69, 1015.35, 1022.59, 1029.76, 1036.86, 1043.87, 1032.96, 1020.96, 1009.1, 997.36, 985.76, 974.28, 962.94, 951.73, 940.64, 929.69, 918.88, 915.44, 913.16, 910.93, 908.75, 906.6, 904.5, 902.45, 900.43, 898.46, 896.54, 894.66, 897.29, 901.09, 904.92, 908.77, 912.64, 916.53, 920.45, 924.39, 928.35, 932.34, 936.34, 938.59, 940.14, 941.69, 943.24, 944.78, 946.33, 947.87, 949.41, 950.94, 952.48, 954.01, 959.06, 966.07, 973.1, 980.16, 987.23, 994.34, 1001.46, 1008.61, 1015.78, 1022.97, 1030.19, 1030.23, 1024.62, 1018.81, 1012.8, 1006.6, 1000.2, 993.6, 986.8, 979.81, 972.61, 965.22, 959.34, 955.16, 950.92, 946.63, 942.27, 937.86, 933.39, 928.85, 924.26, 919.61, 914.89, 907.85, 897.52, 887.3, 877.19, 867.19, 857.3, 847.52, 837.85, 828.29, 818.84, 809.5, 802.66, 801.06, 799.62, 798.34, 797.23, 796.27, 795.48, 794.85, 794.38, 794.08, 793.94, 792.42, 785.8, 779.01, 772.06, 764.95, 757.68, 750.24, 742.64, 734.87, 726.94, 718.85, 710.72, 703.34, 696.04, 688.82, 681.68, 674.62, 667.64, 660.74, 653.91, 647.17, 640.51, 634.44, 637.43, 640.36, 643.23, 646.05, 648.8, 651.5, 654.14, 656.71, 659.23, 661.69, 664.09, 657.61, 650.74, 643.86, 636.97, 630.06, 623.14, 616.2, 609.25, 602.28, 595.31, 588.31, 595.84, 605.69, 615.55, 625.43, 635.33, 645.25, 655.18, 665.14, 675.11, 685.1, 695.1, 700.88, 705.34, 709.77, 714.17, 718.55, 722.9, 727.23, 731.53, 735.81, 740.06, 744.29, 747.54, 750.31, 1011.47, 1018.25, 1024.94, 1031.56, 1038.11, 1029.35, 1018.17, 1007.11, 996.18, 985.39, 974.72, 964.19, 953.79, 943.52, 933.38, 923.37, 919.63, 917.62, 915.65, 913.73, 911.85, 910.01, 908.22, 906.47, 904.77, 903.11, 901.49, 903.82, 907.76, 911.72, 915.7, 919.7, 923.73, 927.77, 931.84, 935.94, 940.05, 944.19, 946.68, 948.21, 949.74, 951.27, 952.8, 954.32, 955.84, 957.36, 958.88, 960.39, 961.91, 966.57, 973.7, 980.87, 988.05, 995.26, 1002.49, 1009.75, 1017.03, 1024.33, 1031.66, 1039.0, 1039.64, 1032.88, 1025.93, 1018.78, 1011.43, 1003.88, 996.14, 988.2, 980.06, 971.72, 963.19, 956.18, 951.67, 947.09, 942.46, 937.77, 933.02, 928.21, 923.34, 918.4, 913.41, 908.36, 901.78, 892.07, 882.47, 872.98, 863.6, 854.33, 845.17, 836.12, 827.18, 818.35, 809.63, 802.81, 802.11, 801.57, 801.19, 800.97, 800.92, 801.03, 801.3, 801.73, 802.33, 803.09, 802.81, 795.28, 787.6, 779.75, 771.74, 763.57, 755.23, 746.73, 738.07, 729.24, 720.25, 711.21, 704.26, 697.39, 690.6, 683.89, 677.26, 670.71, 664.24, 657.85, 651.54, 645.31, 639.16, 641.49, 644.1, 646.66, 649.16, 651.59, 653.97, 656.29, 658.55, 660.75, 662.89, 664.97, 659.24, 652.3, 645.35, 638.38, 631.39, 624.4, 617.39, 610.36, 603.32, 596.27, 589.2, 595.2, 605.13, 615.09, 625.06, 635.06, 645.07, 655.09, 665.14, 675.2, 685.28, 695.38, 701.63, 705.96, 710.25, 714.53, 718.77, 723.0, 727.19, 731.37, 735.51, 739.63, 743.89, 749.1, 754.14, 1007.67, 1013.97, 1020.2, 1026.35, 1032.42, 1025.61, 1015.24, 1004.99, 994.87, 984.89, 975.03, 965.31, 955.72, 946.26, 936.93, 927.73, 923.78, 922.04, 920.33, 918.67, 917.05, 915.48, 913.95, 912.47, 911.02, 909.63, 908.27, 910.34, 914.41, 918.5, 922.61, 926.74, 930.9, 935.08, 939.28, 943.5, 947.75, 952.02, 954.77, 956.28, 957.8, 959.31, 960.81, 962.32, 963.82, 965.32, 966.82, 968.32, 969.81, 974.05, 981.32, 988.61, 995.93, 1003.27, 1010.63, 1018.02, 1025.43, 1032.86, 1040.32, 1047.8, 1049.24, 1041.34, 1033.24, 1024.95, 1016.46, 1007.77, 998.88, 989.8, 980.51, 971.03, 961.35, 953.08, 948.23, 943.32, 938.35, 933.32, 928.23, 923.09, 917.88, 912.61, 907.28, 901.89, 895.6, 886.51, 877.53, 868.66, 859.9, 851.25, 842.71, 834.28, 825.96, 817.75, 809.65, 802.8, 803.0, 803.36, 803.88, 804.56, 805.41, 806.41, 807.58, 808.92, 810.41, 812.07, 813.35, 804.93, 796.34, 787.6, 778.69, 769.62, 760.38, 750.99, 741.42, 731.7, 721.81, 711.75, 705.1, 698.66, 692.31, 686.03, 679.83, 673.71, 667.67, 661.71, 655.83, 650.03, 644.32, 645.61, 647.91, 650.14, 652.32, 654.44, 656.5, 658.5, 660.44, 662.32, 664.14, 665.91, 660.89, 653.87, 646.84, 639.8, 632.74, 625.67, 618.59, 611.49, 604.37, 597.25, 590.11, 594.53, 604.56, 614.61, 624.68, 634.76, 644.87, 654.99, 665.12, 675.28, 685.45, 695.64, 702.44, 706.94, 711.72, 716.78, 722.14, 727.78, 733.7, 739.92, 746.42, 753.21, 760.29, 767.3, 774.26, 1003.94, 1009.77, 1015.53, 1021.21, 1026.81, 1021.74, 1012.18, 1002.74, 993.43, 984.26, 975.21, 966.3, 957.52, 948.87, 940.34, 931.96, 927.89, 926.41, 924.97, 923.57, 922.22, 920.91, 919.64, 918.42, 917.24, 916.11, 915.01, 916.83, 921.03, 925.25, 929.49, 933.76, 938.05, 942.36, 946.69, 951.04, 955.42, 959.82, 962.86, 964.36, 965.85, 967.34, 968.83, 970.32, 971.8, 973.29, 974.76, 976.24, 977.72, 981.51, 988.91, 996.33, 1003.78, 1011.25, 1018.75, 1026.26, 1033.8, 1041.37, 1048.95, 1056.56, 1059.05, 1050.0, 1040.76, 1031.32, 1021.68, 1011.85, 1001.82, 991.59, 981.16, 970.54, 959.72, 950.04, 944.86, 939.61, 934.3, 928.94, 923.51, 918.02, 912.48, 906.87, 901.21, 895.48, 889.31, 880.85, 872.49, 864.24, 856.09, 848.06, 840.14, 832.33, 824.63, 817.03, 809.55, 802.63, 803.73, 804.99, 806.41, 807.99, 809.73, 811.64, 813.71, 815.94, 818.34, 820.89, 823.62, 814.74, 805.25, 795.61, 785.8, 775.83, 765.7, 755.41, 744.95, 734.32, 723.53, 712.58, 705.86, 699.86, 693.93, 688.08, 682.31, 676.63, 671.02, 665.49, 660.04, 654.68, 649.39, 649.79, 651.77, 653.69, 655.55, 657.35, 659.09, 660.77, 662.39, 663.96, 665.46, 666.9, 662.55, 655.46, 648.35, 641.24, 634.1, 626.96, 619.8, 612.63, 605.44, 598.24, 591.02, 593.85, 604.03, 614.2, 624.38, 634.56, 644.75, 654.93, 665.13, 675.32, 685.52, 695.72, 703.8, 709.77, 716.03, 722.57, 729.4, 736.51, 743.91, 751.6, 759.58, 767.85, 776.4, 785.13, 794.02, 1000.29, 1005.65, 1010.93, 1016.14, 1021.28, 1017.74, 1008.99, 1000.36, 991.86, 983.5, 975.26, 967.16, 959.19, 951.34, 943.63, 936.05, 931.96, 930.74, 929.56, 928.42, 927.33, 926.29, 925.29, 924.33, 923.41, 922.54, 921.71, 923.3, 927.63, 931.99, 936.36, 940.76, 945.17, 949.62, 954.08, 958.57, 963.07, 967.61, 970.96, 972.44, 973.91, 975.39, 976.86, 978.32, 979.79, 981.25, 982.71, 984.17, 985.63, 988.95, 996.48, 1004.03, 1011.61, 1019.21, 1026.84, 1034.49, 1042.16, 1049.85, 1057.57, 1065.31, 1069.05, 1058.86, 1048.47, 1037.89, 1027.11, 1016.13, 1004.96, 993.59, 982.02, 970.25, 958.28, 947.06, 941.54, 935.95, 930.31, 924.61, 918.84, 913.02, 907.14, 901.2, 895.2, 889.13, 882.92, 875.07, 867.33, 859.7, 852.18, 844.76, 837.46, 830.27, 823.19, 816.21, 809.35, 802.6, 804.29, 806.45, 808.77, 811.25, 813.9, 816.7, 819.67, 822.8, 826.1, 829.55, 833.17, 824.71, 814.33, 803.79, 793.08, 782.21, 771.18, 759.99, 748.63, 737.11, 725.43, 713.58, 706.54, 700.97, 695.47, 690.06, 684.72, 679.46, 674.29, 669.19, 664.17, 659.24, 654.38, 654.03, 655.69, 657.29, 658.83, 660.31, 661.74, 663.1, 664.41, 665.65, 666.84, 667.96, 664.22, 656.6, 648.87, 641.29, 633.84, 626.54, 619.37, 612.35, 605.47, 598.73, 592.14, 593.75, 603.94, 614.13, 624.32, 634.52, 644.73, 654.93, 665.14, 675.36, 685.57, 695.79, 704.88, 712.32, 720.05, 728.07, 736.37, 744.96, 753.84, 763.0, 772.45, 782.19, 792.22, 802.58, 813.39, 996.71, 1001.6, 1006.42, 1011.15, 1015.82, 1013.61, 1005.67, 997.85, 990.16, 982.61, 975.18, 967.89, 960.72, 953.69, 946.79, 940.02, 935.98, 935.02, 934.11, 933.24, 932.41, 931.63, 930.89, 930.19, 929.54, 928.93, 928.37, 929.75, 934.21, 938.7, 943.2, 947.73, 952.28, 956.85, 961.45, 966.06, 970.7, 975.37, 979.06, 980.52, 981.98, 983.43, 984.88, 986.33, 987.78, 989.22, 990.66, 992.1, 993.54, 996.36, 1004.03, 1011.71, 1019.42, 1027.15, 1034.91, 1042.69, 1050.49, 1058.31, 1066.16, 1074.03, 1079.24, 1067.91, 1056.38, 1044.65, 1032.73, 1020.61, 1008.29, 995.78, 983.06, 970.15, 957.04, 944.14, 938.28, 932.36, 926.38, 920.34, 914.24, 908.08, 901.86, 895.58, 889.24, 882.84, 876.38, 869.18, 862.06, 855.05, 848.15, 841.35, 834.67, 828.1, 821.63, 815.28, 809.04, 802.9, 804.7, 807.76, 810.98, 814.36, 817.9, 821.6, 825.47, 829.5, 833.69, 838.05, 842.57, 834.84, 823.57, 812.13, 800.52, 788.76, 776.83, 764.74, 752.48, 740.06, 727.48, 714.73, 707.15, 702.0, 696.94, 691.95, 687.05, 682.22, 677.47, 672.81, 668.22, 663.72, 659.29, 658.33, 659.15, 659.3, 659.48, 659.67, 659.9, 660.14, 660.41, 660.7, 661.02, 661.36, 657.98, 650.86, 643.88, 637.05, 630.35, 623.79, 617.38, 611.1, 604.97, 598.98, 593.14, 593.64, 603.84, 614.05, 624.27, 634.48, 644.71, 654.93, 665.16, 675.39, 685.62, 695.86, 705.67, 714.59, 723.8, 733.29, 743.06, 753.13, 763.48, 774.11, 785.04, 796.25, 807.76, 819.66, 832.39, 993.21, 997.63, 1001.97, 1006.24, 1010.43, 1009.35, 1002.22, 995.21, 988.33, 981.59, 974.97, 968.48, 962.13, 955.9, 949.81, 943.85, 939.96, 939.26, 938.61, 938.01, 937.44, 936.92, 936.45, 936.01, 935.62, 935.28, 934.98, 936.18, 940.77, 945.39, 950.02, 954.68, 959.36, 964.07, 968.79, 973.54, 978.31, 983.11, 987.17, 988.61, 990.05, 991.48, 992.91, 994.34, 995.77, 997.19, 998.62, 1000.04, 1001.45, 1003.76, 1011.55, 1019.37, 1027.21, 1035.07, 1042.96, 1050.86, 1058.8, 1066.75, 1074.73, 1082.73, 1089.64, 1077.16, 1064.49, 1051.62, 1038.55, 1025.29, 1011.83, 998.17, 984.31, 970.26, 956.0, 941.55, 935.08, 928.82, 922.51, 916.13, 909.69, 903.19, 896.64, 890.02, 883.35, 876.61, 869.82, 863.19, 856.68, 850.29, 844.01, 837.83, 831.77, 825.82, 819.97, 814.24, 808.61, 803.1, 804.94, 808.9, 813.02, 817.3, 821.74, 826.34, 831.11, 836.04, 841.13, 846.38, 851.8, 845.14, 832.96, 820.63, 808.13, 795.46, 782.64, 769.65, 756.5, 743.18, 729.7, 716.06, 707.67, 702.84, 697.76, 692.55, 687.23, 681.8, 676.26, 670.6, 664.82, 658.93, 652.92, 650.0, 650.25, 650.54, 650.84, 651.17, 651.52, 651.9, 652.3, 652.72, 653.17, 653.64, 651.35, 644.98, 638.75, 632.66, 626.71, 620.91, 615.24, 609.71, 604.33, 599.09, 593.99, 593.52, 603.75, 613.98, 624.21, 634.44, 644.68, 654.92, 665.17, 675.42, 685.67, 695.92, 706.18, 716.57, 727.25, 738.22, 749.47, 761.0, 772.83, 784.94, 797.34, 810.03, 823.01, 836.36, 851.01, 989.78, 993.73, 997.6, 1001.4, 1005.12, 1004.96, 998.63, 992.44, 986.37, 980.43, 974.63, 968.95, 963.4, 957.99, 952.71, 947.55, 943.89, 943.46, 943.08, 942.73, 942.43, 942.17, 941.96, 941.79, 941.67, 941.58, 941.55, 942.59, 947.31, 952.06, 956.82, 961.61, 966.43, 971.26, 976.12, 981.0, 985.9, 990.82, 995.28, 996.7, 998.12, 999.53, 1000.94, 1002.35, 1003.76, 1005.17, 1006.57, 1007.97, 1009.37, 1011.13, 1019.06, 1027.0, 1034.97, 1042.96, 1050.98, 1059.02, 1067.08, 1075.17, 1083.28, 1091.41, 1099.57, 1086.6, 1072.79, 1058.78, 1044.57, 1030.16, 1015.56, 1000.76, 985.76, 970.56, 955.16, 939.57, 931.94, 925.35, 918.69, 911.98, 905.2, 898.37, 891.48, 884.52, 877.51, 870.44, 863.31, 857.08, 851.2, 845.42, 839.76, 834.21, 828.76, 823.43, 818.2, 813.08, 808.08, 803.18, 805.03, 809.88, 814.9, 820.08, 825.42, 830.92, 836.58, 842.41, 848.4, 854.55, 860.87, 855.6, 842.02, 827.62, 813.31, 799.07, 784.92, 770.84, 756.85, 742.94, 729.12, 715.37, 705.84, 700.24, 694.54, 688.72, 682.78, 676.74, 670.57, 664.3, 657.9, 651.39, 644.77, 640.95, 641.34, 641.75, 642.19, 642.65, 643.13, 643.64, 644.17, 644.72, 645.3, 645.9, 644.59, 638.96, 633.48, 628.14, 622.94, 617.88, 612.96, 608.18, 603.55, 599.06, 594.7, 593.41, 603.65, 613.89, 624.14, 634.4, 644.65, 654.91, 665.17, 675.44, 685.71, 695.98, 706.4, 718.27, 730.42, 742.86, 755.59, 768.6, 781.9, 795.48, 809.35, 823.52, 837.97, 852.7, 869.26, 986.43, 989.91, 993.31, 996.64, 999.89, 1000.44, 994.92, 989.54, 984.28, 979.15, 974.15, 969.28, 964.55, 959.94, 955.47, 951.12, 947.79, 947.62, 947.5, 947.41, 947.38, 947.38, 947.43, 947.53, 947.66, 947.85, 948.07, 948.98, 953.83, 958.7, 963.6, 968.52, 973.47, 978.43, 983.42, 988.43, 993.46, 998.52, 1003.39, 1004.79, 1006.19, 1007.59, 1008.98, 1010.37, 1011.76, 1013.15, 1014.53, 1015.91, 1017.29, 1018.67, 1026.54, 1034.61, 1042.71, 1050.84, 1058.98, 1067.15, 1075.35, 1083.56, 1091.8, 1100.07, 1108.35, 1096.25, 1081.29, 1066.13, 1050.78, 1035.23, 1019.49, 1003.54, 987.4, 971.06, 954.52, 937.78, 928.86, 921.93, 914.94, 907.88, 900.77, 893.6, 886.37, 879.09, 871.74, 864.33, 856.86, 850.86, 845.6, 840.45, 835.4, 830.47, 825.64, 820.92, 816.32, 811.82, 807.43, 803.16, 804.95, 810.45, 815.73, 820.95, 826.12, 831.22, 836.26, 841.23, 846.15, 851.0, 855.8, 850.86, 836.83, 822.89, 809.02, 795.23, 781.53, 767.91, 754.37, 740.91, 727.54, 714.24, 703.97, 697.76, 691.44, 685.0, 678.45, 671.78, 665.0, 658.11, 651.1, 643.97, 636.73, 631.87, 632.39, 632.94, 633.5, 634.09, 634.71, 635.35, 636.01, 636.69, 637.4, 638.14, 637.67, 632.8, 628.07, 623.47, 619.02, 614.71, 610.54, 606.51, 602.62, 598.88, 595.28, 593.29, 603.55, 613.81, 624.08, 634.35, 644.62, 654.9, 665.18, 675.46, 685.75, 696.04, 706.34, 719.68, 733.31, 747.22, 761.42, 775.9, 790.67, 805.73, 821.08, 836.71, 852.64, 868.85, 887.13, 983.16, 986.16, 989.09, 991.95, 994.73, 995.79, 991.08, 986.51, 982.06, 977.74, 973.55, 969.49, 965.56, 961.76, 958.1, 954.56, 951.64, 951.73, 951.87, 952.05, 952.28, 952.55, 952.86, 953.22, 953.62, 954.06, 954.55, 955.34, 960.32, 965.33, 970.36, 975.41, 980.48, 985.58, 990.7, 995.84, 1001.0, 1006.19, 1011.4, 1012.89, 1014.27, 1015.64, 1017.02, 1018.39, 1019.76, 1021.13, 1022.5, 1023.86, 1025.22, 1026.58, 1033.99, 1042.2, 1050.43, 1058.69, 1066.96, 1075.26, 1083.59, 1091.93, 1100.3, 1108.7, 1117.11, 1106.09, 1089.99, 1073.69, 1057.19, 1040.5, 1023.61, 1006.53, 989.24, 971.76, 954.08, 936.2, 925.84, 918.57, 911.24, 903.85, 896.4, 888.9, 881.33, 873.71, 866.02, 858.27, 850.47, 844.54, 840.26, 836.01, 831.71, 827.38, 823.02, 818.61, 814.17, 809.69, 805.18, 800.63, 801.07, 806.07, 811.02, 815.91, 820.73, 825.49, 830.2, 834.84, 839.41, 843.93, 848.39, 845.14, 831.56, 818.07, 804.65, 791.32, 778.06, 764.89, 751.8, 738.8, 725.87, 713.03, 702.21, 695.39, 688.45, 681.39, 674.22, 666.94, 659.55, 652.04, 644.41, 636.67, 628.81, 622.77, 623.42, 624.1, 624.8, 625.52, 626.26, 627.03, 627.83, 628.64, 629.48, 630.35, 630.62, 626.5, 622.51, 618.67, 614.96, 611.4, 607.98, 604.7, 601.56, 598.56, 595.7, 593.16, 603.44, 613.72, 624.0, 634.29, 644.58, 654.88, 665.18, 675.48, 685.78, 696.09, 706.4, 720.8, 735.91, 751.29, 766.96, 782.92, 799.17, 815.7, 832.52, 849.63, 867.02, 884.71, 904.63, 979.95, 982.49, 984.95, 987.33, 989.64, 991.01, 987.11, 983.34, 979.7, 976.19, 972.81, 969.56, 966.44, 963.45, 960.6, 957.87, 955.44, 955.8, 956.21, 956.65, 957.14, 957.67, 958.25, 958.87, 959.53, 960.24, 960.99, 961.79, 966.8, 971.93, 977.09, 982.28, 987.48, 992.71, 997.96, 1003.23, 1008.52, 1013.84, 1019.18, 1020.98, 1022.35, 1023.7, 1025.06, 1026.41, 1027.77, 1029.11, 1030.46, 1031.81, 1033.15, 1034.49, 1041.43, 1049.77, 1058.13, 1066.51, 1074.92, 1083.35, 1091.81, 1100.28, 1108.78, 1117.31, 1125.86, 1116.12, 1098.88, 1081.44, 1063.8, 1045.97, 1027.94, 1009.71, 991.28, 972.66, 953.84, 934.82, 923.07, 916.01, 908.88, 901.69, 894.44, 887.13, 879.75, 872.32, 864.82, 857.26, 849.63, 843.69, 839.26, 834.8, 830.3, 825.76, 821.19, 816.58, 811.93, 807.25, 802.53, 797.77, 797.09, 801.76, 806.37, 810.92, 815.41, 819.83, 824.2, 828.5, 832.74, 836.92, 841.04, 839.34, 826.21, 813.17, 800.2, 787.32, 774.52, 761.8, 749.16, 736.6, 724.12, 711.73, 700.57, 693.13, 685.57, 677.9, 670.12, 662.22, 654.2, 646.08, 637.83, 629.48, 621.0, 613.65, 614.43, 615.24, 616.07, 616.92, 617.8, 618.7, 619.62, 620.57, 621.54, 622.53, 623.43, 620.05, 616.82, 613.72, 610.76, 607.95, 605.27, 602.74, 600.35, 598.1, 595.99, 594.02, 603.33, 613.63, 623.93, 634.24, 644.55, 654.86, 665.17, 675.49, 685.82, 696.14, 706.47, 721.65, 738.22, 755.08, 772.23, 789.66, 807.38, 825.38, 843.67, 862.25, 881.12, 900.28, 921.75, 976.83, 978.89, 980.88, 982.8, 984.64, 986.1, 983.01, 980.05, 977.22, 974.52, 971.95, 969.5, 967.19, 965.01, 962.96, 961.05, 959.26, 959.83, 960.5, 961.21, 961.96, 962.75, 963.59, 964.47, 965.4, 966.37, 967.38, 968.44, 973.25, 978.52, 983.81, 989.12, 994.46, 999.81, 1005.19, 1010.6, 1016.02, 1021.47, 1026.94, 1029.09, 1030.43, 1031.77, 1033.1, 1034.44, 1035.77, 1037.1, 1038.43, 1039.76, 1041.08, 1042.4, 1048.84, 1057.31, 1065.81, 1074.32, 1082.86, 1091.42, 1100.0, 1108.61, 1117.24, 1125.9, 1134.56, 1124.77, 1105.21, 1085.86, 1066.72, 1047.79, 1029.06, 1010.54, 992.22, 974.12, 956.22, 938.54, 926.44, 919.02, 911.54, 904.0, 896.39, 888.72, 880.99, 873.2, 865.35, 857.43, 849.45, 842.93, 838.29, 833.63, 828.92, 824.18, 819.4, 814.59, 809.73, 804.84, 799.92, 794.96, 793.18, 797.51, 801.78, 805.99, 810.14, 814.23, 818.26, 822.22, 826.13, 829.97, 833.75, 833.46, 820.78, 808.18, 795.67, 783.24, 770.88, 758.61, 746.43, 734.32, 722.3, 710.35, 699.04, 690.98, 682.81, 674.52, 666.12, 657.61, 648.98, 640.23, 631.37, 622.4, 613.31, 604.5, 605.42, 606.35, 607.31, 608.3, 609.3, 610.33, 611.39, 612.46, 613.57, 614.69, 615.84, 613.47, 610.98, 608.63, 606.42, 604.35, 602.43, 600.64, 599.0, 597.49, 596.13, 594.91, 603.22, 613.53, 623.85, 634.18, 644.5, 654.83, 665.17, 675.5, 685.84, 696.19, 706.54, 722.2, 740.25, 758.58, 777.2, 796.11, 815.3, 834.77, 854.54, 874.59, 894.93, 915.56, 938.49, 973.77, 975.37, 976.89, 978.33, 979.7, 981.0, 978.78, 976.63, 974.61, 972.71, 970.95, 969.32, 967.81, 966.44, 965.2, 964.09, 963.11, 963.82, 964.74, 965.72, 966.73, 967.79, 968.89, 970.04, 971.23, 972.46, 973.74, 975.06, 979.68, 985.08, 990.5, 995.94, 1001.41, 1006.9, 1012.41, 1017.94, 1023.5, 1029.08, 1034.68, 1037.19, 1038.51, 1039.83, 1041.15, 1042.47, 1043.78, 1045.09, 1046.4, 1047.71, 1049.01, 1049.15, 1053.65, 1060.75, 1067.83, 1074.88, 1081.9, 1088.9, 1095.87, 1102.81, 1109.72, 1116.61, 1123.48, 1116.56, 1098.2, 1080.06, 1062.11, 1044.38, 1026.85, 1009.53, 992.42, 975.52, 958.83, 942.34, 929.88, 922.1, 914.26, 906.36, 898.4, 890.38, 882.29, 874.15, 865.94, 857.66, 849.33, 842.2, 837.36, 832.49, 827.58, 822.63, 817.65, 812.63, 807.57, 802.47, 797.34, 792.18, 789.33, 793.32, 797.26, 801.13, 804.94, 808.69, 812.38, 816.01, 819.58, 823.08, 826.52, 827.49, 815.27, 803.12, 791.06, 779.07, 767.17, 755.35, 743.61, 731.96, 720.38, 708.89, 697.62, 688.95, 680.16, 671.26, 662.24, 653.11, 643.86, 634.5, 625.03, 615.44, 605.73, 595.91, 596.38, 597.44, 598.53, 599.65, 600.78, 601.95, 603.13, 604.34, 605.57, 606.83, 608.11, 606.74, 605.0, 603.4, 601.94, 600.62, 599.44, 598.4, 597.5, 596.75, 596.13, 595.66, 603.1, 613.44, 623.77, 634.11, 644.46, 654.81, 665.16, 675.51, 685.87, 696.23, 706.6, 722.48, 742.0, 761.8, 781.89, 802.27, 822.93, 843.88, 865.12, 886.64, 908.46, 930.56, 954.86, 970.8, 971.92, 972.97, 973.95, 974.85, 975.67, 974.42, 973.08, 971.86, 970.78, 969.82, 969.0, 968.3, 967.74, 967.31, 967.01, 966.84, 967.76, 968.95, 970.18, 971.46, 972.78, 974.15, 975.56, 977.01, 978.5, 980.04, 981.63, 986.08, 991.61, 997.17, 1002.74, 1008.34, 1013.96, 1019.6, 1025.27, 1030.95, 1035.04, 1038.6, 1039.44, 1038.9, 1038.49, 1038.21, 1038.06, 1038.04, 1038.15, 1038.4, 1038.78, 1039.29, 1039.93, 1044.1, 1051.05, 1057.98, 1064.88, 1071.75, 1078.59, 1085.41, 1092.21, 1098.97, 1105.71, 1112.42, 1108.14, 1090.98, 1074.04, 1057.3, 1040.77, 1024.44, 1008.32, 992.41, 976.71, 961.22, 945.94, 933.37, 925.24, 917.05, 908.79, 900.48, 892.1, 883.66, 875.15, 866.59, 857.96, 849.27, 841.51, 836.47, 831.39, 826.27, 821.12, 815.93, 810.7, 805.44, 800.14, 794.81, 789.43, 785.53, 789.19, 792.79, 796.32, 799.8, 803.21, 806.56, 809.85, 813.08, 816.25, 819.36, 821.45, 809.67, 797.97, 786.36, 774.83, 763.38, 752.01, 740.72, 729.51, 718.39, 707.35, 696.39, 687.03, 677.62, 668.1, 658.47, 648.72, 638.86, 628.89, 618.79, 608.59, 598.27, 587.83, 587.31, 588.51, 589.73, 590.97, 592.24, 593.53, 594.85, 596.19, 597.55, 598.94, 600.35, 599.88, 598.88, 598.03, 597.31, 596.74, 596.31, 596.02, 595.87, 595.86, 595.99, 596.27, 602.98, 613.33, 623.69, 634.05, 644.41, 654.77, 665.14, 675.51, 685.89, 696.27, 706.65, 722.47, 743.46, 764.74, 786.3, 808.15, 830.28, 852.7, 875.41, 898.41, 921.69, 945.26, 970.86, 967.9, 968.55, 969.13, 969.63, 970.06, 970.42, 969.93, 969.4, 968.99, 968.71, 968.57, 968.55, 968.66, 968.91, 969.28, 969.79, 970.43, 971.66, 973.11, 974.61, 976.15, 977.73, 979.36, 981.03, 982.75, 983.41, 983.88, 984.2, 986.87, 990.78, 994.58, 998.27, 1001.84, 1005.3, 1008.65, 1011.88, 1014.99, 1018.0, 1020.89, 1022.21, 1022.45, 1022.83, 1023.34, 1023.98, 1024.75, 1025.65, 1026.69, 1027.85, 1029.15, 1030.58, 1034.58, 1041.38, 1048.15, 1054.9, 1061.62, 1068.32, 1074.99, 1081.63, 1088.24, 1094.83, 1101.39, 1099.51, 1083.56, 1067.82, 1052.28, 1036.95, 1021.82, 1006.91, 992.2, 977.7, 963.41, 949.32, 936.93, 928.44, 919.89, 911.28, 902.61, 893.88, 885.08, 876.22, 867.3, 858.32, 849.27, 840.86, 835.61, 830.33, 825.0, 819.65, 814.25, 808.82, 803.35, 797.85, 792.3, 786.72, 781.8, 785.12, 788.38, 791.58, 794.72, 797.79, 800.81, 803.76, 806.65, 809.48, 812.25, 814.96, 803.99, 792.75, 781.58, 770.5, 759.5, 748.58, 737.74, 726.99, 716.31, 705.72, 695.21, 685.22, 675.2, 665.07, 654.82, 644.45, 633.97, 623.38, 612.68, 601.85, 590.92, 579.87, 578.22, 579.55, 580.9, 582.28, 583.68, 585.1, 586.55, 588.01, 589.51, 591.03, 592.57, 592.87, 592.62, 592.51, 592.55, 592.72, 593.04, 593.49, 594.09, 594.83, 595.71, 596.73, 602.86, 613.23, 623.6, 633.98, 644.36, 654.74, 665.13, 675.52, 685.91, 696.31, 706.71, 722.17, 744.63, 767.38, 790.42, 813.74, 837.35, 861.24, 885.42, 909.89, 934.64, 959.69, 986.48, 965.07, 965.25, 965.36, 965.4, 965.36, 965.24, 965.32, 965.59, 965.99, 966.52, 967.18, 967.97, 968.83, 969.65, 970.54, 971.51, 972.55, 973.49, 974.23, 974.84, 975.3, 975.61, 975.78, 975.8, 975.68, 975.41, 975.0, 974.44, 976.03, 979.27, 982.39, 985.4, 988.29, 991.08, 993.75, 996.3, 998.74, 1001.07, 1003.28, 1004.85, 1005.88, 1007.05, 1008.34, 1009.77, 1011.33, 1013.02, 1014.84, 1016.79, 1018.88, 1021.1, 1025.08, 1031.73, 1038.35, 1044.95, 1051.52, 1058.07, 1064.59, 1071.08, 1077.54, 1083.98, 1090.39, 1090.68, 1075.93, 1061.39, 1047.05, 1032.92, 1019.0, 1005.28, 991.77, 978.47, 965.38, 952.5, 940.54, 931.7, 922.8, 913.83, 904.81, 895.72, 886.57, 877.36, 868.08, 858.74, 849.34, 840.24, 834.79, 829.3, 823.77, 818.21, 812.61, 806.97, 801.3, 795.59, 789.84, 784.05, 778.23, 781.11, 784.04, 786.9, 789.7, 792.44, 795.11, 797.73, 800.28, 802.78, 805.21, 807.58, 798.23, 787.44, 776.72, 766.09, 755.54, 745.07, 734.68, 724.38, 714.15, 704.01, 693.95, 683.53, 672.89, 662.14, 651.28, 640.3, 629.2, 617.99, 606.67, 595.23, 583.68, 572.02, 569.11, 570.57, 572.05, 573.56, 575.09, 576.64, 578.22, 579.82, 581.44, 583.09, 584.76, 585.72, 586.22, 586.86, 587.64, 588.56, 589.62, 590.83, 592.17, 593.66, 595.29, 597.06, 602.73, 613.12, 623.51, 633.9, 644.3, 654.7, 665.1, 675.51, 685.92, 696.34, 706.76, 721.59, 745.53, 769.75, 794.25, 819.05, 844.12, 869.49, 895.14, 921.08, 947.3, 973.82, 1001.72, 962.14, 961.73, 961.31, 960.9, 960.48, 960.06, 960.44, 961.27, 962.18, 963.16, 964.21, 965.34, 966.54, 967.81, 969.15, 970.56, 972.05, 972.7, 972.56, 972.28, 971.86, 971.29, 970.57, 969.71, 968.71, 967.56, 966.26, 964.82, 965.3, 967.86, 970.31, 972.64, 974.86, 976.97, 978.96, 980.84, 982.61, 984.26, 985.79, 987.35, 989.17, 991.13, 993.21, 995.43, 997.77, 1000.25, 1002.86, 1005.6, 1008.47, 1011.48, 1015.61, 1022.11, 1028.58, 1035.03, 1041.45, 1047.84, 1054.21, 1060.55, 1066.86, 1073.15, 1079.41, 1081.64, 1068.09, 1054.75, 1041.61, 1028.68, 1015.96, 1003.45, 991.14, 979.04, 967.15, 955.47, 944.22, 935.03, 925.77, 916.45, 907.07, 897.62, 888.12, 878.55, 868.92, 859.23, 849.47, 839.66, 834.0, 828.31, 822.58, 816.81, 811.0, 805.16, 799.28, 793.36, 787.41, 781.42, 775.39, 777.16, 779.75, 782.27, 784.74, 787.14, 789.48, 791.76, 793.98, 796.13, 798.23, 800.26, 792.39, 782.05, 771.78, 761.6, 751.5, 741.48, 731.54, 721.68, 711.91, 702.22, 692.61, 681.95, 670.7, 659.33, 647.85, 636.25, 624.54, 612.72, 600.78, 588.73, 576.56, 564.28, 559.97, 561.56, 563.17, 564.81, 566.47, 568.15, 569.86, 571.59, 573.35, 575.13, 576.93, 578.43, 579.68, 581.07, 582.59, 584.26, 586.07, 588.02, 590.11, 592.34, 594.72, 597.23, 602.6, 613.01, 623.42, 633.83, 644.24, 654.66, 665.08, 675.51, 685.93, 696.37, 706.8, 720.73, 746.13, 771.83, 797.81, 824.07, 850.62, 877.45, 904.57, 931.98, 959.68, 987.67, 1016.59, 956.9, 956.49, 956.07, 955.66, 955.24, 954.82, 955.36, 956.65, 958.01, 959.44, 960.94, 962.52, 964.17, 965.89, 967.68, 969.55, 971.49, 972.05, 971.03, 969.87, 968.56, 967.11, 965.51, 963.77, 961.89, 959.85, 957.68, 955.35, 954.68, 956.56, 958.34, 959.99, 961.54, 962.97, 964.29, 965.49, 966.58, 967.56, 968.42, 969.73, 972.34, 975.08, 977.95, 980.95, 984.09, 987.35, 990.75, 994.28, 997.94, 1001.73, 1006.16, 1012.51, 1018.84, 1025.13, 1031.4, 1037.64, 1043.86, 1050.05, 1056.21, 1062.35, 1068.46, 1072.39, 1060.04, 1047.9, 1035.97, 1024.24, 1012.72, 1001.4, 990.3, 979.4, 968.71, 958.23, 947.97, 938.41, 928.8, 919.13, 909.39, 899.59, 889.73, 879.81, 869.82, 859.77, 849.66, 839.49, 833.26, 827.35, 821.42, 815.44, 809.43, 803.38, 797.3, 791.18, 785.02, 778.82, 772.59, 773.28, 775.52, 777.71, 779.84, 781.9, 783.91, 785.85, 787.73, 789.55, 791.31, 793.0, 786.47, 776.57, 766.76, 757.03, 747.38, 737.81, 728.32, 718.91, 709.59, 700.34, 691.18, 680.48, 668.61, 656.63, 644.53, 632.32, 620.0, 607.56, 595.01, 582.34, 569.55, 556.66, 550.81, 552.53, 554.27, 556.04, 557.83, 559.64, 561.48, 563.34, 565.23, 567.14, 569.07, 571.0, 573.0, 575.13, 577.4, 579.82, 582.37, 585.07, 587.91, 590.89, 594.01, 597.27, 602.47, 612.89, 623.32, 633.75, 644.18, 654.61, 665.05, 675.5, 685.94, 696.39, 706.85, 719.58, 746.46, 773.62, 801.07, 828.81, 856.82, 885.13, 913.72, 942.6, 971.77, 1001.23, 1031.09, 951.66, 951.25, 950.83, 950.42, 950.0, 949.58, 950.22, 951.96, 953.77, 955.65, 957.6, 959.63, 961.73, 963.9, 966.14, 968.46, 970.85, 971.55, 969.65, 967.6, 965.41, 963.08, 960.6, 957.98, 955.21, 952.29, 949.23, 946.03, 944.17, 945.38, 946.48, 947.46, 948.33, 949.08, 949.73, 950.25, 950.67, 950.97, 951.15, 951.97, 955.37, 958.9, 962.55, 966.34, 970.27, 974.32, 978.5, 982.82, 987.26, 991.84, 996.74, 1002.94, 1009.11, 1015.26, 1021.38, 1027.47, 1033.54, 1039.58, 1045.59, 1051.57, 1057.53, 1062.94, 1051.79, 1040.85, 1030.12, 1019.59, 1009.27, 999.15, 989.25, 979.55, 970.06, 960.78, 951.71, 941.86, 931.9, 921.87, 911.77, 901.62, 891.4, 881.13, 870.79, 860.38, 849.92, 839.39, 832.54, 826.44, 820.29, 814.11, 807.9, 801.64, 795.35, 789.03, 782.66, 776.26, 769.83, 769.45, 771.36, 773.21, 775.0, 776.73, 778.39, 780.0, 781.54, 783.02, 784.45, 785.81, 780.46, 771.02, 761.65, 752.37, 743.17, 734.05, 725.01, 716.05, 707.18, 698.39, 689.68, 679.13, 666.65, 654.05, 641.33, 628.51, 615.57, 602.51, 589.34, 576.06, 562.66, 549.15, 541.62, 543.47, 545.35, 547.24, 549.17, 551.11, 553.08, 555.07, 557.09, 559.13, 561.19, 563.43, 566.17, 569.05, 572.07, 575.24, 578.54, 581.98, 585.56, 589.29, 593.16, 597.17, 602.33, 612.77, 623.22, 633.66, 644.11, 654.57, 665.02, 675.48, 685.95, 696.41, 706.89, 718.15, 746.5, 775.13, 804.05, 833.26, 862.75, 892.53, 923.77, 955.35, 987.26, 1019.5, 1052.07, 946.42, 946.01, 945.59, 945.17, 944.76, 944.34, 945.0, 947.19, 949.45, 951.78, 954.19, 956.66, 959.21, 961.84, 964.53, 967.3, 970.14, 971.19, 968.4, 965.48, 962.41, 959.19, 955.83, 952.32, 948.67, 944.88, 940.94, 936.85, 933.78, 934.31, 934.73, 935.04, 935.23, 935.31, 935.28, 935.13, 934.87, 934.49, 934.0, 934.08, 938.26, 942.58, 947.03, 951.61, 956.31, 961.15, 966.12, 971.23, 976.46, 981.83, 987.35, 993.4, 999.42, 1005.41, 1011.38, 1017.32, 1023.24, 1029.13, 1034.99, 1040.83, 1046.63, 1052.42, 1043.33, 1033.59, 1024.06, 1014.73, 1005.61, 996.7, 987.99, 979.49, 971.2, 963.12, 955.25, 945.37, 935.05, 924.67, 914.22, 903.71, 893.14, 882.51, 871.81, 861.06, 850.24, 839.36, 831.87, 825.55, 819.21, 812.82, 806.4, 799.94, 793.45, 786.91, 780.35, 773.74, 767.1, 765.68, 767.25, 768.77, 770.22, 771.61, 772.94, 774.21, 775.42, 776.56, 777.65, 778.67, 774.38, 765.38, 756.47, 747.64, 738.88, 730.21, 721.62, 713.12, 704.69, 696.35, 688.09, 677.89, 664.79, 651.58, 638.25, 624.81, 611.25, 597.58, 583.8, 569.9, 555.88, 541.75, 532.41, 534.39, 536.4, 538.43, 540.48, 542.55, 544.65, 546.78, 548.92, 551.09, 553.29, 555.72, 559.21, 562.84, 566.6, 570.51, 574.56, 578.75, 583.08, 587.55, 592.16, 596.92, 602.19, 612.65, 623.11, 633.58, 644.04, 654.51, 664.99, 675.76, 687.02, 698.46, 710.1, 721.93, 751.93, 783.21, 814.83, 846.77, 879.05, 911.66, 944.59, 977.86, 1011.46, 1045.4, 1079.67, 941.18, 940.76, 940.35, 939.93, 939.52, 939.1, 939.71, 942.35, 945.06, 947.85, 950.7, 953.63, 956.63, 959.7, 962.84, 966.06, 969.35, 970.97, 967.31, 963.5, 959.54, 955.45, 951.2, 946.82, 942.28, 937.61, 932.78, 927.81, 923.49, 923.35, 923.09, 922.73, 922.24, 921.65, 920.94, 920.12, 919.18, 918.13, 916.97, 916.06, 921.03, 926.13, 931.37, 936.73, 942.23, 947.85, 953.61, 959.5, 965.52, 971.68, 977.96, 983.88, 989.75, 995.59, 1001.41, 1007.2, 1012.97, 1018.71, 1024.42, 1030.1, 1035.76, 1041.39, 1034.67, 1026.13, 1017.79, 1009.67, 1001.75, 994.03, 986.52, 979.23, 972.13, 965.25, 958.58, 948.95, 938.27, 927.53, 916.73, 905.87, 894.94, 883.95, 872.9, 861.79, 850.62, 839.38, 831.23, 824.71, 818.16, 811.56, 804.94, 798.27, 791.57, 784.84, 778.06, 771.25, 764.4, 761.97, 763.21, 764.39, 765.5, 766.56, 767.55, 768.48, 769.35, 770.16, 770.91, 771.59, 768.21, 759.66, 751.2, 742.82, 734.51, 726.29, 718.15, 710.1, 702.12, 694.23, 686.41, 676.76, 663.05, 649.22, 635.27, 621.22, 607.05, 592.76, 578.36, 563.85, 549.22, 534.47, 523.18, 525.29, 527.42, 529.58, 531.77, 533.97, 536.2, 538.45, 540.73, 543.03, 545.36, 547.87, 552.1, 556.48, 560.99, 565.65, 570.44, 575.38, 580.46, 585.83, 591.22, 596.65, 602.1, 612.88, 623.99, 635.28, 646.76, 658.44, 670.3, 682.35, 694.6, 707.03, 719.66, 732.48, 762.25, 795.24, 828.55, 862.18, 896.15, 930.45, 965.08, 1000.05, 1035.34, 1070.97, 1106.93, 935.94, 935.52, 935.11, 934.69, 934.28, 933.86, 934.35, 937.44, 940.6, 943.84, 947.14, 950.52, 953.97, 957.49, 961.09, 964.75, 968.49, 970.9, 966.35, 961.66, 956.82, 951.85, 946.72, 941.45, 936.04, 930.48, 924.78, 918.92, 913.32, 912.5, 911.57, 910.53, 909.37, 908.1, 906.72, 905.22, 903.61, 901.88, 900.04, 898.09, 903.67, 909.56, 915.58, 921.73, 928.01, 934.42, 940.97, 947.65, 954.45, 961.39, 968.47, 974.38, 980.1, 985.8, 991.47, 997.11, 1002.72, 1008.31, 1013.87, 1019.4, 1024.91, 1030.39, 1025.8, 1018.46, 1011.32, 1004.39, 997.67, 991.16, 984.85, 978.75, 972.86, 967.18, 961.7, 952.58, 941.55, 930.45, 919.3, 908.08, 896.8, 885.46, 874.06, 862.59, 851.06, 839.47, 830.62, 823.9, 817.14, 810.35, 803.51, 796.64, 789.74, 782.8, 775.82, 768.8, 761.75, 758.33, 759.23, 760.07, 760.84, 761.56, 762.22, 762.81, 763.35, 763.82, 764.23, 764.58, 761.96, 753.86, 745.85, 737.92, 730.06, 722.29, 714.6, 706.99, 699.47, 692.02, 684.66, 675.75, 661.42, 646.97, 632.41, 617.74, 602.96, 588.06, 573.04, 557.91, 542.67, 527.31, 513.92, 516.16, 518.43, 520.71, 523.03, 525.36, 527.72, 530.11, 533.02, 536.09, 539.13, 542.14, 547.45, 552.8, 558.18, 563.58, 569.01, 574.47, 579.96, 585.48, 591.02, 596.6, 602.2, 613.36, 625.45, 637.73, 650.21, 662.87, 675.72, 688.76, 701.99, 715.41, 729.03, 742.83, 772.26, 806.93, 841.93, 877.27, 912.93, 948.92, 985.25, 1021.9, 1058.89, 1096.21, 1133.86, 930.7, 930.28, 929.87, 929.45, 929.03, 928.62, 928.92, 932.46, 936.07, 939.75, 943.51, 947.34, 951.24, 955.21, 959.25, 963.37, 967.56, 970.97, 965.54, 959.97, 954.25, 948.39, 942.38, 936.23, 929.94, 923.5, 916.91, 910.18, 903.3, 901.77, 900.16, 898.44, 896.61, 894.66, 892.6, 890.43, 888.15, 885.74, 883.23, 880.6, 886.17, 892.85, 899.65, 906.59, 913.66, 920.86, 928.19, 935.66, 943.25, 950.98, 958.83, 964.92, 970.49, 976.03, 981.55, 987.04, 992.5, 997.94, 1003.35, 1008.73, 1014.09, 1019.42, 1016.72, 1010.58, 1004.64, 998.91, 993.39, 988.08, 982.97, 978.07, 973.38, 968.89, 964.62, 956.28, 944.89, 933.44, 921.93, 910.36, 898.72, 887.03, 875.27, 863.45, 851.57, 839.62, 830.05, 823.13, 816.16, 809.16, 802.13, 795.05, 787.94, 780.79, 773.61, 766.39, 759.13, 754.74, 755.3, 755.81, 756.25, 756.63, 756.95, 757.21, 757.4, 757.54, 757.61, 757.63, 755.63, 747.98, 740.42, 732.93, 725.53, 718.21, 710.97, 703.81, 696.73, 689.74, 682.82, 674.85, 659.9, 644.84, 629.67, 614.38, 598.98, 583.46, 567.83, 552.44, 537.35, 522.34, 507.85, 511.01, 514.13, 517.22, 520.28, 523.31, 526.3, 529.27, 532.2, 535.11, 537.98, 540.82, 546.04, 551.53, 557.06, 562.61, 568.19, 573.8, 579.44, 585.1, 590.8, 596.52, 602.27, 613.65, 626.73, 640.0, 653.46, 667.11, 680.95, 694.97, 709.19, 723.6, 738.2, 752.99, 781.93, 818.3, 855.0, 892.02, 929.38, 967.06, 1005.08, 1043.43, 1082.11, 1121.12, 1160.46, 925.46, 925.04, 924.63, 924.21, 923.79, 923.38, 923.41, 927.4, 931.46, 935.6, 939.8, 944.08, 948.43, 952.85, 957.35, 961.92, 966.56, 971.18, 964.87, 958.42, 951.82, 945.08, 938.19, 931.16, 923.98, 916.66, 909.2, 901.58, 893.82, 891.14, 888.86, 886.47, 883.96, 881.34, 878.6, 875.76, 872.8, 869.72, 866.53, 863.23, 868.54, 876.0, 883.6, 891.32, 899.18, 907.17, 915.28, 923.53, 931.91, 940.43, 949.07, 955.47, 960.89, 966.29, 971.65, 976.99, 982.31, 987.59, 992.85, 998.09, 1003.29, 1008.47, 1007.43, 1002.49, 997.76, 993.23, 988.9, 984.79, 980.88, 977.18, 973.68, 970.4, 967.32, 960.04, 948.29, 936.49, 924.63, 912.7, 900.71, 888.66, 876.55, 864.37, 852.13, 839.83, 829.52, 822.39, 815.22, 808.02, 800.77, 793.49, 786.18, 778.83, 771.44, 764.01, 756.55, 751.21, 751.44, 751.61, 751.71, 751.76, 751.74, 751.66, 751.52, 751.32, 751.06, 750.73, 749.22, 742.02, 734.9, 727.87, 720.92, 714.04, 707.25, 700.54, 693.94, 687.34, 680.69, 673.54, 658.26, 643.05, 627.92, 612.86, 597.88, 582.97, 568.14, 553.39, 538.71, 524.11, 509.59, 511.4, 514.36, 517.28, 520.17, 523.03, 525.86, 528.66, 531.42, 534.16, 536.86, 539.53, 544.59, 550.24, 555.91, 561.61, 567.34, 573.1, 578.89, 584.7, 590.54, 596.41, 602.31, 613.75, 627.82, 642.08, 656.52, 671.16, 685.98, 701.0, 716.2, 731.6, 747.19, 762.96, 791.28, 829.34, 867.73, 906.45, 945.5, 984.87, 1024.58, 1064.62, 1104.99, 1145.7, 1186.73, 920.21, 919.8, 919.38, 918.97, 918.55, 918.14, 917.84, 922.28, 926.79, 931.37, 936.03, 940.75, 945.55, 950.43, 955.37, 960.39, 965.48, 970.64, 964.35, 957.01, 949.53, 941.91, 934.14, 926.23, 918.17, 909.97, 901.62, 893.13, 884.49, 880.63, 877.67, 874.6, 871.42, 868.13, 864.72, 861.19, 857.56, 853.81, 849.94, 845.97, 850.78, 859.03, 867.41, 875.92, 884.56, 893.34, 902.24, 911.28, 920.44, 929.74, 939.17, 946.06, 951.33, 956.57, 961.78, 966.97, 972.14, 977.27, 982.38, 987.47, 992.52, 997.55, 997.94, 994.2, 990.66, 987.33, 984.21, 981.29, 978.58, 976.08, 973.78, 971.7, 969.82, 963.86, 951.76, 939.6, 927.38, 915.1, 902.76, 890.35, 877.89, 865.36, 852.76, 840.11, 829.03, 821.69, 814.32, 806.91, 799.46, 791.97, 784.45, 776.9, 769.3, 761.67, 754.0, 747.75, 747.64, 747.47, 747.24, 746.94, 746.59, 746.17, 745.7, 745.42, 745.36, 745.37, 744.95, 738.5, 731.99, 725.42, 718.8, 712.11, 705.37, 698.58, 691.72, 684.8, 677.83, 670.8, 656.25, 641.45, 626.73, 612.09, 597.52, 583.02, 568.61, 554.27, 540.0, 525.81, 511.7, 511.83, 514.61, 517.37, 520.09, 522.78, 525.44, 528.07, 530.67, 533.24, 535.77, 538.27, 543.12, 548.92, 554.74, 560.59, 566.47, 572.37, 578.31, 584.27, 590.26, 596.28, 602.33, 613.67, 628.72, 643.97, 659.4, 675.02, 690.83, 706.84, 723.03, 739.41, 755.98, 772.75, 800.3, 840.05, 880.14, 920.55, 961.29, 1002.36, 1043.76, 1085.49, 1127.55, 1169.95, 1212.68, 914.97, 914.56, 914.14, 913.73, 913.31, 912.89, 912.48, 917.08, 922.04, 927.07, 932.18, 937.35, 942.6, 947.93, 953.32, 958.79, 964.33, 969.94, 963.97, 955.75, 947.39, 938.89, 930.24, 921.45, 912.51, 903.43, 894.2, 884.82, 875.3, 870.23, 866.6, 862.85, 859.0, 855.03, 850.94, 846.75, 842.43, 838.01, 833.47, 828.82, 832.89, 841.92, 851.09, 860.39, 869.82, 879.38, 889.07, 898.89, 908.84, 918.93, 929.14, 936.67, 941.78, 946.88, 951.94, 956.98, 961.99, 966.98, 971.94, 976.87, 981.78, 986.66, 988.24, 985.7, 983.36, 981.23, 979.31, 977.59, 976.08, 974.77, 973.67, 972.79, 972.11, 967.74, 955.29, 942.77, 930.2, 917.57, 904.87, 892.11, 879.29, 866.4, 853.46, 840.45, 828.57, 821.03, 813.45, 805.83, 798.18, 790.49, 782.76, 775.0, 767.01, 759.03, 751.08, 743.94, 743.7, 743.54, 743.44, 743.42, 743.46, 743.58, 743.76, 744.02, 744.35, 744.75, 745.22, 738.55, 731.72, 724.84, 717.9, 710.9, 703.84, 696.73, 689.55, 682.32, 675.03, 667.68, 654.16, 639.78, 625.47, 611.24, 597.08, 583.0, 569.0, 555.07, 541.21, 527.44, 513.74, 512.29, 514.9, 517.49, 520.05, 522.57, 525.06, 527.52, 529.95, 532.35, 534.71, 537.05, 541.62, 547.56, 553.53, 559.53, 565.56, 571.62, 577.7, 583.81, 589.95, 596.12, 602.31, 613.39, 629.43, 645.67, 662.09, 678.69, 695.49, 712.48, 729.66, 747.03, 764.59, 782.34, 809.0, 850.44, 892.21, 934.32, 976.75, 1019.51, 1062.6, 1106.03, 1149.78, 1193.87, 1238.29, 909.73, 909.32, 908.9, 908.49, 908.07, 907.65, 907.24, 911.8, 917.21, 922.7, 928.25, 933.88, 939.58, 945.35, 951.2, 957.12, 963.11, 969.17, 963.73, 954.63, 945.39, 936.01, 926.48, 916.81, 906.99, 897.02, 886.91, 876.66, 866.26, 859.94, 855.63, 851.21, 846.68, 842.04, 837.28, 832.41, 827.42, 822.32, 817.11, 811.78, 814.86, 824.69, 834.64, 844.72, 854.94, 865.28, 875.76, 886.37, 897.11, 907.98, 918.98, 927.3, 932.27, 937.21, 942.13, 947.01, 951.88, 956.71, 961.52, 966.3, 971.06, 975.79, 978.34, 977.0, 975.86, 974.92, 974.2, 973.68, 973.36, 973.26, 973.36, 973.67, 974.19, 971.68, 958.88, 946.01, 933.08, 920.09, 907.04, 893.93, 880.55, 866.61, 852.73, 838.93, 825.91, 817.83, 809.8, 801.8, 793.84, 785.92, 778.04, 770.2, 762.4, 754.63, 746.91, 739.32, 739.48, 739.71, 740.0, 740.37, 740.81, 741.32, 741.9, 742.55, 743.27, 744.06, 744.92, 738.66, 731.52, 724.32, 717.06, 709.74, 702.37, 694.94, 687.45, 679.9, 672.29, 664.63, 652.0, 638.03, 624.13, 610.31, 596.57, 582.9, 569.31, 555.79, 542.35, 528.99, 515.7, 512.78, 515.23, 517.64, 520.03, 522.39, 524.71, 527.0, 529.26, 531.49, 533.69, 535.86, 540.09, 546.18, 552.3, 558.45, 564.63, 570.83, 577.06, 583.32, 589.61, 595.93, 602.27, 612.93, 629.96, 647.18, 664.58, 682.18, 699.96, 717.94, 736.1, 754.46, 773.0, 791.74, 817.36, 860.5, 903.97, 947.76, 991.88, 1036.34, 1081.12, 1126.23, 1171.68, 1217.46, 1263.57, 904.49, 904.08, 903.66, 903.25, 902.83, 902.41, 902.0, 906.46, 912.32, 918.25, 924.26, 930.34, 936.49, 942.71, 949.0, 955.37, 961.81, 968.32, 963.64, 953.66, 943.54, 933.27, 922.86, 912.31, 901.61, 890.77, 879.78, 868.64, 857.36, 849.76, 844.78, 839.68, 834.48, 829.16, 823.73, 818.18, 812.52, 806.75, 800.86, 794.86, 796.71, 807.32, 818.06, 828.93, 839.93, 851.06, 862.32, 873.71, 885.24, 896.89, 908.68, 917.96, 922.78, 927.57, 932.33, 937.07, 941.78, 946.47, 951.13, 955.76, 960.37, 964.94, 968.23, 968.09, 968.14, 968.41, 968.88, 969.56, 970.44, 970.79, 971.0, 971.24, 971.53, 969.67, 955.63, 941.67, 927.78, 913.97, 900.22, 886.54, 872.94, 859.41, 845.95, 832.56, 819.45, 811.6, 803.79, 796.01, 788.28, 780.59, 772.93, 765.31, 757.74, 750.2, 742.7, 735.25, 735.19, 735.81, 736.5, 737.26, 738.09, 738.99, 739.96, 741.0, 742.11, 743.29, 744.55, 738.83, 731.37, 723.85, 716.28, 708.64, 700.95, 693.2, 685.4, 677.53, 669.61, 661.63, 649.76, 636.2, 622.72, 609.31, 595.98, 582.72, 569.54, 556.44, 543.41, 530.46, 517.58, 513.3, 515.58, 517.83, 520.05, 522.23, 524.39, 526.51, 528.61, 530.67, 532.7, 534.69, 538.54, 544.78, 551.04, 557.34, 563.66, 570.02, 576.4, 582.81, 589.24, 595.71, 602.2, 612.28, 630.3, 648.5, 666.89, 685.47, 704.25, 723.21, 742.36, 761.7, 781.23, 800.95, 825.4, 870.23, 915.39, 960.88, 1006.69, 1052.83, 1099.31, 1146.11, 1193.25, 1240.72, 1288.52, 899.25, 898.83, 898.42, 898.0, 897.59, 897.17, 896.75, 901.04, 907.35, 913.74, 920.19, 926.72, 933.32, 939.99, 946.73, 953.55, 960.44, 967.4, 963.69, 952.83, 941.83, 930.68, 919.39, 907.96, 896.38, 884.65, 872.78, 860.77, 848.61, 839.69, 834.04, 828.27, 822.39, 816.4, 810.29, 804.07, 797.74, 791.29, 784.73, 778.05, 778.42, 789.81, 801.34, 813.0, 824.78, 836.7, 848.75, 860.93, 873.24, 885.68, 898.25, 908.64, 913.31, 917.95, 922.57, 927.16, 931.72, 936.15, 940.42, 944.61, 948.73, 952.77, 956.07, 956.24, 956.45, 956.69, 956.98, 957.3, 957.66, 958.05, 958.49, 958.96, 959.47, 959.09, 945.47, 931.92, 918.45, 905.04, 891.71, 878.45, 865.25, 852.14, 839.09, 826.11, 813.21, 805.33, 797.74, 790.19, 782.68, 775.21, 767.78, 760.39, 753.04, 745.73, 738.46, 731.22, 730.82, 731.84, 732.92, 734.07, 735.29, 736.59, 737.95, 739.38, 740.89, 742.46, 744.11, 739.06, 731.28, 723.45, 715.55, 707.6, 699.6, 691.53, 683.41, 675.22, 666.98, 658.68, 647.45, 634.3, 621.23, 608.23, 595.31, 582.47, 569.7, 557.01, 544.39, 531.85, 519.39, 513.85, 515.96, 518.04, 520.09, 522.11, 524.1, 526.06, 527.98, 529.87, 531.74, 533.57, 536.95, 543.34, 549.76, 556.2, 562.67, 569.17, 575.7, 582.26, 588.85, 595.46, 602.1, 611.44, 630.44, 649.63, 669.01, 688.58, 708.34, 728.28, 748.42, 768.75, 789.27, 809.97, 833.12, 879.64, 926.49, 973.66, 1021.17, 1069.0, 1117.17, 1165.66, 1214.49, 1263.65, 1313.14, 894.01, 893.59, 893.18, 892.76, 892.35, 891.93, 891.51, 895.55, 902.31, 909.15, 916.05, 923.03, 930.08, 937.2, 944.39, 951.66, 959.0, 966.41, 963.88, 952.14, 940.26, 928.24, 916.07, 903.75, 891.29, 878.69, 865.94, 853.04, 840.0, 829.74, 823.41, 816.96, 810.41, 803.74, 796.96, 790.07, 783.06, 775.94, 768.7, 761.35, 760.0, 772.18, 784.49, 796.93, 809.51, 822.39, 835.58, 848.71, 861.79, 874.82, 887.8, 898.99, 903.27, 907.48, 911.62, 915.68, 919.67, 923.58, 927.41, 931.18, 934.86, 938.47, 941.72, 942.12, 942.55, 943.02, 943.53, 944.08, 944.66, 945.28, 945.94, 946.64, 947.37, 948.14, 935.24, 922.1, 909.04, 896.05, 883.12, 870.27, 857.5, 844.79, 832.16, 819.59, 807.1, 799.01, 791.65, 784.33, 777.04, 769.8, 762.59, 755.43, 748.3, 741.22, 734.17, 727.16, 726.39, 727.79, 729.27, 730.81, 732.43, 734.11, 735.87, 737.69, 739.59, 741.56, 743.6, 739.34, 731.25, 723.1, 714.89, 706.62, 698.29, 689.91, 681.47, 672.97, 664.42, 655.8, 645.06, 632.32, 619.66, 607.08, 594.57, 582.14, 569.78, 557.5, 545.3, 533.17, 521.12, 514.43, 516.38, 518.29, 520.17, 522.02, 523.84, 525.63, 527.39, 529.11, 530.81, 532.47, 535.34, 541.88, 548.44, 555.03, 561.66, 568.31, 574.98, 581.69, 588.42, 595.18, 601.97, 610.41, 630.4, 650.58, 670.94, 691.5, 712.24, 733.17, 754.3, 775.61, 797.11, 818.81, 840.69, 888.72, 937.26, 986.12, 1035.32, 1084.85, 1134.7, 1184.89, 1235.4, 1286.25, 1337.43, 888.77, 888.35, 887.94, 887.52, 887.11, 886.69, 886.27, 889.99, 897.2, 904.48, 911.84, 919.26, 926.76, 934.33, 941.98, 949.69, 957.48, 965.34, 964.22, 951.6, 938.84, 925.93, 912.88, 899.69, 886.35, 872.86, 859.23, 845.46, 831.54, 819.89, 812.89, 805.77, 798.54, 791.35, 784.35, 777.43, 770.58, 763.8, 757.09, 750.45, 748.39, 761.53, 774.62, 787.66, 800.64, 813.58, 826.46, 839.29, 852.08, 864.8, 877.48, 889.08, 892.93, 896.71, 900.41, 904.04, 907.6, 911.08, 914.48, 917.81, 921.07, 924.25, 927.34, 927.96, 928.62, 929.31, 930.05, 930.82, 931.62, 932.47, 933.35, 934.27, 935.23, 936.23, 924.93, 912.21, 899.56, 886.98, 874.47, 862.03, 849.67, 837.37, 825.15, 813.0, 800.92, 792.66, 785.52, 778.42, 771.37, 764.35, 757.37, 750.43, 743.53, 736.66, 729.84, 723.06, 721.89, 723.68, 725.55, 727.49, 729.49, 731.57, 733.72, 735.94, 738.23, 740.58, 743.02, 739.68, 731.27, 722.8, 714.28, 705.69, 697.05, 688.35, 679.6, 670.78, 661.91, 652.98, 642.59, 630.27, 618.02, 605.85, 593.76, 581.74, 569.79, 557.92, 546.13, 534.41, 522.78, 515.05, 516.82, 518.57, 520.28, 521.97, 523.62, 525.24, 526.83, 528.38, 529.91, 531.4, 533.7, 540.38, 547.1, 553.84, 560.61, 567.41, 574.23, 581.09, 587.97, 594.88, 601.82, 609.2, 630.17, 651.34, 672.69, 694.23, 715.96, 737.87, 759.98, 782.28, 804.77, 827.45, 850.32, 897.47, 947.7, 998.26, 1049.14, 1100.36, 1151.9, 1203.78, 1255.99, 1308.52, 1361.39, 883.53, 883.11, 882.7, 882.28, 881.87, 881.45, 881.03, 884.36, 892.02, 899.75, 907.55, 915.43, 923.38, 931.4, 939.49, 947.65, 955.89, 964.2, 964.7, 951.2, 937.56, 923.01, 908.19, 893.49, 878.92, 864.47, 850.16, 835.97, 821.91, 809.76, 802.92, 796.15, 789.44, 782.81, 776.25, 769.76, 763.34, 756.99, 750.71, 744.5, 741.13, 753.97, 766.76, 779.49, 792.18, 804.81, 817.4, 829.93, 842.41, 854.84, 867.21, 879.24, 882.66, 886.01, 889.28, 892.48, 895.6, 898.65, 901.63, 904.53, 907.35, 910.1, 912.78, 913.76, 914.65, 915.57, 916.52, 917.52, 918.55, 919.62, 920.73, 921.87, 923.05, 924.28, 914.55, 902.24, 890.0, 877.84, 865.74, 853.72, 841.76, 829.88, 818.07, 806.33, 794.67, 786.27, 779.35, 772.48, 765.65, 758.85, 752.1, 745.38, 738.71, 732.07, 725.47, 718.92, 717.31, 719.5, 721.76, 724.09, 726.49, 728.96, 731.5, 734.11, 736.79, 739.54, 742.36, 740.08, 731.35, 722.57, 713.73, 704.83, 695.87, 686.85, 677.78, 668.65, 659.46, 650.21, 640.05, 628.14, 616.31, 604.55, 592.86, 581.25, 569.72, 558.27, 546.88, 535.58, 524.35, 515.69, 517.3, 518.88, 520.42, 521.94, 523.42, 524.87, 526.3, 527.68, 529.04, 530.37, 532.03, 538.86, 545.73, 552.62, 559.54, 566.48, 573.46, 580.46, 587.49, 594.55, 601.63, 608.75, 629.76, 651.91, 674.24, 696.77, 719.48, 742.38, 765.48, 788.76, 812.24, 835.9, 859.75, 905.9, 957.82, 1010.06, 1062.64, 1115.54, 1168.78, 1222.34, 1276.24, 1330.47, 1385.03, 878.28, 877.87, 877.46, 877.04, 876.62, 876.21, 875.79, 878.65, 886.66, 893.85, 900.97, 908.03, 915.04, 921.98, 928.87, 935.69, 942.46, 949.17, 949.88, 935.46, 921.17, 907.01, 892.97, 879.06, 865.28, 851.62, 838.1, 824.7, 811.43, 799.43, 793.02, 786.68, 780.41, 774.21, 768.07, 762.02, 756.03, 750.11, 744.26, 738.49, 733.92, 746.46, 758.95, 771.38, 783.77, 796.1, 808.38, 820.62, 832.79, 844.92, 857.0, 869.02, 872.46, 875.38, 878.22, 880.99, 883.68, 886.3, 888.84, 891.31, 893.71, 896.03, 898.27, 899.53, 900.64, 901.78, 902.96, 904.18, 905.44, 906.73, 908.06, 909.43, 910.84, 912.28, 904.1, 892.2, 880.38, 868.62, 856.94, 845.33, 833.79, 822.32, 810.92, 799.6, 788.34, 779.83, 773.15, 766.5, 759.89, 753.32, 746.79, 740.3, 733.85, 727.44, 721.07, 714.73, 712.67, 715.25, 717.9, 720.62, 723.41, 726.27, 729.2, 732.21, 735.28, 738.42, 741.64, 740.54, 731.49, 722.39, 713.23, 704.02, 694.74, 685.41, 676.02, 666.57, 657.06, 647.5, 637.44, 625.94, 614.51, 603.17, 591.89, 580.7, 569.58, 558.53, 547.56, 536.67, 525.85, 516.37, 517.81, 519.22, 520.6, 521.94, 523.26, 524.54, 525.8, 527.02, 528.21, 529.36, 530.49, 537.32, 544.33, 551.37, 558.43, 565.53, 572.65, 579.8, 586.98, 594.19, 601.42, 608.69, 629.15, 652.29, 675.61, 699.12, 722.82, 746.71, 770.79, 795.05, 819.51, 844.16, 869.0, 914.0, 967.61, 1021.54, 1075.81, 1130.4, 1185.33, 1240.58, 1296.17, 1352.08, 1408.33, 869.13, 867.89, 866.64, 865.4, 864.15, 862.9, 861.65, 862.91, 869.79, 876.62, 883.38, 890.09, 896.74, 903.33, 909.86, 916.32, 922.73, 929.08, 931.38, 917.75, 904.25, 890.87, 877.62, 864.5, 851.51, 838.64, 825.9, 813.29, 800.81, 789.03, 783.05, 777.14, 771.3, 765.53, 759.83, 754.2, 748.64, 743.16, 737.74, 732.4, 727.13, 739.0, 751.19, 763.32, 775.41, 787.44, 799.42, 811.35, 823.23, 835.06, 846.83, 858.55, 862.34, 864.82, 867.23, 869.57, 871.83, 874.02, 876.14, 878.17, 880.14, 882.03, 883.84, 885.25, 886.59, 887.95, 889.36, 890.8, 892.28, 893.8, 895.36, 896.95, 898.58, 900.25, 893.57, 882.09, 870.68, 859.34, 848.07, 836.87, 825.74, 814.68, 803.7, 792.79, 781.95, 773.36, 766.9, 760.48, 754.09, 747.75, 741.44, 735.18, 728.95, 722.76, 716.62, 710.51, 707.95, 710.92, 713.97, 717.08, 720.26, 723.52, 726.84, 730.24, 733.7, 737.24, 740.84, 741.05, 731.69, 722.27, 712.8, 703.26, 693.67, 684.02, 674.32, 664.55, 654.73, 644.85, 634.75, 623.66, 612.65, 601.71, 590.85, 580.06, 569.35, 558.72, 548.16, 537.68, 527.28, 517.08, 518.35, 519.59, 520.8, 521.98, 523.13, 524.24, 525.33, 526.38, 527.4, 528.39, 529.35, 535.74, 542.9, 550.09, 557.3, 564.55, 571.82, 579.12, 586.44, 593.8, 601.18, 608.59, 628.36, 652.48, 676.79, 701.28, 725.97, 750.84, 775.9, 801.16, 826.6, 852.23, 878.06, 921.77, 977.07, 1032.69, 1088.65, 1144.93, 1201.55, 1258.49, 1315.76, 1373.37, 1431.3, 853.41, 852.16, 850.92, 849.67, 848.43, 847.18, 845.93, 846.45, 852.98, 859.45, 865.86, 872.21, 878.5, 884.73, 890.9, 897.01, 903.07, 909.06, 912.75, 899.91, 887.19, 874.61, 862.15, 849.81, 837.61, 825.53, 813.58, 801.76, 790.07, 778.55, 773.0, 767.53, 762.12, 756.78, 751.51, 746.31, 741.19, 736.13, 731.15, 726.24, 721.4, 731.59, 743.48, 755.31, 767.1, 778.83, 790.51, 802.14, 813.72, 825.24, 836.72, 848.14, 852.28, 854.34, 856.32, 858.23, 860.06, 861.82, 863.5, 865.11, 866.64, 868.1, 869.48, 870.94, 872.5, 874.09, 875.72, 877.39, 879.09, 880.84, 882.62, 884.43, 886.29, 888.18, 882.98, 871.91, 860.91, 849.98, 839.12, 828.33, 817.62, 806.98, 796.4, 785.9, 775.47, 766.85, 760.61, 754.41, 748.26, 742.14, 736.06, 730.01, 724.01, 718.05, 712.13, 706.25, 703.17, 706.53, 709.97, 713.47, 717.05, 720.69, 724.41, 728.19, 732.05, 735.98, 739.98, 741.62, 731.94, 722.21, 712.42, 702.57, 692.66, 682.7, 672.67, 662.59, 652.45, 642.25, 631.98, 621.3, 610.7, 600.18, 589.73, 579.35, 569.06, 558.84, 548.69, 538.62, 528.63, 518.71, 518.92, 520.0, 521.04, 522.05, 523.03, 523.98, 524.89, 525.78, 526.63, 527.45, 528.24, 534.14, 541.44, 548.78, 556.14, 563.53, 570.95, 578.4, 585.88, 593.38, 600.91, 608.47, 627.38, 652.48, 677.77, 703.26, 728.92, 754.78, 780.83, 807.07, 833.5, 860.12, 886.92, 929.22, 986.21, 1043.52, 1101.16, 1159.14, 1217.44, 1276.07, 1335.03, 1394.32, 1453.95, 837.68, 836.44, 835.2, 833.95, 832.7, 831.45, 830.2, 830.05, 836.23, 842.34, 848.39, 854.38, 860.32, 866.19, 872.0, 877.76, 883.46, 889.09, 893.99, 881.94, 870.01, 858.21, 846.54, 835.0, 823.58, 812.29, 801.13, 790.09, 779.19, 768.41, 762.89, 757.84, 752.86, 747.96, 743.12, 738.36, 733.66, 729.04, 724.49, 720.0, 715.59, 724.24, 735.82, 747.35, 758.84, 770.27, 781.65, 792.98, 804.25, 815.48, 826.65, 837.77, 842.31, 843.93, 845.48, 846.96, 848.36, 849.69, 850.94, 852.12, 853.22, 854.25, 855.2, 856.59, 858.37, 860.19, 862.04, 863.93, 865.86, 867.83, 869.84, 871.88, 873.96, 876.07, 872.31, 861.65, 851.07, 840.55, 830.1, 819.73, 809.43, 799.19, 789.04, 778.95, 768.93, 760.3, 754.28, 748.31, 742.38, 736.48, 730.63, 724.81, 719.04, 713.3, 707.6, 701.94, 698.31, 702.07, 705.9, 709.79, 713.76, 717.8, 721.9, 726.08, 730.33, 734.65, 739.04, 742.25, 732.25, 722.2, 712.1, 701.93, 691.71, 681.43, 671.09, 660.69, 650.23, 639.72, 629.15, 618.87, 608.68, 598.57, 588.53, 578.57, 568.68, 558.87, 549.14, 539.48, 529.9, 520.39, 519.53, 520.43, 521.31, 522.15, 522.96, 523.74, 524.49, 525.21, 525.89, 526.54, 527.17, 532.5, 539.96, 547.44, 554.96, 562.5, 570.06, 577.66, 585.28, 592.94, 600.62, 608.33, 626.21, 652.3, 678.58, 705.04, 731.7, 758.54, 785.57, 812.79, 840.21, 867.81, 895.6, 936.34, 995.01, 1054.61, 1113.22, 1170.75, 1227.19, 1282.54, 1336.81, 1389.98, 1442.06, 821.96, 820.72, 819.47, 818.23, 816.98, 815.73, 814.48, 813.71, 819.53, 825.28, 830.98, 836.61, 842.19, 847.71, 853.17, 858.56, 863.9, 869.18, 874.4, 863.84, 852.7, 841.69, 830.81, 820.05, 809.42, 798.92, 788.55, 778.3, 768.18, 758.2, 752.7, 748.09, 743.54, 739.07, 734.66, 730.33, 726.06, 721.87, 717.75, 713.7, 709.72, 716.93, 728.21, 739.45, 750.63, 761.76, 772.84, 783.87, 794.84, 805.77, 816.64, 827.46, 832.4, 833.6, 834.72, 835.76, 836.73, 837.63, 838.45, 839.2, 839.87, 840.47, 840.99, 842.2, 844.2, 846.25, 848.33, 850.44, 852.6, 854.79, 857.02, 859.28, 861.59, 863.93, 861.57, 851.33, 841.15, 831.05, 821.01, 811.05, 801.16, 791.34, 781.59, 771.92, 762.31, 753.71, 747.92, 742.17, 736.46, 730.79, 725.16, 719.57, 714.02, 708.51, 703.03, 697.6, 693.39, 697.54, 701.75, 706.04, 710.4, 714.83, 719.33, 723.9, 728.54, 733.25, 738.03, 742.88, 732.62, 722.26, 711.83, 701.35, 690.81, 680.21, 669.56, 658.84, 648.07, 637.24, 626.35, 616.36, 606.59, 596.88, 587.26, 577.71, 568.23, 558.83, 549.51, 540.26, 531.09, 522.0, 520.16, 520.9, 521.6, 522.28, 522.92, 523.53, 524.11, 524.66, 525.18, 525.67, 526.12, 530.84, 538.45, 546.08, 553.74, 561.43, 569.15, 576.89, 584.66, 592.46, 600.29, 608.15, 624.85, 651.93, 679.74, 708.63, 737.95, 767.71, 797.91, 828.53, 859.6, 891.1, 923.03, 962.97, 1018.13, 1072.22, 1125.22, 1177.14, 1227.97, 1277.72, 1326.37, 1373.94, 1420.41, 806.24, 804.99, 803.75, 802.5, 801.26, 800.01, 798.76, 797.51, 802.89, 808.28, 813.62, 818.9, 824.12, 829.28, 834.39, 839.43, 844.41, 849.33, 854.19, 845.61, 835.26, 825.04, 814.94, 804.98, 795.14, 785.42, 775.84, 766.38, 757.05, 747.85, 742.45, 738.26, 734.15, 730.1, 726.13, 722.23, 718.39, 714.63, 710.94, 707.32, 703.77, 709.67, 720.66, 731.59, 742.47, 753.3, 764.08, 774.81, 785.48, 796.11, 806.68, 817.2, 822.57, 823.33, 824.02, 824.64, 825.18, 825.64, 826.04, 826.35, 826.59, 826.76, 826.85, 827.77, 830.0, 832.27, 834.57, 836.91, 839.29, 841.7, 844.16, 846.65, 849.17, 851.74, 850.76, 840.93, 831.16, 821.47, 811.85, 802.3, 792.82, 783.42, 774.08, 764.82, 755.63, 747.07, 741.51, 735.99, 730.5, 725.06, 719.65, 714.29, 708.96, 703.67, 698.42, 693.22, 688.39, 692.93, 697.54, 702.22, 706.97, 711.79, 716.68, 721.65, 726.68, 731.78, 736.95, 742.2, 733.05, 722.37, 711.63, 700.83, 689.97, 679.06, 668.09, 657.06, 645.97, 634.82, 623.62, 613.78, 604.42, 595.13, 585.91, 576.77, 567.71, 558.72, 549.81, 540.97, 532.21, 523.53, 520.83, 521.4, 521.93, 522.44, 522.92, 523.36, 523.77, 524.15, 524.5, 524.82, 525.11, 529.16, 536.91, 544.7, 552.85, 561.18, 569.71, 578.44, 587.36, 596.47, 605.78, 615.29, 632.45, 662.73, 693.44, 724.58, 756.16, 788.17, 820.61, 853.49, 886.8, 920.55, 954.74, 992.88, 1042.44, 1090.91, 1138.31, 1184.62, 1229.84, 1273.98, 1317.03, 1358.99, 1399.86, 790.51, 789.27, 788.03, 786.78, 785.53, 784.29, 783.04, 781.78, 786.3, 791.34, 796.33, 801.25, 806.11, 810.92, 815.66, 820.35, 824.97, 829.54, 834.04, 827.26, 817.7, 808.26, 798.95, 789.77, 780.72, 771.79, 763.0, 754.33, 745.78, 737.37, 732.12, 728.36, 724.68, 721.07, 717.52, 714.05, 710.65, 707.32, 704.06, 700.87, 697.76, 702.47, 713.15, 723.78, 734.36, 744.89, 755.37, 765.8, 776.17, 786.49, 796.77, 806.99, 812.81, 813.14, 813.4, 813.59, 813.7, 813.73, 813.7, 813.58, 813.39, 813.13, 812.79, 813.3, 815.76, 818.25, 820.78, 823.34, 825.94, 828.58, 831.26, 833.97, 836.72, 839.51, 839.88, 830.46, 821.11, 811.83, 802.62, 793.48, 784.41, 775.42, 766.5, 757.64, 748.86, 740.4, 735.07, 729.77, 724.51, 719.29, 714.11, 708.96, 703.86, 698.8, 693.77, 688.79, 683.85, 688.26, 693.26, 698.33, 703.47, 708.69, 713.97, 719.32, 724.75, 730.24, 735.8, 741.44, 733.53, 722.53, 711.48, 700.36, 689.19, 677.96, 666.67, 655.33, 643.92, 632.46, 620.94, 611.12, 602.17, 593.29, 584.49, 575.76, 567.11, 558.53, 550.03, 541.6, 533.26, 524.98, 521.52, 521.92, 522.29, 522.6, 522.89, 523.16, 523.43, 523.68, 523.93, 524.16, 524.38, 528.08, 536.86, 545.84, 555.0, 564.36, 573.92, 583.67, 593.61, 603.75, 614.08, 624.61, 641.22, 673.75, 706.72, 740.11, 773.94, 808.2, 842.89, 878.02, 913.58, 949.58, 986.01, 1023.88, 1067.82, 1110.69, 1152.48, 1193.18, 1232.8, 1271.33, 1308.78, 1345.13, 1380.39, 774.79, 773.55, 772.3, 771.06, 769.81, 768.56, 767.31, 766.06, 769.78, 774.46, 779.09, 783.65, 788.16, 792.61, 797.0, 801.33, 805.59, 809.8, 813.95, 808.77, 800.0, 791.35, 782.83, 774.44, 766.17, 758.04, 750.03, 742.14, 734.39, 726.76, 721.72, 718.4, 715.14, 711.96, 708.85, 705.81, 702.84, 699.94, 697.11, 694.35, 691.67, 695.31, 705.69, 716.02, 726.3, 736.53, 746.71, 756.84, 766.91, 776.93, 786.91, 796.83, 803.12, 803.03, 802.86, 802.61, 802.29, 801.9, 801.43, 800.88, 800.27, 799.57, 798.8, 798.8, 801.48, 804.19, 806.94, 809.73, 812.56, 815.42, 818.32, 821.26, 824.24, 827.25, 828.92, 819.91, 810.98, 802.11, 793.31, 784.59, 775.93, 767.35, 758.84, 750.4, 742.03, 733.73, 728.58, 723.51, 718.47, 713.47, 708.52, 703.6, 698.72, 693.88, 689.09, 684.33, 679.61, 683.51, 688.91, 694.37, 699.91, 705.51, 711.18, 716.93, 722.74, 728.63, 734.58, 740.61, 734.08, 722.76, 711.39, 699.96, 688.47, 676.92, 665.32, 653.66, 641.94, 630.16, 618.32, 608.39, 599.85, 591.38, 582.95, 574.58, 566.27, 558.04, 549.87, 541.77, 533.74, 525.78, 521.62, 521.89, 522.14, 522.38, 522.61, 522.83, 523.04, 523.24, 523.42, 523.6, 523.77, 526.97, 536.78, 546.77, 556.96, 567.35, 577.93, 588.7, 599.67, 610.83, 622.18, 633.74, 649.57, 684.35, 719.56, 755.2, 791.28, 827.79, 864.73, 902.11, 939.92, 978.17, 1016.85, 1055.95, 1094.29, 1131.56, 1167.74, 1202.84, 1236.85, 1269.78, 1301.62, 1332.36, 1362.02, 759.07, 757.82, 756.58, 755.34, 754.09, 752.84, 751.59, 750.34, 753.31, 757.64, 761.91, 766.12, 770.27, 774.36, 778.39, 782.36, 786.27, 790.13, 793.92, 790.16, 782.18, 774.32, 766.58, 758.98, 751.5, 744.15, 736.93, 729.83, 722.86, 716.03, 711.25, 708.36, 705.53, 702.78, 700.1, 697.49, 694.95, 692.48, 690.09, 687.76, 685.5, 688.2, 698.29, 708.32, 718.3, 728.23, 738.1, 747.93, 757.7, 767.43, 777.1, 786.72, 793.51, 792.99, 792.38, 791.71, 790.96, 790.13, 789.23, 788.26, 787.21, 786.09, 784.89, 784.26, 787.16, 790.09, 793.07, 796.08, 799.13, 802.22, 805.35, 808.51, 811.71, 814.95, 817.9, 809.3, 800.77, 792.32, 783.93, 775.62, 767.38, 759.21, 751.11, 743.08, 735.12, 727.24, 722.05, 717.2, 712.39, 707.62, 702.89, 698.2, 693.55, 688.93, 684.36, 679.82, 675.33, 678.7, 684.49, 690.34, 696.27, 702.26, 708.33, 714.46, 720.67, 726.94, 733.29, 739.71, 734.67, 723.04, 711.35, 699.07, 686.68, 674.42, 662.29, 650.3, 638.43, 626.7, 615.1, 605.23, 597.03, 588.89, 580.82, 572.82, 564.89, 557.03, 549.24, 541.51, 533.86, 526.27, 521.58, 521.78, 521.98, 522.16, 522.34, 522.5, 522.66, 522.8, 522.93, 523.05, 523.16, 525.67, 536.5, 547.52, 558.73, 570.14, 581.74, 593.54, 605.53, 617.71, 630.09, 642.67, 657.48, 694.51, 731.98, 769.87, 808.19, 846.95, 886.14, 925.77, 965.83, 1006.32, 1047.25, 1088.62, 1121.85, 1153.5, 1184.08, 1213.57, 1241.98, 1269.31, 1295.54, 1320.69, 1344.74, 743.34, 742.1, 740.86, 739.61, 738.37, 737.12, 735.87, 734.62, 736.9, 740.87, 744.78, 748.63, 752.43, 756.16, 759.84, 763.46, 767.01, 770.51, 773.95, 771.42, 764.22, 757.15, 750.21, 743.39, 736.7, 730.13, 723.7, 717.39, 711.21, 705.16, 700.71, 698.24, 695.85, 693.53, 691.28, 689.1, 687.0, 684.96, 682.99, 681.09, 679.27, 681.15, 690.93, 700.66, 710.34, 719.97, 729.55, 739.07, 748.55, 757.97, 767.34, 776.66, 783.97, 783.02, 781.98, 780.88, 779.7, 778.44, 777.11, 775.71, 774.23, 772.68, 771.05, 769.67, 772.8, 775.96, 779.16, 782.4, 785.67, 788.98, 792.33, 795.72, 799.14, 802.6, 806.1, 798.61, 790.5, 782.46, 774.48, 766.58, 758.75, 750.99, 743.3, 735.69, 728.14, 720.67, 715.49, 710.86, 706.28, 701.73, 697.22, 692.76, 688.33, 683.94, 679.59, 675.28, 671.01, 673.82, 679.99, 686.24, 692.14, 697.89, 703.55, 709.12, 714.6, 719.98, 725.27, 730.47, 726.14, 714.09, 702.16, 690.36, 678.7, 667.17, 655.77, 644.5, 633.37, 622.36, 611.49, 601.92, 594.09, 586.32, 578.63, 571.0, 563.45, 555.96, 548.54, 541.19, 533.91, 526.69, 521.54, 521.69, 521.83, 521.96, 522.08, 522.19, 522.28, 522.37, 522.45, 522.51, 522.57, 524.18, 536.03, 548.07, 560.31, 572.74, 585.36, 598.18, 611.2, 624.4, 637.81, 651.41, 665.2, 704.25, 743.96, 784.1, 824.68, 865.68, 907.12, 949.0, 991.3, 1034.05, 1077.23, 1120.84, 1150.48, 1176.53, 1201.51, 1225.4, 1248.2, 1269.92, 1290.56, 1310.1, 1328.56, 727.62, 726.38, 725.13, 723.89, 722.64, 721.4, 720.15, 718.9, 720.54, 724.16, 727.71, 731.21, 734.65, 738.03, 741.35, 744.61, 747.81, 750.95, 754.03, 752.55, 746.14, 739.86, 733.7, 727.67, 721.77, 715.99, 710.34, 704.82, 699.43, 694.16, 690.09, 688.06, 686.1, 684.21, 682.39, 680.64, 678.97, 677.36, 675.82, 674.36, 672.96, 674.14, 683.63, 693.06, 702.43, 711.76, 721.04, 730.26, 739.44, 748.56, 757.63, 766.65, 774.51, 773.12, 771.66, 770.12, 768.51, 766.83, 765.07, 763.23, 761.33, 759.34, 757.28, 755.15, 758.4, 761.79, 765.21, 768.67, 772.17, 775.71, 779.28, 782.89, 786.54, 790.22, 793.95, 787.86, 780.15, 772.52, 764.96, 757.47, 750.05, 742.7, 735.43, 728.22, 721.09, 714.03, 708.88, 704.48, 700.09, 695.61, 691.16, 686.75, 682.36, 678.01, 673.69, 669.4, 665.15, 666.73, 672.25, 677.68, 683.02, 688.27, 693.43, 698.49, 703.47, 708.35, 713.14, 717.83, 715.12, 703.79, 692.59, 681.52, 670.59, 659.78, 649.11, 638.57, 628.17, 617.89, 607.75, 598.53, 591.08, 583.69, 576.37, 569.11, 561.93, 554.81, 547.77, 540.79, 533.88, 527.05, 521.52, 521.61, 521.7, 521.77, 521.83, 521.88, 521.92, 521.96, 521.98, 521.99, 521.99, 522.49, 535.36, 548.43, 561.69, 575.14, 588.79, 602.63, 616.67, 630.9, 645.33, 659.95, 674.76, 713.56, 755.52, 797.91, 840.73, 883.98, 927.67, 971.79, 1016.35, 1061.34, 1106.76, 1152.63, 1180.2, 1200.65, 1220.02, 1238.31, 1255.51, 1271.63, 1286.66, 1300.61, 1313.46, 711.9, 710.66, 709.41, 708.17, 706.92, 705.67, 704.42, 703.17, 704.25, 707.5, 710.7, 713.85, 716.93, 719.95, 722.91, 725.82, 728.66, 731.44, 734.17, 733.56, 727.93, 722.44, 717.07, 711.82, 706.71, 701.72, 696.85, 692.12, 687.51, 683.04, 679.41, 677.81, 676.28, 674.82, 673.43, 672.11, 670.87, 669.69, 668.58, 667.55, 666.59, 667.19, 676.37, 685.5, 694.58, 703.61, 712.58, 721.51, 730.38, 739.2, 747.97, 756.69, 765.11, 763.3, 761.4, 759.44, 757.4, 755.28, 753.1, 750.83, 748.49, 746.08, 743.59, 741.03, 743.97, 747.58, 751.22, 754.91, 758.63, 762.39, 766.19, 770.02, 773.89, 777.8, 781.75, 777.03, 769.74, 762.52, 755.34, 748.21, 741.1, 734.03, 727.0, 720.0, 713.04, 706.11, 700.81, 696.45, 692.12, 687.83, 683.57, 679.34, 675.14, 670.97, 666.84, 662.74, 658.67, 659.22, 664.24, 669.17, 674.0, 678.75, 683.4, 687.96, 692.43, 696.8, 701.09, 705.28, 703.96, 693.36, 682.89, 672.55, 662.34, 652.27, 642.32, 632.51, 622.83, 613.29, 603.88, 595.08, 588.0, 580.98, 574.03, 567.15, 560.34, 553.6, 546.93, 540.33, 533.79, 527.33, 521.5, 521.54, 521.57, 521.59, 521.59, 521.59, 521.58, 521.55, 521.52, 521.47, 521.41, 521.35, 534.51, 548.6, 562.88, 577.36, 592.03, 606.89, 621.95, 637.2, 652.65, 668.29, 684.13, 722.44, 766.64, 811.28, 856.35, 901.85, 947.79, 994.16, 1040.96, 1088.2, 1135.87, 1183.98, 1211.0, 1225.84, 1239.61, 1252.3, 1263.9, 1274.42, 1283.85, 1292.2, 1299.45, 696.17, 694.93, 693.69, 692.44, 691.2, 689.95, 688.7, 687.45, 688.01, 690.91, 693.75, 696.54, 699.26, 701.93, 704.54, 707.08, 709.57, 712.0, 714.37, 714.43, 709.6, 704.89, 700.3, 695.85, 691.52, 687.31, 683.24, 679.29, 675.47, 671.78, 668.65, 667.48, 666.39, 665.36, 664.4, 663.51, 662.7, 661.95, 661.27, 660.67, 660.14, 660.29, 669.17, 678.0, 686.77, 695.5, 704.18, 712.8, 721.38, 729.9, 738.37, 746.79, 755.15, 753.55, 751.22, 748.83, 746.36, 743.81, 741.2, 738.5, 735.73, 732.89, 729.97, 726.98, 729.49, 733.6, 738.45, 743.11, 747.57, 751.85, 755.93, 759.83, 763.53, 767.04, 770.35, 766.55, 759.52, 752.51, 745.55, 738.61, 731.72, 724.85, 718.03, 711.23, 704.48, 697.76, 692.43, 688.26, 684.12, 680.01, 675.94, 671.89, 667.88, 663.9, 659.96, 656.05, 652.17, 651.8, 656.32, 660.74, 665.07, 669.31, 673.46, 677.52, 681.48, 685.35, 689.13, 692.82, 692.66, 682.79, 673.05, 663.44, 653.96, 644.62, 635.4, 626.32, 617.37, 608.55, 599.87, 591.56, 584.85, 578.2, 571.63, 565.13, 558.69, 552.32, 546.02, 539.79, 533.63, 527.54, 521.52, 521.48, 521.45, 521.41, 521.37, 521.31, 521.24, 521.16, 521.06, 520.96, 520.85, 520.73, 533.46, 548.57, 563.87, 579.37, 595.07, 610.95, 627.03, 643.31, 659.78, 676.45, 693.31, 730.89, 777.34, 824.23, 871.54, 919.29, 967.47, 1016.09, 1065.14, 1114.62, 1164.54, 1214.9, 1242.88, 1252.12, 1260.29, 1267.38, 1273.38, 1278.3, 1282.14, 1284.88, 1286.54, 680.45, 679.21, 677.97, 676.72, 675.48, 674.23, 672.98, 671.73, 671.82, 674.37, 676.86, 679.29, 681.66, 683.97, 686.22, 688.41, 690.54, 692.61, 694.62, 695.18, 691.13, 687.21, 683.41, 679.74, 676.2, 672.78, 669.49, 666.33, 663.3, 660.4, 657.83, 657.09, 656.42, 655.82, 655.3, 654.84, 654.45, 654.14, 653.89, 653.72, 653.61, 653.58, 662.01, 670.54, 679.02, 687.45, 695.82, 704.15, 712.42, 720.64, 728.81, 736.93, 745.0, 743.92, 743.46, 742.72, 741.69, 740.38, 738.78, 736.89, 734.72, 732.26, 729.52, 726.48, 728.57, 732.51, 736.26, 739.81, 743.18, 746.35, 749.33, 752.12, 754.72, 757.13, 759.34, 756.1, 749.27, 742.47, 735.71, 728.99, 722.29, 715.64, 709.02, 702.43, 695.88, 689.37, 684.02, 680.04, 676.08, 672.16, 668.27, 664.42, 660.59, 656.8, 653.04, 649.32, 645.62, 644.48, 648.49, 652.4, 656.23, 659.97, 663.61, 667.16, 670.62, 673.99, 677.27, 680.45, 681.24, 672.1, 663.08, 654.2, 645.45, 636.83, 628.35, 619.99, 611.77, 603.68, 595.72, 587.97, 581.63, 575.36, 569.16, 563.03, 556.97, 550.97, 545.05, 539.19, 533.4, 527.68, 522.03, 521.43, 521.35, 521.25, 521.15, 521.03, 520.91, 520.77, 520.62, 520.47, 520.3, 520.12, 532.21, 548.35, 564.68, 581.2, 597.91, 614.82, 631.93, 649.22, 666.72, 684.41, 702.29, 738.91, 787.61, 836.74, 886.31, 936.3, 986.73, 1037.59, 1088.89, 1140.62, 1192.78, 1245.38, 1275.84, 1279.48, 1282.05, 1283.54, 1283.95, 1283.27, 1281.5, 1278.65, 1274.71, 664.73, 663.49, 662.24, 661.0, 659.75, 658.51, 657.26, 656.01, 655.7, 657.89, 660.02, 662.09, 664.11, 666.06, 667.95, 669.79, 671.57, 673.28, 674.94, 675.8, 672.54, 669.4, 666.39, 663.51, 660.75, 658.12, 655.62, 653.25, 651.0, 648.88, 646.93, 646.62, 646.38, 646.22, 646.12, 646.09, 646.14, 646.25, 646.44, 646.69, 647.02, 647.42, 658.59, 670.35, 681.8, 692.95, 703.79, 714.32, 724.55, 734.46, 744.06, 753.35, 762.33, 762.89, 760.75, 758.32, 755.61, 752.62, 749.34, 745.77, 741.92, 737.78, 733.35, 728.63, 728.78, 731.61, 734.26, 736.71, 738.97, 741.04, 742.92, 744.61, 746.1, 747.41, 748.52, 745.61, 738.99, 732.4, 725.84, 719.32, 712.84, 706.39, 699.97, 693.59, 687.25, 680.94, 675.58, 671.78, 668.01, 664.28, 660.58, 656.91, 653.27, 649.67, 646.1, 642.56, 639.05, 637.24, 640.74, 644.16, 647.48, 650.71, 653.85, 656.9, 659.86, 662.72, 665.49, 668.17, 669.68, 661.27, 652.98, 644.83, 636.81, 628.92, 621.16, 613.53, 606.04, 598.68, 591.45, 584.35, 578.34, 572.45, 566.62, 560.86, 555.17, 549.55, 544.0, 538.52, 533.1, 527.76, 522.48, 521.39, 521.25, 521.1, 520.94, 520.77, 520.59, 520.4, 520.19, 519.98, 519.76, 519.52, 530.78, 547.93, 565.29, 582.83, 600.57, 618.5, 636.62, 654.94, 673.46, 692.17, 711.07, 746.5, 797.45, 848.83, 900.64, 952.88, 1005.55, 1058.66, 1112.2, 1166.18, 1220.59, 1275.44, 1309.88, 1307.93, 1304.9, 1300.79, 1295.6, 1289.32, 1281.96, 1273.51, 1263.98, 649.0, 647.76, 646.52, 645.28, 644.03, 642.78, 641.53, 640.28, 639.63, 641.46, 643.24, 644.96, 646.61, 648.21, 649.75, 651.23, 652.65, 654.01, 655.31, 656.29, 653.82, 651.47, 649.24, 647.15, 645.18, 643.33, 641.62, 640.03, 638.57, 637.45, 638.39, 640.9, 643.86, 647.2, 650.92, 655.02, 659.51, 664.37, 669.62, 675.26, 681.27, 687.68, 697.21, 707.12, 716.72, 726.02, 735.01, 743.69, 752.06, 760.12, 767.87, 775.31, 782.44, 782.14, 778.32, 774.21, 769.82, 765.15, 760.18, 754.93, 749.4, 743.58, 737.47, 731.07, 729.17, 730.9, 732.44, 733.8, 734.95, 735.92, 736.7, 737.29, 737.68, 737.88, 737.89, 735.09, 728.67, 722.29, 715.94, 709.62, 703.34, 697.1, 690.89, 684.72, 678.58, 672.47, 667.1, 663.49, 659.91, 656.36, 652.85, 649.37, 645.92, 642.5, 639.11, 635.76, 632.44, 630.09, 633.09, 636.0, 638.82, 641.55, 644.19, 646.73, 649.18, 651.54, 653.81, 655.99, 658.0, 650.31, 642.75, 635.33, 628.03, 620.87, 613.84, 606.94, 600.17, 593.54, 587.04, 580.67, 574.99, 569.47, 564.01, 558.63, 553.31, 548.06, 542.88, 537.77, 532.73, 527.76, 522.86, 521.37, 521.17, 520.96, 520.75, 520.52, 520.28, 520.03, 519.77, 519.5, 519.23, 518.93, 529.15, 547.33, 565.7, 584.27, 603.03, 621.98, 641.13, 660.47, 680.01, 699.74, 719.66, 753.66, 806.86, 860.48, 914.54, 969.03, 1023.95, 1079.3, 1135.09, 1191.31, 1247.96, 1305.06, 1345.0, 1337.45, 1328.83, 1319.12, 1308.33, 1296.46, 1283.5, 1269.46, 1254.33, 633.28, 632.04, 630.8, 629.55, 628.31, 627.06, 625.81, 624.56, 623.62, 625.1, 626.52, 627.88, 629.18, 630.42, 631.6, 632.73, 633.79, 635.55, 638.37, 641.25, 641.57, 642.02, 642.64, 643.42, 644.37, 645.48, 646.77, 648.22, 649.83, 651.62, 653.57, 658.07, 663.31, 668.93, 674.94, 681.32, 688.09, 695.23, 702.76, 710.68, 718.98, 727.66, 736.14, 744.2, 751.95, 759.4, 766.53, 773.36, 779.88, 786.09, 791.99, 797.58, 802.85, 801.69, 796.18, 790.39, 784.32, 777.96, 771.32, 764.39, 757.17, 749.67, 741.87, 733.8, 729.76, 730.39, 730.82, 731.07, 731.13, 731.0, 730.67, 730.16, 729.45, 728.55, 727.45, 724.53, 718.32, 712.14, 706.0, 699.89, 693.81, 687.78, 681.77, 675.8, 669.87, 663.97, 658.59, 655.17, 651.77, 648.41, 645.09, 641.79, 638.53, 635.3, 632.1, 628.93, 625.8, 623.04, 625.53, 627.94, 630.25, 632.48, 634.61, 636.65, 638.6, 640.46, 642.22, 643.89, 645.47, 639.22, 632.39, 625.69, 619.12, 612.69, 606.39, 600.21, 594.18, 588.27, 582.49, 576.85, 571.57, 566.42, 561.34, 556.32, 551.38, 546.51, 541.7, 536.96, 532.29, 527.7, 523.17, 521.35, 521.1, 520.83, 520.56, 520.28, 519.98, 519.68, 519.36, 519.04, 518.7, 518.36, 527.33, 546.53, 565.92, 585.51, 605.29, 625.27, 645.44, 665.8, 686.36, 707.11, 728.06, 760.4, 815.84, 871.71, 928.01, 984.75, 1041.91, 1099.51, 1157.54, 1216.01, 1274.91, 1334.24, 1381.2, 1368.06, 1353.84, 1338.53, 1322.15, 1304.68, 1286.13, 1266.5, 1245.77, 617.56, 616.32, 615.08, 613.83, 612.59, 612.71, 612.78, 612.71, 612.59, 615.24, 617.96, 620.73, 623.56, 626.45, 629.39, 632.4, 635.46, 638.58, 641.75, 644.99, 646.47, 647.94, 649.56, 651.36, 653.32, 655.44, 657.74, 660.2, 662.82, 665.62, 668.58, 674.86, 682.38, 690.29, 698.57, 707.24, 716.29, 725.71, 735.53, 745.72, 756.3, 767.26, 775.39, 781.59, 787.49, 793.09, 798.37, 803.35, 808.01, 812.37, 816.42, 820.16, 823.59, 821.51, 814.32, 806.85, 799.1, 791.06, 782.74, 774.12, 765.23, 756.04, 746.57, 736.81, 730.53, 730.06, 729.39, 728.54, 727.5, 726.26, 724.83, 723.22, 721.41, 719.4, 717.21, 713.93, 707.93, 701.95, 696.02, 690.12, 684.25, 678.42, 672.62, 666.86, 661.13, 655.44, 650.05, 646.81, 643.6, 640.43, 637.29, 634.18, 631.11, 628.06, 625.05, 622.07, 619.13, 616.21, 618.06, 619.97, 621.78, 623.5, 625.12, 626.66, 628.11, 629.46, 630.72, 631.89, 632.97, 627.99, 621.89, 615.92, 610.08, 604.37, 598.8, 593.36, 588.04, 582.87, 577.82, 572.9, 568.07, 563.3, 558.59, 553.95, 549.38, 544.88, 540.45, 536.08, 531.79, 527.56, 523.4, 521.34, 521.03, 520.71, 520.38, 520.05, 519.7, 519.34, 518.97, 518.59, 518.19, 517.79, 525.31, 545.54, 565.96, 586.56, 607.37, 628.36, 649.55, 670.94, 692.52, 714.29, 736.26, 766.7, 824.39, 882.51, 941.05, 1000.03, 1059.44, 1119.29, 1179.56, 1240.27, 1301.42, 1363.0, 1418.49, 1399.75, 1379.93, 1359.03, 1337.05, 1313.99, 1289.85, 1264.62, 1238.3, 617.64, 617.53, 617.26, 616.85, 616.29, 615.58, 614.72, 613.72, 612.56, 615.37, 618.44, 621.57, 624.76, 628.0, 631.3, 634.66, 638.07, 641.55, 645.08, 648.67, 651.22, 653.69, 656.33, 659.13, 662.1, 665.24, 668.54, 672.01, 675.65, 679.45, 683.42, 691.27, 701.08, 711.27, 721.83, 732.78, 744.11, 755.81, 767.91, 780.38, 793.24, 806.48, 814.94, 819.3, 823.34, 827.09, 830.52, 833.64, 836.46, 838.97, 841.17, 843.05, 844.63, 841.62, 832.75, 823.6, 814.17, 804.45, 794.44, 784.15, 773.57, 762.71, 751.56, 740.12, 731.49, 729.92, 728.15, 726.2, 724.05, 721.71, 719.19, 716.47, 713.56, 710.45, 707.16, 703.3, 697.5, 691.73, 686.0, 680.31, 674.65, 669.02, 663.43, 657.87, 652.35, 646.87, 641.47, 638.42, 635.4, 632.42, 629.46, 626.54, 623.65, 620.79, 617.97, 615.18, 612.42, 609.69, 610.68, 612.08, 613.39, 614.6, 615.73, 616.76, 617.7, 618.55, 619.31, 619.98, 620.55, 616.64, 611.26, 606.02, 600.91, 595.93, 591.08, 586.36, 581.78, 577.33, 573.01, 568.82, 564.51, 560.11, 555.78, 551.51, 547.31, 543.19, 539.12, 535.13, 531.21, 527.36, 523.57, 521.34, 520.98, 520.6, 520.22, 519.83, 519.42, 519.0, 518.58, 518.14, 517.69, 517.24, 523.11, 544.35, 565.79, 587.42, 609.25, 631.27, 653.48, 675.88, 698.48, 721.28, 744.27, 772.58, 832.51, 892.88, 953.67, 1014.89, 1076.55, 1138.63, 1201.15, 1264.11, 1327.49, 1391.32, 1455.58, 1432.51, 1407.1, 1380.61, 1353.04, 1324.39, 1294.65, 1263.83, 1231.92, 625.35, 624.3, 623.1, 621.75, 620.25, 618.61, 616.81, 614.87, 612.78, 615.44, 618.87, 622.35, 625.89, 629.49, 633.15, 636.86, 640.63, 644.46, 648.35, 652.29, 655.8, 659.28, 662.93, 666.74, 670.72, 674.87, 679.18, 683.66, 688.31, 693.12, 698.1, 707.31, 719.4, 731.87, 744.71, 757.94, 771.55, 785.54, 799.91, 814.66, 829.8, 845.32, 854.81, 857.31, 859.51, 861.4, 862.98, 864.25, 865.22, 865.88, 866.22, 866.26, 865.99, 862.02, 851.47, 840.64, 829.52, 818.12, 806.44, 794.46, 782.21, 769.66, 756.83, 743.71, 732.65, 729.97, 727.1, 724.05, 720.8, 717.36, 713.73, 709.91, 705.9, 701.69, 697.3, 692.63, 687.04, 681.48, 675.95, 670.46, 665.01, 659.59, 654.2, 648.85, 643.54, 638.26, 633.01, 630.0, 627.17, 624.37, 621.6, 618.87, 616.16, 613.49, 610.86, 608.25, 605.68, 603.14, 603.39, 604.29, 605.09, 605.8, 606.43, 606.96, 607.39, 607.74, 607.99, 608.16, 608.23, 605.15, 600.5, 595.99, 591.6, 587.35, 583.23, 579.24, 575.38, 571.66, 568.07, 564.61, 560.88, 556.86, 552.89, 549.0, 545.18, 541.42, 537.73, 534.12, 530.57, 527.09, 523.67, 521.35, 520.94, 520.51, 520.07, 519.62, 519.15, 518.68, 518.2, 517.71, 517.21, 516.69, 520.71, 542.98, 565.44, 588.09, 610.93, 633.97, 657.21, 680.63, 704.25, 728.07, 752.08, 778.03, 840.21, 902.82, 965.85, 1029.32, 1093.22, 1157.55, 1222.31, 1287.51, 1353.14, 1419.21, 1485.71, 1466.36, 1435.36, 1403.27, 1370.11, 1335.87, 1300.54, 1264.13, 1225.58, 633.21, 631.22, 629.08, 626.8, 624.37, 621.78, 619.05, 616.17, 613.14, 615.46, 619.24, 623.08, 626.98, 630.93, 634.94, 639.01, 643.13, 647.32, 651.56, 655.86, 660.21, 664.7, 669.36, 674.19, 679.17, 684.33, 689.65, 695.14, 700.8, 706.62, 712.61, 722.98, 737.34, 752.09, 767.22, 782.72, 798.61, 814.88, 831.53, 848.56, 865.98, 883.78, 894.98, 895.63, 895.98, 896.02, 895.75, 895.17, 894.29, 893.1, 891.59, 889.78, 887.66, 882.7, 870.47, 857.96, 845.16, 832.08, 818.72, 805.07, 791.13, 776.9, 762.39, 747.59, 733.99, 730.21, 726.24, 722.08, 717.73, 713.19, 708.46, 703.54, 698.43, 693.12, 687.63, 681.94, 676.54, 671.19, 665.87, 660.58, 655.33, 650.12, 644.94, 639.79, 634.69, 629.61, 624.57, 621.54, 618.9, 616.29, 613.71, 611.16, 608.64, 606.16, 603.71, 601.29, 598.91, 596.55, 596.19, 596.58, 596.88, 597.09, 597.21, 597.24, 597.17, 597.02, 596.77, 596.43, 595.99, 593.53, 589.61, 585.82, 582.16, 578.64, 575.24, 571.98, 568.85, 565.85, 562.99, 560.26, 557.19, 553.53, 549.94, 546.42, 542.97, 539.59, 536.27, 533.03, 529.85, 526.74, 523.7, 521.38, 520.9, 520.42, 519.92, 519.42, 518.9, 518.37, 517.83, 517.29, 516.73, 516.16, 518.12, 541.41, 564.89, 588.56, 612.43, 636.49, 660.74, 685.19, 709.83, 734.67, 759.7, 784.93, 847.48, 912.33, 977.61, 1043.32, 1109.46, 1176.03, 1243.04, 1310.48, 1373.86, 1436.04, 1497.7, 1478.48, 1438.32, 1398.05, 1357.67, 1317.18, 1276.58, 1235.87, 1195.04, 641.21, 638.29, 635.21, 631.99, 628.63, 625.11, 621.44, 617.63, 613.66, 615.42, 619.55, 623.75, 628.0, 632.31, 636.67, 641.1, 645.58, 650.12, 654.71, 659.37, 664.46, 669.96, 675.63, 681.46, 687.46, 693.63, 699.96, 706.46, 713.12, 719.96, 726.96, 738.26, 754.91, 771.93, 789.34, 807.13, 825.29, 843.84, 862.77, 882.08, 901.78, 921.86, 935.46, 934.26, 932.76, 930.95, 928.83, 926.41, 923.67, 920.63, 917.28, 913.62, 909.64, 903.67, 889.76, 875.57, 861.09, 846.33, 831.28, 815.95, 800.34, 784.43, 768.24, 751.76, 735.52, 730.64, 725.57, 720.31, 714.86, 709.22, 703.39, 697.37, 691.15, 684.75, 678.15, 671.36, 666.01, 660.86, 655.75, 650.67, 645.62, 640.61, 635.64, 630.7, 625.8, 620.93, 616.1, 613.05, 610.6, 608.17, 605.78, 603.42, 601.09, 598.79, 596.53, 594.3, 592.1, 589.93, 589.08, 588.97, 588.77, 588.47, 588.09, 587.61, 587.04, 586.38, 585.63, 584.79, 583.85, 581.78, 578.58, 575.52, 572.59, 569.8, 567.13, 564.59, 562.19, 559.92, 557.78, 555.77, 553.42, 550.14, 546.92, 543.78, 540.7, 537.69, 534.74, 531.87, 529.07, 526.33, 523.67, 521.41, 520.88, 520.34, 519.79, 519.23, 518.65, 518.07, 517.48, 516.87, 516.26, 515.63, 515.34, 539.65, 564.15, 588.84, 613.73, 638.81, 664.08, 689.55, 715.21, 740.77, 765.95, 791.12, 849.69, 913.34, 976.47, 1039.07, 1101.16, 1162.72, 1223.76, 1284.27, 1344.26, 1403.73, 1462.66, 1451.92, 1411.21, 1370.39, 1329.45, 1288.41, 1247.25, 1205.99, 1164.61, 649.36, 645.5, 641.5, 637.34, 633.04, 628.58, 623.98, 619.23, 614.33, 615.32, 619.81, 624.36, 628.97, 633.63, 638.35, 643.13, 647.96, 652.86, 657.81, 662.82, 668.54, 675.06, 681.73, 688.58, 695.59, 702.76, 710.1, 717.61, 725.28, 733.13, 741.13, 753.17, 772.1, 791.4, 811.09, 831.15, 851.6, 872.42, 893.63, 915.22, 937.19, 959.55, 976.26, 973.21, 969.85, 966.19, 962.23, 957.95, 953.37, 948.48, 943.27, 937.76, 931.94, 924.92, 909.33, 893.46, 877.3, 860.86, 844.14, 827.13, 809.83, 792.25, 774.38, 756.22, 737.78, 731.26, 725.09, 718.73, 712.18, 705.44, 698.51, 691.38, 684.07, 676.56, 668.86, 660.97, 655.44, 650.5, 645.59, 640.72, 635.88, 631.07, 626.31, 621.57, 616.88, 612.21, 607.59, 604.53, 602.26, 600.02, 597.82, 595.64, 593.5, 591.39, 589.31, 587.27, 585.26, 583.28, 582.06, 581.45, 580.74, 579.95, 579.06, 578.08, 577.01, 575.84, 574.59, 573.24, 571.8, 569.89, 567.43, 565.09, 562.89, 560.82, 558.88, 557.07, 555.39, 553.85, 552.44, 551.16, 549.59, 546.68, 543.83, 541.06, 538.35, 535.72, 533.15, 530.65, 528.22, 525.85, 523.56, 521.46, 520.87, 520.27, 519.66, 519.05, 518.42, 517.78, 517.13, 516.47, 515.82, 515.35, 514.91, 538.12, 563.35, 588.56, 613.75, 638.93, 664.09, 689.23, 714.35, 739.45, 764.54, 789.61, 842.38, 903.31, 963.72, 1023.6, 1082.97, 1141.81, 1200.13, 1257.92, 1315.19, 1371.94, 1428.16, 1425.47, 1384.21, 1342.83, 1301.34, 1259.74, 1218.04, 1176.22, 1134.29, 657.66, 652.87, 647.92, 642.83, 637.59, 632.21, 626.67, 620.98, 615.15, 615.16, 620.01, 624.91, 629.87, 634.89, 639.97, 645.1, 650.29, 655.54, 660.85, 666.21, 672.47, 679.99, 687.67, 695.53, 703.54, 711.73, 720.08, 728.59, 737.28, 746.13, 755.15, 767.7, 788.91, 810.49, 832.46, 854.8, 877.52, 900.63, 924.11, 947.98, 972.23, 996.87, 1017.36, 1012.46, 1007.26, 1001.75, 995.93, 989.81, 983.37, 976.63, 969.58, 962.22, 954.55, 946.45, 929.19, 911.64, 893.8, 875.68, 857.28, 838.59, 819.61, 800.35, 780.81, 760.97, 740.85, 732.07, 724.8, 717.34, 709.69, 701.84, 693.81, 685.59, 677.17, 668.57, 659.77, 650.78, 644.83, 640.1, 635.39, 630.73, 626.09, 621.5, 616.93, 612.41, 607.92, 603.46, 599.04, 595.98, 593.89, 591.84, 589.82, 587.83, 585.88, 583.96, 582.07, 580.21, 578.38, 576.59, 575.14, 574.02, 572.81, 571.51, 570.11, 568.63, 567.06, 565.39, 563.63, 561.78, 559.84, 557.88, 556.14, 554.53, 553.06, 551.71, 550.5, 549.42, 548.47, 547.65, 546.96, 546.41, 545.68, 543.15, 540.68, 538.27, 535.94, 533.68, 531.48, 529.35, 527.29, 525.3, 523.35, 521.42, 520.77, 520.19, 519.65, 519.15, 518.69, 518.27, 517.9, 517.56, 517.27, 517.02, 516.81, 537.58, 562.72, 587.83, 612.93, 638.01, 663.08, 688.12, 713.15, 738.16, 763.15, 788.12, 835.6, 893.8, 951.49, 1008.65, 1065.3, 1121.42, 1177.02, 1232.1, 1286.65, 1340.67, 1394.17, 1399.13, 1357.31, 1315.38, 1273.34, 1231.19, 1188.93, 1146.56, 1104.07, 666.11, 660.38, 654.5, 648.48, 642.3, 635.98, 629.51, 622.89, 616.12, 614.95, 620.15, 625.41, 630.73, 636.1, 641.53, 647.02, 652.57, 658.17, 663.83, 669.55, 676.22, 684.75, 693.45, 702.31, 711.34, 720.53, 729.89, 739.41, 749.11, 758.96, 768.99, 781.86, 805.34, 829.21, 853.45, 878.07, 903.07, 928.45, 954.21, 980.36, 1006.89, 1033.8, 1058.77, 1052.02, 1044.97, 1037.61, 1029.95, 1021.97, 1013.69, 1005.1, 996.2, 987.0, 977.48, 967.65, 949.33, 930.1, 910.59, 890.79, 870.71, 850.34, 829.68, 808.75, 787.52, 766.01, 744.2, 733.07, 724.7, 716.14, 707.38, 698.44, 689.31, 679.99, 670.47, 660.76, 650.86, 640.77, 634.19, 629.66, 625.17, 620.7, 616.28, 611.88, 607.53, 603.21, 598.92, 594.67, 590.45, 587.39, 585.49, 583.63, 581.79, 579.99, 578.23, 576.49, 574.79, 573.11, 571.47, 569.87, 568.3, 566.67, 564.96, 563.16, 561.26, 559.28, 557.2, 555.03, 552.77, 550.42, 547.97, 545.73, 544.72, 543.84, 543.09, 542.47, 541.98, 541.63, 541.4, 541.31, 541.35, 541.74, 542.06, 540.02, 538.01, 536.02, 534.05, 532.11, 530.19, 528.29, 526.42, 524.57, 522.75, 520.95, 520.4, 520.04, 519.72, 519.44, 519.21, 519.01, 518.86, 518.75, 518.68, 518.66, 518.67, 537.07, 562.11, 587.13, 612.13, 637.11, 662.08, 687.03, 711.96, 736.88, 761.77, 786.65, 829.33, 884.81, 939.78, 994.23, 1048.15, 1101.56, 1154.44, 1206.79, 1258.63, 1309.93, 1360.71, 1372.89, 1330.52, 1288.04, 1245.44, 1202.74, 1159.93, 1117.0, 1073.97, 674.71, 668.04, 661.23, 654.27, 647.16, 639.9, 632.5, 624.94, 617.23, 614.68, 620.24, 625.85, 631.52, 637.25, 643.04, 648.88, 654.78, 660.74, 666.75, 672.83, 679.82, 689.36, 699.06, 708.93, 718.96, 729.17, 739.53, 750.07, 760.77, 771.63, 782.67, 795.64, 821.4, 847.54, 874.06, 900.96, 928.24, 955.9, 983.94, 1012.36, 1041.17, 1070.36, 1099.93, 1091.89, 1082.99, 1073.79, 1064.27, 1054.45, 1044.32, 1033.88, 1023.14, 1012.08, 1000.71, 989.04, 969.76, 948.85, 927.66, 906.18, 884.42, 862.37, 840.04, 817.42, 794.52, 771.33, 747.85, 734.26, 724.79, 715.12, 705.27, 695.23, 685.0, 674.57, 663.96, 653.15, 642.15, 630.96, 623.52, 619.19, 614.9, 610.64, 606.42, 602.24, 598.08, 593.97, 589.89, 585.84, 581.83, 578.77, 577.06, 575.38, 573.73, 572.12, 570.54, 568.99, 567.47, 565.99, 564.53, 563.11, 561.55, 559.42, 557.21, 554.9, 552.5, 550.01, 547.43, 544.76, 542.0, 539.14, 537.04, 535.28, 535.75, 536.2, 536.66, 537.1, 537.54, 537.97, 538.4, 538.82, 539.23, 539.63, 540.03, 538.22, 536.34, 534.48, 532.64, 530.83, 529.04, 527.27, 525.53, 523.82, 522.12, 520.45, 519.98, 519.85, 519.75, 519.7, 519.68, 519.71, 519.78, 519.9, 520.05, 520.25, 520.49, 536.57, 561.51, 586.44, 611.34, 636.23, 661.11, 685.96, 710.8, 735.61, 760.41, 785.19, 823.58, 876.34, 928.59, 980.32, 1031.53, 1082.21, 1132.37, 1182.01, 1231.13, 1279.72, 1327.78, 1346.76, 1303.84, 1260.8, 1217.66, 1174.4, 1131.03, 1087.55, 1043.97, 683.45, 675.85, 668.11, 660.21, 652.17, 643.97, 635.63, 627.14, 618.5, 614.35, 620.26, 626.23, 632.26, 638.34, 644.48, 650.68, 656.94, 663.25, 669.62, 676.05, 683.24, 693.79, 704.51, 715.38, 726.43, 737.64, 749.01, 760.55, 772.26, 784.14, 796.18, 809.05, 837.09, 865.5, 894.3, 923.47, 953.03, 982.96, 1013.28, 1043.98, 1075.06, 1106.53, 1138.38, 1132.08, 1121.33, 1110.27, 1098.91, 1087.24, 1075.26, 1062.98, 1050.38, 1037.48, 1024.26, 1010.74, 990.47, 967.88, 945.01, 921.86, 898.42, 874.7, 850.69, 826.39, 801.81, 776.94, 751.78, 735.64, 725.06, 714.3, 703.35, 692.21, 680.87, 669.35, 657.64, 645.73, 633.63, 621.34, 612.8, 608.68, 604.6, 600.55, 596.53, 592.55, 588.61, 584.7, 580.82, 576.98, 573.17, 570.12, 568.59, 567.1, 565.64, 564.21, 562.82, 561.45, 560.12, 558.83, 557.63, 556.89, 555.94, 553.93, 551.92, 549.9, 547.87, 545.84, 543.8, 541.75, 539.7, 537.64, 535.57, 533.55, 533.98, 534.4, 534.81, 535.22, 535.62, 536.02, 536.41, 536.79, 537.16, 537.53, 537.89, 536.4, 534.65, 532.92, 531.21, 529.53, 527.87, 526.23, 524.62, 523.04, 521.47, 519.93, 519.52, 519.61, 519.74, 519.91, 520.12, 520.37, 520.66, 521.0, 521.38, 521.8, 522.26, 536.09, 560.94, 585.77, 610.58, 635.37, 660.15, 684.91, 709.65, 734.37, 759.07, 783.76, 818.34, 868.39, 917.92, 966.93, 1015.42, 1063.39, 1110.83, 1157.75, 1204.15, 1250.02, 1295.37, 1320.74, 1277.26, 1233.67, 1189.97, 1146.17, 1102.25, 1058.22, 1014.07, 692.35, 683.81, 675.13, 666.3, 657.32, 648.19, 638.92, 629.49, 619.92, 613.97, 620.23, 626.56, 632.94, 639.38, 645.87, 652.42, 659.03, 665.7, 672.43, 679.21, 686.51, 698.06, 709.79, 721.67, 733.73, 745.94, 758.33, 770.88, 783.59, 796.48, 809.53, 822.75, 852.39, 883.08, 914.16, 945.61, 977.44, 1009.65, 1042.25, 1075.22, 1108.58, 1142.32, 1176.45, 1172.57, 1159.97, 1147.07, 1133.86, 1120.34, 1106.51, 1092.38, 1077.94, 1063.19, 1048.13, 1032.76, 1011.47, 987.2, 962.65, 937.82, 912.7, 887.3, 861.62, 835.64, 809.39, 782.84, 756.01, 737.2, 725.53, 713.67, 701.62, 689.38, 676.94, 664.32, 651.51, 638.5, 625.3, 611.91, 602.06, 598.14, 594.26, 590.42, 586.61, 582.83, 579.09, 575.39, 571.72, 568.42, 565.48, 563.08, 562.34, 561.6, 560.86, 560.12, 559.39, 558.65, 557.91, 557.17, 556.43, 555.69, 554.84, 552.8, 550.75, 548.7, 546.63, 544.56, 542.49, 540.4, 538.31, 536.21, 534.11, 532.0, 532.21, 532.6, 532.98, 533.35, 533.71, 534.07, 534.42, 534.76, 535.1, 535.43, 535.75, 534.55, 532.93, 531.33, 529.75, 528.2, 526.67, 525.17, 523.69, 522.23, 520.8, 519.39, 519.02, 519.33, 519.68, 520.08, 520.51, 520.99, 521.5, 522.06, 522.66, 523.31, 523.99, 535.62, 560.38, 585.11, 609.83, 634.53, 659.21, 683.87, 708.52, 733.14, 757.75, 782.34, 813.63, 860.96, 907.77, 954.06, 999.83, 1045.08, 1089.81, 1134.01, 1177.69, 1220.85, 1263.48, 1294.83, 1250.79, 1206.65, 1162.4, 1118.04, 1073.57, 1028.98, 984.29, 701.39, 691.92, 682.3, 672.54, 662.63, 652.56, 642.35, 631.99, 621.49, 613.53, 620.15, 626.83, 633.56, 640.35, 647.2, 654.11, 661.08, 668.1, 675.18, 682.32, 689.61, 702.17, 714.9, 727.8, 740.86, 754.08, 767.48, 781.03, 794.76, 808.65, 822.71, 836.94, 867.32, 900.29, 933.64, 967.37, 1001.47, 1035.96, 1070.83, 1106.08, 1141.72, 1177.73, 1214.13, 1213.37, 1198.92, 1184.17, 1169.11, 1153.75, 1138.08, 1122.1, 1105.81, 1089.21, 1072.3, 1055.09, 1032.75, 1006.8, 980.58, 954.07, 927.28, 900.2, 872.83, 845.19, 817.25, 789.03, 760.52, 738.96, 726.19, 713.23, 700.08, 686.73, 673.2, 659.48, 645.57, 631.47, 617.68, 603.96, 593.46, 590.45, 587.46, 584.49, 581.55, 578.62, 575.73, 572.85, 570.0, 567.17, 564.36, 561.88, 561.14, 560.4, 559.66, 558.92, 558.18, 557.45, 556.71, 555.97, 555.23, 554.49, 553.75, 551.68, 549.59, 547.5, 545.4, 543.29, 541.18, 539.06, 536.93, 534.8, 532.65, 530.51, 530.46, 530.81, 531.15, 531.48, 531.81, 532.13, 532.44, 532.75, 533.05, 533.34, 533.62, 532.68, 531.19, 529.72, 528.27, 526.85, 525.45, 524.08, 522.73, 521.4, 520.1, 518.82, 518.48, 519.02, 519.59, 520.2, 520.86, 521.56, 522.3, 523.08, 523.91, 524.77, 525.68, 535.18, 559.84, 584.48, 609.1, 633.7, 658.29, 682.85, 707.4, 731.93, 756.45, 780.94, 809.44, 854.05, 898.14, 941.72, 984.77, 1027.3, 1069.31, 1110.8, 1151.76, 1192.2, 1232.12, 1269.02, 1224.43, 1179.74, 1134.93, 1090.02, 1044.99, 999.86, 954.61, 710.58, 700.18, 689.62, 678.93, 668.08, 657.08, 645.94, 634.64, 623.2, 613.03, 620.01, 627.04, 634.13, 641.28, 648.48, 655.74, 663.06, 670.44, 677.87, 685.36, 692.92, 706.12, 719.86, 733.76, 747.83, 762.06, 776.46, 791.03, 805.76, 820.66, 835.72, 850.96, 881.87, 917.12, 952.75, 988.75, 1025.13, 1061.89, 1099.04, 1136.57, 1174.47, 1212.77, 1251.44, 1254.48, 1238.18, 1221.59, 1204.68, 1187.47, 1169.95, 1152.13, 1133.99, 1115.55, 1096.79, 1077.73, 1054.31, 1026.69, 998.79, 970.61, 942.14, 913.38, 884.34, 855.01, 824.59, 793.83, 763.1, 738.43, 724.49, 710.62, 696.82, 683.08, 669.41, 655.81, 642.28, 628.81, 615.41, 602.08, 591.0, 588.12, 585.26, 582.43, 579.62, 576.82, 574.06, 571.31, 568.59, 565.89, 563.21, 560.67, 559.94, 559.2, 558.46, 557.72, 556.98, 556.24, 555.51, 554.77, 554.03, 553.29, 552.55, 550.56, 548.44, 546.31, 544.17, 542.03, 539.88, 537.72, 535.56, 533.38, 531.21, 529.02, 528.71, 529.02, 529.32, 529.62, 529.91, 530.19, 530.47, 530.74, 531.0, 531.25, 531.5, 530.78, 529.42, 528.08, 526.77, 525.48, 524.21, 522.96, 521.74, 520.55, 519.38, 518.23, 517.9, 518.66, 519.45, 520.29, 521.17, 522.09, 523.06, 524.06, 525.11, 526.2, 527.33, 534.75, 559.31, 583.86, 608.38, 632.89, 657.38, 681.85, 706.31, 730.74, 755.16, 779.56, 805.76, 847.65, 889.03, 929.89, 970.22, 1010.04, 1049.34, 1088.11, 1126.35, 1164.08, 1201.28, 1237.95, 1198.18, 1152.93, 1107.57, 1062.11, 1016.53, 970.84, 925.04, 719.92, 708.58, 697.09, 685.46, 673.68, 661.75, 649.67, 637.44, 625.07, 612.54, 619.81, 627.19, 634.64, 642.14, 649.7, 657.31, 664.99, 672.72, 680.51, 688.35, 696.26, 709.9, 724.64, 739.55, 754.63, 769.87, 785.28, 800.85, 816.59, 832.5, 848.57, 864.81, 896.05, 933.57, 971.47, 1009.75, 1048.41, 1087.45, 1126.87, 1166.67, 1206.85, 1247.42, 1288.37, 1295.9, 1277.76, 1259.31, 1240.56, 1221.5, 1202.14, 1182.46, 1160.77, 1137.67, 1114.59, 1091.52, 1065.16, 1034.32, 1003.53, 972.76, 942.03, 911.34, 880.68, 850.06, 819.47, 788.91, 758.39, 732.55, 719.01, 705.53, 692.12, 678.78, 665.5, 652.3, 639.16, 626.08, 613.08, 600.14, 588.53, 585.78, 583.05, 580.34, 577.66, 575.0, 572.36, 569.75, 567.16, 564.59, 562.04, 559.52, 558.73, 558.0, 557.26, 556.52, 555.78, 555.04, 554.3, 553.57, 552.83, 552.09, 551.35, 549.45, 547.29, 545.13, 542.95, 540.77, 538.58, 536.39, 534.19, 531.98, 529.76, 527.54, 526.97, 527.24, 527.51, 527.77, 528.02, 528.26, 528.5, 528.73, 528.96, 529.18, 529.39, 528.86, 527.63, 526.42, 525.24, 524.08, 522.94, 521.83, 520.74, 519.67, 518.63, 517.61, 517.28, 518.25, 519.27, 520.34, 521.44, 522.58, 523.77, 525.0, 526.27, 527.58, 528.94, 534.34, 558.81, 583.26, 607.69, 632.1, 656.5, 680.87, 705.23, 729.57, 753.89, 778.2, 802.6, 841.78, 880.44, 918.58, 956.2, 993.3, 1029.88, 1065.94, 1101.47, 1136.48, 1170.96, 1204.92, 1172.03, 1126.23, 1080.32, 1034.3, 988.17, 941.93, 895.58, 729.4, 717.13, 704.71, 692.15, 679.43, 666.57, 653.55, 640.39, 627.08, 613.62, 619.55, 627.29, 635.09, 642.95, 650.86, 658.83, 666.86, 674.94, 683.09, 691.29, 699.55, 713.51, 729.27, 745.18, 761.27, 777.52, 793.93, 810.51, 827.26, 844.17, 861.25, 878.5, 909.85, 949.65, 989.83, 1030.38, 1071.31, 1112.62, 1153.06, 1192.24, 1230.87, 1268.97, 1306.52, 1314.32, 1291.21, 1268.12, 1245.04, 1221.97, 1198.92, 1175.89, 1152.87, 1129.86, 1106.87, 1083.9, 1058.33, 1027.71, 997.12, 966.56, 936.04, 905.56, 875.11, 844.69, 814.31, 783.96, 753.65, 726.61, 713.46, 700.38, 687.36, 674.41, 661.53, 648.71, 635.97, 623.29, 610.67, 598.13, 586.03, 583.41, 580.81, 578.24, 575.68, 573.16, 570.65, 568.17, 565.7, 563.27, 560.85, 558.46, 557.53, 556.79, 556.06, 555.32, 554.58, 553.84, 553.1, 552.36, 551.63, 550.89, 550.15, 548.35, 546.15, 543.95, 541.74, 539.52, 537.3, 535.06, 532.82, 530.58, 528.33, 526.07, 525.23, 525.47, 525.7, 525.92, 526.13, 526.34, 526.54, 526.74, 526.92, 527.1, 527.28, 526.92, 525.82, 524.74, 523.69, 522.65, 521.65, 520.66, 519.71, 518.77, 517.86, 516.97, 516.61, 517.81, 519.05, 520.34, 521.67, 523.03, 524.44, 525.89, 527.39, 528.92, 530.5, 533.95, 558.32, 582.68, 607.01, 631.33, 655.63, 679.91, 704.17, 728.42, 752.65, 776.85, 801.05, 836.42, 872.36, 907.79, 942.7, 977.08, 1010.94, 1044.29, 1077.1, 1109.4, 1141.17, 1172.41, 1145.99, 1099.64, 1053.18, 1006.6, 959.92, 913.13, 866.23, 739.04, 725.83, 712.48, 698.98, 685.33, 671.53, 657.59, 643.49, 629.25, 614.86, 619.24, 627.33, 635.48, 643.69, 651.96, 660.29, 668.67, 677.11, 685.61, 694.16, 702.78, 716.96, 733.73, 750.65, 767.74, 785.0, 803.53, 822.35, 841.15, 859.91, 878.63, 897.33, 927.51, 966.66, 1005.28, 1043.36, 1080.91, 1117.92, 1154.38, 1190.31, 1225.7, 1260.54, 1294.84, 1305.73, 1282.72, 1259.72, 1236.73, 1213.76, 1190.81, 1167.87, 1144.95, 1122.04, 1099.14, 1076.26, 1051.47, 1021.05, 990.67, 960.33, 930.01, 899.74, 869.49, 839.29, 809.11, 778.98, 748.87, 720.61, 707.85, 695.16, 682.53, 669.97, 657.49, 645.06, 632.71, 620.42, 608.2, 596.05, 583.97, 581.02, 578.55, 576.11, 573.68, 571.29, 568.91, 566.56, 564.23, 561.92, 559.64, 557.38, 556.33, 555.59, 554.85, 554.12, 553.38, 552.64, 551.9, 551.16, 550.42, 549.69, 548.95, 547.26, 545.02, 542.78, 540.53, 538.28, 536.01, 533.75, 531.47, 529.19, 526.9, 524.6, 523.5, 523.7, 523.89, 524.08, 524.26, 524.43, 524.59, 524.75, 524.9, 525.04, 525.18, 524.95, 523.98, 523.03, 522.11, 521.21, 520.33, 519.48, 518.65, 517.85, 517.07, 516.31, 515.9, 517.33, 518.79, 520.3, 521.85, 523.44, 525.07, 526.75, 528.46, 530.22, 532.02, 533.87, 557.86, 582.11, 606.35, 630.57, 654.78, 678.96, 703.13, 727.28, 751.41, 775.53, 799.62, 831.58, 864.81, 897.52, 929.71, 961.38, 992.53, 1023.16, 1053.26, 1082.84, 1111.9, 1140.43, 1120.05, 1073.15, 1026.14, 979.01, 931.78, 884.43, 836.98, 748.82, 734.68, 720.39, 705.96, 691.38, 676.65, 661.77, 646.74, 631.56, 616.24, 618.86, 627.31, 635.82, 644.76, 655.42, 666.1, 676.82, 687.57, 698.36, 709.18, 720.03, 735.77, 754.54, 773.28, 791.99, 810.66, 829.31, 847.93, 866.51, 885.06, 903.58, 922.08, 948.89, 984.79, 1020.16, 1054.99, 1089.28, 1123.04, 1156.25, 1188.93, 1221.06, 1252.65, 1283.7, 1297.13, 1274.21, 1251.3, 1228.41, 1205.54, 1182.68, 1159.84, 1137.01, 1114.2, 1091.4, 1068.61, 1044.57, 1014.36, 984.19, 954.05, 923.95, 893.88, 863.85, 833.85, 803.88, 773.95, 744.06, 714.53, 702.16, 689.87, 677.63, 665.47, 653.37, 641.34, 629.38, 617.49, 605.66, 593.9, 582.21, 578.6, 576.26, 573.95, 571.66, 569.39, 567.15, 564.93, 562.73, 560.55, 558.4, 556.27, 555.13, 554.39, 553.65, 552.91, 552.18, 551.44, 550.7, 549.96, 549.22, 548.49, 547.75, 546.17, 543.9, 541.62, 539.33, 537.04, 534.74, 532.43, 530.12, 527.8, 525.48, 523.14, 521.78, 521.94, 522.1, 522.24, 522.38, 522.52, 522.65, 522.77, 522.88, 522.98, 523.08, 522.96, 522.12, 521.3, 520.51, 519.74, 518.99, 518.27, 517.57, 516.9, 516.25, 515.62, 515.15, 516.8, 518.49, 520.22, 521.99, 523.81, 525.66, 527.56, 529.5, 531.48, 533.5, 535.57, 557.4, 581.57, 605.71, 629.84, 653.95, 678.04, 702.11, 726.16, 750.2, 774.22, 798.22, 827.26, 857.77, 887.77, 917.25, 946.2, 974.64, 1002.55, 1029.94, 1056.81, 1083.15, 1108.97, 1094.23, 1046.77, 999.21, 951.53, 903.74, 855.85, 807.84, 758.75, 744.33, 730.2, 716.23, 702.41, 688.73, 675.21, 661.85, 648.63, 635.57, 638.65, 649.4, 660.17, 670.98, 681.82, 692.69, 703.59, 714.53, 725.5, 736.51, 747.54, 762.62, 781.18, 799.71, 818.21, 836.68, 855.12, 873.53, 891.91, 910.25, 928.57, 946.85, 970.81, 1003.46, 1035.58, 1067.16, 1098.2, 1128.7, 1158.66, 1188.08, 1216.97, 1245.31, 1273.1, 1288.5, 1265.68, 1242.87, 1220.08, 1197.3, 1174.54, 1151.79, 1129.06, 1106.34, 1083.64, 1060.95, 1037.64, 1007.64, 977.68, 947.75, 917.85, 887.99, 858.16, 828.37, 798.62, 768.89, 739.21, 709.56, 696.41, 684.51, 672.67, 660.9, 649.19, 637.56, 625.99, 614.49, 603.05, 591.69, 580.39, 576.16, 573.96, 571.78, 569.62, 567.48, 565.37, 563.27, 561.21, 559.16, 557.14, 555.14, 553.93, 553.19, 552.45, 551.71, 550.98, 550.24, 549.5, 548.76, 548.02, 547.28, 546.55, 545.08, 542.78, 540.46, 538.14, 535.81, 533.47, 531.13, 528.78, 526.42, 524.06, 521.69, 520.06, 520.19, 520.31, 520.42, 520.52, 520.62, 520.71, 520.79, 520.87, 520.93, 521.0, 520.94, 520.23, 519.55, 518.88, 518.24, 517.63, 517.03, 516.47, 515.92, 515.4, 514.91, 514.43, 516.23, 518.15, 520.1, 522.09, 524.13, 526.21, 528.33, 530.49, 532.7, 534.94, 537.23, 556.97, 581.04, 605.09, 629.12, 653.13, 677.13, 701.11, 725.06, 749.01, 772.93, 796.84, 823.45, 851.25, 878.54, 905.3, 931.54, 957.26, 982.46, 1007.14, 1031.29, 1054.92, 1078.03, 1068.51, 1020.5, 972.38, 924.15, 875.81, 827.37, 778.81, 777.2, 763.86, 750.67, 737.64, 724.75, 712.01, 699.43, 687.0, 674.72, 662.59, 664.28, 675.21, 686.17, 697.16, 708.19, 719.24, 730.33, 741.46, 752.61, 763.8, 775.03, 789.5, 807.86, 826.18, 844.48, 862.74, 880.97, 899.17, 917.34, 935.48, 953.59, 971.66, 993.27, 1022.67, 1051.53, 1079.86, 1107.65, 1134.9, 1161.61, 1187.78, 1213.41, 1238.5, 1263.05, 1279.87, 1257.14, 1234.42, 1211.73, 1189.04, 1166.38, 1143.72, 1121.09, 1098.46, 1075.86, 1053.27, 1030.67, 1000.88, 971.12, 941.4, 911.72, 882.06, 852.45, 822.86, 793.31, 763.8, 734.32, 704.88, 690.6, 679.08, 667.64, 656.26, 644.95, 633.7, 622.53, 611.42, 600.38, 589.4, 578.5, 573.7, 571.63, 569.58, 567.55, 565.54, 563.56, 561.6, 559.66, 557.74, 555.85, 553.98, 552.72, 551.99, 551.25, 550.51, 549.77, 549.04, 548.3, 547.56, 546.82, 546.08, 545.34, 544.01, 541.66, 539.31, 536.95, 534.58, 532.21, 529.83, 527.44, 525.05, 522.65, 520.24, 518.36, 518.44, 518.52, 518.6, 518.66, 518.72, 518.78, 518.82, 518.86, 518.89, 518.91, 518.9, 518.32, 517.77, 517.23, 516.72, 516.24, 515.78, 515.34, 514.92, 514.53, 514.17, 513.83, 515.63, 517.76, 519.94, 522.15, 524.41, 526.71, 529.06, 531.44, 533.87, 536.34, 538.85, 556.56, 580.53, 604.48, 628.42, 652.34, 676.24, 700.12, 723.98, 747.83, 771.66, 795.47, 820.17, 845.26, 869.82, 893.87, 917.4, 940.41, 962.9, 984.86, 1006.3, 1027.22, 1047.61, 1042.89, 994.33, 945.66, 896.88, 847.99, 798.99, 749.88, 795.65, 783.25, 771.0, 758.89, 746.94, 735.14, 723.49, 712.0, 700.65, 689.46, 689.88, 700.99, 712.14, 723.31, 734.52, 745.77, 757.04, 768.35, 779.7, 791.07, 802.48, 816.42, 834.57, 852.68, 870.77, 888.83, 906.85, 924.85, 942.81, 960.74, 978.64, 996.51, 1016.27, 1042.41, 1068.03, 1093.1, 1117.64, 1141.64, 1165.1, 1188.02, 1210.4, 1232.24, 1253.54, 1271.21, 1248.58, 1225.96, 1203.36, 1180.77, 1158.2, 1135.64, 1113.1, 1090.58, 1068.06, 1045.57, 1023.09, 994.09, 964.54, 935.02, 905.55, 876.1, 846.69, 817.32, 787.98, 758.67, 729.4, 700.16, 684.71, 673.59, 662.54, 651.55, 640.63, 629.78, 619.0, 608.28, 597.63, 587.05, 576.54, 571.22, 569.28, 567.36, 565.46, 563.58, 561.73, 559.9, 558.09, 556.31, 554.54, 552.81, 551.52, 550.79, 550.05, 549.31, 548.57, 547.83, 547.1, 546.36, 545.62, 544.88, 544.14, 542.94, 540.56, 538.17, 535.77, 533.37, 530.96, 528.54, 526.12, 523.69, 521.25, 518.8, 516.65, 516.7, 516.75, 516.78, 516.81, 516.84, 516.85, 516.86, 516.86, 516.85, 516.84, 516.82, 516.39, 515.96, 515.56, 515.18, 514.83, 514.49, 514.19, 513.9, 513.64, 513.41, 513.2, 514.97, 517.33, 519.73, 522.17, 524.65, 527.18, 529.74, 532.35, 535.0, 537.69, 540.43, 556.16, 580.04, 603.9, 627.74, 651.56, 675.36, 699.15, 722.92, 746.67, 770.4, 794.12, 817.4, 839.77, 861.63, 882.97, 903.78, 924.08, 943.85, 963.11, 981.84, 1000.04, 1017.72, 1017.39, 968.27, 919.05, 869.96, 821.66, 773.87, 726.59, 813.95, 802.49, 791.17, 780.0, 768.99, 758.12, 747.41, 736.85, 726.44, 716.18, 715.44, 726.74, 738.07, 749.44, 760.83, 772.26, 783.73, 795.22, 806.75, 818.31, 829.9, 843.37, 861.31, 879.22, 897.1, 914.95, 932.77, 950.55, 968.31, 986.03, 1003.72, 1021.39, 1039.8, 1062.7, 1085.06, 1106.88, 1128.17, 1148.92, 1169.13, 1188.8, 1207.94, 1226.53, 1244.57, 1262.08, 1240.0, 1217.48, 1194.98, 1172.48, 1150.01, 1127.55, 1105.1, 1082.67, 1060.25, 1037.85, 1015.47, 987.26, 957.92, 928.61, 899.34, 870.1, 840.9, 811.74, 782.6, 753.51, 724.44, 695.42, 678.76, 668.03, 657.37, 646.77, 636.25, 625.79, 615.4, 605.07, 594.82, 584.63, 574.51, 568.71, 566.9, 565.11, 563.34, 561.6, 559.88, 558.18, 556.5, 554.85, 553.21, 551.61, 550.32, 549.58, 548.85, 548.11, 547.37, 546.63, 545.9, 545.16, 544.42, 543.68, 542.94, 541.88, 539.46, 537.03, 534.6, 532.16, 529.71, 527.26, 524.8, 522.33, 519.85, 517.37, 514.96, 514.97, 514.98, 514.98, 514.97, 514.95, 514.93, 514.9, 514.87, 514.82, 514.77, 514.72, 514.43, 514.14, 513.86, 513.61, 513.39, 513.19, 513.01, 512.86, 512.73, 512.62, 512.54, 514.28, 516.86, 519.49, 522.15, 524.85, 527.6, 530.39, 533.22, 536.09, 539.01, 541.96, 555.78, 579.56, 603.33, 627.07, 650.8, 674.51, 698.2, 721.87, 745.53, 769.17, 792.79, 815.15, 834.81, 853.95, 872.58, 889.9, 906.69, 923.18, 939.36, 955.22, 970.76, 986.0, 990.36, 943.18, 896.5, 850.32, 804.64, 759.48, 714.81, 832.11, 821.57, 811.19, 800.96, 790.88, 780.95, 771.18, 761.55, 752.08, 742.76, 740.98, 752.46, 763.98, 775.53, 787.11, 798.73, 810.38, 822.06, 833.77, 845.52, 857.3, 870.35, 888.09, 905.79, 923.46, 941.1, 958.72, 976.29, 993.84, 1011.36, 1028.85, 1046.3, 1063.87, 1083.51, 1102.63, 1121.2, 1139.24, 1156.74, 1173.7, 1190.13, 1206.01, 1221.35, 1236.15, 1250.41, 1231.41, 1208.99, 1186.58, 1164.18, 1141.8, 1119.43, 1097.08, 1074.75, 1052.43, 1030.12, 1007.83, 980.39, 951.26, 922.16, 893.1, 864.07, 835.08, 806.12, 777.19, 748.3, 719.45, 690.63, 672.74, 662.4, 652.13, 641.93, 631.8, 621.73, 611.73, 601.8, 591.94, 582.14, 572.41, 566.19, 564.5, 562.84, 561.21, 559.59, 558.0, 556.43, 554.88, 553.36, 551.86, 550.38, 549.12, 548.38, 547.64, 546.91, 546.17, 545.43, 544.69, 543.96, 543.22, 542.48, 541.74, 540.82, 538.36, 535.9, 533.43, 530.95, 528.47, 525.98, 523.48, 520.98, 518.46, 515.95, 513.42, 513.25, 513.22, 513.18, 513.13, 513.08, 513.02, 512.95, 512.88, 512.8, 512.71, 512.62, 512.45, 512.29, 512.14, 512.02, 511.93, 511.86, 511.81, 511.79, 511.79, 511.81, 511.86, 513.55, 516.35, 519.2, 522.08, 525.01, 527.98, 530.99, 534.04, 537.14, 540.28, 543.45, 555.42, 579.11, 602.78, 626.43, 649.95, 673.07, 695.83, 718.22, 740.25, 761.91, 783.21, 803.09, 819.52, 835.63, 851.44, 866.93, 882.12, 896.99, 911.54, 925.79, 939.72, 953.34, 962.36, 917.79, 873.73, 830.17, 787.12, 744.57, 702.53, 850.11, 840.52, 831.07, 821.78, 812.63, 803.64, 794.79, 786.1, 777.57, 769.18, 766.49, 778.16, 789.86, 801.59, 813.36, 825.16, 837.0, 848.86, 860.76, 872.69, 884.66, 897.37, 914.9, 932.39, 949.86, 967.29, 984.7, 1002.07, 1019.41, 1036.72, 1054.0, 1071.25, 1088.46, 1104.87, 1120.73, 1136.06, 1150.85, 1165.1, 1178.81, 1191.99, 1204.62, 1216.72, 1228.27, 1239.28, 1222.81, 1200.48, 1178.16, 1155.86, 1133.58, 1111.31, 1089.05, 1066.81, 1044.58, 1022.37, 1000.18, 973.49, 944.57, 915.68, 886.82, 858.0, 829.22, 800.47, 771.75, 743.07, 714.42, 685.81, 666.65, 656.71, 646.83, 637.02, 627.28, 617.6, 608.0, 598.46, 588.99, 579.58, 570.25, 563.63, 562.08, 560.55, 559.05, 557.56, 556.1, 554.66, 553.25, 551.85, 550.48, 549.14, 547.92, 547.18, 546.44, 545.71, 544.97, 544.23, 543.49, 542.75, 542.02, 541.28, 540.54, 539.77, 537.28, 534.78, 532.27, 529.75, 527.23, 524.71, 522.17, 519.63, 517.08, 514.53, 511.97, 511.53, 511.46, 511.38, 511.3, 511.21, 511.12, 511.01, 510.9, 510.79, 510.66, 510.53, 510.45, 510.41, 510.4, 510.41, 510.44, 510.5, 510.59, 510.69, 510.82, 510.98, 511.16, 512.78, 515.8, 518.87, 521.98, 525.18, 528.75, 532.41, 536.16, 539.99, 543.91, 547.92, 558.29, 580.94, 603.23, 625.16, 646.72, 667.93, 688.77, 709.25, 729.36, 749.11, 768.49, 786.9, 801.71, 816.21, 830.4, 844.28, 857.85, 871.1, 884.04, 896.67, 908.99, 920.99, 932.68, 891.9, 850.46, 809.52, 769.09, 729.16, 689.73, 867.97, 859.31, 850.8, 842.44, 834.23, 826.17, 818.26, 810.51, 802.9, 795.45, 791.96, 803.82, 815.71, 827.63, 839.58, 851.57, 863.59, 875.64, 887.72, 899.84, 911.99, 924.42, 941.74, 959.03, 976.29, 993.52, 1010.71, 1027.88, 1045.01, 1062.12, 1079.19, 1096.23, 1113.24, 1126.76, 1139.38, 1151.45, 1162.99, 1174.0, 1184.47, 1194.39, 1203.78, 1212.63, 1220.93, 1228.69, 1214.18, 1191.95, 1169.73, 1147.52, 1125.33, 1103.16, 1081.0, 1058.86, 1036.73, 1014.61, 992.51, 966.55, 937.84, 909.16, 880.51, 851.9, 823.32, 794.78, 766.27, 737.8, 709.36, 680.95, 660.49, 650.94, 641.46, 632.04, 622.69, 613.41, 604.19, 595.05, 585.97, 576.96, 568.01, 561.06, 559.64, 558.24, 556.86, 555.51, 554.18, 552.87, 551.59, 550.32, 549.08, 547.87, 546.72, 545.98, 545.24, 544.5, 543.77, 543.03, 542.29, 541.55, 540.81, 540.08, 539.34, 538.6, 536.2, 533.66, 531.12, 528.56, 526.01, 523.44, 520.87, 518.29, 515.71, 513.12, 510.52, 509.82, 509.71, 509.6, 509.48, 509.35, 509.22, 509.08, 508.93, 508.78, 508.61, 508.45, 508.42, 508.51, 508.63, 508.77, 508.94, 509.45, 509.95, 510.46, 510.96, 511.47, 511.97, 513.78, 517.47, 521.24, 525.1, 529.05, 533.09, 537.21, 541.42, 545.72, 550.11, 554.58, 563.08, 583.82, 604.19, 624.2, 643.85, 663.14, 682.07, 700.64, 718.84, 736.67, 754.14, 771.02, 784.22, 797.1, 809.67, 821.94, 833.89, 845.53, 856.85, 867.87, 878.57, 888.96, 899.03, 865.5, 826.68, 788.36, 750.55, 713.24, 676.43, 885.68, 877.95, 870.38, 862.95, 855.68, 848.55, 841.58, 834.76, 828.09, 821.57, 817.41, 829.45, 841.53, 853.63, 865.77, 877.94, 890.15, 902.38, 914.65, 926.96, 939.29, 951.66, 968.62, 985.7, 1002.75, 1019.77, 1036.76, 1053.72, 1070.65, 1087.54, 1104.41, 1121.24, 1138.04, 1149.19, 1158.56, 1167.39, 1175.68, 1183.44, 1190.66, 1197.34, 1203.48, 1209.08, 1214.14, 1218.65, 1205.55, 1183.41, 1161.28, 1139.17, 1117.08, 1095.0, 1072.93, 1050.89, 1028.85, 1006.83, 984.83, 959.58, 931.08, 902.6, 874.16, 845.76, 817.39, 789.05, 760.75, 732.49, 704.26, 676.06, 654.27, 645.11, 636.02, 626.99, 618.04, 609.15, 600.32, 591.57, 582.88, 574.26, 565.71, 558.46, 557.17, 555.9, 554.66, 553.44, 552.24, 551.06, 549.9, 548.77, 547.66, 546.57, 545.51, 544.78, 544.04, 543.3, 542.56, 541.83, 541.09, 540.35, 539.61, 538.88, 538.14, 537.4, 535.12, 532.55, 529.97, 527.38, 524.79, 522.18, 519.58, 516.96, 514.34, 511.71, 509.07, 508.11, 507.97, 507.82, 507.66, 507.5, 507.73, 507.94, 508.14, 508.33, 508.5, 508.66, 508.99, 509.49, 510.0, 510.5, 511.01, 511.51, 512.02, 512.52, 513.03, 513.53, 514.04, 515.71, 519.86, 524.1, 528.42, 532.84, 537.34, 541.92, 546.6, 551.36, 556.21, 561.15, 568.23, 587.05, 605.51, 623.61, 641.35, 658.72, 675.74, 692.39, 708.67, 724.6, 740.15, 755.35, 767.03, 778.3, 789.25, 799.9, 810.24, 820.26, 829.97, 839.37, 848.46, 857.23, 865.69, 838.6, 802.4, 766.7, 731.5, 696.81, 662.62, 903.24, 896.45, 889.81, 883.32, 876.98, 870.79, 864.75, 858.86, 853.13, 847.55, 842.83, 855.06, 867.31, 879.61, 891.93, 904.29, 916.68, 929.1, 941.56, 954.04, 966.57, 979.12, 995.53, 1012.41, 1029.25, 1046.07, 1062.85, 1079.6, 1096.32, 1113.01, 1129.67, 1146.29, 1162.89, 1172.16, 1178.27, 1183.86, 1188.9, 1193.41, 1197.39, 1200.82, 1203.72, 1206.07, 1207.88, 1209.15, 1196.89, 1174.85, 1152.82, 1130.8, 1108.81, 1086.82, 1064.85, 1042.9, 1020.96, 999.04, 977.13, 952.58, 924.28, 896.01, 867.78, 839.58, 811.42, 783.3, 755.2, 727.14, 699.12, 671.13, 647.98, 639.21, 630.51, 621.88, 613.32, 604.82, 596.39, 588.02, 579.73, 571.5, 563.34, 555.84, 554.68, 553.55, 552.43, 551.34, 550.27, 549.22, 548.2, 547.19, 546.22, 545.26, 544.33, 543.57, 542.84, 542.1, 541.36, 540.63, 539.89, 539.15, 538.41, 537.67, 536.94, 536.2, 534.06, 531.45, 528.83, 526.2, 523.61, 521.29, 519.0, 516.73, 514.5, 512.3, 510.13, 509.35, 509.55, 509.75, 509.93, 510.09, 510.24, 510.38, 510.51, 510.62, 510.72, 510.8, 511.05, 511.56, 512.06, 512.57, 513.07, 513.58, 514.08, 514.59, 515.09, 515.6, 516.1, 517.55, 522.17, 526.87, 531.66, 536.54, 541.5, 546.55, 551.69, 556.91, 562.23, 567.63, 573.74, 590.65, 607.2, 623.38, 639.21, 654.67, 669.77, 684.5, 698.88, 712.89, 726.53, 739.81, 750.15, 759.8, 769.14, 778.18, 786.9, 795.31, 803.4, 811.19, 818.66, 825.82, 832.66, 811.19, 777.61, 744.52, 711.94, 679.87, 648.3, 920.65, 914.8, 909.09, 903.53, 898.13, 892.88, 887.77, 882.82, 878.02, 873.37, 868.87, 880.63, 893.07, 905.55, 918.06, 930.6, 943.18, 955.79, 968.43, 981.1, 993.81, 1006.55, 1022.48, 1039.15, 1055.78, 1072.39, 1088.97, 1105.51, 1122.02, 1138.5, 1154.96, 1171.37, 1187.76, 1195.66, 1198.53, 1200.86, 1202.67, 1203.93, 1204.66, 1204.85, 1204.49, 1203.6, 1202.17, 1200.2, 1188.22, 1166.27, 1144.34, 1122.42, 1100.52, 1078.63, 1056.76, 1034.9, 1013.05, 991.23, 969.41, 945.54, 917.44, 889.39, 861.36, 833.37, 805.42, 777.5, 749.62, 721.77, 693.95, 666.17, 641.62, 633.25, 624.94, 616.7, 608.53, 600.42, 592.38, 584.41, 576.51, 568.67, 560.9, 553.2, 552.17, 551.16, 550.18, 549.22, 548.28, 547.36, 546.47, 545.6, 544.75, 543.92, 543.12, 542.37, 541.64, 540.9, 540.16, 539.55, 539.09, 538.6, 538.08, 537.53, 536.95, 536.34, 534.6, 532.32, 530.08, 527.86, 525.67, 523.51, 521.39, 519.29, 517.23, 515.19, 513.19, 512.25, 512.38, 512.5, 512.6, 512.69, 512.77, 512.84, 512.89, 512.92, 512.95, 512.96, 513.12, 513.62, 514.13, 514.63, 515.14, 515.64, 516.15, 516.65, 517.16, 517.66, 518.17, 519.31, 524.39, 529.55, 534.81, 540.15, 545.57, 551.09, 556.69, 562.38, 568.16, 574.02, 579.97, 594.61, 609.24, 623.51, 637.42, 650.97, 664.16, 676.98, 689.44, 701.54, 713.27, 724.64, 733.58, 741.61, 749.34, 756.76, 763.87, 770.66, 777.14, 783.31, 789.17, 794.72, 799.95, 783.28, 752.31, 721.84, 691.88, 662.43, 633.47, 937.92, 933.0, 928.22, 923.6, 919.13, 914.81, 910.64, 906.63, 902.76, 899.04, 895.48, 906.17, 918.8, 931.46, 944.16, 956.89, 969.65, 982.44, 995.27, 1008.13, 1021.02, 1033.94, 1049.46, 1065.92, 1082.35, 1098.75, 1115.12, 1131.45, 1147.76, 1164.04, 1180.28, 1196.49, 1212.67, 1219.69, 1219.32, 1218.41, 1216.97, 1214.98, 1212.47, 1209.41, 1205.81, 1201.68, 1197.0, 1191.78, 1179.54, 1157.68, 1135.84, 1114.02, 1092.21, 1070.42, 1048.64, 1026.88, 1005.13, 983.4, 961.68, 938.46, 910.58, 882.73, 854.91, 827.13, 799.38, 771.67, 743.99, 716.35, 688.74, 661.17, 635.19, 627.21, 619.3, 611.45, 603.67, 595.95, 588.31, 580.73, 573.22, 565.77, 558.39, 551.09, 549.64, 548.76, 547.91, 547.07, 546.75, 546.48, 546.15, 545.79, 545.39, 544.94, 544.45, 543.96, 543.45, 542.91, 542.34, 541.74, 541.11, 540.45, 539.76, 539.05, 538.3, 537.52, 535.96, 533.85, 531.77, 529.72, 527.7, 525.71, 523.75, 521.82, 519.93, 518.06, 516.22, 515.16, 515.22, 515.26, 515.29, 515.31, 515.31, 515.3, 515.28, 515.24, 515.19, 515.13, 515.18, 515.69, 516.19, 516.7, 517.2, 517.71, 518.21, 518.72, 519.22, 519.73, 520.23, 520.98, 526.52, 532.15, 537.87, 543.67, 549.56, 555.54, 561.61, 567.76, 574.0, 580.33, 586.74, 598.93, 611.65, 624.0, 636.0, 647.64, 658.91, 669.82, 680.37, 690.56, 700.38, 709.83, 717.31, 723.74, 729.85, 735.65, 741.15, 746.33, 751.2, 755.75, 759.99, 763.92, 767.54, 754.86, 726.51, 698.66, 671.31, 644.47, 618.14, 955.03, 951.05, 947.21, 943.52, 939.99, 936.6, 933.37, 930.28, 927.35, 924.57, 921.94, 931.68, 944.5, 957.35, 970.23, 983.14, 996.09, 1009.07, 1022.08, 1035.12, 1048.2, 1061.31, 1076.47, 1092.73, 1108.95, 1125.14, 1141.3, 1157.43, 1173.53, 1189.6, 1205.64, 1221.64, 1237.62, 1244.27, 1240.65, 1236.49, 1231.8, 1226.58, 1220.82, 1214.51, 1207.67, 1200.29, 1192.37, 1183.91, 1170.83, 1149.08, 1127.33, 1105.6, 1083.89, 1062.19, 1040.51, 1018.84, 997.19, 975.55, 953.93, 931.35, 903.67, 876.03, 848.42, 820.85, 793.31, 765.8, 738.34, 710.9, 683.5, 656.14, 628.81, 621.11, 613.59, 606.13, 599.14, 592.68, 586.23, 579.81, 573.4, 567.02, 560.65, 554.31, 552.95, 552.6, 552.2, 551.77, 551.29, 550.77, 550.21, 549.6, 548.96, 548.27, 547.54, 546.85, 546.17, 545.46, 544.72, 543.96, 543.16, 542.33, 541.48, 540.59, 539.68, 538.74, 537.28, 535.34, 533.43, 531.55, 529.7, 527.87, 526.08, 524.32, 522.59, 520.89, 519.22, 518.09, 518.07, 518.04, 518.0, 517.94, 517.87, 517.78, 517.68, 517.57, 517.45, 517.31, 517.25, 517.75, 518.26, 518.76, 519.27, 519.77, 520.28, 520.78, 521.29, 521.79, 522.3, 522.8, 528.57, 534.66, 540.84, 547.11, 553.46, 559.91, 566.44, 573.05, 579.76, 586.55, 593.43, 603.61, 614.41, 624.86, 634.94, 644.67, 654.03, 663.03, 671.67, 679.94, 687.85, 695.39, 701.36, 706.17, 710.67, 714.86, 718.73, 722.3, 725.56, 728.5, 731.13, 733.44, 735.45, 725.93, 700.2, 674.97, 650.24, 626.02, 602.3, 972.0, 968.95, 966.05, 963.29, 960.69, 958.24, 955.94, 953.79, 951.79, 949.94, 948.25, 957.17, 970.17, 983.2, 996.27, 1009.37, 1022.5, 1035.66, 1048.86, 1062.09, 1075.35, 1088.65, 1103.52, 1119.56, 1135.58, 1151.57, 1167.52, 1183.44, 1199.34, 1215.2, 1231.03, 1246.83, 1262.59, 1269.38, 1262.51, 1255.11, 1247.18, 1238.71, 1229.7, 1220.16, 1210.07, 1199.45, 1188.29, 1176.58, 1162.11, 1140.45, 1118.81, 1097.17, 1075.56, 1053.95, 1032.37, 1010.79, 989.24, 967.69, 946.17, 924.2, 896.73, 869.3, 842.0, 815.2, 788.56, 762.07, 735.73, 709.55, 683.52, 657.64, 631.92, 623.75, 617.34, 610.94, 604.57, 598.22, 591.89, 585.57, 579.28, 573.01, 566.76, 560.52, 558.5, 557.9, 557.26, 556.59, 555.87, 555.1, 554.3, 553.45, 552.57, 551.64, 550.66, 549.76, 548.92, 548.04, 547.14, 546.2, 545.24, 544.25, 543.22, 542.17, 541.09, 539.98, 538.58, 536.81, 535.06, 533.35, 531.66, 530.01, 528.39, 526.79, 525.23, 523.7, 522.2, 521.03, 520.94, 520.83, 520.71, 520.58, 520.43, 520.28, 520.1, 519.92, 519.72, 519.51, 519.32, 519.82, 520.32, 520.83, 521.33, 521.84, 522.34, 522.85, 523.35, 523.86, 524.36, 524.87, 530.53, 537.09, 543.73, 550.46, 557.28, 564.19, 571.18, 578.26, 585.42, 592.68, 600.02, 608.64, 617.54, 626.07, 634.25, 642.06, 649.51, 656.6, 663.32, 669.68, 675.68, 681.31, 685.71, 688.91, 691.79, 694.37, 696.63, 698.58, 700.23, 701.56, 702.57, 703.28, 703.67, 696.5, 673.39, 650.77, 628.66, 607.05, 585.95, 988.83, 986.71, 984.74, 982.92, 981.25, 979.73, 978.36, 977.15, 976.08, 975.17, 974.41, 982.62, 995.81, 1009.03, 1022.28, 1035.56, 1048.88, 1062.23, 1075.61, 1089.02, 1102.47, 1115.95, 1130.6, 1146.44, 1162.25, 1178.03, 1193.77, 1209.49, 1225.18, 1240.83, 1256.45, 1272.05, 1287.61, 1295.02, 1284.91, 1274.27, 1263.09, 1251.38, 1239.13, 1226.34, 1213.01, 1199.15, 1184.74, 1169.79, 1153.38, 1131.81, 1110.45, 1090.4, 1070.07, 1049.46, 1028.57, 1007.39, 985.93, 964.19, 942.17, 919.86, 893.54, 867.33, 841.28, 815.38, 789.64, 764.04, 738.6, 713.32, 688.18, 663.21, 638.38, 628.82, 622.52, 616.24, 609.98, 603.74, 597.52, 591.32, 585.14, 578.98, 572.84, 566.72, 564.08, 563.25, 562.37, 561.45, 560.49, 559.48, 558.44, 557.35, 556.22, 555.05, 553.83, 552.71, 551.69, 550.65, 549.58, 548.48, 547.35, 546.19, 545.0, 543.78, 542.53, 541.25, 539.85, 538.24, 536.66, 535.12, 533.6, 532.11, 530.66, 529.23, 527.84, 526.47, 525.14, 523.98, 523.82, 523.64, 523.44, 523.24, 523.02, 522.78, 522.54, 522.28, 522.01, 521.72, 521.42, 521.88, 522.39, 522.89, 523.4, 523.9, 524.41, 524.91, 525.42, 525.92, 526.43, 526.93, 532.4, 539.43, 546.53, 553.73, 561.01, 568.38, 575.83, 583.38, 591.01, 598.72, 606.53, 614.04, 621.03, 627.65, 633.91, 639.81, 645.35, 650.53, 655.34, 659.79, 663.88, 667.6, 670.37, 671.95, 673.22, 674.19, 674.84, 675.18, 675.21, 674.92, 674.33, 673.42, 672.19, 666.57, 646.07, 626.07, 606.57, 587.58, 569.09, 1005.5, 1004.31, 1003.28, 1002.39, 1001.66, 1001.07, 1000.64, 1000.35, 1000.22, 1000.24, 1000.42, 1008.04, 1021.42, 1034.82, 1048.26, 1061.73, 1075.23, 1088.76, 1102.33, 1115.93, 1129.56, 1143.23, 1157.71, 1173.34, 1188.95, 1204.52, 1220.06, 1235.57, 1251.05, 1266.5, 1281.91, 1297.3, 1312.65, 1321.2, 1307.69, 1293.37, 1279.06, 1264.76, 1250.48, 1236.2, 1221.94, 1207.69, 1193.44, 1179.21, 1164.49, 1143.32, 1121.88, 1100.15, 1078.15, 1055.86, 1033.29, 1010.44, 987.31, 963.89, 940.19, 916.21, 890.88, 865.57, 840.42, 815.42, 790.57, 765.87, 741.33, 716.94, 692.7, 668.62, 644.69, 633.88, 627.69, 621.52, 615.37, 609.25, 603.14, 597.05, 590.98, 584.93, 578.91, 572.9, 569.71, 568.63, 567.51, 566.35, 565.15, 563.9, 562.61, 561.29, 559.91, 558.5, 557.04, 555.68, 554.5, 553.29, 552.05, 550.79, 549.49, 548.16, 546.8, 545.42, 544.0, 542.55, 541.08, 539.64, 538.24, 536.86, 535.51, 534.19, 532.9, 531.64, 530.41, 529.22, 528.05, 526.95, 526.71, 526.45, 526.19, 525.91, 525.61, 525.3, 524.98, 524.65, 524.3, 523.94, 523.57, 523.95, 524.45, 524.96, 525.46, 525.97, 526.47, 526.98, 527.48, 527.99, 528.49, 529.0, 534.19, 541.68, 549.25, 556.91, 564.65, 572.48, 580.4, 588.41, 596.5, 604.68, 612.95, 619.8, 624.88, 629.59, 633.94, 637.93, 641.55, 644.82, 647.72, 650.26, 652.44, 654.25, 655.34, 655.31, 654.97, 654.32, 653.35, 652.08, 650.5, 648.6, 646.39, 643.87, 641.03, 636.13, 618.24, 600.86, 583.98, 567.6, 551.72, 1022.02, 1021.77, 1021.67, 1021.72, 1021.92, 1022.26, 1022.76, 1023.41, 1024.22, 1025.17, 1026.27, 1033.44, 1047.0, 1060.58, 1074.21, 1087.86, 1101.55, 1115.27, 1129.02, 1142.81, 1156.62, 1170.48, 1184.93, 1200.34, 1215.49, 1230.38, 1245.0, 1259.37, 1273.48, 1287.33, 1300.91, 1314.24, 1327.3, 1335.86, 1321.58, 1307.32, 1293.07, 1278.84, 1264.61, 1250.39, 1236.18, 1221.99, 1207.8, 1193.63, 1179.47, 1156.7, 1133.58, 1110.18, 1086.5, 1062.54, 1038.29, 1013.77, 988.96, 963.87, 938.5, 912.84, 888.07, 863.66, 839.4, 815.3, 791.34, 767.54, 743.89, 720.4, 697.06, 673.87, 650.84, 638.91, 632.84, 626.78, 620.75, 614.73, 608.73, 602.76, 596.8, 590.87, 584.95, 579.06, 575.38, 574.06, 572.7, 571.3, 569.85, 568.37, 566.84, 565.26, 563.65, 562.0, 560.3, 558.69, 557.34, 555.96, 554.56, 553.12, 551.66, 550.16, 548.64, 547.08, 545.5, 543.89, 542.29, 541.02, 539.78, 538.56, 537.38, 536.23, 535.11, 534.02, 532.96, 531.93, 530.93, 529.96, 529.61, 529.29, 528.94, 528.59, 528.22, 527.84, 527.44, 527.04, 526.62, 526.18, 525.73, 526.02, 526.52, 527.02, 527.53, 528.03, 528.54, 529.04, 529.55, 530.05, 530.56, 531.06, 535.9, 543.84, 551.88, 560.0, 568.21, 576.5, 584.88, 593.35, 601.91, 610.55, 619.28, 625.92, 629.08, 631.88, 634.32, 636.4, 638.12, 639.47, 640.47, 641.09, 641.36, 641.26, 640.62, 638.97, 637.02, 634.75, 632.18, 629.29, 626.1, 622.59, 618.77, 614.63, 610.18, 605.19, 589.91, 575.14, 560.88, 547.11, 533.85, 1038.4, 1039.08, 1039.92, 1040.9, 1042.03, 1043.31, 1044.74, 1046.32, 1048.06, 1049.95, 1053.29, 1061.19, 1075.87, 1090.38, 1104.74, 1118.94, 1132.98, 1146.86, 1160.58, 1174.14, 1187.54, 1200.78, 1214.08, 1227.89, 1241.43, 1254.71, 1267.74, 1280.5, 1293.0, 1305.24, 1317.22, 1328.94, 1340.4, 1349.69, 1335.47, 1321.27, 1307.08, 1292.9, 1278.73, 1264.57, 1250.42, 1236.28, 1222.15, 1208.04, 1193.93, 1170.36, 1145.56, 1120.49, 1095.13, 1069.5, 1043.58, 1017.38, 990.9, 964.14, 937.09, 909.76, 885.11, 861.59, 838.23, 815.02, 791.96, 769.06, 746.31, 723.71, 701.27, 678.98, 656.84, 643.93, 637.96, 632.02, 626.1, 620.2, 614.31, 608.45, 602.6, 596.78, 590.98, 585.19, 581.1, 579.54, 577.93, 576.29, 574.6, 572.87, 571.1, 569.29, 567.43, 565.53, 563.59, 561.72, 560.21, 558.66, 557.09, 555.49, 553.85, 552.19, 550.5, 548.78, 547.03, 545.25, 543.47, 542.36, 541.29, 540.24, 539.23, 538.25, 537.29, 536.37, 535.47, 534.61, 533.78, 532.98, 532.53, 532.13, 531.71, 531.29, 530.84, 530.39, 529.92, 529.44, 528.94, 528.43, 527.91, 528.08, 528.59, 529.09, 529.59, 530.1, 530.6, 531.11, 531.61, 532.12, 532.62, 533.13, 537.51, 545.92, 554.42, 563.0, 571.67, 580.43, 589.28, 598.21, 607.23, 616.34, 625.53, 632.4, 633.65, 634.54, 635.07, 635.24, 635.05, 634.49, 633.57, 632.29, 630.65, 628.64, 626.2, 622.94, 619.37, 615.5, 611.31, 606.81, 602.0, 596.88, 591.45, 585.71, 579.65, 573.27, 561.08, 548.92, 537.27, 526.12, 515.47, 1064.88, 1068.62, 1072.22, 1075.68, 1078.99, 1082.16, 1085.19, 1088.07, 1090.81, 1093.4, 1095.85, 1101.88, 1115.54, 1129.05, 1142.39, 1155.58, 1168.61, 1181.48, 1194.2, 1206.75, 1219.14, 1231.37, 1243.5, 1255.69, 1267.63, 1279.31, 1290.73, 1301.89, 1312.79, 1323.42, 1333.8, 1343.91, 1353.77, 1363.35, 1349.36, 1335.21, 1321.07, 1306.95, 1292.84, 1278.73, 1264.64, 1250.56, 1236.49, 1222.43, 1208.38, 1184.3, 1157.83, 1131.08, 1104.05, 1076.74, 1049.15, 1021.28, 993.12, 964.69, 935.97, 906.96, 881.99, 859.37, 836.91, 814.6, 792.43, 770.43, 748.57, 726.87, 705.32, 683.93, 662.68, 648.92, 643.07, 637.24, 631.43, 625.64, 619.87, 614.12, 608.39, 602.68, 596.98, 591.31, 586.85, 585.05, 583.2, 581.32, 579.39, 577.42, 575.41, 573.35, 571.25, 569.12, 566.93, 564.79, 563.1, 561.39, 559.65, 557.88, 556.08, 554.25, 552.4, 550.51, 548.59, 546.64, 544.66, 543.68, 542.77, 541.89, 541.05, 540.23, 539.44, 538.69, 537.96, 537.26, 536.6, 535.97, 535.47, 534.99, 534.5, 534.0, 533.48, 532.95, 532.41, 531.85, 531.28, 530.7, 530.1, 530.15, 530.65, 531.15, 531.66, 532.16, 532.67, 533.17, 533.68, 534.18, 534.69, 535.19, 539.04, 547.92, 556.88, 565.92, 575.06, 584.28, 593.59, 602.98, 612.46, 622.03, 631.69, 639.24, 638.58, 637.56, 636.18, 634.44, 632.33, 629.87, 627.04, 623.85, 620.3, 616.38, 612.09, 607.22, 602.04, 596.55, 590.75, 584.64, 578.22, 571.49, 564.45, 557.09, 549.42, 541.43, 531.74, 522.2, 513.16, 504.62, 496.59, 1116.92, 1119.73, 1122.4, 1124.92, 1127.3, 1129.54, 1131.63, 1133.58, 1135.38, 1137.04, 1138.56, 1142.72, 1155.37, 1167.87, 1180.21, 1192.39, 1204.41, 1216.28, 1227.98, 1239.52, 1250.91, 1262.13, 1273.17, 1283.76, 1294.1, 1304.17, 1313.98, 1323.54, 1332.83, 1341.87, 1350.64, 1359.15, 1367.4, 1375.38, 1363.23, 1349.14, 1335.06, 1320.99, 1306.94, 1292.89, 1278.86, 1264.83, 1250.82, 1236.82, 1222.83, 1198.52, 1170.37, 1141.95, 1113.25, 1084.27, 1055.0, 1025.46, 995.63, 965.52, 935.12, 904.45, 878.73, 857.0, 835.43, 814.02, 792.75, 771.64, 750.68, 729.87, 709.22, 688.72, 668.38, 653.9, 648.16, 642.44, 636.75, 631.07, 625.41, 619.77, 614.15, 608.55, 602.97, 597.41, 592.65, 590.61, 588.52, 586.39, 584.22, 582.01, 579.75, 577.46, 575.12, 572.74, 570.32, 567.88, 566.03, 564.15, 562.24, 560.31, 558.34, 556.34, 554.32, 552.26, 550.18, 548.06, 545.92, 544.96, 544.22, 543.51, 542.83, 542.18, 541.56, 540.97, 540.42, 539.89, 539.39, 538.92, 538.41, 537.86, 537.3, 536.72, 536.13, 535.52, 534.91, 534.28, 533.63, 532.98, 532.3, 532.21, 532.72, 533.22, 533.72, 534.23, 534.73, 535.24, 535.74, 536.25, 536.75, 537.26, 540.49, 549.82, 559.25, 568.76, 578.35, 588.04, 597.81, 607.66, 617.61, 627.64, 637.76, 646.44, 643.87, 640.94, 637.65, 634.0, 629.98, 625.61, 620.87, 615.77, 610.31, 604.48, 598.29, 591.81, 585.02, 577.91, 570.5, 562.78, 554.75, 546.41, 537.75, 528.78, 519.5, 509.91, 501.9, 494.97, 488.54, 482.62, 477.19, 1169.11, 1170.99, 1172.72, 1174.31, 1175.76, 1177.06, 1178.22, 1179.24, 1180.11, 1180.84, 1181.42, 1183.72, 1195.37, 1206.86, 1218.19, 1229.36, 1240.37, 1251.23, 1261.92, 1272.46, 1282.84, 1293.05, 1303.11, 1312.09, 1320.82, 1329.29, 1337.5, 1345.45, 1353.14, 1360.57, 1367.74, 1374.65, 1381.29, 1387.67, 1377.09, 1363.06, 1349.04, 1335.03, 1321.03, 1307.04, 1293.06, 1279.1, 1265.14, 1251.2, 1237.26, 1213.02, 1183.2, 1153.1, 1122.73, 1092.07, 1061.13, 1029.92, 998.41, 966.63, 934.56, 902.21, 875.31, 854.48, 833.81, 813.28, 792.92, 772.7, 752.63, 732.72, 712.97, 693.36, 673.91, 658.86, 653.23, 647.63, 642.04, 636.47, 630.93, 625.4, 619.89, 614.41, 608.94, 603.49, 598.49, 596.2, 593.88, 591.51, 589.09, 586.64, 584.14, 581.61, 579.03, 576.41, 573.74, 571.04, 568.99, 566.94, 564.87, 562.76, 560.63, 558.46, 556.27, 554.05, 551.8, 549.52, 547.21, 546.22, 545.64, 545.1, 544.59, 544.11, 543.65, 543.23, 542.84, 542.48, 542.15, 541.85, 541.37, 540.75, 540.11, 539.46, 538.79, 538.11, 537.42, 536.72, 536.0, 535.27, 534.52, 534.28, 534.78, 535.29, 535.79, 536.29, 536.8, 537.3, 537.81, 538.31, 538.82, 539.32, 541.84, 551.64, 561.53, 571.5, 581.56, 591.71, 601.94, 612.26, 622.67, 633.16, 643.75, 654.0, 649.52, 644.68, 639.48, 633.92, 628.0, 621.71, 615.07, 608.06, 600.69, 592.95, 584.85, 576.7, 568.3, 559.59, 550.56, 541.23, 531.59, 521.63, 511.37, 500.79, 490.5, 482.17, 477.67, 475.31, 472.98, 470.66, 468.36, 1221.45, 1222.4, 1223.2, 1223.85, 1224.37, 1224.74, 1224.96, 1225.05, 1224.99, 1224.78, 1224.43, 1224.89, 1235.53, 1246.01, 1256.33, 1266.49, 1276.5, 1286.35, 1296.03, 1305.56, 1314.93, 1324.14, 1333.18, 1340.69, 1347.81, 1354.68, 1361.28, 1367.63, 1373.72, 1379.54, 1385.11, 1390.41, 1395.45, 1400.23, 1390.94, 1376.97, 1363.0, 1349.05, 1335.11, 1321.18, 1307.26, 1293.35, 1279.45, 1265.56, 1251.69, 1227.8, 1196.31, 1164.54, 1132.49, 1100.16, 1067.55, 1034.66, 1001.48, 968.03, 934.29, 900.27, 871.74, 851.81, 832.03, 812.4, 792.93, 773.61, 754.44, 735.42, 716.56, 697.85, 679.3, 663.8, 658.28, 652.79, 647.32, 641.86, 636.43, 631.01, 625.62, 620.24, 614.89, 609.55, 604.37, 601.85, 599.28, 596.66, 594.01, 591.32, 588.58, 585.8, 582.98, 580.11, 577.21, 574.26, 571.97, 569.76, 567.52, 565.25, 562.95, 560.62, 558.26, 555.87, 553.45, 551.0, 548.52, 547.44, 547.04, 546.66, 546.31, 546.0, 545.71, 545.46, 545.24, 545.04, 544.88, 544.75, 544.35, 543.65, 542.93, 542.21, 541.47, 540.72, 539.95, 539.17, 538.38, 537.57, 536.75, 536.34, 536.85, 537.35, 537.85, 538.36, 538.86, 539.37, 539.87, 540.38, 540.88, 541.39, 543.12, 553.38, 563.73, 574.16, 584.68, 595.29, 605.99, 616.77, 627.64, 638.6, 649.65, 660.78, 655.53, 648.78, 641.67, 634.2, 626.37, 618.18, 609.62, 600.71, 591.43, 581.78, 572.36, 563.58, 555.02, 546.52, 538.08, 529.7, 521.39, 513.13, 504.93, 496.79, 488.72, 480.7, 475.83, 473.57, 471.33, 469.1, 466.89, 1273.94, 1273.95, 1273.82, 1273.54, 1273.12, 1272.56, 1271.85, 1271.0, 1270.01, 1268.87, 1267.59, 1266.22, 1275.85, 1285.32, 1294.63, 1303.79, 1312.79, 1321.63, 1330.3, 1338.82, 1347.18, 1355.38, 1363.42, 1369.54, 1375.06, 1380.32, 1385.33, 1390.07, 1394.55, 1398.77, 1402.73, 1406.43, 1409.87, 1413.05, 1404.78, 1390.87, 1376.96, 1363.07, 1349.18, 1335.31, 1321.44, 1307.59, 1293.75, 1279.92, 1266.1, 1242.86, 1209.69, 1176.25, 1142.53, 1108.53, 1074.25, 1039.68, 1004.84, 969.71, 934.3, 898.6, 868.01, 848.98, 830.1, 811.36, 792.79, 774.36, 756.09, 737.97, 720.0, 702.18, 684.52, 668.72, 663.32, 657.93, 652.57, 647.23, 641.91, 636.6, 631.32, 626.06, 620.82, 615.59, 610.39, 607.53, 604.72, 601.86, 598.97, 596.03, 593.05, 590.03, 586.97, 583.87, 580.72, 577.53, 574.99, 572.61, 570.2, 567.76, 565.29, 562.79, 560.27, 557.71, 555.13, 552.51, 549.87, 548.64, 548.4, 548.19, 548.01, 547.86, 547.74, 547.66, 547.6, 547.57, 547.58, 547.61, 547.33, 546.56, 545.77, 544.97, 544.16, 543.33, 542.49, 541.64, 540.77, 539.89, 539.0, 538.41, 538.91, 539.42, 539.92, 540.42, 540.93, 541.43, 541.94, 542.44, 542.95, 543.45, 544.3, 555.03, 565.84, 576.74, 587.72, 598.79, 609.95, 621.2, 632.53, 643.95, 655.46, 666.53, 660.36, 651.06, 641.77, 632.48, 623.19, 613.9, 604.62, 595.34, 586.07, 576.8, 567.53, 558.89, 550.65, 542.47, 534.35, 526.28, 518.28, 510.34, 502.46, 494.64, 486.88, 479.18, 473.98, 471.81, 469.66, 467.53, 465.41, 1326.57, 1325.65, 1324.59, 1323.38, 1322.03, 1320.53, 1318.89, 1317.11, 1315.18, 1313.11, 1310.9, 1308.54, 1316.33, 1324.79, 1333.1, 1341.25, 1349.24, 1357.07, 1364.74, 1372.25, 1379.6, 1386.8, 1393.83, 1398.66, 1402.57, 1406.23, 1409.63, 1412.77, 1415.65, 1418.27, 1420.63, 1422.72, 1424.56, 1426.13, 1418.62, 1404.76, 1390.91, 1377.07, 1363.24, 1349.43, 1335.62, 1321.83, 1308.04, 1294.27, 1280.51, 1258.2, 1223.36, 1188.25, 1152.85, 1117.18, 1081.22, 1044.99, 1008.47, 971.67, 934.59, 897.22, 864.14, 846.0, 828.01, 810.18, 792.49, 774.96, 757.58, 740.36, 723.28, 706.37, 689.6, 673.62, 668.33, 663.06, 657.81, 652.58, 647.37, 642.18, 637.01, 631.86, 626.73, 621.61, 616.52, 613.25, 610.2, 607.11, 603.97, 600.79, 597.57, 594.31, 591.01, 587.66, 584.27, 580.84, 578.03, 575.49, 572.91, 570.3, 567.67, 565.0, 562.31, 559.59, 556.83, 554.05, 551.24, 549.8, 549.73, 549.69, 549.68, 549.7, 549.75, 549.83, 549.94, 550.08, 550.25, 550.45, 550.33, 549.48, 548.62, 547.75, 546.86, 545.96, 545.05, 544.12, 543.18, 542.22, 541.26, 540.47, 540.98, 541.48, 541.99, 542.49, 542.99, 543.5, 544.0, 544.51, 545.01, 545.52, 546.03, 556.49, 567.43, 578.24, 588.91, 599.45, 609.86, 620.12, 630.26, 640.25, 650.11, 659.83, 655.33, 646.05, 636.77, 627.5, 618.24, 608.97, 599.71, 590.45, 581.19, 571.94, 562.69, 554.15, 546.22, 538.36, 530.55, 522.8, 515.12, 507.49, 499.93, 492.42, 484.97, 477.59, 472.11, 470.03, 467.98, 465.94, 463.91, 1379.36, 1377.5, 1375.5, 1373.36, 1371.08, 1368.65, 1366.08, 1363.36, 1360.51, 1357.5, 1354.36, 1351.07, 1356.97, 1364.43, 1371.73, 1378.87, 1385.85, 1392.67, 1399.34, 1405.84, 1412.19, 1418.37, 1424.4, 1428.03, 1430.34, 1432.4, 1434.2, 1435.73, 1437.01, 1438.03, 1438.78, 1439.28, 1439.51, 1439.48, 1432.44, 1418.64, 1404.85, 1391.07, 1377.3, 1363.54, 1349.79, 1336.05, 1322.32, 1308.61, 1294.9, 1273.82, 1237.31, 1200.52, 1163.46, 1126.11, 1088.48, 1050.58, 1012.39, 973.91, 935.16, 896.12, 860.11, 842.87, 825.77, 808.83, 792.04, 775.41, 758.93, 742.59, 726.42, 710.39, 694.52, 678.81, 673.32, 668.16, 663.02, 657.91, 652.81, 647.73, 642.67, 637.63, 632.61, 627.62, 622.64, 619.02, 615.73, 612.39, 609.01, 605.59, 602.13, 598.63, 595.08, 591.5, 587.87, 584.19, 581.11, 578.39, 575.65, 572.88, 570.07, 567.24, 564.38, 561.49, 558.57, 555.62, 552.64, 550.93, 551.03, 551.16, 551.31, 551.5, 551.72, 551.96, 552.24, 552.55, 552.89, 553.25, 553.35, 552.42, 551.49, 550.54, 549.58, 548.6, 547.61, 546.61, 545.6, 544.57, 543.53, 542.56, 543.2, 543.82, 544.42, 544.99, 545.53, 546.06, 546.56, 547.04, 547.49, 547.92, 548.32, 557.17, 567.39, 577.47, 587.42, 597.24, 606.92, 616.46, 625.87, 635.14, 644.28, 653.27, 650.29, 641.03, 631.78, 622.53, 613.28, 604.03, 594.79, 585.55, 576.32, 567.08, 557.85, 549.35, 541.74, 534.18, 526.69, 519.26, 511.89, 504.58, 497.33, 490.14, 483.01, 475.94, 470.22, 468.24, 466.27, 464.33, 462.4, 1431.25, 1429.5, 1426.57, 1423.49, 1420.28, 1416.92, 1413.41, 1409.77, 1405.98, 1402.04, 1397.96, 1393.74, 1397.77, 1404.22, 1410.51, 1416.65, 1422.62, 1428.44, 1434.1, 1439.59, 1444.93, 1450.11, 1455.13, 1457.67, 1458.38, 1458.83, 1459.02, 1458.96, 1458.63, 1458.05, 1457.2, 1456.09, 1454.72, 1453.09, 1446.25, 1432.51, 1418.77, 1405.05, 1391.34, 1377.64, 1363.95, 1350.27, 1336.6, 1322.94, 1309.29, 1289.72, 1251.54, 1213.08, 1174.34, 1135.32, 1096.03, 1056.45, 1016.58, 976.44, 936.02, 895.31, 855.93, 839.58, 823.39, 807.34, 791.45, 775.7, 760.12, 744.68, 729.4, 714.27, 699.29, 684.47, 678.29, 673.25, 668.22, 663.22, 658.23, 653.26, 648.32, 643.39, 638.48, 633.6, 628.73, 624.83, 621.29, 617.72, 614.1, 610.44, 606.73, 602.99, 599.2, 595.37, 591.5, 587.59, 584.21, 581.33, 578.42, 575.48, 572.51, 569.51, 566.48, 563.43, 560.34, 557.22, 554.08, 552.04, 552.3, 552.6, 552.92, 553.27, 553.66, 554.07, 554.51, 554.99, 555.49, 556.03, 556.48, 556.0, 555.43, 554.78, 554.04, 553.21, 552.29, 551.29, 550.2, 549.02, 547.76, 546.4, 546.82, 547.31, 547.77, 548.21, 548.63, 549.03, 549.4, 549.74, 550.07, 550.37, 550.64, 557.98, 567.48, 576.84, 586.07, 595.16, 604.11, 612.93, 621.62, 630.16, 638.58, 646.85, 645.25, 636.02, 626.78, 617.55, 608.32, 599.09, 589.87, 580.65, 571.43, 562.22, 553.01, 544.48, 537.19, 529.95, 522.78, 515.66, 508.61, 501.61, 494.68, 487.8, 480.99, 474.23, 468.31, 466.42, 464.55, 462.7, 460.86, 1482.59, 1481.64, 1477.78, 1473.77, 1469.62, 1465.33, 1460.9, 1456.32, 1451.6, 1446.73, 1441.72, 1436.56, 1438.74, 1444.18, 1449.47, 1454.59, 1459.56, 1464.37, 1469.02, 1473.51, 1477.84, 1482.01, 1486.03, 1487.56, 1486.67, 1485.52, 1484.11, 1482.45, 1480.52, 1478.33, 1475.88, 1473.17, 1470.2, 1466.97, 1460.06, 1446.37, 1432.69, 1419.03, 1405.37, 1391.73, 1378.09, 1364.47, 1350.86, 1337.26, 1323.67, 1305.9, 1266.05, 1225.92, 1185.51, 1144.82, 1103.85, 1062.6, 1021.07, 979.25, 937.16, 894.78, 852.11, 836.15, 820.84, 805.69, 790.69, 775.85, 761.15, 746.61, 732.22, 717.99, 703.91, 689.98, 683.25, 678.31, 673.4, 668.51, 663.63, 658.78, 653.94, 649.13, 644.34, 639.56, 634.81, 630.68, 626.9, 623.09, 619.22, 615.32, 611.38, 607.39, 603.37, 599.3, 595.18, 591.03, 587.34, 584.29, 581.22, 578.11, 574.98, 571.81, 568.62, 565.39, 562.14, 558.86, 555.54, 553.23, 554.0, 554.84, 555.76, 556.76, 557.83, 558.98, 560.21, 561.52, 562.91, 564.37, 565.78, 564.82, 563.77, 562.63, 561.41, 560.1, 558.7, 557.22, 555.64, 553.98, 552.23, 550.4, 550.46, 550.82, 551.15, 551.47, 551.75, 552.02, 552.26, 552.48, 552.67, 552.84, 552.98, 558.93, 567.71, 576.34, 584.85, 593.21, 601.44, 609.54, 617.5, 625.32, 633.01, 640.56, 640.21, 630.99, 621.78, 612.57, 603.36, 594.15, 584.95, 575.75, 566.55, 557.35, 548.16, 539.56, 532.58, 525.66, 518.8, 512.0, 505.26, 498.58, 491.96, 485.4, 478.9, 472.47, 466.39, 464.59, 462.81, 461.05, 459.31, 1533.57, 1533.93, 1529.14, 1524.2, 1519.12, 1513.9, 1508.53, 1503.02, 1497.36, 1491.57, 1485.62, 1479.54, 1479.87, 1484.3, 1488.58, 1492.7, 1496.66, 1500.46, 1504.11, 1507.59, 1510.92, 1514.08, 1517.08, 1517.72, 1515.23, 1512.48, 1509.47, 1506.2, 1502.67, 1498.88, 1494.83, 1490.52, 1485.94, 1481.11, 1473.85, 1460.22, 1446.6, 1432.99, 1419.4, 1405.81, 1392.23, 1378.67, 1365.11, 1351.57, 1338.03, 1322.36, 1280.83, 1239.03, 1196.95, 1154.59, 1111.95, 1069.03, 1025.83, 982.34, 938.58, 894.53, 850.2, 832.56, 818.15, 803.89, 789.79, 775.84, 762.04, 748.39, 734.9, 721.55, 708.37, 695.33, 688.18, 683.36, 678.56, 673.78, 669.02, 664.27, 659.55, 654.85, 650.17, 645.5, 640.86, 636.57, 632.56, 628.5, 624.39, 620.25, 616.07, 611.84, 607.57, 603.26, 598.91, 594.51, 590.68, 587.72, 584.73, 581.73, 578.7, 575.66, 572.6, 569.52, 566.42, 563.3, 560.16, 557.67, 558.86, 560.13, 561.48, 562.9, 564.4, 565.98, 567.64, 569.37, 571.19, 573.08, 575.04, 573.73, 572.2, 570.58, 568.87, 567.08, 565.2, 563.23, 561.17, 559.03, 556.8, 554.48, 554.13, 554.35, 554.56, 554.74, 554.9, 555.03, 555.14, 555.23, 555.29, 555.33, 555.35, 560.02, 568.07, 575.98, 583.76, 591.4, 598.91, 606.28, 613.52, 620.62, 627.58, 634.41, 635.17, 625.97, 616.77, 607.58, 598.39, 589.2, 580.02, 570.84, 561.66, 552.48, 543.31, 534.57, 527.91, 521.31, 514.76, 508.28, 501.85, 495.49, 489.19, 482.94, 476.76, 470.64, 464.57, 462.74, 461.06, 459.39, 457.74, 1584.17, 1586.37, 1580.64, 1574.77, 1568.76, 1562.61, 1556.31, 1549.87, 1543.28, 1536.55, 1529.68, 1522.66, 1521.16, 1524.58, 1527.85, 1530.97, 1533.92, 1536.72, 1539.35, 1541.83, 1544.15, 1546.31, 1548.31, 1548.14, 1544.05, 1539.69, 1535.08, 1530.21, 1525.08, 1519.69, 1514.04, 1508.12, 1501.95, 1495.51, 1487.64, 1474.07, 1460.5, 1446.95, 1433.41, 1419.88, 1406.36, 1392.85, 1379.36, 1365.87, 1352.39, 1338.93, 1295.9, 1252.43, 1208.68, 1164.65, 1120.34, 1075.75, 1030.87, 985.72, 940.28, 894.56, 848.56, 828.82, 815.31, 801.94, 788.73, 775.67, 762.77, 750.02, 737.42, 724.97, 712.67, 700.53, 693.1, 688.39, 683.7, 679.03, 674.38, 669.75, 665.14, 660.55, 655.98, 651.43, 646.8, 642.03, 637.51, 633.06, 628.69, 624.4, 620.18, 616.03, 611.96, 607.97, 604.05, 600.2, 596.61, 593.53, 590.44, 587.32, 584.19, 581.04, 577.86, 574.67, 571.46, 568.23, 564.98, 562.03, 563.65, 565.34, 567.12, 568.97, 570.9, 572.9, 574.98, 577.15, 579.39, 581.7, 584.1, 582.72, 580.71, 578.61, 576.42, 574.14, 571.78, 569.33, 566.79, 564.17, 561.46, 558.66, 557.81, 557.91, 557.99, 558.04, 558.07, 558.07, 558.05, 558.01, 557.95, 557.85, 557.74, 561.24, 568.56, 575.75, 582.81, 589.73, 596.51, 603.16, 609.67, 616.05, 622.29, 628.4, 630.12, 620.94, 611.76, 602.59, 593.42, 584.25, 575.09, 565.92, 556.77, 547.61, 538.46, 529.53, 523.18, 516.89, 510.66, 504.5, 498.39, 492.34, 486.35, 480.42, 474.55, 468.74, 463.0, 460.88, 459.28, 457.71, 456.15, 1634.41, 1638.95, 1632.3, 1625.5, 1618.55, 1611.47, 1604.24, 1596.86, 1589.35, 1581.69, 1573.88, 1565.93, 1562.61, 1565.03, 1567.29, 1569.4, 1571.35, 1573.13, 1574.77, 1576.24, 1577.55, 1578.7, 1579.69, 1578.82, 1573.12, 1567.17, 1560.95, 1554.48, 1547.75, 1540.76, 1533.51, 1525.99, 1518.22, 1510.18, 1501.41, 1487.9, 1474.39, 1460.9, 1447.42, 1433.94, 1420.48, 1407.03, 1393.59, 1380.16, 1366.74, 1353.33, 1311.25, 1266.11, 1220.69, 1174.99, 1129.0, 1082.74, 1036.2, 989.38, 942.27, 894.88, 847.21, 824.93, 812.31, 799.84, 787.52, 775.36, 763.35, 751.49, 739.78, 728.23, 716.83, 704.98, 696.61, 691.37, 686.2, 681.12, 676.11, 671.19, 666.34, 661.58, 656.89, 652.29, 647.77, 643.42, 639.33, 635.31, 631.37, 627.5, 623.71, 619.99, 616.35, 612.78, 609.29, 605.88, 602.56, 599.37, 596.16, 592.94, 589.69, 586.43, 583.14, 579.84, 576.52, 573.18, 569.82, 566.44, 568.35, 570.48, 572.68, 574.96, 577.31, 579.74, 582.25, 584.84, 587.51, 590.25, 593.07, 591.8, 589.31, 586.72, 584.05, 581.3, 578.45, 575.52, 572.5, 569.39, 566.2, 562.92, 561.53, 561.5, 561.44, 561.36, 561.26, 561.14, 560.99, 560.82, 560.62, 560.4, 560.16, 562.59, 569.2, 575.66, 581.99, 588.19, 594.25, 600.18, 605.97, 611.62, 617.14, 622.52, 625.07, 615.91, 606.75, 597.59, 588.44, 579.3, 570.15, 561.01, 551.87, 542.73, 533.6, 524.47, 518.39, 512.42, 506.51, 500.65, 494.86, 489.13, 483.45, 477.84, 472.29, 466.79, 461.36, 458.99, 457.49, 456.01, 454.54, 1684.28, 1691.69, 1684.1, 1676.37, 1668.49, 1660.47, 1652.31, 1644.01, 1635.56, 1626.97, 1618.23, 1609.35, 1604.22, 1605.63, 1606.89, 1607.99, 1608.93, 1609.71, 1610.34, 1610.8, 1611.11, 1611.26, 1611.24, 1609.76, 1602.46, 1594.91, 1587.09, 1579.02, 1570.69, 1562.09, 1553.24, 1544.13, 1534.75, 1525.11, 1515.18, 1501.72, 1488.27, 1474.84, 1461.41, 1448.0, 1434.59, 1421.2, 1407.81, 1394.44, 1381.08, 1367.73, 1326.88, 1280.06, 1232.97, 1185.6, 1137.95, 1090.02, 1041.81, 993.31, 944.54, 895.93, 847.51, 822.97, 811.49, 799.87, 788.11, 776.22, 764.19, 752.03, 739.73, 727.29, 714.72, 702.01, 692.85, 688.07, 683.37, 678.75, 674.21, 669.75, 665.37, 661.07, 656.85, 652.71, 648.65, 644.74, 641.08, 637.49, 633.97, 630.53, 627.17, 623.88, 620.66, 617.52, 614.46, 611.47, 608.53, 605.23, 601.91, 598.57, 595.22, 591.84, 588.45, 585.03, 581.6, 578.15, 574.67, 571.18, 572.99, 575.54, 578.16, 580.87, 583.65, 586.51, 589.45, 592.46, 595.55, 598.72, 601.97, 600.97, 597.99, 594.93, 591.78, 588.54, 585.21, 581.8, 578.3, 574.71, 571.03, 567.27, 565.26, 565.1, 564.92, 564.71, 564.48, 564.22, 563.95, 563.64, 563.32, 562.97, 562.59, 564.08, 569.96, 575.7, 581.31, 586.78, 592.12, 597.32, 602.39, 607.32, 612.12, 616.78, 620.01, 610.87, 601.73, 592.6, 583.47, 574.34, 565.21, 556.09, 546.97, 537.85, 528.74, 519.63, 513.54, 507.89, 502.29, 496.75, 491.27, 485.85, 480.5, 475.2, 469.96, 464.78, 459.66, 457.09, 455.68, 454.29, 452.91, 1733.78, 1744.57, 1736.05, 1727.38, 1718.58, 1709.63, 1700.54, 1691.3, 1681.92, 1672.4, 1662.73, 1652.92, 1645.99, 1646.4, 1646.65, 1646.74, 1646.68, 1646.46, 1646.08, 1645.53, 1644.83, 1643.98, 1642.96, 1640.96, 1632.06, 1622.9, 1613.49, 1603.81, 1593.88, 1583.69, 1573.24, 1562.52, 1551.55, 1540.31, 1528.81, 1515.54, 1502.15, 1488.77, 1475.4, 1462.04, 1448.69, 1435.35, 1422.03, 1407.93, 1393.61, 1379.41, 1339.58, 1291.0, 1242.47, 1193.99, 1145.57, 1097.19, 1048.87, 1000.6, 952.38, 904.21, 856.09, 828.12, 815.84, 803.42, 790.87, 778.18, 765.35, 752.39, 739.29, 726.06, 712.68, 699.18, 689.01, 684.69, 680.46, 676.3, 672.22, 668.23, 664.32, 660.48, 656.73, 653.05, 649.46, 645.99, 642.75, 639.59, 636.5, 633.49, 630.55, 627.69, 624.9, 622.19, 619.55, 616.99, 614.51, 611.1, 607.68, 604.23, 600.76, 597.27, 593.77, 590.24, 586.7, 583.13, 579.55, 575.95, 577.54, 580.52, 583.57, 586.71, 589.91, 593.2, 596.56, 600.0, 603.52, 607.12, 610.79, 610.23, 606.77, 603.22, 599.59, 595.87, 592.06, 588.16, 584.18, 580.11, 575.95, 571.71, 569.02, 568.73, 568.42, 568.08, 567.72, 567.34, 566.93, 566.5, 566.04, 565.56, 565.06, 565.71, 570.86, 575.88, 580.77, 585.52, 590.13, 594.61, 598.95, 603.16, 607.23, 611.17, 614.95, 605.83, 596.71, 587.6, 578.48, 569.37, 560.27, 551.16, 542.06, 532.97, 523.87, 514.78, 508.63, 503.29, 498.01, 492.79, 487.62, 482.52, 477.48, 472.49, 467.57, 462.71, 457.91, 455.17, 453.85, 452.55, 451.27, 1782.91, 1797.59, 1788.14, 1778.55, 1768.81, 1758.93, 1748.91, 1738.74, 1728.43, 1717.98, 1707.38, 1696.64, 1687.92, 1687.33, 1686.57, 1685.66, 1684.59, 1683.36, 1681.97, 1680.43, 1678.72, 1676.86, 1674.83, 1672.42, 1661.92, 1651.16, 1640.15, 1628.87, 1617.34, 1605.55, 1593.49, 1579.4, 1565.02, 1550.65, 1536.3, 1521.66, 1507.08, 1492.63, 1478.31, 1464.11, 1450.04, 1436.09, 1422.28, 1408.58, 1395.02, 1381.59, 1345.09, 1296.81, 1248.58, 1200.41, 1152.29, 1104.21, 1056.19, 1008.22, 960.3, 912.44, 864.62, 833.4, 820.32, 807.11, 793.76, 780.27, 766.64, 752.89, 738.99, 724.96, 710.79, 696.48, 685.08, 681.23, 677.46, 673.77, 670.16, 666.63, 663.18, 659.81, 656.52, 653.31, 650.19, 647.16, 644.35, 641.61, 638.95, 636.37, 633.86, 631.42, 629.06, 626.78, 624.57, 622.44, 620.38, 617.0, 613.46, 609.9, 606.32, 602.73, 599.11, 595.47, 591.82, 588.14, 584.45, 580.74, 582.02, 585.43, 588.91, 592.46, 596.1, 599.81, 603.6, 607.47, 611.41, 615.44, 619.54, 619.57, 615.63, 611.6, 607.49, 603.28, 598.99, 594.62, 590.15, 585.6, 580.96, 576.23, 572.81, 572.39, 571.94, 571.48, 570.99, 570.47, 569.93, 569.37, 568.79, 568.18, 567.54, 567.47, 571.9, 576.19, 580.36, 584.38, 588.27, 592.03, 595.65, 599.14, 602.48, 605.7, 608.77, 600.79, 591.69, 582.59, 573.5, 564.41, 555.32, 546.24, 537.16, 528.08, 519.01, 509.93, 503.67, 498.64, 493.67, 488.76, 483.91, 479.13, 474.4, 469.73, 465.12, 460.58, 456.09, 453.23, 452.0, 450.8, 449.61, 1831.68, 1850.77, 1840.38, 1829.86, 1819.19, 1808.38, 1797.43, 1786.33, 1775.09, 1763.71, 1752.18, 1740.51, 1730.02, 1728.42, 1726.66, 1724.74, 1722.66, 1720.43, 1717.3, 1711.54, 1705.65, 1699.65, 1693.52, 1687.27, 1673.29, 1658.91, 1644.53, 1630.17, 1615.83, 1601.5, 1587.18, 1572.88, 1558.6, 1544.32, 1530.07, 1516.07, 1502.26, 1488.57, 1475.01, 1461.58, 1448.27, 1435.09, 1422.03, 1409.11, 1396.31, 1383.64, 1350.54, 1302.57, 1254.65, 1206.77, 1158.95, 1111.18, 1063.46, 1015.8, 968.18, 920.62, 873.1, 838.82, 824.95, 810.93, 796.78, 782.5, 768.07, 753.52, 738.82, 723.99, 709.02, 693.92, 681.08, 677.69, 674.39, 671.16, 668.02, 664.95, 661.97, 659.06, 656.24, 653.49, 650.83, 648.25, 645.87, 643.56, 641.33, 639.17, 637.09, 635.08, 633.15, 631.29, 629.51, 627.8, 626.18, 622.92, 619.27, 615.59, 611.91, 608.2, 604.47, 600.72, 596.96, 593.17, 589.37, 585.54, 586.43, 590.26, 594.16, 598.15, 602.21, 606.35, 610.56, 614.86, 619.23, 623.68, 628.2, 629.01, 624.58, 620.07, 615.47, 610.79, 606.02, 601.16, 596.21, 591.18, 586.06, 580.85, 576.62, 576.07, 575.49, 574.9, 574.28, 573.63, 572.96, 572.27, 571.56, 570.82, 570.06, 569.36, 573.07, 576.64, 580.08, 583.38, 586.55, 589.59, 592.48, 595.25, 597.87, 600.36, 602.72, 595.74, 586.66, 577.58, 568.51, 559.44, 550.37, 541.31, 532.25, 523.19, 514.13, 505.08, 498.64, 493.92, 489.27, 484.68, 480.15, 475.67, 471.26, 466.91, 462.61, 458.38, 454.21, 451.27, 450.14, 449.03, 447.93, 1880.08, 1904.09, 1892.77, 1881.32, 1869.72, 1857.98, 1843.5, 1828.15, 1812.66, 1797.02, 1781.24, 1765.3, 1749.72, 1743.93, 1738.03, 1732.0, 1725.85, 1719.59, 1713.2, 1706.69, 1700.06, 1693.31, 1686.44, 1679.44, 1666.1, 1651.81, 1637.53, 1623.26, 1609.01, 1594.78, 1580.56, 1566.35, 1552.16, 1537.98, 1523.82, 1510.35, 1497.3, 1484.38, 1471.59, 1458.92, 1446.37, 1433.96, 1421.67, 1409.5, 1397.47, 1385.56, 1355.95, 1308.28, 1260.66, 1213.09, 1165.57, 1118.1, 1070.68, 1023.32, 976.0, 928.74, 881.53, 844.38, 829.7, 814.89, 799.94, 784.86, 769.64, 754.29, 738.79, 723.17, 707.4, 691.5, 676.99, 674.07, 671.23, 668.47, 665.79, 663.19, 660.67, 658.23, 655.87, 653.59, 651.4, 649.28, 647.32, 645.44, 643.63, 641.9, 640.25, 638.67, 637.16, 635.73, 634.38, 633.1, 631.9, 628.85, 625.09, 621.31, 617.51, 613.69, 609.85, 605.99, 602.11, 598.22, 594.3, 590.37, 590.75, 595.01, 599.34, 603.75, 608.24, 612.81, 617.45, 622.17, 626.96, 631.84, 636.79, 638.53, 633.62, 628.63, 623.55, 618.38, 613.13, 607.79, 602.36, 596.84, 591.24, 585.55, 580.45, 579.77, 579.07, 578.34, 577.59, 576.81, 576.02, 575.2, 574.35, 573.48, 572.59, 571.67, 574.37, 577.22, 579.94, 582.52, 584.97, 587.28, 589.45, 591.49, 593.4, 595.16, 596.79, 590.69, 581.63, 572.57, 563.52, 554.47, 545.42, 536.37, 527.33, 518.29, 509.26, 500.22, 493.55, 489.15, 484.81, 480.53, 476.32, 472.16, 468.06, 464.02, 460.04, 456.13, 452.27, 449.3, 448.26, 447.24, 446.23, 1905.98, 1929.56, 1914.02, 1898.33, 1882.5, 1866.52, 1850.39, 1834.12, 1817.7, 1801.13, 1784.41, 1767.55, 1750.53, 1743.67, 1737.02, 1730.25, 1723.36, 1716.34, 1709.21, 1701.96, 1694.58, 1687.09, 1679.47, 1671.74, 1658.9, 1644.7, 1630.51, 1616.34, 1602.19, 1588.04, 1573.92, 1559.81, 1545.71, 1531.63, 1517.56, 1504.51, 1492.22, 1480.07, 1468.03, 1456.13, 1444.35, 1432.7, 1421.17, 1409.77, 1398.5, 1387.36, 1361.3, 1313.93, 1266.62, 1219.35, 1172.13, 1124.97, 1077.85, 1030.79, 983.78, 936.82, 889.91, 850.07, 834.6, 818.99, 803.24, 787.36, 771.34, 755.19, 738.9, 722.48, 705.91, 689.21, 672.83, 670.38, 668.0, 665.7, 663.49, 661.35, 659.3, 657.32, 655.43, 653.61, 651.88, 650.23, 648.69, 647.24, 645.86, 644.56, 643.33, 642.18, 641.1, 640.1, 639.17, 638.32, 637.54, 634.81, 630.93, 627.04, 623.13, 619.2, 615.25, 611.28, 607.29, 603.28, 599.26, 595.21, 595.0, 599.69, 604.45, 609.28, 614.2, 619.19, 624.26, 629.4, 634.62, 639.93, 645.3, 648.13, 642.74, 637.27, 631.71, 626.06, 620.33, 614.5, 608.59, 602.6, 596.51, 590.34, 584.3, 583.49, 582.66, 581.81, 580.93, 580.02, 579.1, 578.14, 577.17, 576.17, 575.15, 574.1, 575.81, 577.94, 579.93, 581.79, 583.51, 585.1, 586.56, 587.87, 589.05, 590.1, 591.01, 585.63, 576.59, 567.55, 558.52, 549.49, 540.46, 531.43, 522.41, 513.39, 504.38, 495.36, 488.4, 484.31, 480.29, 476.33, 472.43, 468.58, 464.8, 461.08, 457.41, 453.81, 450.27, 447.3, 446.36, 445.43, 444.52, 1916.75, 1941.25, 1924.78, 1908.16, 1891.4, 1874.49, 1857.44, 1840.23, 1822.88, 1805.38, 1787.74, 1769.94, 1752.0, 1743.52, 1736.13, 1728.61, 1720.98, 1713.22, 1705.35, 1697.35, 1689.23, 1680.99, 1672.63, 1664.15, 1651.67, 1637.57, 1623.48, 1609.4, 1595.34, 1581.29, 1567.26, 1553.25, 1539.24, 1525.25, 1511.28, 1498.54, 1487.02, 1475.62, 1464.36, 1453.21, 1442.2, 1431.31, 1420.55, 1409.91, 1399.4, 1389.02, 1366.61, 1319.54, 1272.53, 1225.56, 1178.65, 1131.78, 1084.97, 1038.21, 991.5, 944.84, 898.24, 855.9, 839.63, 823.22, 806.68, 790.0, 773.18, 756.23, 739.14, 721.92, 704.56, 687.06, 669.43, 666.6, 664.69, 662.85, 661.1, 659.43, 657.84, 656.33, 654.9, 653.55, 652.28, 651.09, 649.99, 648.96, 648.01, 647.14, 646.34, 645.61, 644.96, 644.38, 643.88, 643.46, 643.11, 640.78, 636.8, 632.79, 628.77, 624.73, 620.67, 616.59, 612.49, 608.37, 604.23, 600.08, 599.18, 604.29, 609.47, 614.74, 620.08, 625.49, 630.99, 636.56, 642.21, 647.93, 653.74, 657.82, 651.96, 646.0, 639.96, 633.83, 627.61, 621.31, 614.92, 608.44, 601.87, 595.22, 588.48, 587.24, 586.28, 585.3, 584.29, 583.25, 582.2, 581.12, 580.01, 578.88, 577.73, 576.56, 577.38, 578.79, 580.06, 581.2, 582.2, 583.06, 583.79, 584.39, 584.85, 585.17, 585.36, 580.58, 571.55, 562.54, 553.52, 544.51, 535.5, 526.49, 517.49, 508.49, 499.49, 490.5, 483.19, 479.42, 475.71, 472.06, 468.48, 464.95, 461.48, 458.07, 454.72, 451.44, 448.21, 445.29, 444.44, 443.6, 442.79, 1927.96, 1952.8, 1935.69, 1918.14, 1900.45, 1882.61, 1864.63, 1846.49, 1828.21, 1809.79, 1791.21, 1772.48, 1753.61, 1743.5, 1735.36, 1727.1, 1718.72, 1710.22, 1701.6, 1692.86, 1684.0, 1675.01, 1665.91, 1656.69, 1644.44, 1630.43, 1616.43, 1602.45, 1588.48, 1574.53, 1560.59, 1546.67, 1532.76, 1518.87, 1504.99, 1492.44, 1481.68, 1471.05, 1460.55, 1450.17, 1439.92, 1429.8, 1419.8, 1409.93, 1400.18, 1390.56, 1371.86, 1325.1, 1278.38, 1231.72, 1185.11, 1138.55, 1092.04, 1045.58, 999.17, 952.82, 906.51, 861.86, 844.79, 827.59, 810.24, 792.77, 775.16, 757.41, 739.53, 721.5, 703.35, 685.05, 666.62, 662.74, 661.29, 659.93, 658.64, 657.43, 656.31, 655.26, 654.29, 653.41, 652.6, 651.88, 651.22, 650.62, 650.09, 649.64, 649.27, 648.97, 648.75, 648.6, 648.53, 648.53, 648.61, 646.78, 642.68, 638.57, 634.43, 630.28, 626.11, 621.92, 617.71, 613.48, 609.23, 604.96, 603.28, 608.81, 614.42, 620.11, 625.88, 631.72, 637.64, 643.64, 649.71, 655.87, 662.1, 667.61, 661.25, 654.82, 648.29, 641.68, 634.99, 628.2, 621.33, 614.37, 607.32, 600.19, 592.97, 591.02, 589.93, 588.81, 587.67, 586.51, 585.32, 584.11, 582.88, 581.62, 580.34, 579.04, 579.09, 579.77, 580.32, 580.74, 581.02, 581.16, 581.17, 581.04, 580.78, 580.38, 579.85, 575.51, 566.51, 557.51, 548.52, 539.52, 530.53, 521.55, 512.56, 503.58, 494.61, 485.63, 477.92, 474.47, 471.07, 467.74, 464.47, 461.33, 458.68, 456.0, 453.3, 450.58, 447.83, 445.1, 444.54, 443.97, 443.38, 1939.6, 1961.65, 1946.74, 1928.27, 1909.65, 1890.88, 1871.96, 1852.9, 1833.69, 1814.34, 1794.83, 1775.18, 1755.37, 1743.59, 1734.71, 1725.7, 1716.58, 1707.34, 1697.97, 1688.49, 1678.88, 1669.16, 1659.31, 1649.34, 1637.18, 1623.27, 1609.37, 1595.48, 1581.61, 1567.75, 1553.91, 1540.08, 1526.26, 1512.46, 1498.68, 1486.22, 1476.23, 1466.36, 1456.62, 1447.0, 1437.52, 1428.15, 1418.92, 1409.81, 1400.83, 1391.98, 1377.07, 1330.6, 1284.19, 1237.83, 1191.52, 1145.26, 1099.05, 1052.9, 1006.79, 960.74, 914.74, 868.79, 850.09, 832.09, 813.95, 795.68, 777.27, 758.72, 740.04, 721.22, 702.27, 683.18, 663.95, 658.8, 657.82, 656.92, 656.09, 655.35, 654.69, 654.11, 653.61, 653.18, 652.84, 652.58, 652.37, 652.19, 652.1, 652.07, 652.13, 652.26, 652.46, 652.74, 653.09, 653.52, 654.03, 652.79, 648.58, 644.36, 640.11, 635.85, 631.57, 627.27, 622.95, 618.61, 614.25, 609.87, 607.3, 613.26, 619.3, 625.41, 631.6, 637.87, 644.22, 650.64, 657.14, 663.72, 670.38, 677.11, 670.64, 663.72, 656.72, 649.63, 642.45, 635.18, 627.83, 620.39, 612.86, 605.24, 597.54, 594.81, 593.59, 592.35, 591.08, 589.79, 588.47, 587.13, 585.77, 584.38, 582.97, 581.54, 580.93, 580.89, 580.72, 580.41, 579.97, 579.39, 578.68, 577.83, 576.84, 575.72, 574.47, 570.45, 561.47, 552.49, 543.51, 534.54, 525.57, 517.08, 508.81, 500.57, 492.38, 484.23, 476.86, 474.19, 471.51, 468.8, 466.08, 463.33, 460.56, 457.77, 454.96, 452.13, 449.27, 446.4, 445.6, 444.97, 444.33, 1951.69, 1970.94, 1957.95, 1938.54, 1918.99, 1899.3, 1879.45, 1859.46, 1839.32, 1819.04, 1798.6, 1778.02, 1757.29, 1743.8, 1734.18, 1724.43, 1714.56, 1704.58, 1694.47, 1684.24, 1673.89, 1663.42, 1652.83, 1642.12, 1629.92, 1616.09, 1602.29, 1588.49, 1574.72, 1560.95, 1547.2, 1533.47, 1519.75, 1506.05, 1492.36, 1479.87, 1470.64, 1461.54, 1452.56, 1443.71, 1434.99, 1426.39, 1417.91, 1409.57, 1401.35, 1393.26, 1382.22, 1336.06, 1289.95, 1243.89, 1197.88, 1151.93, 1106.02, 1060.16, 1014.36, 968.61, 922.91, 877.26, 855.53, 836.73, 817.79, 798.72, 779.51, 760.17, 740.69, 721.08, 701.33, 681.44, 661.42, 654.78, 654.26, 653.83, 653.47, 653.19, 652.99, 652.87, 652.84, 652.88, 653.0, 653.21, 653.44, 653.7, 654.03, 654.43, 654.91, 655.47, 656.1, 656.8, 657.58, 658.44, 659.37, 658.82, 654.51, 650.17, 645.81, 641.44, 637.05, 632.63, 628.2, 623.75, 619.28, 614.79, 611.24, 617.63, 624.09, 630.63, 637.25, 643.95, 650.72, 657.57, 664.5, 671.5, 678.58, 685.74, 680.12, 672.72, 665.23, 657.66, 650.0, 642.25, 634.42, 626.49, 618.49, 610.39, 602.2, 598.64, 597.28, 595.91, 594.51, 593.09, 591.65, 590.18, 588.68, 587.17, 585.63, 584.06, 582.91, 582.15, 581.25, 580.22, 579.06, 577.76, 576.4, 575.11, 573.83, 572.57, 571.32, 568.28, 559.97, 551.7, 543.48, 535.3, 527.16, 519.06, 511.01, 503.0, 495.03, 487.1, 479.43, 476.65, 473.86, 471.04, 468.2, 465.34, 462.47, 459.56, 456.64, 453.7, 450.74, 447.75, 446.67, 445.98, 445.29, 1964.22, 1980.68, 1969.3, 1948.96, 1928.48, 1907.86, 1887.09, 1866.17, 1845.1, 1823.88, 1802.52, 1781.01, 1759.35, 1744.14, 1733.77, 1723.28, 1712.66, 1701.93, 1691.08, 1680.11, 1669.02, 1657.81, 1646.47, 1635.02, 1622.63, 1608.9, 1595.19, 1581.49, 1567.81, 1554.14, 1540.49, 1526.85, 1513.22, 1499.61, 1486.02, 1473.4, 1464.93, 1456.59, 1448.38, 1440.29, 1432.33, 1424.49, 1416.78, 1409.2, 1401.75, 1394.42, 1387.22, 1341.46, 1295.66, 1249.9, 1204.19, 1158.54, 1112.93, 1067.38, 1021.88, 976.43, 931.03, 885.68, 861.1, 841.5, 821.77, 801.9, 781.9, 761.76, 741.48, 721.07, 700.53, 679.84, 659.02, 650.68, 650.63, 650.66, 650.76, 650.95, 651.21, 651.56, 651.99, 652.49, 653.08, 653.75, 654.44, 655.12, 655.88, 656.71, 657.62, 658.6, 659.66, 660.79, 662.0, 663.28, 664.64, 664.88, 660.45, 656.0, 651.54, 647.05, 642.55, 638.02, 633.48, 628.92, 624.34, 619.74, 615.12, 621.93, 628.82, 635.78, 642.82, 649.95, 657.14, 664.42, 671.77, 679.2, 686.71, 694.29, 689.68, 681.8, 673.83, 665.77, 657.63, 649.41, 641.09, 632.69, 624.2, 615.62, 606.96, 602.48, 601.0, 599.5, 597.97, 596.42, 594.84, 593.11, 591.07, 589.07, 587.12, 585.2, 583.54, 582.24, 580.95, 579.68, 578.42, 577.18, 575.95, 574.73, 573.53, 572.34, 571.16, 568.88, 560.79, 552.75, 544.75, 536.79, 528.88, 521.0, 513.17, 505.38, 497.64, 489.93, 482.27, 479.13, 476.23, 473.3, 470.35, 467.38, 464.39, 461.38, 458.35, 455.29, 452.22, 449.12, 447.75, 447.01, 446.26, 1977.2, 1990.86, 1980.79, 1959.53, 1938.12, 1916.57, 1894.87, 1873.02, 1851.02, 1828.88, 1806.59, 1784.15, 1761.56, 1744.59, 1733.47, 1722.24, 1710.89, 1699.41, 1687.82, 1676.1, 1664.27, 1652.31, 1640.24, 1628.04, 1615.33, 1601.7, 1588.08, 1574.48, 1560.89, 1547.31, 1533.75, 1520.21, 1506.68, 1493.16, 1479.66, 1466.79, 1459.09, 1451.52, 1444.07, 1436.74, 1429.54, 1422.47, 1415.52, 1408.71, 1402.01, 1395.45, 1389.01, 1346.82, 1301.31, 1255.86, 1210.45, 1165.1, 1119.8, 1074.55, 1029.35, 984.2, 939.1, 894.06, 866.8, 846.41, 825.88, 805.22, 784.42, 763.48, 742.41, 721.2, 699.86, 678.38, 656.76, 646.5, 646.91, 647.41, 647.98, 648.63, 649.36, 650.17, 651.06, 652.03, 653.08, 654.21, 655.37, 656.48, 657.66, 658.92, 660.25, 661.66, 663.14, 664.7, 666.34, 668.05, 669.83, 670.95, 666.41, 661.85, 657.28, 652.68, 648.06, 643.43, 638.78, 634.1, 629.41, 624.7, 619.97, 626.14, 633.46, 640.85, 648.32, 655.87, 663.49, 671.19, 678.97, 686.82, 694.76, 702.77, 699.33, 690.96, 682.52, 673.98, 665.36, 656.65, 647.71, 638.27, 628.68, 618.93, 609.02, 602.92, 600.86, 598.84, 596.86, 594.93, 593.03, 591.17, 589.36, 587.58, 585.85, 584.16, 582.62, 581.4, 580.19, 578.99, 577.8, 576.63, 575.48, 574.33, 573.21, 572.09, 570.99, 569.44, 561.58, 553.76, 545.98, 538.25, 530.55, 522.9, 515.29, 507.73, 500.2, 492.72, 485.28, 481.63, 478.62, 475.58, 472.52, 469.44, 466.34, 463.22, 460.07, 456.91, 453.73, 450.52, 448.84, 448.04, 447.24, 1990.61, 2001.47, 1992.44, 1970.25, 1947.91, 1925.43, 1902.8, 1880.02, 1857.09, 1834.02, 1810.8, 1787.43, 1763.92, 1745.16, 1733.3, 1721.32, 1709.23, 1697.01, 1684.67, 1672.22, 1659.64, 1646.94, 1634.12, 1621.18, 1608.02, 1594.48, 1580.96, 1567.45, 1553.95, 1540.47, 1527.01, 1513.56, 1500.12, 1486.7, 1473.29, 1460.07, 1453.13, 1446.32, 1439.63, 1433.07, 1426.63, 1420.32, 1414.14, 1408.08, 1402.15, 1396.35, 1390.68, 1352.12, 1306.92, 1261.76, 1216.66, 1171.61, 1126.61, 1081.66, 1036.76, 991.92, 947.12, 902.38, 872.64, 851.45, 830.13, 808.67, 787.07, 765.34, 743.47, 721.47, 699.33, 677.05, 654.64, 642.24, 643.12, 644.08, 645.11, 646.22, 647.42, 648.69, 650.05, 651.48, 653.0, 654.59, 656.22, 657.76, 659.37, 661.05, 662.81, 664.65, 666.56, 668.54, 670.6, 672.74, 674.95, 677.04, 672.39, 667.72, 663.03, 658.33, 653.6, 648.86, 644.09, 639.31, 634.51, 629.68, 624.84, 630.29, 638.03, 645.85, 653.74, 661.71, 669.76, 678.07, 686.51, 694.68, 702.59, 710.22, 707.43, 698.07, 688.56, 678.89, 669.06, 659.09, 648.95, 638.66, 628.22, 617.62, 606.86, 599.61, 597.78, 595.98, 594.22, 592.51, 590.83, 589.2, 587.61, 586.06, 584.54, 583.07, 581.69, 580.54, 579.41, 578.28, 577.17, 576.08, 574.99, 573.93, 572.87, 571.83, 570.8, 569.79, 562.32, 554.72, 547.17, 539.66, 532.19, 524.76, 517.37, 510.03, 502.73, 495.47, 488.25, 484.15, 481.03, 477.88, 474.71, 471.52, 468.31, 465.07, 461.82, 458.55, 455.25, 451.93, 449.94, 449.09, 448.23, 2004.46, 2012.53, 2004.23, 1981.11, 1957.84, 1934.43, 1910.87, 1887.17, 1863.32, 1839.31, 1815.17, 1790.87, 1766.43, 1745.85, 1733.25, 1720.53, 1707.69, 1694.73, 1681.65, 1668.45, 1655.13, 1641.68, 1628.12, 1614.44, 1600.69, 1587.24, 1573.82, 1560.4, 1547.0, 1533.61, 1520.24, 1506.89, 1493.54, 1480.22, 1466.9, 1453.61, 1447.04, 1440.99, 1435.07, 1429.27, 1423.59, 1418.05, 1412.63, 1407.33, 1402.17, 1397.13, 1392.22, 1357.37, 1312.47, 1267.62, 1222.82, 1178.07, 1133.37, 1088.72, 1044.13, 999.58, 955.09, 910.65, 878.62, 856.63, 834.51, 812.25, 789.86, 767.33, 744.67, 721.87, 698.93, 675.86, 652.65, 637.91, 639.25, 640.67, 642.16, 643.74, 645.4, 647.14, 648.96, 650.86, 652.84, 654.9, 657.0, 658.96, 661.0, 663.11, 665.3, 667.56, 669.89, 672.31, 674.79, 677.36, 679.99, 682.71, 678.39, 673.61, 668.81, 664.0, 659.16, 654.3, 649.93, 646.14, 642.74, 639.71, 637.07, 642.86, 651.41, 659.69, 667.71, 675.46, 682.95, 690.17, 697.12, 703.81, 710.23, 716.39, 713.93, 703.72, 693.36, 682.84, 672.16, 661.33, 650.35, 639.21, 627.91, 616.46, 604.85, 596.26, 594.65, 593.08, 591.54, 590.05, 588.6, 587.18, 585.81, 584.48, 583.2, 581.95, 580.75, 579.68, 578.61, 577.56, 576.53, 575.51, 574.5, 573.5, 572.52, 571.56, 570.6, 569.66, 563.02, 555.65, 548.31, 541.02, 533.78, 526.57, 519.41, 512.29, 505.21, 498.17, 491.18, 486.7, 483.46, 480.2, 476.92, 473.62, 470.3, 466.95, 463.59, 460.2, 456.8, 453.37, 451.05, 450.15, 449.23, 2018.76, 2024.04, 2016.17, 1992.12, 1967.93, 1943.59, 1919.1, 1894.46, 1869.68, 1844.75, 1819.68, 1794.45, 1769.08, 1746.66, 1733.32, 1719.85, 1706.27, 1692.57, 1678.74, 1664.8, 1650.74, 1636.55, 1622.25, 1607.82, 1593.34, 1579.99, 1566.66, 1553.34, 1540.03, 1526.74, 1513.46, 1500.2, 1486.95, 1473.72, 1460.5, 1447.3, 1440.83, 1435.54, 1430.38, 1425.34, 1420.43, 1415.64, 1410.99, 1406.46, 1402.05, 1397.78, 1393.63, 1362.58, 1317.98, 1273.43, 1228.93, 1184.48, 1140.08, 1095.74, 1051.44, 1007.2, 963.01, 918.87, 884.73, 861.95, 839.03, 815.98, 792.79, 769.46, 746.0, 722.41, 698.67, 674.8, 650.8, 633.49, 635.29, 637.17, 639.14, 641.18, 643.3, 645.5, 647.79, 650.15, 652.59, 655.12, 657.7, 660.09, 662.56, 665.09, 667.71, 670.39, 673.14, 675.62, 678.05, 680.44, 682.79, 685.09, 681.78, 677.84, 674.29, 671.11, 668.31, 665.9, 663.86, 662.2, 660.93, 660.04, 659.53, 664.13, 671.19, 677.99, 684.52, 690.79, 696.79, 702.53, 708.0, 713.21, 718.15, 722.82, 720.59, 709.53, 698.31, 686.94, 675.41, 663.73, 651.9, 639.91, 627.76, 615.46, 603.0, 592.87, 591.48, 590.13, 588.82, 587.55, 586.32, 585.13, 583.98, 582.87, 581.81, 580.78, 579.8, 578.79, 577.81, 576.83, 575.87, 574.92, 573.99, 573.07, 572.16, 571.27, 570.39, 569.52, 563.68, 556.53, 549.42, 542.35, 535.32, 528.34, 521.4, 514.5, 507.65, 500.83, 494.06, 489.26, 485.91, 482.54, 479.15, 475.74, 472.31, 468.85, 465.38, 461.88, 458.37, 454.83, 452.17, 451.21, 450.24, 2033.49, 2035.98, 2028.25, 2003.28, 1978.15, 1952.89, 1927.47, 1901.91, 1876.2, 1850.34, 1824.34, 1798.19, 1771.89, 1747.59, 1733.5, 1719.29, 1704.97, 1690.52, 1675.96, 1661.27, 1646.47, 1631.54, 1616.49, 1601.32, 1586.03, 1572.73, 1559.49, 1546.26, 1533.05, 1519.85, 1506.67, 1493.5, 1480.35, 1467.21, 1454.09, 1440.98, 1434.48, 1429.96, 1425.56, 1421.29, 1417.14, 1413.12, 1409.22, 1405.45, 1401.81, 1398.3, 1394.91, 1367.73, 1323.43, 1279.18, 1234.98, 1190.84, 1146.74, 1102.7, 1058.71, 1014.76, 970.87, 927.04, 890.98, 867.4, 843.68, 819.83, 795.85, 771.73, 747.47, 723.08, 698.55, 673.89, 649.09, 628.99, 631.26, 633.6, 636.03, 638.54, 641.12, 644.06, 647.13, 650.13, 653.04, 655.88, 658.64, 661.13, 663.58, 665.98, 668.34, 670.66, 672.94, 675.17, 677.36, 679.51, 681.62, 683.68, 682.53, 680.73, 679.31, 678.26, 677.6, 677.31, 677.41, 677.88, 678.74, 679.98, 681.6, 685.66, 691.24, 696.55, 701.6, 706.39, 710.91, 715.16, 719.15, 722.88, 726.33, 729.52, 727.4, 715.49, 703.42, 691.2, 678.82, 666.29, 653.6, 640.76, 627.77, 614.61, 601.3, 589.44, 588.27, 587.14, 586.05, 585.01, 584.0, 583.03, 582.11, 581.22, 580.38, 579.57, 578.81, 577.9, 576.98, 576.08, 575.2, 574.32, 573.46, 572.62, 571.78, 570.97, 570.16, 569.37, 564.29, 557.36, 550.48, 543.63, 536.83, 530.07, 523.35, 516.68, 510.04, 503.45, 496.91, 491.84, 488.39, 484.91, 481.4, 477.88, 474.34, 470.78, 467.19, 463.58, 459.96, 456.31, 453.31, 452.29, 451.27, 2048.67, 2048.36, 2040.49, 2014.58, 1988.53, 1962.33, 1935.99, 1909.5, 1882.86, 1856.08, 1829.15, 1802.07, 1774.84, 1748.64, 1733.81, 1718.86, 1703.79, 1688.6, 1673.29, 1657.86, 1642.32, 1626.65, 1610.86, 1594.94, 1578.91, 1565.45, 1552.3, 1539.17, 1526.05, 1512.95, 1499.86, 1486.79, 1473.73, 1460.68, 1447.65, 1434.64, 1428.02, 1424.25, 1420.62, 1417.11, 1413.72, 1410.46, 1407.33, 1404.32, 1401.44, 1398.69, 1396.06, 1372.83, 1328.83, 1284.89, 1240.99, 1197.15, 1153.35, 1109.61, 1065.92, 1022.28, 978.69, 935.15, 897.36, 872.98, 848.47, 823.83, 799.05, 774.14, 749.31, 724.65, 700.16, 675.84, 651.7, 630.24, 633.32, 636.33, 639.26, 642.11, 644.89, 647.59, 650.21, 652.76, 655.22, 657.62, 659.93, 662.18, 664.38, 666.54, 668.66, 670.74, 672.78, 674.77, 676.72, 678.63, 680.5, 682.32, 682.91, 683.24, 683.95, 685.04, 686.5, 688.35, 690.58, 693.18, 696.17, 699.54, 703.29, 707.45, 711.55, 715.38, 718.95, 722.25, 725.29, 728.06, 730.57, 732.81, 734.78, 736.49, 734.37, 721.6, 708.68, 695.61, 682.38, 669.0, 655.47, 641.77, 627.93, 613.92, 599.76, 585.97, 585.02, 584.12, 583.25, 582.42, 581.64, 580.89, 580.19, 579.53, 578.9, 578.32, 577.78, 576.99, 576.15, 575.32, 574.51, 573.71, 572.92, 572.15, 571.39, 570.65, 569.92, 569.2, 564.87, 558.16, 551.5, 544.87, 538.29, 531.76, 525.26, 518.81, 512.4, 506.03, 499.7, 494.45, 490.88, 487.29, 483.68, 480.05, 476.39, 472.72, 469.02, 465.31, 461.57, 457.81, 454.45, 453.38, 452.3, 2064.28, 2061.18, 2052.86, 2026.03, 1999.05, 1971.93, 1944.66, 1917.24, 1889.68, 1861.96, 1834.1, 1806.1, 1777.94, 1749.8, 1734.23, 1718.54, 1702.73, 1686.8, 1670.75, 1654.58, 1638.29, 1621.87, 1605.34, 1588.69, 1571.91, 1558.15, 1545.1, 1532.06, 1519.04, 1506.03, 1493.03, 1480.06, 1467.09, 1454.14, 1441.2, 1428.28, 1421.42, 1418.42, 1415.55, 1412.8, 1410.18, 1407.68, 1405.31, 1403.06, 1400.95, 1398.95, 1397.09, 1377.88, 1334.18, 1290.54, 1246.95, 1203.12, 1158.47, 1114.0, 1069.71, 1025.61, 981.7, 937.97, 898.4, 873.72, 849.2, 824.86, 800.7, 776.7, 752.87, 729.22, 705.73, 682.42, 659.29, 636.51, 639.14, 641.71, 644.19, 646.6, 648.93, 651.18, 653.36, 655.46, 657.49, 659.43, 661.3, 663.27, 665.23, 667.15, 669.03, 670.87, 672.66, 674.41, 676.12, 677.79, 679.41, 680.99, 682.91, 685.37, 688.21, 691.43, 695.03, 699.01, 703.37, 708.1, 713.22, 718.72, 724.61, 729.51, 732.13, 734.48, 736.56, 738.38, 739.94, 741.23, 742.25, 743.01, 743.5, 743.73, 741.49, 727.87, 714.1, 700.18, 686.1, 671.87, 657.48, 642.94, 628.24, 613.39, 598.38, 583.21, 581.73, 581.05, 580.4, 579.8, 579.24, 578.71, 578.23, 577.79, 577.39, 577.03, 576.72, 576.07, 575.3, 574.55, 573.81, 573.08, 572.37, 571.67, 570.99, 570.32, 569.66, 569.02, 565.4, 558.91, 552.47, 546.07, 539.72, 533.4, 527.13, 520.9, 514.71, 508.56, 502.46, 497.08, 493.4, 489.7, 485.97, 482.23, 478.47, 474.68, 470.88, 467.05, 463.2, 459.33, 455.61, 454.48, 453.34, 2080.34, 2074.45, 2065.39, 2037.63, 2009.72, 1981.67, 1953.47, 1925.13, 1896.64, 1868.0, 1839.21, 1810.28, 1781.19, 1751.96, 1734.77, 1718.34, 1701.79, 1685.11, 1668.32, 1651.41, 1634.38, 1617.22, 1599.95, 1582.55, 1565.03, 1550.84, 1537.88, 1524.94, 1512.01, 1499.09, 1486.19, 1473.31, 1460.44, 1447.58, 1434.74, 1421.91, 1414.71, 1412.47, 1410.35, 1408.39, 1406.35, 1404.01, 1401.38, 1398.45, 1395.23, 1391.72, 1387.91, 1370.42, 1326.14, 1282.03, 1238.11, 1194.37, 1150.82, 1107.45, 1064.27, 1021.26, 978.45, 935.82, 895.76, 872.09, 848.58, 825.24, 802.08, 779.09, 756.26, 733.61, 711.14, 688.83, 666.7, 644.74, 645.05, 647.16, 649.2, 651.16, 653.05, 654.86, 656.59, 658.25, 659.83, 661.33, 662.75, 664.4, 666.12, 667.8, 669.44, 671.03, 672.58, 674.1, 675.56, 676.99, 678.37, 679.71, 682.53, 687.13, 692.1, 697.45, 703.18, 709.29, 715.78, 722.64, 729.89, 737.53, 745.54, 751.84, 752.97, 753.84, 754.44, 754.78, 754.85, 754.66, 754.2, 753.48, 752.49, 751.23, 748.76, 734.3, 719.68, 704.9, 689.97, 674.89, 659.65, 644.26, 628.71, 613.01, 597.15, 581.14, 578.4, 577.94, 577.52, 577.14, 576.79, 576.49, 576.23, 576.02, 575.84, 575.7, 575.61, 575.13, 574.44, 573.76, 573.1, 572.44, 571.81, 571.18, 570.57, 569.98, 569.39, 568.82, 565.89, 559.63, 553.41, 547.23, 541.1, 535.0, 528.95, 522.95, 516.98, 511.06, 505.18, 499.73, 495.94, 492.12, 488.29, 484.44, 480.56, 476.67, 472.75, 468.81, 464.85, 460.88, 456.87, 455.59, 454.4, 2096.83, 2088.15, 2078.06, 2049.37, 2020.54, 1991.56, 1962.43, 1933.16, 1903.74, 1874.18, 1844.46, 1814.6, 1784.59, 1754.43, 1735.44, 1718.26, 1700.97, 1683.55, 1666.02, 1648.36, 1630.59, 1612.69, 1594.67, 1576.54, 1558.28, 1543.51, 1530.65, 1518.4, 1506.64, 1494.82, 1482.94, 1471.01, 1459.02, 1446.97, 1434.86, 1422.7, 1414.78, 1411.85, 1408.63, 1405.11, 1401.3, 1397.2, 1392.81, 1388.12, 1383.13, 1377.85, 1372.28, 1357.09, 1313.9, 1270.9, 1228.08, 1185.44, 1142.99, 1100.72, 1058.63, 1016.73, 975.01, 933.48, 892.96, 870.28, 847.78, 825.45, 803.29, 781.3, 759.48, 737.84, 716.37, 695.06, 673.94, 652.98, 651.03, 652.7, 654.29, 655.81, 657.25, 658.61, 659.9, 661.11, 662.24, 663.3, 664.28, 665.57, 667.05, 668.49, 669.89, 671.24, 672.55, 673.82, 675.05, 676.23, 677.37, 678.47, 681.78, 688.5, 695.61, 703.09, 710.95, 719.19, 727.81, 736.81, 746.19, 755.95, 766.09, 774.43, 774.08, 773.46, 772.59, 771.44, 770.03, 768.36, 766.42, 764.22, 761.75, 759.01, 756.0, 740.87, 725.4, 709.78, 694.0, 678.07, 661.98, 645.74, 629.34, 612.79, 596.08, 579.21, 575.03, 574.79, 574.59, 574.43, 574.31, 574.23, 574.19, 574.2, 574.24, 574.33, 574.45, 574.18, 573.56, 572.96, 572.37, 571.79, 571.23, 570.68, 570.14, 569.62, 569.11, 568.61, 566.34, 560.3, 554.3, 548.35, 542.43, 536.56, 530.74, 524.95, 519.21, 513.51, 507.85, 502.39, 498.49, 494.57, 490.63, 486.66, 482.68, 478.67, 474.65, 470.6, 466.53, 462.44, 458.33, 456.71, 455.46, 2113.77, 2102.3, 2090.39, 2061.26, 2031.5, 2001.6, 1971.54, 1941.34, 1911.0, 1880.5, 1849.86, 1819.08, 1788.14, 1757.05, 1736.22, 1720.7, 1704.86, 1688.64, 1672.03, 1655.04, 1637.67, 1619.9, 1601.75, 1583.22, 1564.29, 1549.25, 1537.24, 1525.19, 1513.07, 1500.9, 1488.67, 1476.38, 1464.04, 1451.64, 1439.18, 1426.66, 1417.09, 1412.39, 1407.41, 1402.13, 1396.55, 1390.69, 1384.53, 1378.07, 1371.32, 1364.28, 1356.94, 1343.57, 1301.49, 1259.58, 1217.86, 1176.32, 1134.97, 1093.8, 1052.81, 1012.01, 971.39, 930.96, 890.72, 868.31, 846.81, 825.49, 804.33, 783.35, 762.53, 741.89, 721.42, 701.13, 681.0, 661.05, 657.08, 658.31, 659.45, 660.53, 661.52, 662.44, 663.28, 664.05, 664.73, 665.35, 665.88, 666.79, 668.03, 669.22, 670.38, 671.49, 672.56, 673.59, 674.57, 675.52, 676.42, 677.28, 680.65, 689.51, 698.74, 708.35, 718.34, 728.71, 739.46, 750.59, 762.1, 773.99, 786.26, 797.29, 795.45, 793.36, 791.0, 788.37, 785.48, 782.33, 778.91, 775.22, 771.27, 767.05, 762.56, 747.6, 731.29, 714.81, 698.18, 681.4, 664.46, 647.37, 630.12, 612.72, 595.16, 577.45, 571.62, 571.6, 571.62, 571.68, 571.79, 571.93, 572.11, 572.34, 572.61, 572.91, 573.26, 573.22, 572.67, 572.14, 571.63, 571.12, 570.63, 570.16, 569.69, 569.25, 568.81, 568.39, 566.74, 560.93, 555.15, 549.42, 543.73, 538.08, 532.48, 526.91, 521.39, 515.91, 510.48, 505.09, 501.07, 497.04, 492.99, 488.91, 484.82, 480.7, 476.56, 472.41, 468.23, 464.03, 459.8, 457.84, 456.54, 2131.14, 2116.89, 2103.17, 2076.05, 2047.3, 2018.6, 1989.94, 1961.32, 1932.74, 1904.2, 1875.71, 1847.26, 1818.85, 1790.49, 1769.65, 1751.79, 1733.56, 1714.94, 1695.93, 1676.55, 1656.77, 1636.62, 1616.07, 1595.14, 1573.82, 1556.79, 1544.43, 1532.02, 1519.56, 1507.03, 1494.45, 1481.81, 1469.12, 1456.36, 1443.55, 1430.69, 1419.69, 1413.23, 1406.48, 1399.43, 1392.09, 1384.46, 1376.54, 1368.32, 1359.81, 1351.0, 1341.9, 1329.87, 1288.88, 1248.08, 1207.46, 1167.02, 1126.76, 1086.69, 1046.81, 1007.1, 967.59, 928.25, 889.11, 866.17, 845.68, 825.35, 805.2, 785.22, 765.41, 745.77, 726.31, 707.02, 687.89, 668.95, 663.21, 663.99, 664.7, 665.32, 665.87, 666.35, 666.74, 667.06, 667.31, 667.47, 667.56, 668.04, 669.04, 670.0, 670.91, 671.78, 672.61, 673.4, 674.14, 674.85, 675.51, 676.12, 679.14, 690.13, 701.49, 713.24, 725.36, 737.85, 750.73, 763.99, 777.63, 791.65, 806.05, 820.41, 817.09, 813.52, 809.67, 805.57, 801.2, 796.56, 791.66, 786.49, 781.06, 775.36, 769.39, 754.49, 737.32, 720.0, 702.52, 684.89, 667.1, 649.16, 631.06, 612.81, 594.4, 575.84, 568.16, 568.37, 568.61, 568.9, 569.22, 569.59, 569.99, 570.44, 570.93, 571.46, 572.03, 572.24, 571.77, 571.31, 570.87, 570.44, 570.02, 569.62, 569.23, 568.86, 568.5, 568.15, 567.11, 561.51, 555.96, 550.45, 544.98, 539.56, 534.17, 528.83, 523.53, 518.28, 513.06, 507.89, 503.67, 499.53, 495.37, 491.18, 486.98, 482.75, 478.5, 474.23, 469.94, 465.63, 461.3, 458.98, 457.62, 2162.32, 2148.91, 2135.47, 2109.72, 2081.22, 2052.76, 2024.34, 1995.97, 1967.64, 1939.35, 1911.1, 1882.89, 1854.73, 1826.61, 1803.52, 1783.27, 1762.64, 1741.62, 1720.22, 1698.44, 1676.27, 1653.72, 1630.78, 1607.45, 1583.73, 1564.39, 1551.68, 1538.92, 1526.1, 1513.22, 1500.29, 1487.3, 1474.25, 1461.15, 1447.99, 1434.77, 1422.58, 1414.35, 1405.84, 1397.03, 1387.93, 1378.54, 1368.85, 1358.87, 1348.59, 1338.02, 1327.16, 1315.98, 1276.1, 1236.39, 1196.87, 1157.53, 1118.37, 1079.4, 1040.61, 1002.01, 963.59, 925.36, 887.31, 863.86, 844.37, 825.05, 805.9, 786.92, 768.12, 749.49, 731.02, 712.73, 694.62, 676.67, 669.42, 669.76, 670.01, 670.2, 670.3, 670.33, 670.28, 670.16, 669.95, 669.67, 669.32, 669.34, 670.1, 670.81, 671.49, 672.12, 672.71, 673.25, 673.76, 674.22, 674.64, 675.01, 677.26, 690.38, 703.87, 717.74, 731.99, 746.62, 761.63, 777.01, 792.78, 808.93, 825.46, 842.37, 839.0, 833.94, 828.62, 823.03, 817.18, 811.06, 804.68, 798.03, 791.12, 783.94, 776.49, 761.53, 743.51, 725.34, 707.01, 688.53, 669.89, 651.1, 632.16, 613.05, 593.8, 574.38, 564.67, 565.1, 565.56, 566.07, 566.61, 567.2, 567.83, 568.5, 569.21, 569.96, 570.75, 571.25, 570.85, 570.47, 570.1, 569.75, 569.4, 569.08, 568.76, 568.46, 568.17, 567.9, 567.43, 562.06, 556.73, 551.44, 546.2, 540.99, 535.83, 530.71, 525.64, 520.6, 515.61, 510.66, 506.3, 502.04, 497.77, 493.47, 489.16, 484.82, 480.46, 476.08, 471.68, 467.26, 462.82, 460.14, 458.72, 2195.17, 2181.51, 2167.81, 2143.36, 2115.1, 2086.89, 2058.71, 2030.58, 2002.49, 1974.45, 1946.44, 1918.48, 1890.56, 1862.69, 1837.78, 1815.13, 1792.1, 1768.69, 1744.9, 1720.72, 1696.15, 1671.2, 1645.87, 1620.15, 1594.04, 1572.04, 1558.99, 1545.87, 1532.7, 1519.47, 1506.19, 1492.85, 1479.45, 1465.99, 1452.48, 1438.91, 1425.76, 1415.77, 1405.49, 1394.92, 1384.05, 1372.9, 1361.45, 1349.71, 1337.67, 1325.33, 1312.71, 1299.78, 1263.13, 1224.52, 1186.1, 1147.86, 1109.8, 1071.93, 1034.24, 996.73, 959.41, 922.28, 885.33, 861.37, 842.89, 824.57, 806.43, 788.46, 770.65, 753.02, 735.57, 718.28, 701.16, 684.22, 675.71, 675.6, 675.41, 675.15, 674.81, 674.39, 673.9, 673.33, 672.68, 671.96, 671.15, 670.68, 671.2, 671.67, 672.1, 672.49, 672.84, 673.15, 673.41, 673.63, 673.81, 673.94, 675.0, 690.25, 705.87, 721.87, 738.25, 755.01, 772.14, 789.66, 807.55, 825.83, 844.49, 863.53, 861.17, 854.63, 847.83, 840.76, 833.43, 825.83, 817.97, 809.84, 801.44, 792.78, 783.86, 768.73, 749.86, 730.83, 711.66, 692.33, 672.84, 653.2, 633.4, 613.45, 593.35, 573.08, 561.14, 561.78, 562.47, 563.2, 563.97, 564.78, 565.63, 566.52, 567.45, 568.42, 569.44, 570.24, 569.92, 569.61, 569.32, 569.04, 568.77, 568.51, 568.27, 568.05, 567.84, 567.64, 567.45, 562.56, 557.45, 552.39, 547.36, 542.38, 537.44, 532.55, 527.69, 522.88, 518.11, 513.38, 508.94, 504.57, 500.19, 495.78, 491.36, 486.91, 482.44, 477.95, 473.44, 468.91, 464.36, 461.17, 459.47, 2228.07, 2214.15, 2200.19, 2176.95, 2148.94, 2120.97, 2093.04, 2065.16, 2037.31, 2009.51, 1981.75, 1954.04, 1926.36, 1898.73, 1872.42, 1847.38, 1821.95, 1796.15, 1769.96, 1743.38, 1716.42, 1689.08, 1661.35, 1633.23, 1604.73, 1579.76, 1566.35, 1552.88, 1539.36, 1525.78, 1512.14, 1498.45, 1484.7, 1470.89, 1457.03, 1443.1, 1429.23, 1417.48, 1405.43, 1393.1, 1380.47, 1367.55, 1354.34, 1340.84, 1327.04, 1312.94, 1298.55, 1283.86, 1249.97, 1212.47, 1175.14, 1138.0, 1101.04, 1064.27, 1027.68, 991.27, 955.05, 919.01, 883.16, 858.72, 841.24, 823.93, 806.79, 789.82, 773.02, 756.39, 739.94, 723.65, 707.54, 691.6, 682.07, 681.52, 680.88, 680.17, 679.39, 678.53, 677.59, 676.57, 675.48, 674.31, 673.07, 672.07, 672.34, 672.57, 672.76, 672.91, 673.02, 673.08, 673.11, 673.09, 673.02, 672.92, 672.77, 689.74, 707.49, 725.62, 744.13, 763.01, 782.28, 801.92, 821.94, 842.35, 863.13, 884.3, 883.61, 875.58, 867.3, 858.75, 849.94, 840.86, 831.52, 821.91, 812.04, 801.9, 791.49, 776.07, 756.36, 736.48, 716.46, 696.28, 675.94, 655.45, 634.81, 614.01, 593.05, 571.94, 557.56, 558.43, 559.34, 560.29, 561.28, 562.31, 563.38, 564.5, 565.65, 566.84, 568.08, 569.23, 568.98, 568.74, 568.52, 568.31, 568.12, 567.94, 567.77, 567.62, 567.48, 567.36, 567.25, 563.02, 558.14, 553.29, 548.49, 543.73, 539.02, 534.34, 529.71, 525.12, 520.57, 516.07, 511.57, 506.71, 501.85, 497.01, 492.18, 487.36, 482.55, 477.75, 472.96, 468.18, 463.41, 459.86, 458.21, 2261.01, 2246.83, 2232.61, 2210.51, 2182.74, 2155.02, 2127.33, 2099.69, 2072.09, 2044.54, 2017.02, 1989.55, 1962.12, 1934.73, 1907.45, 1880.01, 1852.19, 1823.99, 1795.4, 1766.43, 1737.08, 1707.34, 1677.22, 1646.71, 1615.81, 1587.53, 1573.77, 1559.95, 1546.07, 1532.14, 1518.16, 1504.11, 1490.01, 1475.85, 1461.63, 1447.36, 1433.03, 1419.48, 1405.67, 1391.57, 1377.18, 1362.5, 1347.53, 1332.26, 1316.7, 1300.84, 1284.69, 1268.24, 1236.63, 1200.23, 1164.0, 1127.96, 1092.1, 1056.42, 1020.93, 985.62, 950.5, 915.56, 880.81, 855.89, 839.42, 823.11, 806.97, 791.01, 775.21, 759.59, 744.13, 728.85, 713.75, 698.81, 688.51, 687.51, 686.43, 685.28, 684.05, 682.74, 681.36, 679.9, 678.36, 676.75, 675.06, 673.49, 673.53, 673.52, 673.47, 673.37, 673.24, 673.06, 672.84, 672.58, 672.28, 671.93, 671.55, 688.86, 708.74, 729.0, 749.63, 770.64, 792.04, 813.81, 835.96, 858.49, 881.4, 904.7, 906.31, 896.81, 887.04, 877.01, 866.72, 856.16, 845.34, 834.25, 822.9, 811.28, 799.39, 783.58, 763.01, 742.29, 721.41, 700.38, 679.2, 657.86, 636.37, 614.72, 592.91, 570.95, 553.94, 555.03, 556.17, 557.34, 558.55, 559.8, 561.1, 562.43, 563.81, 565.22, 566.68, 568.18, 568.02, 567.86, 567.71, 567.58, 567.46, 567.35, 567.26, 567.18, 567.11, 567.06, 567.03, 563.44, 558.43, 553.34, 548.31, 543.34, 538.42, 533.56, 528.75, 524.0, 519.3, 514.66, 509.95, 505.14, 500.34, 495.56, 490.78, 486.02, 481.26, 476.52, 471.79, 467.06, 462.35, 458.56, 456.94, 2293.99, 2279.55, 2265.07, 2244.03, 2216.51, 2189.03, 2161.59, 2134.19, 2106.83, 2079.52, 2052.25, 2025.02, 1997.84, 1970.69, 1942.85, 1913.02, 1882.81, 1852.21, 1821.23, 1789.87, 1758.13, 1725.99, 1693.48, 1660.57, 1627.28, 1595.35, 1581.24, 1567.07, 1552.85, 1538.57, 1524.23, 1509.83, 1495.38, 1480.87, 1466.3, 1451.67, 1436.99, 1421.77, 1406.2, 1390.34, 1374.19, 1357.74, 1341.01, 1323.98, 1306.65, 1289.04, 1271.12, 1252.91, 1223.11, 1187.8, 1152.67, 1117.73, 1082.97, 1048.39, 1014.0, 979.79, 945.76, 911.92, 878.27, 852.9, 837.43, 822.12, 806.99, 792.02, 777.23, 762.61, 748.16, 733.88, 719.78, 705.85, 695.03, 693.58, 692.06, 690.46, 688.78, 687.03, 685.2, 683.3, 681.32, 679.26, 677.13, 674.96, 674.75, 674.5, 674.21, 673.88, 673.5, 673.09, 672.63, 672.12, 671.58, 670.99, 670.36, 687.6, 709.61, 732.0, 754.76, 777.9, 801.42, 825.31, 849.59, 874.25, 899.29, 924.71, 929.27, 918.29, 907.05, 895.54, 883.77, 871.73, 859.43, 846.86, 834.03, 820.93, 807.56, 791.23, 769.82, 748.25, 726.52, 704.64, 682.61, 660.42, 638.08, 615.58, 592.93, 570.12, 550.29, 551.6, 552.95, 554.35, 555.78, 557.25, 558.77, 560.33, 561.92, 563.56, 565.24, 566.96, 567.05, 566.65, 565.99, 565.36, 564.78, 564.24, 563.74, 563.28, 562.85, 562.48, 562.14, 558.77, 553.92, 549.13, 544.4, 539.72, 535.09, 530.53, 526.01, 521.56, 517.16, 512.81, 508.31, 503.56, 498.82, 494.09, 489.37, 484.66, 479.97, 475.28, 470.6, 465.93, 461.28, 457.24, 455.66, 2327.01, 2312.31, 2297.57, 2277.51, 2250.23, 2222.99, 2195.8, 2168.65, 2141.54, 2114.47, 2087.44, 2060.46, 2033.52, 2006.62, 1978.65, 1946.42, 1913.81, 1880.83, 1847.45, 1813.7, 1779.56, 1745.03, 1710.12, 1674.83, 1639.14, 1603.24, 1588.78, 1574.26, 1559.68, 1545.04, 1530.35, 1515.61, 1500.8, 1485.94, 1471.02, 1456.04, 1441.01, 1424.35, 1407.02, 1389.4, 1371.48, 1353.28, 1334.78, 1315.99, 1296.9, 1277.52, 1257.85, 1237.88, 1209.41, 1175.19, 1141.17, 1107.32, 1073.66, 1040.18, 1006.88, 973.77, 940.84, 908.1, 875.54, 849.74, 835.26, 820.96, 806.83, 792.87, 779.08, 765.46, 752.02, 738.74, 725.64, 712.71, 701.62, 699.73, 697.76, 695.72, 693.6, 691.4, 689.13, 686.78, 684.35, 681.85, 679.27, 676.61, 676.02, 675.53, 675.0, 674.42, 673.81, 673.15, 672.45, 671.71, 670.92, 670.09, 669.22, 685.97, 710.1, 734.62, 759.5, 784.77, 810.42, 836.44, 862.85, 889.63, 916.8, 944.35, 952.5, 940.04, 927.32, 914.33, 901.08, 887.56, 873.78, 859.74, 845.42, 830.85, 816.0, 799.04, 776.78, 754.36, 731.79, 709.06, 686.18, 663.14, 639.95, 616.61, 593.1, 569.45, 546.59, 548.12, 549.5, 550.61, 551.72, 552.84, 553.98, 555.12, 556.27, 557.43, 558.6, 559.78, 559.59, 559.1, 558.66, 558.26, 557.89, 557.57, 557.29, 557.05, 556.86, 556.7, 556.58, 553.9, 549.35, 544.86, 540.42, 536.04, 531.71, 527.44, 523.22, 519.06, 514.96, 510.91, 506.67, 501.98, 497.29, 492.62, 487.95, 483.3, 478.66, 474.03, 469.4, 464.79, 460.19, 455.92, 454.37, 2360.07, 2345.11, 2330.11, 2310.95, 2283.92, 2256.92, 2229.97, 2203.07, 2176.2, 2149.37, 2122.59, 2095.85, 2069.16, 2042.5, 2014.83, 1980.21, 1945.2, 1909.82, 1874.06, 1837.91, 1801.38, 1764.46, 1727.16, 1689.47, 1651.39, 1612.93, 1596.37, 1581.49, 1566.57, 1551.58, 1536.54, 1521.44, 1506.28, 1491.07, 1475.8, 1460.47, 1445.09, 1427.22, 1408.13, 1388.75, 1369.07, 1349.11, 1328.85, 1308.29, 1287.45, 1266.31, 1244.87, 1223.14, 1195.52, 1162.4, 1129.47, 1096.72, 1064.16, 1031.78, 999.58, 967.56, 935.73, 904.09, 872.63, 846.4, 832.93, 819.63, 806.5, 793.55, 780.76, 768.14, 755.7, 743.43, 731.33, 719.4, 708.29, 705.95, 703.54, 701.05, 698.49, 695.85, 693.13, 690.34, 687.47, 684.52, 681.49, 678.39, 677.33, 676.6, 675.83, 675.01, 674.15, 673.26, 672.32, 671.33, 670.31, 669.24, 668.13, 683.96, 710.22, 736.86, 763.88, 791.27, 819.04, 847.19, 875.72, 904.63, 933.93, 963.6, 976.0, 962.06, 947.85, 933.39, 918.66, 903.66, 888.4, 872.88, 857.09, 841.03, 824.71, 807.01, 783.89, 759.7, 734.39, 709.3, 684.43, 659.78, 635.36, 611.15, 587.17, 563.42, 539.88, 539.71, 540.86, 542.02, 543.19, 544.36, 545.55, 546.75, 547.95, 549.17, 550.4, 551.63, 551.77, 551.51, 551.29, 551.11, 550.97, 550.87, 550.81, 550.79, 550.81, 550.88, 550.98, 548.99, 544.73, 540.53, 536.39, 532.31, 528.27, 524.3, 520.38, 516.51, 512.7, 508.95, 505.02, 500.38, 495.75, 491.13, 486.52, 481.93, 477.34, 472.76, 468.2, 463.64, 459.1, 454.59, 453.08, 2393.17, 2377.95, 2362.7, 2344.35, 2317.56, 2290.82, 2264.11, 2237.45, 2210.82, 2184.24, 2157.71, 2131.21, 2104.76, 2078.35, 2051.39, 2014.37, 1976.98, 1939.2, 1901.05, 1862.51, 1823.58, 1784.27, 1744.58, 1704.5, 1664.03, 1623.17, 1604.01, 1588.79, 1573.51, 1558.18, 1542.78, 1527.33, 1511.82, 1496.26, 1480.64, 1464.96, 1449.22, 1430.38, 1409.53, 1388.39, 1366.95, 1345.22, 1323.2, 1300.89, 1278.28, 1255.38, 1232.19, 1208.69, 1181.44, 1149.43, 1117.59, 1085.94, 1054.48, 1023.19, 992.09, 961.17, 930.44, 899.89, 869.53, 842.9, 830.43, 818.13, 806.01, 794.05, 782.27, 770.65, 759.21, 747.94, 736.84, 725.92, 715.16, 712.25, 709.39, 706.46, 703.45, 700.37, 697.21, 693.97, 690.65, 687.26, 683.79, 680.25, 678.68, 677.71, 676.7, 675.64, 674.54, 673.4, 672.22, 671.0, 669.73, 668.42, 667.07, 681.57, 709.96, 738.73, 767.87, 797.39, 827.29, 857.56, 888.22, 919.26, 950.67, 982.47, 999.76, 984.34, 966.5, 947.17, 927.8, 908.4, 888.97, 869.49, 849.99, 830.44, 810.87, 790.82, 766.3, 741.99, 717.9, 694.04, 670.39, 646.97, 623.76, 600.78, 578.03, 555.49, 533.18, 531.0, 532.2, 533.42, 534.64, 535.87, 537.12, 538.37, 539.63, 540.9, 542.18, 543.47, 543.91, 543.87, 543.87, 543.92, 544.0, 544.12, 544.28, 544.49, 544.73, 545.02, 545.35, 544.01, 540.06, 536.15, 532.31, 528.52, 524.78, 521.1, 517.48, 513.91, 510.39, 506.94, 503.35, 498.77, 494.2, 489.63, 485.08, 480.54, 476.01, 471.49, 466.98, 462.48, 457.99, 453.51, 451.78, 2426.31, 2410.83, 2395.32, 2377.72, 2351.17, 2324.67, 2298.21, 2271.79, 2245.41, 2219.07, 2192.78, 2166.53, 2140.32, 2114.15, 2088.03, 2048.93, 2009.14, 1968.97, 1928.42, 1887.49, 1846.17, 1804.47, 1762.39, 1719.91, 1677.05, 1633.81, 1611.72, 1596.15, 1580.51, 1564.83, 1549.08, 1533.28, 1517.42, 1501.51, 1485.53, 1469.5, 1453.42, 1433.84, 1411.22, 1388.32, 1365.12, 1341.63, 1317.85, 1293.78, 1269.41, 1244.75, 1219.8, 1194.54, 1167.19, 1136.27, 1105.53, 1074.98, 1044.61, 1014.42, 984.42, 954.6, 924.96, 895.51, 866.25, 839.22, 827.76, 816.46, 805.34, 794.39, 783.6, 772.99, 762.55, 752.28, 742.19, 732.26, 722.51, 718.63, 715.33, 711.95, 708.5, 704.97, 701.36, 697.68, 693.92, 690.08, 686.17, 682.18, 680.08, 678.86, 677.61, 676.31, 674.98, 673.6, 672.17, 670.71, 669.2, 667.65, 666.06, 678.81, 709.33, 740.34, 770.89, 800.91, 830.42, 859.4, 887.86, 915.8, 943.22, 970.12, 986.25, 966.81, 947.33, 927.81, 908.26, 888.68, 869.06, 849.41, 829.72, 809.99, 790.23, 770.43, 747.15, 724.06, 701.2, 678.55, 656.13, 633.93, 611.95, 590.19, 568.66, 547.34, 526.25, 522.27, 523.54, 524.81, 526.08, 527.37, 528.67, 529.98, 531.3, 532.62, 533.96, 535.3, 536.02, 536.2, 536.42, 536.68, 536.99, 537.33, 537.72, 538.14, 538.61, 539.12, 539.66, 538.98, 535.32, 531.72, 528.17, 524.67, 521.23, 517.85, 514.52, 511.24, 508.03, 504.87, 501.68, 497.15, 492.63, 488.13, 483.63, 479.14, 474.67, 470.2, 465.75, 461.31, 456.87, 452.45, 450.47, 2459.49, 2443.76, 2427.99, 2411.04, 2384.74, 2358.48, 2332.26, 2306.09, 2279.95, 2253.86, 2227.81, 2201.81, 2175.84, 2149.92, 2124.04, 2083.86, 2041.68, 1999.12, 1956.18, 1912.86, 1869.15, 1825.06, 1780.58, 1735.72, 1690.47, 1644.83, 1619.48, 1603.56, 1587.57, 1571.54, 1555.44, 1539.29, 1523.08, 1506.81, 1490.49, 1474.11, 1457.67, 1437.58, 1413.21, 1388.54, 1363.59, 1338.34, 1312.8, 1286.96, 1260.84, 1234.41, 1207.7, 1180.69, 1152.74, 1122.92, 1093.29, 1063.83, 1034.56, 1005.47, 976.56, 947.84, 919.3, 890.95, 862.78, 835.37, 824.91, 814.62, 804.5, 794.55, 784.77, 775.16, 765.72, 756.46, 747.36, 738.44, 729.69, 725.08, 721.33, 717.51, 713.62, 709.64, 705.59, 701.47, 697.26, 692.98, 688.63, 684.19, 681.51, 680.69, 680.74, 680.68, 680.49, 680.18, 679.75, 679.2, 678.53, 677.73, 676.81, 686.7, 715.35, 743.49, 771.1, 798.2, 824.78, 850.84, 876.37, 901.39, 925.88, 949.85, 967.48, 947.85, 928.19, 908.49, 888.76, 868.99, 849.19, 829.35, 809.48, 789.57, 769.63, 749.65, 727.77, 705.91, 684.27, 662.85, 641.65, 620.67, 599.91, 579.37, 559.06, 538.97, 519.1, 513.54, 514.86, 516.18, 517.52, 518.86, 520.22, 521.58, 522.95, 524.33, 525.72, 527.12, 528.08, 528.48, 528.93, 529.41, 529.94, 530.5, 531.11, 531.75, 532.44, 533.17, 533.94, 533.9, 530.53, 527.22, 523.97, 520.77, 517.62, 514.54, 511.5, 508.52, 505.6, 502.74, 499.93, 495.52, 491.06, 486.61, 482.17, 477.74, 473.32, 468.91, 464.51, 460.12, 455.75, 451.38, 449.16, 2492.71, 2476.72, 2460.69, 2444.33, 2418.27, 2392.26, 2366.28, 2340.35, 2314.46, 2288.61, 2262.81, 2237.04, 2211.32, 2185.65, 2160.01, 2119.18, 2074.61, 2029.66, 1984.33, 1938.61, 1892.51, 1846.03, 1799.16, 1751.91, 1704.27, 1656.24, 1627.3, 1611.02, 1594.69, 1578.3, 1561.86, 1545.35, 1528.79, 1512.17, 1495.5, 1478.77, 1461.98, 1441.62, 1415.48, 1389.06, 1362.34, 1335.33, 1308.03, 1280.44, 1252.55, 1224.37, 1195.9, 1167.13, 1138.12, 1109.4, 1080.85, 1052.5, 1024.32, 996.33, 968.52, 940.89, 913.45, 886.19, 859.12, 832.23, 821.9, 812.61, 803.49, 794.54, 785.76, 777.16, 768.72, 760.45, 752.36, 744.44, 736.69, 731.61, 727.65, 723.91, 720.39, 717.1, 714.04, 711.2, 708.59, 706.21, 704.05, 702.12, 701.02, 700.49, 699.85, 699.08, 698.19, 697.17, 696.04, 694.79, 693.41, 691.91, 690.29, 696.22, 721.94, 747.15, 771.84, 796.01, 819.66, 842.79, 865.4, 887.49, 909.06, 930.1, 948.74, 928.93, 909.08, 889.2, 869.28, 849.33, 829.35, 809.33, 789.27, 769.18, 749.05, 728.89, 708.17, 687.54, 667.12, 646.92, 626.94, 607.18, 587.65, 568.33, 549.24, 530.37, 511.73, 504.8, 506.17, 507.55, 508.94, 510.34, 511.75, 513.17, 514.6, 516.03, 517.48, 518.94, 520.1, 520.72, 521.39, 522.1, 522.84, 523.63, 524.46, 525.33, 526.24, 527.19, 528.18, 528.76, 525.69, 522.67, 519.71, 516.81, 513.96, 511.17, 508.43, 505.75, 503.12, 500.55, 498.04, 493.88, 489.47, 485.08, 480.69, 476.32, 471.96, 467.6, 463.26, 458.93, 454.61, 450.3, 447.84, 2525.98, 2509.73, 2493.44, 2477.11, 2451.76, 2425.99, 2400.26, 2374.57, 2348.93, 2323.33, 2297.76, 2272.24, 2246.77, 2221.33, 2195.94, 2154.89, 2107.93, 2060.58, 2012.86, 1964.75, 1916.27, 1867.39, 1818.13, 1768.49, 1718.46, 1668.04, 1635.18, 1618.55, 1601.87, 1585.13, 1568.33, 1551.47, 1534.56, 1517.59, 1500.57, 1483.48, 1466.34, 1445.94, 1418.05, 1389.86, 1361.39, 1332.62, 1303.56, 1274.21, 1244.56, 1214.62, 1184.39, 1153.86, 1123.31, 1095.69, 1068.24, 1040.98, 1013.9, 987.0, 960.29, 933.76, 907.42, 881.26, 855.28, 829.49, 819.28, 811.56, 803.87, 796.22, 788.58, 780.98, 773.41, 765.87, 758.35, 750.87, 743.42, 738.92, 736.26, 733.81, 731.6, 729.6, 727.84, 726.3, 724.98, 723.89, 723.03, 722.4, 721.66, 720.43, 719.07, 717.6, 716.01, 714.29, 712.45, 710.5, 708.41, 706.21, 703.89, 706.26, 729.05, 751.33, 773.1, 794.34, 815.07, 835.27, 854.96, 874.12, 892.76, 910.88, 928.47, 910.04, 890.01, 869.94, 849.84, 829.71, 809.54, 789.34, 769.1, 748.83, 728.52, 708.17, 688.36, 668.94, 649.74, 630.77, 612.01, 593.48, 575.16, 557.07, 539.2, 521.55, 504.13, 496.05, 497.48, 498.91, 500.36, 501.81, 503.28, 504.75, 506.23, 507.72, 509.23, 510.74, 512.08, 512.92, 513.81, 514.74, 515.71, 516.72, 517.77, 518.86, 519.99, 521.16, 522.37, 523.56, 520.79, 518.07, 515.41, 512.8, 510.24, 507.75, 505.31, 502.92, 500.59, 498.31, 496.1, 492.23, 487.88, 483.54, 479.21, 474.89, 470.58, 466.29, 462.0, 457.72, 453.46, 449.2, 446.51, 2559.28, 2542.77, 2526.22, 2509.64, 2485.22, 2459.69, 2434.2, 2408.76, 2383.36, 2358.0, 2332.68, 2307.41, 2282.17, 2256.98, 2231.83, 2190.97, 2141.62, 2091.89, 2041.78, 1991.28, 1940.4, 1889.14, 1837.49, 1785.46, 1733.04, 1680.23, 1643.11, 1626.13, 1609.1, 1592.01, 1574.86, 1557.65, 1540.39, 1523.07, 1505.69, 1488.26, 1470.77, 1450.56, 1420.9, 1390.96, 1360.73, 1330.2, 1299.38, 1268.27, 1236.86, 1205.16, 1173.17, 1140.88, 1109.38, 1084.12, 1058.94, 1033.83, 1008.79, 983.83, 958.94, 934.12, 909.37, 884.7, 860.09, 835.56, 824.28, 816.73, 809.21, 801.72, 794.26, 786.82, 779.42, 772.04, 764.7, 757.38, 750.09, 746.01, 744.64, 743.5, 742.58, 741.88, 741.41, 741.17, 741.15, 741.35, 741.79, 742.45, 742.41, 740.48, 738.42, 736.25, 733.95, 731.53, 728.99, 726.33, 723.54, 720.64, 717.61, 716.81, 736.68, 756.03, 774.87, 793.19, 810.99, 828.27, 845.03, 861.27, 876.98, 892.18, 906.84, 891.18, 870.97, 850.72, 830.44, 810.12, 789.77, 769.38, 748.96, 728.5, 708.01, 687.48, 668.32, 650.12, 632.15, 614.39, 596.86, 579.55, 562.45, 545.58, 528.94, 512.51, 496.31, 487.29, 488.77, 490.26, 491.76, 493.27, 494.79, 496.32, 497.86, 499.4, 500.96, 502.53, 504.02, 505.08, 506.19, 507.34, 508.53, 509.76, 511.03, 512.34, 513.7, 515.09, 516.52, 518.0, 515.83, 513.41, 511.04, 508.73, 506.47, 504.27, 502.12, 500.03, 498.0, 496.02, 494.09, 490.57, 486.27, 481.99, 477.72, 473.45, 469.2, 464.96, 460.73, 456.51, 452.3, 448.1, 445.17, 2592.62, 2575.86, 2559.05, 2542.21, 2518.63, 2493.35, 2468.1, 2442.91, 2417.75, 2392.63, 2367.56, 2342.53, 2317.54, 2292.59, 2267.69, 2227.44, 2175.7, 2123.58, 2071.08, 2018.19, 1964.92, 1911.27, 1857.23, 1802.81, 1748.0, 1692.81, 1651.1, 1633.77, 1616.39, 1598.94, 1581.45, 1563.89, 1546.28, 1528.61, 1510.88, 1493.09, 1476.11, 1457.01, 1426.56, 1395.91, 1365.07, 1334.04, 1302.82, 1271.4, 1239.79, 1207.99, 1175.99, 1143.8, 1111.42, 1085.99, 1061.23, 1036.55, 1011.94, 987.41, 962.94, 938.55, 914.23, 889.99, 865.82, 841.72, 829.26, 821.88, 814.52, 807.2, 799.9, 792.64, 785.4, 778.19, 771.01, 763.86, 756.74, 752.88, 752.8, 752.96, 753.33, 753.93, 754.76, 755.81, 757.09, 758.59, 760.32, 762.28, 763.29, 760.65, 757.9, 755.02, 752.02, 748.89, 745.65, 742.29, 738.8, 735.19, 731.46, 727.88, 744.82, 761.25, 777.16, 792.56, 807.43, 821.79, 835.62, 848.94, 861.73, 874.0, 885.74, 872.35, 851.96, 831.53, 811.06, 790.56, 770.03, 749.46, 728.86, 708.22, 687.54, 666.83, 648.06, 631.08, 614.33, 597.8, 581.49, 565.39, 549.52, 533.87, 518.45, 503.24, 488.26, 478.51, 480.05, 481.6, 483.15, 484.72, 486.29, 487.88, 489.47, 491.07, 492.69, 494.31, 495.91, 497.2, 498.53, 499.9, 501.31, 502.77, 504.26, 505.79, 507.36, 508.98, 510.64, 512.33, 510.82, 508.69, 506.62, 504.6, 502.64, 500.73, 498.88, 497.09, 495.35, 493.67, 492.04, 488.89, 484.66, 480.43, 476.21, 472.0, 467.81, 463.62, 459.45, 455.28, 451.13, 446.98, 443.83, 2626.01, 2608.98, 2591.92, 2574.82, 2552.01, 2526.97, 2501.97, 2477.01, 2452.1, 2427.23, 2402.4, 2377.61, 2352.87, 2328.16, 2303.5, 2264.3, 2210.17, 2155.66, 2100.76, 2045.49, 1989.83, 1933.79, 1877.37, 1820.79, 1765.31, 1710.61, 1667.36, 1651.56, 1635.37, 1618.82, 1601.89, 1584.58, 1566.89, 1548.83, 1530.39, 1511.57, 1492.37, 1472.09, 1440.47, 1408.65, 1376.65, 1344.45, 1312.07, 1279.48, 1246.71, 1213.74, 1180.58, 1147.23, 1113.68, 1087.78, 1063.45, 1039.2, 1015.02, 990.91, 966.88, 942.92, 919.03, 895.21, 871.46, 847.79, 834.21, 826.99, 819.8, 812.65, 805.52, 798.42, 791.35, 784.31, 777.3, 770.32, 763.37, 759.52, 760.74, 762.19, 763.86, 765.76, 767.88, 770.23, 772.8, 775.6, 778.63, 781.88, 784.29, 780.95, 777.49, 773.91, 770.21, 766.38, 762.43, 758.37, 754.18, 749.86, 745.43, 740.87, 753.48, 766.99, 779.97, 792.44, 804.4, 815.83, 826.74, 837.13, 847.0, 856.34, 865.16, 853.56, 832.98, 812.37, 791.72, 771.04, 750.32, 729.57, 708.78, 687.96, 667.1, 646.21, 627.57, 611.82, 596.29, 580.98, 565.89, 551.02, 536.37, 521.94, 507.73, 493.75, 479.99, 469.73, 471.33, 472.93, 474.54, 476.16, 477.79, 479.43, 481.08, 482.73, 484.4, 486.08, 487.76, 489.28, 490.83, 492.42, 494.05, 495.73, 497.44, 499.2, 500.99, 502.83, 504.7, 506.62, 505.75, 503.92, 502.14, 500.42, 498.76, 497.14, 495.59, 494.09, 492.64, 491.26, 489.92, 487.21, 483.03, 478.86, 474.7, 470.54, 466.4, 462.27, 458.15, 454.05, 449.95, 445.86, 442.48, 2659.43, 2642.15, 2624.83, 2607.47, 2585.34, 2560.55, 2535.79, 2511.08, 2486.41, 2461.78, 2437.2, 2409.84, 2381.53, 2353.19, 2324.81, 2284.24, 2227.36, 2171.25, 2115.91, 2061.34, 2007.54, 1954.51, 1902.26, 1850.78, 1800.08, 1750.15, 1707.18, 1689.05, 1670.55, 1651.68, 1632.43, 1612.8, 1592.8, 1572.42, 1551.67, 1530.53, 1509.01, 1487.12, 1454.57, 1421.59, 1388.42, 1355.06, 1321.51, 1287.76, 1253.82, 1219.69, 1185.36, 1150.84, 1116.13, 1089.5, 1065.6, 1041.78, 1018.03, 994.35, 970.74, 947.21, 923.75, 900.36, 877.04, 853.8, 839.12, 832.08, 825.06, 818.07, 811.11, 804.18, 797.27, 790.4, 783.56, 776.74, 769.96, 765.94, 768.46, 771.2, 774.17, 777.36, 780.78, 784.42, 788.29, 792.39, 796.71, 801.26, 805.42, 801.37, 797.21, 792.92, 788.52, 783.99, 779.34, 774.57, 769.68, 764.66, 759.53, 754.27, 762.66, 773.24, 783.3, 792.85, 801.88, 810.39, 818.37, 825.84, 832.79, 839.21, 845.11, 834.8, 814.04, 793.24, 772.41, 751.55, 730.65, 709.71, 688.74, 667.74, 646.7, 625.62, 606.87, 592.34, 578.03, 563.94, 550.07, 536.42, 522.99, 509.78, 496.8, 484.04, 471.49, 460.94, 462.59, 464.24, 465.91, 467.59, 469.27, 470.97, 472.67, 474.38, 476.11, 477.84, 479.58, 481.32, 483.09, 484.9, 486.75, 488.65, 490.58, 492.56, 494.58, 496.63, 498.73, 500.87, 500.63, 499.09, 497.61, 496.19, 494.81, 493.5, 492.24, 491.03, 489.88, 488.79, 487.75, 485.52, 481.39, 477.28, 473.17, 469.07, 464.99, 460.91, 456.85, 452.8, 448.75, 444.72, 441.12, 2685.03, 2664.07, 2642.94, 2621.64, 2596.41, 2568.13, 2539.82, 2511.48, 2483.09, 2454.67, 2426.22, 2397.73, 2369.2, 2340.63, 2312.03, 2275.26, 2223.16, 2171.83, 2121.26, 2071.46, 2022.44, 1974.18, 1926.7, 1880.0, 1834.07, 1788.91, 1747.37, 1726.92, 1706.11, 1684.92, 1663.35, 1641.41, 1619.09, 1596.4, 1573.32, 1549.87, 1526.04, 1501.83, 1468.86, 1434.72, 1400.38, 1365.86, 1331.14, 1296.23, 1261.13, 1225.83, 1190.34, 1154.65, 1118.78, 1091.14, 1067.68, 1044.28, 1020.96, 997.71, 974.53, 951.43, 928.39, 905.43, 882.55, 859.73, 844.01, 837.13, 830.28, 823.46, 816.67, 809.9, 803.17, 796.46, 789.79, 783.14, 776.52, 772.14, 775.95, 779.99, 784.25, 788.74, 793.46, 798.4, 803.56, 808.95, 814.57, 820.41, 826.48, 821.92, 817.05, 812.06, 806.95, 801.72, 796.37, 790.9, 785.3, 779.58, 773.75, 767.78, 772.36, 780.01, 787.15, 793.77, 799.88, 805.46, 810.53, 815.07, 819.1, 822.6, 825.57, 816.07, 795.13, 774.15, 753.14, 732.09, 711.01, 689.89, 668.74, 647.55, 626.33, 605.07, 585.94, 572.64, 559.55, 546.68, 534.03, 521.6, 509.39, 497.4, 485.64, 474.1, 462.78, 452.14, 453.84, 455.55, 457.27, 459.01, 460.75, 462.5, 464.25, 466.02, 467.8, 469.59, 471.38, 473.31, 475.3, 477.34, 479.41, 481.53, 483.69, 485.88, 488.12, 490.4, 492.72, 495.08, 495.45, 494.21, 493.02, 491.89, 490.82, 489.8, 488.83, 487.92, 487.07, 486.27, 485.53, 483.81, 479.74, 475.68, 471.63, 467.59, 463.56, 459.54, 455.54, 451.54, 447.55, 443.58, 439.76, 2677.91, 2655.86, 2633.64, 2611.24, 2585.87, 2557.37, 2528.84, 2500.28, 2471.67, 2443.03, 2414.36, 2385.65, 2356.9, 2328.11, 2299.29, 2265.51, 2218.19, 2171.63, 2125.84, 2080.82, 2036.56, 1993.08, 1950.37, 1908.44, 1867.28, 1826.9, 1787.93, 1765.17, 1742.04, 1718.53, 1694.65, 1670.39, 1645.76, 1620.75, 1595.36, 1569.59, 1543.45, 1516.92, 1483.34, 1448.04, 1412.54, 1376.85, 1340.97, 1304.89, 1268.62, 1232.16, 1195.51, 1158.66, 1121.62, 1092.72, 1069.68, 1046.71, 1023.82, 1001.0, 978.25, 955.57, 932.97, 910.44, 887.98, 865.59, 848.87, 842.16, 835.48, 828.82, 822.2, 815.6, 809.03, 802.5, 795.99, 789.51, 783.06, 778.11, 783.22, 788.56, 794.12, 799.9, 805.91, 812.14, 818.6, 825.29, 832.2, 839.34, 846.71, 842.58, 837.01, 831.32, 825.51, 819.58, 813.52, 807.35, 801.05, 794.63, 788.09, 781.42, 782.57, 787.3, 791.52, 795.22, 798.4, 801.06, 803.21, 804.83, 805.93, 806.51, 806.56, 797.38, 776.25, 755.09, 733.9, 712.67, 691.4, 670.1, 648.76, 627.39, 605.99, 584.55, 564.8, 552.71, 540.84, 529.19, 517.76, 506.56, 495.57, 484.8, 474.26, 463.93, 453.83, 443.96, 445.08, 446.85, 448.63, 450.41, 452.21, 454.01, 455.83, 457.65, 459.48, 461.33, 463.18, 465.26, 467.48, 469.73, 472.03, 474.37, 476.75, 479.16, 481.62, 484.12, 486.66, 489.24, 490.22, 489.27, 488.38, 487.54, 486.76, 486.04, 485.37, 484.75, 484.2, 483.69, 483.25, 482.1, 478.08, 474.08, 470.08, 466.1, 462.13, 458.16, 454.21, 450.27, 446.34, 442.42, 438.51, 2670.97, 2647.82, 2624.51, 2601.02, 2575.36, 2546.65, 2517.9, 2489.11, 2460.29, 2431.43, 2402.53, 2373.6, 2344.63, 2315.63, 2286.59, 2255.0, 2212.45, 2170.67, 2129.65, 2089.4, 2049.92, 2011.21, 1973.27, 1936.11, 1899.73, 1864.12, 1829.29, 1803.8, 1778.35, 1752.53, 1726.33, 1699.76, 1672.81, 1645.48, 1617.78, 1589.69, 1561.23, 1532.39, 1498.02, 1461.55, 1424.89, 1388.03, 1350.99, 1313.75, 1276.32, 1238.69, 1200.87, 1162.86, 1124.65, 1094.22, 1071.61, 1049.08, 1026.61, 1004.22, 981.9, 959.65, 937.47, 915.37, 893.34, 871.38, 853.71, 847.16, 840.65, 834.16, 827.7, 821.27, 814.87, 808.5, 802.16, 795.85, 789.56, 783.86, 790.27, 796.9, 803.75, 810.83, 818.13, 825.66, 833.42, 841.4, 849.61, 858.04, 866.7, 863.37, 857.1, 850.71, 844.19, 837.56, 830.8, 823.92, 816.92, 809.8, 802.56, 795.19, 793.29, 795.11, 796.4, 797.18, 797.44, 797.18, 796.4, 795.1, 793.28, 790.94, 788.07, 778.72, 757.41, 736.06, 714.69, 693.27, 671.82, 650.34, 628.82, 607.27, 585.68, 564.06, 543.43, 532.56, 521.91, 511.49, 501.28, 491.29, 481.52, 471.98, 462.65, 453.55, 444.67, 436.01, 436.32, 438.14, 439.97, 441.81, 443.66, 445.52, 447.39, 449.27, 451.16, 453.06, 454.96, 457.18, 459.61, 462.09, 464.61, 467.17, 469.76, 472.4, 475.08, 477.8, 480.57, 483.37, 484.93, 484.28, 483.68, 483.14, 482.65, 482.22, 481.85, 481.53, 481.27, 481.06, 480.91, 480.37, 476.41, 472.46, 468.53, 464.6, 460.68, 456.77, 452.86, 448.96, 445.1, 441.27, 437.48, 2664.19, 2639.96, 2615.55, 2590.97, 2564.89, 2535.96, 2506.99, 2477.98, 2448.94, 2419.86, 2390.74, 2361.59, 2332.4, 2303.18, 2273.92, 2243.73, 2205.95, 2168.94, 2132.69, 2097.21, 2062.5, 2028.57, 1995.4, 1963.01, 1931.39, 1900.55, 1870.49, 1842.81, 1815.04, 1786.9, 1758.39, 1729.5, 1700.24, 1670.59, 1640.58, 1610.18, 1579.4, 1548.25, 1512.89, 1475.26, 1437.43, 1399.41, 1361.2, 1322.8, 1284.2, 1245.41, 1206.43, 1167.25, 1127.88, 1095.65, 1073.47, 1051.36, 1029.33, 1007.36, 985.47, 963.65, 941.9, 920.23, 898.62, 877.09, 858.51, 852.13, 845.79, 839.47, 833.17, 826.91, 820.68, 814.48, 808.3, 802.16, 796.04, 789.95, 797.09, 805.02, 813.17, 821.54, 830.14, 838.96, 848.01, 857.29, 866.79, 876.52, 886.47, 884.28, 877.31, 870.21, 863.0, 855.66, 848.2, 840.62, 832.92, 825.09, 817.15, 809.08, 804.54, 803.43, 801.8, 799.66, 797.0, 793.82, 790.12, 785.9, 781.16, 775.89, 770.11, 760.09, 738.6, 717.07, 695.51, 673.91, 652.28, 630.62, 608.92, 587.18, 565.41, 543.6, 521.84, 512.19, 502.77, 493.56, 484.57, 475.8, 467.25, 458.93, 450.82, 442.94, 435.28, 427.84, 427.54, 429.42, 431.3, 433.2, 435.11, 437.02, 438.95, 440.88, 442.82, 444.78, 446.74, 449.05, 451.71, 454.4, 457.14, 459.92, 462.74, 465.6, 468.5, 471.44, 474.43, 477.45, 479.58, 479.23, 478.93, 478.68, 478.49, 478.36, 478.28, 478.25, 478.26, 478.26, 478.25, 478.16, 474.2, 470.28, 466.38, 462.53, 458.71, 454.92, 451.17, 447.46, 443.78, 440.13, 436.53, 2657.59, 2632.26, 2606.76, 2581.09, 2554.45, 2525.3, 2496.11, 2466.88, 2437.62, 2408.32, 2378.99, 2349.62, 2320.21, 2290.76, 2261.28, 2231.69, 2198.68, 2166.44, 2134.96, 2104.26, 2074.32, 2045.15, 2016.76, 1989.13, 1962.29, 1936.22, 1910.93, 1882.19, 1852.11, 1821.65, 1790.83, 1759.62, 1728.04, 1696.09, 1663.75, 1631.04, 1597.95, 1564.48, 1527.95, 1489.15, 1450.16, 1410.98, 1371.61, 1332.04, 1292.28, 1252.33, 1212.18, 1171.84, 1131.31, 1097.01, 1075.26, 1053.58, 1031.97, 1010.44, 988.97, 967.58, 946.26, 925.01, 903.84, 882.74, 863.29, 857.08, 850.9, 844.74, 838.62, 832.53, 826.46, 820.42, 814.42, 808.44, 802.49, 796.57, 803.69, 812.91, 822.36, 832.02, 841.92, 852.03, 862.38, 872.95, 883.74, 894.77, 906.02, 905.32, 897.64, 889.84, 881.92, 873.88, 865.72, 857.44, 849.04, 840.51, 831.86, 823.09, 816.3, 812.27, 807.72, 802.66, 797.07, 790.97, 784.35, 777.21, 769.55, 761.37, 752.66, 741.49, 719.82, 698.11, 676.37, 654.59, 632.77, 610.93, 589.04, 567.12, 545.17, 523.18, 501.16, 491.6, 483.4, 475.41, 467.64, 460.09, 452.76, 445.66, 438.77, 432.11, 425.67, 419.45, 418.75, 420.68, 422.63, 424.58, 426.54, 428.51, 430.49, 432.48, 434.48, 436.48, 438.5, 440.88, 443.76, 446.68, 449.64, 452.64, 455.68, 458.76, 461.88, 465.3, 468.78, 472.22, 475.03, 475.06, 475.07, 475.07, 475.07, 475.06, 475.05, 475.03, 474.99, 474.96, 474.91, 474.86, 471.37, 467.63, 463.92, 460.25, 456.61, 453.01, 449.45, 445.92, 442.42, 438.96, 435.54, 2651.16, 2624.74, 2598.15, 2571.38, 2544.05, 2514.68, 2485.27, 2455.82, 2426.34, 2396.82, 2367.27, 2337.68, 2308.05, 2278.39, 2248.69, 2218.95, 2190.64, 2163.17, 2136.47, 2110.53, 2085.36, 2060.96, 2037.34, 2014.49, 1992.41, 1971.11, 1950.58, 1921.95, 1889.55, 1856.78, 1823.64, 1790.12, 1756.23, 1721.96, 1687.31, 1652.29, 1616.88, 1581.1, 1543.21, 1503.24, 1463.09, 1422.74, 1382.21, 1341.48, 1300.55, 1259.44, 1218.13, 1176.63, 1134.93, 1098.3, 1076.98, 1055.73, 1034.55, 1013.44, 992.4, 971.44, 950.54, 929.73, 908.98, 888.31, 868.04, 861.99, 855.98, 849.99, 844.04, 838.11, 832.21, 826.34, 820.5, 814.69, 808.91, 803.16, 810.07, 820.58, 831.32, 842.28, 853.47, 864.88, 876.52, 888.38, 900.47, 912.79, 925.33, 926.47, 918.09, 909.59, 900.97, 892.23, 883.37, 874.38, 865.28, 856.05, 846.7, 837.23, 828.58, 821.62, 814.16, 806.17, 797.67, 788.65, 779.11, 769.05, 758.47, 747.37, 735.74, 722.93, 701.07, 679.18, 657.25, 635.29, 613.3, 591.27, 569.2, 547.1, 524.96, 502.79, 480.58, 470.79, 463.8, 457.04, 450.49, 444.16, 438.05, 432.16, 426.5, 421.05, 415.83, 410.83, 409.95, 411.94, 413.94, 415.95, 417.96, 419.99, 422.02, 424.07, 426.52, 429.27, 432.0, 435.01, 438.6, 442.14, 445.65, 449.12, 452.55, 455.95, 459.3, 462.62, 465.89, 469.13, 472.07, 472.06, 472.03, 472.0, 471.96, 471.92, 471.87, 471.81, 471.74, 471.66, 471.58, 471.49, 468.5, 464.94, 461.42, 457.94, 454.49, 451.07, 447.69, 444.34, 441.03, 437.76, 434.52, 2644.9, 2617.39, 2589.7, 2561.85, 2533.69, 2504.09, 2474.46, 2444.8, 2415.1, 2385.36, 2355.58, 2325.77, 2295.93, 2266.04, 2236.12, 2206.16, 2181.84, 2159.14, 2137.2, 2116.04, 2095.64, 2076.01, 2057.15, 2039.06, 2021.75, 2005.22, 1989.46, 1962.08, 1927.37, 1892.29, 1856.83, 1821.0, 1784.8, 1748.21, 1711.25, 1673.91, 1636.2, 1598.1, 1558.65, 1517.53, 1476.21, 1434.7, 1393.0, 1351.11, 1309.02, 1266.74, 1224.27, 1181.6, 1138.74, 1099.52, 1078.62, 1057.8, 1037.05, 1016.37, 995.76, 975.22, 954.76, 934.37, 914.05, 893.8, 873.63, 866.88, 861.03, 855.21, 849.42, 843.66, 837.93, 832.23, 826.56, 820.91, 815.3, 809.71, 816.22, 828.03, 840.06, 852.32, 864.8, 877.51, 890.44, 903.6, 916.98, 930.59, 944.43, 947.75, 938.67, 929.47, 920.14, 910.7, 901.14, 891.45, 881.64, 871.71, 861.66, 851.49, 841.37, 831.5, 821.11, 810.2, 798.78, 786.84, 774.38, 761.4, 747.91, 733.89, 719.34, 704.28, 682.36, 660.29, 638.18, 616.03, 593.85, 571.64, 549.39, 527.11, 504.79, 482.44, 460.05, 449.76, 443.99, 438.44, 433.11, 428.0, 423.11, 418.45, 414.0, 410.32, 407.27, 404.25, 404.12, 406.94, 409.73, 412.5, 415.25, 417.97, 420.67, 423.34, 425.99, 428.62, 431.22, 434.0, 437.38, 440.72, 444.03, 447.29, 450.52, 453.71, 456.86, 459.98, 463.05, 466.09, 469.09, 469.07, 469.01, 468.94, 468.86, 468.78, 468.69, 468.59, 468.49, 468.38, 468.26, 468.13, 465.6, 462.23, 458.89, 455.59, 452.32, 449.09, 445.89, 442.73, 439.61, 436.52, 433.46, 2638.82, 2610.21, 2581.43, 2552.48, 2523.35, 2493.54, 2463.69, 2433.81, 2403.89, 2373.93, 2343.93, 2313.9, 2283.84, 2253.73, 2223.59, 2193.42, 2172.27, 2154.34, 2137.17, 2120.77, 2105.14, 2090.28, 2076.19, 2062.87, 2050.33, 2038.56, 2027.57, 2002.59, 1965.57, 1928.17, 1890.41, 1852.26, 1813.74, 1774.84, 1735.57, 1695.92, 1655.89, 1615.48, 1574.29, 1532.0, 1489.52, 1446.85, 1403.98, 1360.93, 1317.68, 1274.24, 1230.6, 1186.77, 1142.75, 1100.66, 1080.2, 1059.8, 1039.48, 1019.22, 999.04, 978.94, 958.9, 938.94, 919.04, 899.23, 879.48, 871.74, 866.06, 860.41, 854.78, 849.19, 843.63, 838.09, 832.59, 827.11, 821.66, 816.24, 822.16, 835.26, 848.58, 862.13, 875.91, 889.91, 904.13, 918.59, 933.26, 948.17, 963.3, 969.15, 959.37, 949.46, 939.44, 929.29, 919.03, 908.64, 898.13, 887.5, 876.75, 865.87, 854.68, 841.88, 828.58, 814.75, 800.41, 785.55, 770.18, 754.28, 737.86, 720.92, 703.46, 685.48, 663.68, 641.42, 619.13, 596.81, 574.44, 552.05, 529.62, 507.15, 485.68, 464.96, 444.32, 434.99, 431.83, 428.72, 425.64, 422.6, 419.61, 416.65, 413.74, 410.86, 408.03, 405.23, 404.65, 407.34, 410.01, 412.65, 415.26, 417.86, 420.42, 422.97, 425.49, 427.99, 430.46, 433.02, 436.2, 439.34, 442.44, 445.5, 448.53, 451.51, 454.46, 457.37, 460.24, 463.08, 465.87, 466.08, 465.98, 465.88, 465.77, 465.65, 465.52, 465.39, 465.25, 465.1, 464.94, 464.78, 462.66, 459.47, 456.32, 453.2, 450.12, 447.08, 444.06, 441.09, 438.15, 435.24, 432.37, 2632.9, 2603.21, 2573.34, 2543.29, 2513.08, 2483.02, 2452.96, 2422.85, 2392.71, 2362.53, 2332.32, 2302.07, 2271.78, 2241.46, 2211.1, 2180.7, 2161.94, 2148.78, 2136.38, 2124.74, 2113.88, 2103.78, 2094.46, 2085.91, 2078.13, 2071.12, 2064.9, 2043.48, 2004.15, 1964.44, 1924.35, 1883.9, 1843.06, 1801.85, 1760.27, 1718.3, 1675.96, 1633.24, 1590.12, 1546.67, 1503.02, 1459.19, 1415.16, 1370.94, 1326.53, 1281.93, 1237.13, 1192.14, 1146.96, 1101.74, 1081.7, 1061.73, 1041.83, 1022.01, 1002.26, 982.58, 962.97, 943.43, 923.97, 904.58, 885.26, 876.57, 871.05, 865.57, 860.12, 854.69, 849.29, 843.92, 838.59, 833.28, 827.99, 822.74, 827.86, 842.26, 856.88, 871.72, 886.79, 902.09, 917.6, 933.35, 949.32, 965.51, 981.94, 990.67, 980.19, 969.58, 958.86, 948.01, 937.04, 925.95, 914.74, 903.41, 891.96, 880.38, 868.5, 852.79, 836.56, 819.82, 802.56, 784.79, 766.49, 747.68, 728.34, 708.93, 689.66, 670.5, 649.89, 628.99, 608.16, 587.41, 566.74, 546.14, 525.62, 505.17, 484.8, 464.51, 444.29, 433.71, 430.78, 427.88, 425.03, 422.21, 419.44, 416.71, 414.01, 411.36, 408.75, 406.17, 405.21, 407.77, 410.3, 412.82, 415.3, 417.77, 420.21, 422.62, 425.01, 427.38, 429.72, 432.08, 435.06, 437.99, 440.89, 443.75, 446.57, 449.35, 452.1, 454.81, 457.48, 460.11, 462.7, 463.1, 462.97, 462.83, 462.68, 462.52, 462.36, 462.19, 462.01, 461.83, 461.63, 461.43, 459.69, 456.68, 453.72, 450.78, 447.89, 445.02, 442.2, 439.41, 436.65, 433.93, 431.24, 2627.16, 2596.37, 2565.41, 2534.28, 2502.97, 2472.54, 2442.25, 2411.93, 2381.57, 2351.17, 2320.74, 2290.27, 2259.76, 2229.22, 2198.64, 2168.03, 2150.84, 2142.45, 2134.81, 2127.94, 2121.84, 2116.51, 2111.96, 2108.17, 2105.15, 2102.92, 2101.45, 2084.75, 2043.1, 2001.08, 1958.68, 1915.91, 1872.76, 1829.24, 1785.35, 1741.07, 1696.42, 1651.38, 1606.14, 1561.53, 1516.72, 1471.72, 1426.53, 1381.15, 1335.58, 1289.81, 1243.85, 1197.7, 1151.35, 1104.81, 1083.13, 1063.59, 1044.12, 1024.72, 1005.4, 986.15, 966.96, 947.86, 928.82, 909.86, 890.97, 881.37, 876.02, 870.71, 865.42, 860.16, 854.93, 849.73, 844.56, 839.41, 834.3, 829.22, 833.35, 849.04, 864.95, 881.09, 897.45, 914.04, 930.85, 947.89, 965.15, 982.64, 1000.36, 1012.32, 1001.13, 989.83, 978.4, 966.85, 955.18, 943.39, 931.48, 918.91, 904.32, 889.81, 875.37, 855.97, 836.46, 817.06, 797.76, 778.56, 759.46, 740.46, 721.57, 702.77, 684.08, 665.49, 645.52, 625.04, 604.64, 584.32, 564.07, 543.9, 523.8, 503.79, 483.84, 463.98, 444.19, 432.4, 429.68, 427.01, 424.38, 421.78, 419.23, 416.72, 414.25, 411.82, 409.42, 407.07, 405.79, 408.22, 410.63, 413.01, 415.37, 417.7, 420.01, 422.3, 424.56, 426.8, 429.01, 431.2, 433.95, 436.68, 439.38, 442.04, 444.65, 447.23, 449.78, 452.28, 454.75, 457.17, 459.56, 460.13, 459.96, 459.78, 459.6, 459.41, 459.21, 459.0, 458.78, 458.56, 458.33, 458.1, 456.68, 453.86, 451.08, 448.33, 445.62, 442.94, 440.3, 437.69, 435.12, 432.58, 430.08, 2621.59, 2589.71, 2557.66, 2525.43, 2493.03, 2462.09, 2431.59, 2401.04, 2370.46, 2339.85, 2309.19, 2278.51, 2247.78, 2217.02, 2186.22, 2155.39, 2138.98, 2135.35, 2132.48, 2130.38, 2129.04, 2128.48, 2128.68, 2129.66, 2131.41, 2133.93, 2137.24, 2126.39, 2082.43, 2038.09, 1993.38, 1948.3, 1902.85, 1857.01, 1810.8, 1764.22, 1717.25, 1669.91, 1622.36, 1576.58, 1530.61, 1484.45, 1438.1, 1391.55, 1344.82, 1297.89, 1250.77, 1203.45, 1155.94, 1108.24, 1084.49, 1065.37, 1046.33, 1027.36, 1008.47, 989.64, 970.89, 952.21, 933.6, 915.07, 896.6, 886.14, 880.96, 875.81, 870.69, 865.6, 860.54, 855.5, 850.5, 845.52, 840.58, 835.66, 838.61, 855.6, 872.8, 890.23, 907.89, 925.77, 943.87, 962.2, 978.86, 994.64, 1010.11, 1021.74, 1007.01, 992.34, 977.74, 963.22, 948.76, 934.37, 920.06, 905.81, 891.63, 877.53, 863.49, 845.11, 826.18, 807.35, 788.62, 770.0, 751.47, 733.05, 714.73, 696.51, 678.39, 660.37, 641.07, 621.02, 601.05, 581.15, 561.33, 541.58, 521.91, 502.32, 482.81, 463.37, 444.01, 431.04, 428.55, 426.1, 423.68, 421.31, 418.98, 416.69, 414.44, 412.23, 410.06, 407.93, 406.4, 408.7, 410.97, 413.23, 415.45, 417.66, 419.84, 422.0, 424.13, 426.24, 428.32, 430.38, 432.88, 435.41, 437.9, 440.36, 442.77, 445.15, 447.49, 449.79, 452.05, 454.28, 456.46, 457.17, 456.96, 456.75, 456.53, 456.3, 456.06, 455.81, 455.56, 455.3, 455.04, 454.76, 453.64, 451.0, 448.4, 445.84, 443.31, 440.82, 438.36, 435.94, 433.55, 431.2, 428.88, 2616.19, 2583.22, 2550.08, 2516.76, 2483.27, 2451.68, 2420.96, 2390.19, 2359.39, 2328.56, 2297.68, 2266.78, 2235.83, 2204.85, 2173.83, 2142.78, 2126.35, 2127.49, 2129.38, 2132.04, 2135.47, 2139.67, 2144.64, 2150.38, 2156.89, 2164.18, 2172.24, 2168.4, 2122.13, 2075.49, 2028.47, 1981.07, 1933.3, 1885.16, 1836.64, 1787.74, 1738.47, 1688.81, 1638.78, 1591.82, 1544.69, 1497.37, 1449.85, 1402.15, 1354.25, 1306.16, 1257.88, 1209.4, 1160.73, 1111.87, 1085.77, 1067.09, 1048.48, 1029.93, 1011.46, 993.07, 974.74, 956.49, 938.31, 920.2, 902.17, 890.89, 885.87, 880.89, 875.94, 871.01, 866.12, 861.25, 856.38, 851.4, 846.3, 841.09, 841.83, 858.28, 874.43, 890.28, 905.83, 921.07, 936.02, 950.66, 965.0, 979.04, 992.77, 1005.31, 990.98, 976.72, 962.53, 948.42, 934.37, 920.39, 906.48, 892.64, 878.87, 865.17, 851.55, 834.14, 815.79, 797.54, 779.39, 761.34, 743.39, 725.54, 707.79, 690.14, 672.6, 655.16, 636.54, 616.92, 597.37, 577.9, 558.51, 539.19, 519.95, 500.78, 481.69, 462.68, 443.75, 429.64, 427.37, 425.14, 422.95, 420.8, 418.69, 416.62, 414.6, 412.61, 410.66, 408.75, 407.03, 409.2, 411.34, 413.47, 415.57, 417.64, 419.69, 421.72, 423.72, 425.7, 427.66, 429.59, 431.85, 434.18, 436.47, 438.72, 440.93, 443.11, 445.24, 447.34, 449.4, 451.42, 453.41, 454.21, 453.97, 453.72, 453.46, 453.19, 452.92, 452.64, 452.35, 452.05, 451.75, 451.44, 450.56, 448.11, 445.69, 443.31, 440.97, 438.66, 436.39, 434.15, 431.94, 429.78, 427.64, 2610.96, 2576.9, 2542.67, 2508.26, 2473.68, 2441.3, 2410.36, 2379.38, 2348.36, 2317.3, 2286.21, 2255.08, 2223.92, 2192.72, 2161.48, 2130.21, 2112.96, 2118.86, 2125.52, 2132.94, 2141.13, 2150.09, 2159.83, 2170.33, 2181.6, 2193.65, 2206.48, 2210.8, 2162.21, 2113.26, 2063.93, 2014.22, 1964.14, 1913.69, 1862.86, 1811.65, 1760.06, 1708.1, 1655.75, 1607.26, 1558.96, 1510.48, 1461.8, 1412.94, 1363.88, 1314.62, 1265.18, 1215.54, 1165.71, 1115.68, 1086.99, 1068.73, 1050.55, 1032.43, 1014.39, 996.42, 978.56, 960.94, 943.55, 926.38, 909.44, 897.86, 893.0, 888.03, 882.95, 877.75, 872.44, 867.02, 861.48, 855.83, 850.07, 844.19, 842.19, 856.9, 871.31, 885.41, 899.22, 912.73, 925.94, 938.84, 951.44, 963.74, 975.73, 987.42, 974.88, 961.03, 947.25, 933.54, 919.9, 906.33, 892.83, 879.4, 866.04, 852.75, 839.53, 823.08, 805.3, 787.62, 770.05, 752.57, 735.19, 717.92, 700.75, 683.68, 666.71, 649.84, 631.94, 612.75, 593.63, 574.58, 555.61, 536.72, 517.91, 499.17, 480.5, 461.92, 443.41, 428.2, 426.16, 424.15, 422.18, 420.25, 418.36, 416.52, 414.71, 412.94, 411.22, 409.53, 407.89, 409.72, 411.74, 413.73, 415.7, 417.65, 419.57, 421.47, 423.34, 425.19, 427.02, 428.82, 430.86, 432.99, 435.07, 437.12, 439.13, 441.1, 443.03, 444.93, 446.79, 448.6, 450.38, 451.26, 450.98, 450.69, 450.4, 450.09, 449.78, 449.47, 449.14, 448.81, 448.47, 448.12, 447.44, 445.18, 442.95, 440.75, 438.59, 436.47, 434.38, 432.32, 430.31, 428.32, 426.37, 2605.91, 2570.75, 2535.43, 2499.93, 2464.26, 2430.96, 2399.8, 2368.59, 2337.36, 2306.08, 2274.77, 2243.42, 2212.04, 2180.62, 2149.16, 2117.67, 2098.81, 2109.46, 2120.89, 2133.07, 2146.03, 2159.75, 2174.24, 2189.5, 2205.54, 2222.35, 2239.94, 2253.56, 2202.67, 2151.4, 2099.76, 2047.75, 1995.36, 1942.59, 1889.45, 1835.93, 1782.04, 1727.76, 1673.11, 1622.88, 1573.43, 1523.78, 1473.94, 1423.92, 1373.1, 1321.24, 1269.71, 1218.51, 1167.63, 1117.09, 1085.99, 1068.37, 1050.96, 1033.78, 1016.81, 1000.07, 983.55, 967.25, 951.17, 935.32, 919.69, 907.74, 902.21, 896.57, 890.82, 884.96, 878.99, 872.9, 866.7, 860.38, 853.95, 847.41, 842.84, 855.81, 868.48, 880.85, 892.92, 904.69, 916.15, 927.32, 938.18, 948.74, 958.99, 968.94, 958.72, 945.27, 931.9, 918.6, 905.37, 892.2, 879.11, 866.09, 853.13, 840.25, 827.44, 811.92, 794.71, 777.61, 760.61, 743.7, 726.9, 710.2, 693.6, 677.1, 660.71, 644.41, 627.26, 608.49, 589.8, 571.18, 552.64, 534.18, 515.79, 497.47, 479.24, 461.08, 443.0, 426.73, 424.9, 423.11, 421.37, 419.66, 417.99, 416.37, 414.78, 413.24, 411.73, 410.27, 408.85, 410.27, 412.16, 414.02, 415.86, 417.68, 419.47, 421.24, 422.99, 424.71, 426.4, 428.08, 429.91, 431.83, 433.71, 435.56, 437.36, 439.13, 440.86, 442.55, 444.21, 445.82, 447.4, 448.32, 448.0, 447.68, 447.35, 447.01, 446.66, 446.3, 445.94, 445.57, 445.19, 444.81, 444.29, 442.21, 440.17, 438.16, 436.18, 434.24, 432.34, 430.47, 428.63, 426.83, 425.07, 2601.02, 2564.78, 2528.36, 2491.77, 2455.01, 2420.65, 2389.27, 2357.85, 2326.39, 2294.9, 2263.37, 2231.8, 2200.2, 2168.56, 2136.88, 2105.17, 2083.89, 2099.31, 2115.49, 2132.44, 2150.15, 2168.64, 2187.89, 2207.91, 2228.71, 2250.28, 2272.62, 2295.74, 2243.5, 2189.93, 2135.98, 2081.65, 2024.71, 1965.72, 1906.62, 1847.42, 1788.1, 1728.68, 1669.15, 1615.39, 1563.88, 1512.69, 1461.83, 1411.3, 1361.1, 1311.22, 1261.68, 1212.46, 1163.57, 1115.01, 1082.87, 1066.56, 1050.47, 1034.6, 1018.95, 1003.53, 988.32, 973.34, 958.58, 944.04, 929.72, 917.72, 911.53, 905.23, 898.81, 892.29, 885.65, 878.89, 872.03, 865.05, 857.95, 850.75, 843.8, 855.03, 865.96, 876.59, 886.92, 896.95, 906.67, 916.1, 925.22, 934.04, 942.56, 950.77, 942.48, 929.44, 916.48, 903.58, 890.76, 878.0, 865.32, 852.7, 840.16, 827.68, 815.28, 800.65, 784.02, 767.49, 751.06, 734.73, 718.51, 702.38, 686.36, 670.43, 654.61, 638.89, 622.51, 604.17, 585.9, 567.71, 549.59, 531.55, 513.59, 495.7, 477.89, 460.16, 442.5, 425.21, 423.6, 422.04, 420.51, 419.03, 417.58, 416.18, 414.81, 413.49, 412.21, 410.97, 409.76, 410.84, 412.6, 414.33, 416.05, 417.73, 419.4, 421.04, 422.65, 424.24, 425.81, 427.36, 428.99, 430.71, 432.39, 434.03, 435.64, 437.2, 438.73, 440.22, 441.67, 443.08, 444.46, 445.39, 445.03, 444.67, 444.3, 443.92, 443.54, 443.15, 442.75, 442.34, 441.93, 441.51, 441.08, 439.21, 437.35, 435.53, 433.74, 431.98, 430.26, 428.57, 426.92, 425.31, 423.73, 2596.31, 2558.98, 2521.47, 2483.79, 2445.94, 2410.38, 2378.78, 2347.14, 2315.46, 2283.75, 2252.0, 2220.21, 2188.39, 2156.53, 2124.63, 2092.7, 2068.2, 2088.38, 2109.33, 2131.04, 2153.89, 2176.13, 2197.69, 2218.56, 2238.75, 2258.26, 2277.07, 2295.2, 2243.28, 2184.08, 2124.77, 2065.35, 2005.82, 1946.19, 1886.45, 1826.6, 1766.64, 1706.58, 1646.4, 1593.13, 1543.6, 1494.4, 1445.53, 1396.98, 1348.76, 1300.87, 1253.31, 1206.08, 1159.18, 1112.6, 1079.53, 1064.53, 1049.76, 1035.21, 1020.87, 1006.76, 992.87, 979.2, 965.76, 952.53, 939.53, 927.82, 920.97, 914.0, 906.92, 899.72, 892.42, 885.0, 877.47, 869.82, 862.07, 854.19, 846.21, 854.54, 863.73, 872.62, 881.22, 889.51, 897.5, 905.18, 912.57, 919.65, 926.43, 932.9, 926.17, 913.54, 900.98, 888.5, 876.08, 863.73, 851.45, 839.25, 827.11, 815.04, 803.05, 789.28, 773.23, 757.27, 741.42, 725.66, 710.01, 694.46, 679.01, 663.66, 648.41, 633.26, 617.68, 599.76, 581.92, 564.15, 546.47, 528.85, 511.32, 493.86, 476.47, 459.17, 441.94, 424.78, 422.27, 420.92, 419.62, 418.36, 417.13, 415.95, 414.81, 413.71, 412.64, 411.62, 410.64, 411.43, 413.06, 414.67, 416.25, 417.81, 419.34, 420.86, 422.34, 423.81, 425.24, 426.66, 428.11, 429.63, 431.11, 432.54, 433.95, 435.31, 436.63, 437.92, 439.17, 440.38, 441.55, 442.46, 442.07, 441.67, 441.26, 440.85, 440.43, 440.0, 439.56, 439.12, 438.67, 438.21, 437.75, 436.18, 434.5, 432.86, 431.25, 429.68, 428.14, 426.64, 425.18, 423.75, 422.35, 2591.77, 2553.34, 2514.75, 2475.98, 2437.03, 2400.14, 2368.32, 2338.67, 2310.71, 2282.55, 2254.21, 2225.68, 2196.96, 2168.06, 2138.96, 2109.68, 2083.48, 2104.21, 2124.26, 2143.63, 2162.32, 2180.33, 2197.66, 2214.31, 2230.27, 2245.55, 2260.15, 2274.05, 2227.07, 2167.22, 2107.27, 2047.21, 1987.04, 1926.76, 1866.38, 1805.89, 1745.29, 1684.58, 1623.77, 1570.53, 1522.99, 1475.78, 1428.89, 1382.33, 1336.1, 1290.19, 1244.61, 1199.37, 1154.45, 1109.86, 1075.97, 1062.29, 1048.83, 1035.59, 1022.57, 1009.77, 997.2, 984.84, 972.71, 960.8, 949.12, 938.03, 930.51, 922.88, 915.13, 907.28, 899.31, 891.22, 883.03, 874.72, 866.29, 857.76, 849.11, 854.36, 861.81, 868.96, 875.81, 882.37, 888.62, 894.57, 900.21, 905.56, 910.6, 915.34, 909.79, 897.57, 885.42, 873.34, 861.33, 849.39, 837.52, 825.72, 813.99, 802.33, 790.74, 777.81, 762.33, 746.95, 731.67, 716.49, 701.41, 686.43, 671.56, 656.78, 642.11, 627.53, 612.77, 595.28, 577.86, 560.53, 543.26, 526.08, 508.97, 491.93, 474.97, 458.09, 441.29, 424.56, 420.89, 419.77, 418.68, 417.64, 416.64, 415.68, 414.76, 413.88, 413.04, 412.24, 411.48, 412.05, 413.55, 415.03, 416.48, 417.91, 419.32, 420.7, 422.06, 423.39, 424.7, 425.99, 427.27, 428.59, 429.86, 431.1, 432.29, 433.45, 434.58, 435.66, 436.7, 437.71, 438.68, 439.54, 439.11, 438.67, 438.23, 437.78, 437.32, 436.85, 436.38, 435.9, 435.42, 434.92, 434.42, 433.11, 431.62, 430.16, 428.74, 427.35, 426.0, 424.68, 423.4, 422.15, 420.94, 2617.53, 2581.9, 2545.86, 2509.42, 2472.58, 2437.27, 2408.47, 2379.49, 2350.32, 2320.96, 2291.42, 2261.69, 2231.77, 2201.66, 2171.37, 2140.88, 2110.2, 2125.99, 2141.81, 2156.96, 2171.43, 2185.21, 2198.32, 2210.74, 2222.48, 2233.54, 2243.91, 2253.59, 2210.96, 2150.47, 2089.87, 2029.17, 1968.36, 1907.44, 1846.42, 1785.28, 1724.04, 1662.69, 1601.24, 1547.61, 1502.06, 1456.83, 1411.93, 1367.35, 1323.1, 1279.18, 1235.59, 1192.33, 1149.39, 1106.79, 1072.18, 1059.82, 1047.67, 1035.75, 1024.05, 1012.56, 1001.3, 990.26, 979.45, 968.85, 958.48, 948.36, 940.17, 931.87, 923.46, 914.94, 906.31, 897.56, 888.7, 879.72, 870.63, 861.43, 852.12, 854.47, 860.18, 865.6, 870.71, 875.53, 880.04, 884.25, 888.16, 891.77, 895.07, 898.07, 893.33, 881.52, 869.78, 858.11, 846.51, 834.98, 823.51, 812.12, 800.8, 789.55, 778.36, 766.24, 751.33, 736.53, 721.82, 707.21, 692.71, 678.31, 664.0, 649.8, 635.7, 621.7, 607.79, 590.72, 573.73, 556.82, 539.98, 523.22, 506.54, 489.93, 473.4, 456.94, 440.57, 424.27, 419.47, 418.57, 417.71, 416.89, 416.11, 415.37, 414.67, 414.01, 413.39, 412.81, 412.28, 412.7, 414.07, 415.41, 416.74, 418.04, 419.31, 420.57, 421.79, 423.0, 424.18, 425.34, 426.47, 427.58, 428.65, 429.68, 430.68, 431.64, 432.56, 433.44, 434.28, 435.08, 435.85, 436.58, 436.16, 435.68, 435.2, 434.72, 434.22, 433.72, 433.21, 432.69, 432.17, 431.64, 431.1, 430.0, 428.7, 427.42, 426.18, 424.98, 423.81, 422.68, 421.58, 420.52, 419.49, 2674.47, 2636.23, 2597.58, 2558.53, 2519.08, 2480.68, 2450.68, 2420.49, 2390.12, 2359.57, 2328.82, 2297.89, 2266.77, 2235.46, 2203.96, 2172.27, 2140.4, 2148.45, 2160.05, 2170.97, 2181.21, 2190.78, 2199.66, 2207.86, 2215.37, 2222.21, 2228.35, 2233.81, 2194.95, 2133.82, 2072.58, 2011.24, 1949.78, 1888.23, 1826.56, 1764.78, 1702.9, 1640.91, 1578.81, 1524.36, 1480.79, 1437.55, 1394.63, 1352.04, 1309.78, 1267.84, 1226.23, 1184.95, 1144.0, 1103.38, 1068.18, 1057.13, 1046.3, 1035.69, 1025.3, 1015.13, 1005.19, 995.46, 985.96, 976.68, 967.62, 958.79, 949.94, 940.98, 931.91, 922.72, 913.42, 904.01, 894.48, 884.84, 875.09, 865.22, 855.24, 854.88, 858.86, 862.53, 865.91, 868.99, 871.76, 874.24, 876.41, 878.28, 879.85, 881.11, 876.81, 865.41, 854.07, 842.81, 831.62, 820.49, 809.44, 798.45, 787.54, 776.69, 765.92, 754.57, 740.24, 726.0, 711.87, 697.84, 683.91, 670.08, 656.35, 642.72, 629.19, 615.77, 602.44, 586.09, 569.52, 553.04, 536.63, 520.29, 504.04, 487.85, 471.75, 455.72, 439.77, 423.89, 418.01, 417.33, 416.7, 416.1, 415.54, 415.02, 414.54, 414.1, 413.71, 413.35, 413.03, 413.36, 414.6, 415.82, 417.02, 418.19, 419.33, 420.46, 421.56, 422.63, 423.68, 424.71, 425.71, 426.61, 427.48, 428.31, 429.1, 429.86, 430.57, 431.25, 431.89, 432.49, 433.06, 433.58, 433.21, 432.7, 432.18, 431.66, 431.13, 430.59, 430.04, 429.49, 428.93, 428.36, 427.79, 426.86, 425.74, 424.65, 423.6, 422.58, 421.59, 420.64, 419.73, 418.85, 418.01, 2731.81, 2690.97, 2649.71, 2608.05, 2565.99, 2524.28, 2493.08, 2461.69, 2430.12, 2398.36, 2366.41, 2334.28, 2301.96, 2269.45, 2236.75, 2203.86, 2170.78, 2171.59, 2178.97, 2185.66, 2191.68, 2197.02, 2201.68, 2205.66, 2208.95, 2211.56, 2213.48, 2214.72, 2179.05, 2117.27, 2055.4, 1993.41, 1931.32, 1869.11, 1806.81, 1744.39, 1681.87, 1619.23, 1556.49, 1500.79, 1459.2, 1417.94, 1377.01, 1336.4, 1296.12, 1256.17, 1216.55, 1177.25, 1138.28, 1099.65, 1063.96, 1054.22, 1044.71, 1035.41, 1026.34, 1017.48, 1008.85, 1000.44, 992.25, 984.28, 976.54, 969.02, 959.83, 950.2, 940.46, 930.61, 920.65, 910.57, 900.38, 890.07, 879.66, 869.13, 858.48, 855.59, 857.83, 859.77, 861.41, 862.75, 863.79, 864.53, 864.96, 865.1, 864.93, 864.45, 860.22, 849.22, 838.3, 827.44, 816.65, 805.93, 795.29, 784.71, 774.2, 763.76, 753.4, 742.8, 729.04, 715.38, 701.82, 688.36, 675.0, 661.74, 648.59, 635.53, 622.58, 609.73, 596.98, 581.38, 565.24, 549.18, 533.2, 517.29, 501.45, 485.7, 470.02, 454.42, 438.89, 423.44, 416.52, 416.06, 415.64, 415.26, 414.93, 414.63, 414.37, 414.15, 413.98, 413.84, 413.75, 414.05, 415.16, 416.25, 417.32, 418.36, 419.38, 420.37, 421.34, 422.29, 423.21, 424.11, 424.98, 425.68, 426.35, 426.97, 427.56, 428.12, 428.63, 429.11, 429.54, 429.94, 430.3, 430.63, 430.27, 429.73, 429.17, 428.61, 428.04, 427.47, 426.89, 426.3, 425.7, 425.09, 424.48, 423.69, 422.75, 421.91, 421.21, 420.52, 419.84, 419.18, 418.52, 417.87, 417.23, 2789.57, 2746.12, 2702.25, 2657.98, 2613.31, 2568.23, 2535.66, 2503.08, 2470.3, 2437.34, 2404.19, 2370.86, 2337.34, 2303.63, 2269.73, 2235.64, 2201.36, 2195.41, 2198.56, 2201.04, 2202.83, 2203.95, 2204.38, 2204.14, 2203.21, 2201.6, 2199.3, 2196.31, 2163.25, 2100.84, 2038.31, 1975.69, 1912.95, 1850.11, 1787.16, 1724.1, 1660.94, 1597.66, 1534.28, 1476.89, 1437.28, 1398.01, 1359.06, 1320.43, 1282.14, 1244.17, 1206.53, 1169.22, 1132.23, 1095.58, 1059.52, 1051.1, 1042.89, 1034.91, 1027.15, 1019.61, 1012.29, 1005.19, 998.32, 991.67, 985.24, 979.03, 969.83, 959.54, 949.13, 938.61, 927.99, 917.24, 906.39, 895.42, 884.34, 873.15, 861.84, 856.6, 857.1, 857.31, 857.21, 856.81, 856.12, 855.12, 853.82, 852.21, 850.31, 848.1, 843.55, 832.96, 822.45, 812.0, 801.62, 791.31, 781.06, 770.89, 760.79, 750.76, 740.8, 730.91, 717.74, 704.65, 691.67, 678.78, 666.0, 653.31, 640.73, 628.25, 615.87, 603.59, 591.41, 576.59, 560.88, 545.25, 529.69, 514.2, 498.8, 483.47, 468.21, 453.04, 437.94, 422.91, 414.98, 414.74, 414.55, 414.39, 414.27, 414.2, 414.16, 414.17, 414.21, 414.3, 414.42, 414.77, 415.75, 416.71, 417.65, 418.56, 419.45, 420.31, 421.15, 421.97, 422.76, 423.53, 424.28, 424.79, 425.25, 425.68, 426.06, 426.41, 426.72, 427.0, 427.23, 427.43, 427.59, 427.71, 427.34, 426.76, 426.17, 425.79, 425.4, 425.0, 424.6, 424.19, 423.77, 423.34, 422.91, 422.38, 421.71, 421.06, 420.41, 419.78, 419.16, 418.54, 417.94, 417.35, 416.77, 2847.74, 2801.67, 2755.2, 2708.33, 2661.04, 2613.35, 2578.44, 2544.65, 2510.68, 2476.52, 2442.17, 2407.63, 2372.91, 2338.0, 2302.9, 2267.61, 2232.13, 2219.91, 2218.84, 2217.09, 2214.66, 2211.56, 2207.77, 2203.3, 2198.15, 2192.32, 2185.8, 2178.6, 2147.56, 2084.5, 2021.34, 1958.07, 1894.69, 1831.21, 1767.62, 1703.92, 1640.11, 1576.2, 1512.17, 1452.65, 1415.04, 1377.74, 1340.78, 1304.14, 1267.82, 1231.84, 1196.18, 1160.85, 1125.85, 1091.18, 1056.84, 1047.75, 1040.86, 1034.19, 1027.74, 1021.52, 1015.51, 1009.73, 1004.17, 998.83, 993.71, 988.81, 979.94, 968.98, 957.91, 946.73, 935.44, 924.03, 912.51, 900.88, 889.14, 877.28, 865.31, 857.91, 856.67, 855.14, 853.31, 851.18, 848.74, 846.01, 842.97, 839.63, 835.99, 832.04, 826.82, 816.64, 806.52, 796.48, 786.51, 776.61, 766.77, 757.01, 747.32, 737.69, 728.14, 718.66, 706.33, 693.82, 681.41, 669.1, 656.89, 644.78, 632.77, 620.86, 609.05, 597.35, 585.74, 571.73, 556.44, 541.23, 526.1, 511.04, 496.06, 481.16, 466.33, 451.58, 436.9, 422.31, 413.4, 413.39, 413.41, 413.47, 413.58, 413.73, 413.91, 414.14, 414.4, 414.71, 415.06, 415.5, 416.36, 417.19, 418.0, 418.78, 419.54, 420.27, 420.99, 421.67, 422.34, 422.98, 423.6, 423.93, 424.19, 424.42, 424.66, 424.91, 425.16, 425.41, 425.66, 425.92, 426.17, 426.42, 426.29, 425.88, 425.47, 425.05, 424.63, 424.19, 423.75, 423.31, 422.85, 422.39, 421.92, 421.4, 420.79, 420.19, 419.61, 419.03, 418.46, 417.9, 417.35, 416.82, 416.29, 2906.32, 2857.64, 2808.56, 2759.08, 2709.19, 2658.89, 2621.41, 2586.42, 2551.24, 2515.88, 2480.33, 2444.6, 2408.67, 2372.56, 2336.26, 2299.77, 2263.09, 2245.08, 2239.79, 2233.82, 2227.17, 2219.84, 2211.84, 2203.15, 2193.78, 2183.73, 2172.99, 2161.56, 2131.97, 2068.27, 2004.47, 1940.56, 1876.54, 1812.41, 1748.18, 1683.84, 1619.39, 1554.84, 1490.17, 1428.1, 1392.46, 1357.15, 1322.17, 1287.51, 1253.18, 1219.18, 1185.5, 1152.16, 1119.14, 1086.45, 1054.09, 1044.18, 1038.6, 1033.25, 1028.12, 1023.2, 1018.51, 1014.04, 1009.79, 1005.76, 1001.96, 998.38, 990.16, 978.54, 966.81, 954.96, 943.01, 930.94, 918.75, 906.46, 894.05, 881.53, 868.89, 859.52, 856.55, 853.28, 849.71, 845.84, 841.67, 837.2, 832.42, 827.35, 821.97, 816.29, 810.01, 800.24, 790.53, 780.9, 771.33, 761.83, 752.41, 743.05, 733.77, 724.55, 715.4, 706.33, 694.83, 682.89, 671.05, 659.31, 647.68, 636.14, 624.7, 613.37, 602.13, 591.0, 579.97, 566.79, 551.93, 537.15, 522.44, 507.81, 493.25, 478.77, 464.37, 450.05, 435.8, 421.62, 411.78, 411.99, 412.23, 412.52, 412.85, 413.21, 413.62, 414.07, 414.55, 415.08, 415.65, 416.26, 416.99, 417.69, 418.37, 418.97, 419.53, 420.1, 420.66, 421.23, 421.79, 422.35, 422.92, 423.24, 423.49, 423.75, 424.0, 424.25, 424.5, 424.76, 425.01, 425.26, 425.51, 425.76, 425.67, 425.23, 424.78, 424.33, 423.87, 423.4, 422.92, 422.44, 421.94, 421.45, 420.94, 420.42, 419.86, 419.32, 418.79, 418.26, 417.75, 417.25, 416.75, 416.27, 415.8, 2965.3, 2914.02, 2862.34, 2810.24, 2757.75, 2704.84, 2664.56, 2628.37, 2592.0, 2555.44, 2518.69, 2481.75, 2444.63, 2407.31, 2369.81, 2332.12, 2294.25, 2270.93, 2261.42, 2251.23, 2240.36, 2228.81, 2216.59, 2203.68, 2190.09, 2175.82, 2160.86, 2145.22, 2116.49, 2052.15, 1987.7, 1923.15, 1858.49, 1793.72, 1728.85, 1663.87, 1598.78, 1533.58, 1468.27, 1403.21, 1369.56, 1336.23, 1303.23, 1270.56, 1238.21, 1206.19, 1174.5, 1143.13, 1112.1, 1081.39, 1051.01, 1040.39, 1036.13, 1032.09, 1028.27, 1024.67, 1021.29, 1018.13, 1015.19, 1012.48, 1009.99, 1007.72, 1000.5, 988.21, 975.82, 963.31, 950.69, 937.95, 925.11, 912.15, 899.07, 885.89, 872.59, 861.42, 856.72, 851.71, 846.41, 840.8, 834.89, 828.69, 822.18, 815.37, 808.26, 800.84, 793.12, 783.77, 774.47, 765.24, 756.08, 746.99, 737.97, 729.02, 720.15, 711.34, 702.6, 693.93, 683.23, 671.86, 660.6, 649.43, 638.36, 627.4, 616.54, 605.77, 595.11, 584.55, 574.1, 561.77, 547.34, 532.98, 518.7, 504.5, 490.37, 476.31, 462.34, 448.44, 434.61, 420.86, 410.12, 410.55, 411.02, 411.53, 412.09, 412.66, 413.23, 413.8, 414.36, 414.92, 415.48, 416.03, 416.59, 417.16, 417.72, 418.29, 418.85, 419.41, 419.98, 420.54, 421.1, 421.67, 422.23, 422.58, 422.84, 423.09, 423.34, 423.59, 423.85, 424.1, 424.35, 424.6, 424.86, 425.11, 425.07, 424.59, 424.1, 423.61, 423.11, 422.6, 422.09, 421.57, 421.04, 420.51, 419.97, 419.42, 418.92, 418.43, 417.96, 417.49, 417.03, 416.58, 416.15, 415.72, 415.3, 3024.7, 2970.81, 2916.52, 2861.82, 2806.71, 2751.2, 2707.91, 2670.52, 2632.94, 2595.18, 2557.23, 2519.1, 2480.77, 2442.26, 2403.56, 2364.67, 2325.59, 2297.46, 2283.73, 2269.32, 2254.23, 2238.46, 2222.02, 2204.89, 2187.09, 2168.59, 2149.42, 2129.56, 2101.11, 2036.13, 1971.04, 1905.85, 1840.54, 1775.14, 1709.62, 1644.0, 1578.27, 1512.43, 1446.48, 1380.43, 1346.33, 1314.99, 1283.97, 1253.28, 1222.91, 1192.87, 1163.16, 1133.78, 1104.72, 1076.0, 1047.6, 1036.38, 1033.43, 1030.71, 1028.2, 1025.91, 1023.84, 1022.0, 1020.37, 1018.97, 1017.79, 1016.84, 1010.95, 998.0, 984.94, 971.76, 958.48, 945.08, 931.57, 917.95, 904.21, 890.36, 876.4, 863.63, 857.19, 850.44, 843.4, 836.06, 828.42, 820.48, 812.24, 803.69, 794.85, 785.7, 776.24, 767.22, 758.33, 749.51, 740.76, 732.08, 723.47, 714.92, 706.45, 698.05, 689.72, 681.46, 671.52, 660.73, 650.04, 639.44, 628.95, 618.56, 608.27, 598.08, 587.99, 578.0, 568.12, 556.68, 542.67, 528.74, 514.89, 501.62, 488.43, 475.17, 461.82, 448.38, 434.87, 421.27, 409.28, 409.85, 410.41, 410.98, 411.54, 412.09, 412.64, 413.19, 413.74, 414.28, 414.82, 415.35, 415.91, 416.47, 417.04, 417.6, 418.16, 418.73, 419.29, 419.86, 420.42, 420.98, 421.55, 421.93, 422.18, 422.43, 422.69, 422.94, 423.19, 423.44, 423.69, 423.95, 424.2, 424.45, 424.47, 423.95, 423.43, 422.9, 422.36, 421.82, 421.27, 420.71, 420.15, 419.58, 419.0, 418.42, 417.97, 417.54, 417.12, 416.7, 416.3, 415.91, 415.53, 415.16, 414.79, 3084.5, 3028.01, 2971.11, 2913.8, 2856.09, 2797.98, 2751.45, 2712.86, 2674.08, 2635.12, 2595.97, 2556.63, 2517.11, 2477.4, 2437.5, 2397.41, 2357.13, 2324.67, 2306.72, 2288.09, 2268.78, 2248.8, 2228.13, 2206.79, 2184.76, 2162.05, 2138.66, 2114.58, 2085.83, 2020.21, 1954.48, 1888.65, 1822.71, 1756.66, 1690.5, 1624.24, 1557.87, 1491.39, 1424.8, 1358.1, 1322.77, 1293.41, 1264.38, 1235.66, 1207.28, 1179.22, 1151.49, 1124.09, 1097.02, 1070.27, 1043.86, 1032.15, 1030.52, 1029.1, 1027.91, 1026.93, 1026.18, 1025.64, 1025.33, 1025.24, 1025.38, 1025.73, 1021.51, 1007.9, 994.17, 980.33, 966.39, 952.32, 938.15, 923.86, 909.46, 894.95, 880.32, 866.13, 857.95, 849.48, 840.7, 831.63, 822.25, 812.57, 802.6, 792.32, 781.73, 770.85, 759.66, 750.61, 742.13, 733.71, 725.37, 717.09, 708.89, 700.75, 692.69, 684.69, 676.77, 668.91, 659.71, 649.49, 639.37, 629.5, 621.41, 613.17, 604.79, 596.25, 587.56, 578.72, 569.74, 559.39, 546.08, 532.68, 519.2, 505.64, 491.99, 478.26, 464.45, 450.56, 436.58, 422.52, 408.81, 409.36, 409.9, 410.45, 410.99, 411.52, 412.06, 412.59, 413.11, 413.64, 414.16, 414.67, 415.22, 415.79, 416.35, 416.92, 417.48, 418.04, 418.61, 419.17, 419.73, 420.3, 420.86, 421.27, 421.52, 421.78, 422.03, 422.28, 422.53, 422.79, 423.04, 423.29, 423.54, 423.79, 423.87, 423.32, 422.76, 422.2, 421.62, 421.04, 420.46, 419.86, 419.26, 418.65, 418.04, 417.41, 417.01, 416.63, 416.26, 415.91, 415.56, 415.22, 414.9, 414.58, 414.28, 3144.72, 3085.62, 3026.11, 2966.2, 2905.88, 2845.16, 2795.17, 2755.38, 2715.41, 2675.24, 2634.9, 2594.36, 2553.64, 2512.73, 2471.63, 2430.34, 2388.86, 2352.55, 2330.38, 2307.53, 2284.01, 2259.81, 2234.93, 2209.36, 2183.12, 2156.2, 2128.59, 2100.29, 2070.66, 2004.4, 1938.03, 1871.55, 1804.97, 1738.28, 1671.49, 1604.58, 1537.57, 1470.45, 1403.22, 1335.89, 1298.89, 1271.51, 1244.45, 1217.72, 1191.32, 1165.24, 1139.5, 1114.07, 1088.98, 1064.22, 1039.79, 1027.71, 1027.38, 1027.28, 1027.4, 1027.73, 1028.29, 1029.07, 1030.07, 1031.29, 1032.74, 1034.41, 1032.18, 1017.91, 1003.52, 989.02, 974.4, 959.68, 944.84, 929.89, 914.83, 899.65, 884.36, 868.96, 859.02, 848.81, 838.3, 827.49, 816.38, 804.97, 793.26, 781.24, 768.93, 756.31, 743.38, 733.93, 725.85, 717.84, 711.56, 706.74, 701.77, 696.65, 691.39, 685.98, 680.42, 674.71, 667.68, 659.2, 650.56, 641.77, 632.84, 623.75, 614.52, 605.13, 595.6, 585.91, 576.08, 565.33, 551.55, 537.69, 523.75, 509.73, 495.63, 481.44, 467.17, 452.82, 438.38, 423.86, 409.26, 408.87, 409.4, 409.92, 410.44, 410.96, 411.48, 411.99, 412.5, 413.0, 413.5, 414.0, 414.54, 415.1, 415.67, 416.23, 416.79, 417.36, 417.92, 418.49, 419.05, 419.61, 420.17, 420.61, 420.87, 421.12, 421.37, 421.62, 421.88, 422.13, 422.38, 422.63, 422.88, 423.14, 423.29, 422.7, 422.1, 421.5, 420.89, 420.27, 419.65, 419.02, 418.38, 417.74, 417.08, 416.42, 416.04, 415.72, 415.4, 415.1, 414.81, 414.53, 414.26, 413.99, 413.74, 3205.34, 3143.63, 3081.52, 3019.01, 2956.08, 2892.76, 2839.09, 2798.1, 2756.92, 2715.56, 2674.01, 2632.28, 2590.36, 2548.25, 2505.95, 2463.46, 2420.79, 2381.11, 2354.72, 2327.66, 2299.92, 2271.5, 2242.4, 2212.62, 2182.16, 2151.02, 2119.2, 2086.69, 2053.49, 1988.69, 1921.68, 1854.57, 1787.34, 1720.01, 1652.57, 1585.03, 1517.38, 1449.62, 1381.75, 1313.77, 1274.68, 1249.28, 1224.21, 1199.46, 1175.03, 1150.94, 1127.17, 1103.73, 1080.62, 1057.83, 1035.38, 1023.04, 1024.03, 1025.24, 1026.66, 1028.31, 1030.18, 1032.27, 1034.59, 1037.12, 1039.88, 1042.86, 1042.97, 1028.03, 1012.98, 997.81, 982.54, 967.15, 951.65, 936.03, 920.31, 904.47, 888.51, 872.45, 860.39, 848.44, 836.47, 826.52, 816.54, 806.54, 796.5, 786.45, 776.36, 766.25, 756.12, 749.23, 744.0, 738.62, 733.1, 727.42, 721.61, 715.64, 709.53, 703.27, 696.86, 690.31, 682.65, 673.31, 663.83, 654.19, 644.41, 634.48, 624.4, 614.17, 603.78, 593.25, 582.57, 571.35, 557.11, 542.79, 528.39, 513.91, 499.35, 484.7, 469.97, 455.16, 440.27, 425.29, 410.23, 408.38, 408.89, 409.4, 409.9, 410.4, 410.9, 411.39, 411.88, 412.37, 412.85, 413.33, 413.86, 414.42, 414.98, 415.55, 416.11, 416.67, 417.24, 417.8, 418.36, 418.93, 419.49, 419.96, 420.21, 420.46, 420.72, 420.97, 421.22, 421.47, 421.72, 421.98, 422.23, 422.48, 422.71, 422.08, 421.45, 420.81, 420.16, 419.51, 418.85, 418.18, 417.51, 416.82, 416.14, 415.44, 415.06, 414.79, 414.53, 414.28, 414.05, 413.82, 413.6, 413.4, 413.2, 3266.37, 3202.06, 3137.34, 3072.22, 3006.7, 2940.76, 2883.19, 2841.0, 2798.63, 2756.07, 2713.32, 2670.39, 2627.27, 2583.96, 2540.46, 2496.77, 2452.9, 2410.35, 2379.74, 2348.46, 2316.5, 2283.87, 2250.55, 2216.56, 2181.89, 2146.53, 2110.49, 2073.77, 2036.36, 1973.09, 1905.44, 1837.68, 1769.82, 1701.85, 1633.77, 1565.58, 1497.29, 1428.89, 1360.38, 1291.76, 1250.14, 1226.72, 1203.63, 1180.86, 1158.42, 1136.3, 1114.51, 1093.05, 1071.92, 1051.12, 1030.65, 1018.15, 1020.45, 1022.97, 1025.71, 1028.67, 1031.86, 1035.26, 1038.88, 1042.73, 1046.8, 1051.09, 1053.87, 1038.27, 1022.55, 1008.94, 995.41, 981.65, 967.66, 953.43, 938.96, 924.26, 909.31, 894.13, 883.17, 873.1, 863.01, 852.89, 842.75, 832.58, 822.38, 812.16, 801.91, 791.63, 781.33, 773.46, 767.38, 761.15, 754.78, 748.26, 741.59, 734.78, 727.82, 720.71, 713.46, 706.05, 697.77, 687.58, 677.25, 666.77, 656.14, 645.36, 634.43, 623.35, 612.12, 600.74, 589.21, 577.45, 562.75, 547.97, 533.12, 518.17, 503.15, 488.05, 472.86, 457.59, 442.23, 426.79, 411.27, 407.9, 408.39, 408.88, 409.36, 409.85, 410.32, 410.8, 411.27, 411.74, 412.2, 412.66, 413.17, 413.73, 414.3, 414.86, 415.43, 415.99, 416.55, 417.12, 417.68, 418.24, 418.8, 419.3, 419.55, 419.81, 420.06, 420.31, 420.56, 420.82, 421.07, 421.32, 421.57, 421.82, 422.07, 421.47, 420.8, 420.13, 419.44, 418.75, 418.06, 417.35, 416.64, 415.92, 415.2, 414.46, 414.06, 413.85, 413.65, 413.46, 413.27, 413.1, 412.94, 412.79, 412.65, 3327.81, 3260.89, 3193.57, 3125.85, 3057.72, 2989.18, 2927.49, 2884.1, 2840.52, 2796.76, 2752.82, 2708.69, 2664.37, 2619.86, 2575.16, 2530.28, 2485.21, 2440.26, 2405.44, 2369.94, 2333.77, 2296.92, 2259.39, 2221.18, 2182.29, 2142.72, 2102.47, 2061.53, 2019.91, 1957.59, 1889.3, 1820.9, 1752.4, 1683.79, 1615.07, 1546.24, 1477.31, 1408.27, 1339.12, 1269.86, 1225.28, 1203.84, 1182.72, 1161.94, 1141.47, 1121.34, 1101.53, 1082.05, 1062.9, 1044.07, 1025.58, 1013.04, 1016.66, 1022.53, 1029.16, 1036.05, 1043.19, 1050.58, 1058.23, 1066.14, 1074.3, 1082.72, 1091.36, 1077.13, 1062.68, 1047.99, 1033.06, 1017.9, 1002.5, 986.87, 971.0, 954.9, 938.55, 921.97, 910.06, 899.83, 889.57, 879.29, 868.98, 858.65, 848.28, 837.9, 827.48, 817.04, 806.58, 797.83, 790.9, 783.83, 776.61, 769.24, 761.73, 754.06, 746.26, 738.3, 730.2, 721.95, 713.03, 702.0, 690.82, 679.49, 668.01, 656.39, 644.61, 632.69, 620.61, 608.38, 596.01, 583.48, 568.47, 553.24, 537.92, 522.52, 507.04, 491.47, 475.82, 460.09, 444.28, 428.38, 412.41, 407.42, 407.89, 408.36, 408.83, 409.29, 409.75, 410.21, 410.66, 411.11, 411.56, 412.0, 412.49, 413.05, 413.61, 414.18, 414.74, 415.3, 415.87, 416.43, 416.99, 417.56, 418.12, 418.64, 418.9, 419.15, 419.4, 419.65, 419.91, 420.16, 420.41, 420.66, 420.91, 421.17, 421.42, 420.87, 420.16, 419.45, 418.73, 418.0, 417.27, 416.53, 415.78, 415.02, 414.26, 413.49, 413.06, 412.9, 412.76, 412.62, 412.49, 412.38, 412.27, 412.17, 412.09, 3389.66, 3320.14, 3250.21, 3179.88, 3109.15, 3038.01, 2971.97, 2927.38, 2882.61, 2837.65, 2792.51, 2747.18, 2701.66, 2655.95, 2610.06, 2563.98, 2517.7, 2471.24, 2431.81, 2392.1, 2351.71, 2310.65, 2268.9, 2226.48, 2183.38, 2139.6, 2095.13, 2049.98, 2004.14, 1942.2, 1873.27, 1804.23, 1735.08, 1665.83, 1596.47, 1527.01, 1457.43, 1387.75, 1317.96, 1248.07, 1200.08, 1183.28, 1166.99, 1150.93, 1135.12, 1119.54, 1104.19, 1089.09, 1074.23, 1059.6, 1045.22, 1034.91, 1042.53, 1050.4, 1058.53, 1066.91, 1075.54, 1084.43, 1093.57, 1102.98, 1112.63, 1122.55, 1132.72, 1119.23, 1103.37, 1087.28, 1070.95, 1054.39, 1037.59, 1020.55, 1003.28, 985.78, 968.03, 950.05, 936.99, 926.59, 916.17, 905.72, 895.24, 884.74, 874.22, 863.67, 853.09, 842.48, 831.85, 822.36, 814.58, 806.65, 798.59, 790.37, 782.01, 773.5, 764.84, 756.04, 747.09, 737.99, 728.45, 716.57, 704.54, 692.37, 680.04, 667.57, 654.94, 642.17, 629.25, 616.18, 602.95, 589.58, 574.28, 558.58, 542.81, 526.95, 511.0, 494.98, 478.87, 462.68, 446.41, 430.06, 413.62, 406.95, 407.4, 407.85, 408.3, 408.74, 409.18, 409.62, 410.06, 410.49, 410.91, 411.34, 411.8, 412.37, 412.93, 413.49, 414.06, 414.62, 415.18, 415.75, 416.31, 416.87, 417.43, 417.99, 418.24, 418.49, 418.75, 419.0, 419.25, 419.5, 419.75, 420.01, 420.26, 420.51, 420.76, 420.27, 419.53, 418.78, 418.02, 417.26, 416.49, 415.71, 414.93, 414.14, 413.34, 412.53, 412.05, 411.94, 411.85, 411.77, 411.7, 411.64, 411.59, 411.55, 411.52, 3451.91, 3379.79, 3307.26, 3234.33, 3160.99, 3087.25, 3016.64, 2970.86, 2924.89, 2878.73, 2832.39, 2785.86, 2739.14, 2692.24, 2645.14, 2597.86, 2550.39, 2502.74, 2458.86, 2414.93, 2370.33, 2325.06, 2279.1, 2232.47, 2185.15, 2137.16, 2088.48, 2039.11, 1989.02, 1926.97, 1858.04, 1789.56, 1721.51, 1653.91, 1586.75, 1520.03, 1453.75, 1387.92, 1322.53, 1257.59, 1210.55, 1195.45, 1180.58, 1165.95, 1151.55, 1137.4, 1123.48, 1109.8, 1096.36, 1083.16, 1070.2, 1059.54, 1068.65, 1078.02, 1087.64, 1097.51, 1107.64, 1118.03, 1128.67, 1139.56, 1150.71, 1162.12, 1173.79, 1161.56, 1144.3, 1126.8, 1109.07, 1091.11, 1072.91, 1054.48, 1035.8, 1016.9, 997.75, 978.37, 963.94, 953.38, 942.79, 932.18, 921.54, 910.87, 900.18, 889.46, 878.72, 867.95, 857.15, 847.03, 838.4, 829.63, 820.71, 811.65, 802.44, 793.08, 783.58, 773.93, 764.13, 754.19, 744.01, 731.29, 718.41, 705.39, 692.22, 678.9, 665.43, 651.81, 638.04, 624.12, 610.05, 595.83, 580.17, 564.01, 547.78, 531.46, 515.06, 498.57, 482.01, 465.36, 448.63, 431.81, 414.92, 406.47, 406.91, 407.34, 407.77, 408.2, 408.62, 409.04, 409.45, 409.87, 410.27, 410.68, 411.12, 411.68, 412.24, 412.81, 413.37, 413.93, 414.5, 415.06, 415.62, 416.19, 416.75, 417.31, 417.58, 417.84, 418.09, 418.34, 418.59, 418.85, 419.1, 419.35, 419.6, 419.85, 420.1, 419.68, 418.9, 418.12, 417.32, 416.52, 415.72, 414.9, 414.08, 413.25, 412.42, 411.57, 411.02, 410.98, 410.94, 410.91, 410.89, 410.89, 410.89, 410.91, 410.93, 3514.57, 3439.85, 3364.72, 3289.19, 3213.25, 3136.9, 3061.5, 3014.52, 2967.35, 2920.0, 2872.46, 2824.73, 2776.81, 2728.71, 2680.42, 2631.94, 2580.84, 2528.28, 2477.41, 2427.01, 2376.56, 2326.06, 2275.52, 2224.93, 2174.3, 2123.62, 2072.89, 2022.12, 1971.3, 1912.08, 1845.86, 1780.09, 1714.75, 1649.85, 1585.39, 1521.38, 1457.81, 1394.68, 1332.0, 1269.77, 1221.05, 1207.37, 1193.93, 1180.72, 1167.75, 1155.02, 1142.53, 1130.27, 1118.25, 1106.48, 1094.94, 1083.91, 1094.52, 1105.38, 1116.5, 1127.87, 1139.49, 1151.37, 1163.51, 1175.9, 1188.54, 1201.45, 1214.61, 1204.13, 1185.47, 1166.57, 1147.44, 1128.07, 1108.47, 1088.64, 1068.56, 1048.25, 1027.71, 1006.92, 990.92, 980.19, 969.44, 958.66, 947.86, 937.03, 926.17, 915.29, 904.38, 893.44, 882.48, 871.84, 862.37, 852.75, 842.99, 833.08, 823.02, 812.82, 802.47, 791.97, 781.32, 770.53, 759.59, 746.15, 732.43, 718.56, 704.55, 690.38, 676.06, 661.6, 646.98, 632.22, 617.3, 602.23, 586.14, 569.52, 552.83, 536.05, 519.19, 502.25, 485.22, 468.12, 450.93, 433.65, 416.3, 406.0, 406.42, 406.84, 407.25, 407.65, 408.06, 408.46, 408.86, 409.25, 409.64, 410.03, 410.43, 411.0, 411.56, 412.12, 412.69, 413.25, 413.81, 414.38, 414.94, 415.5, 416.06, 416.63, 416.93, 417.18, 417.43, 417.68, 417.94, 418.19, 418.44, 418.69, 418.94, 419.2, 419.45, 419.1, 418.29, 417.46, 416.63, 415.8, 414.95, 414.1, 413.24, 412.38, 411.51, 410.63, 409.99, 410.0, 410.01, 410.04, 410.08, 410.13, 410.19, 410.26, 410.34, 3577.65, 3500.32, 3422.59, 3342.75, 3260.35, 3178.7, 3097.82, 3045.48, 2993.8, 2941.81, 2889.49, 2836.86, 2783.9, 2730.63, 2677.03, 2623.11, 2568.87, 2514.3, 2462.46, 2411.79, 2361.06, 2310.3, 2259.48, 2208.62, 2157.71, 2106.76, 2055.76, 2004.72, 1953.63, 1896.76, 1833.25, 1770.17, 1707.54, 1645.35, 1583.6, 1522.29, 1461.42, 1401.0, 1341.02, 1281.49, 1231.32, 1219.06, 1207.04, 1195.26, 1183.71, 1172.41, 1161.34, 1150.5, 1139.91, 1129.56, 1119.44, 1109.57, 1120.14, 1132.5, 1145.11, 1157.97, 1171.09, 1184.46, 1198.09, 1211.98, 1226.12, 1240.52, 1255.17, 1246.93, 1226.87, 1206.57, 1186.04, 1165.28, 1144.28, 1123.04, 1101.56, 1079.85, 1057.91, 1035.72, 1017.93, 1007.04, 996.12, 985.17, 974.2, 963.21, 952.19, 941.14, 930.07, 918.97, 907.84, 896.81, 886.49, 876.02, 865.41, 854.65, 843.75, 832.7, 821.5, 810.16, 798.66, 787.03, 775.24, 761.17, 746.6, 731.89, 717.02, 702.01, 686.85, 671.54, 656.07, 640.46, 624.7, 608.79, 592.19, 575.12, 557.96, 540.72, 523.41, 506.0, 488.52, 470.95, 453.31, 435.57, 417.76, 405.54, 405.94, 406.33, 406.73, 407.11, 407.5, 407.88, 408.26, 408.64, 409.01, 409.38, 409.75, 410.31, 410.87, 411.44, 412.0, 412.56, 413.13, 413.69, 414.25, 414.82, 415.38, 415.94, 416.27, 416.52, 416.78, 417.03, 417.28, 417.53, 417.78, 418.04, 418.29, 418.54, 418.79, 418.53, 417.67, 416.81, 415.95, 415.07, 414.19, 413.31, 412.41, 411.51, 410.6, 409.68, 408.94, 409.0, 409.08, 409.16, 409.26, 409.36, 409.47, 409.6, 409.74, 3575.49, 3495.65, 3416.56, 3338.23, 3260.66, 3183.86, 3107.81, 3051.93, 2998.25, 2944.24, 2889.91, 2835.27, 2780.3, 2725.02, 2669.41, 2613.48, 2557.22, 2500.65, 2447.56, 2396.61, 2345.61, 2294.57, 2243.49, 2192.35, 2141.17, 2089.95, 2038.68, 1987.36, 1936.0, 1880.99, 1820.19, 1759.82, 1699.89, 1640.4, 1581.36, 1522.75, 1464.59, 1406.87, 1349.6, 1292.77, 1241.35, 1230.52, 1219.92, 1209.56, 1199.44, 1189.56, 1179.91, 1170.5, 1161.33, 1152.4, 1143.71, 1135.26, 1145.51, 1159.36, 1173.46, 1187.82, 1202.44, 1217.3, 1232.43, 1247.81, 1263.44, 1279.33, 1295.48, 1289.98, 1268.52, 1246.82, 1224.89, 1202.72, 1180.32, 1157.68, 1134.8, 1111.69, 1088.35, 1064.76, 1044.96, 1033.91, 1022.82, 1011.72, 1000.58, 989.42, 978.23, 967.02, 955.78, 944.52, 933.23, 921.92, 910.75, 899.44, 887.98, 876.38, 864.63, 852.73, 840.68, 828.49, 816.15, 803.67, 791.03, 776.34, 760.92, 745.36, 729.65, 713.79, 697.78, 681.62, 665.32, 648.86, 632.25, 615.49, 598.32, 580.79, 563.18, 545.48, 527.7, 509.84, 491.9, 473.88, 455.77, 437.58, 419.31, 405.08, 405.46, 405.83, 406.21, 406.58, 406.95, 407.31, 407.67, 408.03, 408.38, 408.73, 409.08, 409.63, 410.19, 410.75, 411.32, 411.88, 412.44, 413.01, 413.57, 414.13, 414.69, 415.26, 415.61, 415.87, 416.12, 416.37, 416.62, 416.88, 417.13, 417.38, 417.63, 417.88, 418.13, 417.96, 417.07, 416.17, 415.27, 414.36, 413.44, 412.52, 411.59, 410.65, 409.7, 408.75, 407.89, 408.0, 408.13, 408.27, 408.42, 408.58, 408.75, 408.93, 409.12, 3555.7, 3480.69, 3406.44, 3332.95, 3260.22, 3188.25, 3117.04, 3058.71, 3003.01, 2946.99, 2890.65, 2834.0, 2777.02, 2719.72, 2662.1, 2604.16, 2545.9, 2487.31, 2432.7, 2381.47, 2330.21, 2278.89, 2227.53, 2176.13, 2124.68, 2073.18, 2021.64, 1970.05, 1918.41, 1864.79, 1806.69, 1749.02, 1691.8, 1635.02, 1578.67, 1522.77, 1467.31, 1412.3, 1357.73, 1303.61, 1251.15, 1241.74, 1232.57, 1223.63, 1214.93, 1206.47, 1198.25, 1190.26, 1182.52, 1175.01, 1167.74, 1160.71, 1170.62, 1185.97, 1201.57, 1217.42, 1233.53, 1249.89, 1266.51, 1283.38, 1300.51, 1317.9, 1335.54, 1333.26, 1310.4, 1287.3, 1263.97, 1240.4, 1216.6, 1192.56, 1168.28, 1143.77, 1119.03, 1094.04, 1072.03, 1060.81, 1049.56, 1038.29, 1026.99, 1015.66, 1004.31, 992.93, 981.53, 970.1, 958.64, 947.16, 935.17, 923.01, 910.7, 898.25, 885.65, 872.91, 860.02, 846.98, 833.79, 820.46, 806.98, 791.65, 775.39, 758.99, 742.43, 725.72, 708.87, 691.86, 674.71, 657.41, 639.95, 622.35, 604.54, 586.55, 568.48, 550.32, 532.08, 513.77, 495.37, 476.88, 458.32, 439.67, 420.94, 404.62, 404.98, 405.34, 405.69, 406.05, 406.39, 406.74, 407.08, 407.42, 407.75, 408.09, 408.41, 408.94, 409.51, 410.07, 410.63, 411.19, 411.76, 412.32, 412.88, 413.45, 414.01, 414.57, 414.96, 415.21, 415.46, 415.71, 415.97, 416.22, 416.47, 416.72, 416.97, 417.23, 417.48, 417.4, 416.47, 415.54, 414.6, 413.65, 412.7, 411.74, 410.77, 409.79, 408.81, 407.82, 406.82, 406.99, 407.18, 407.37, 407.57, 407.79, 408.01, 408.25, 408.49, 3535.14, 3464.97, 3395.56, 3326.91, 3259.01, 3191.88, 3125.51, 3065.8, 3008.09, 2950.06, 2891.71, 2833.05, 2774.06, 2714.75, 2655.12, 2595.17, 2534.9, 2474.3, 2417.88, 2366.38, 2314.84, 2263.26, 2211.63, 2159.95, 2108.23, 2056.46, 2004.64, 1952.78, 1900.87, 1848.15, 1792.75, 1737.79, 1683.27, 1629.19, 1575.55, 1522.35, 1469.6, 1417.28, 1365.42, 1314.0, 1263.02, 1252.72, 1244.97, 1237.46, 1230.19, 1223.15, 1216.35, 1209.79, 1203.46, 1197.38, 1191.53, 1185.93, 1195.49, 1212.33, 1229.43, 1246.77, 1264.37, 1282.23, 1300.34, 1318.71, 1337.33, 1356.21, 1375.34, 1376.78, 1352.52, 1328.02, 1303.29, 1278.32, 1253.12, 1227.68, 1202.0, 1176.09, 1149.95, 1123.56, 1099.12, 1087.73, 1076.32, 1064.88, 1053.42, 1041.93, 1030.41, 1018.87, 1007.3, 995.71, 984.09, 972.44, 959.73, 946.72, 933.57, 920.27, 906.82, 893.23, 879.5, 865.61, 851.58, 837.4, 823.07, 807.12, 790.01, 772.76, 755.36, 737.81, 720.1, 702.25, 684.25, 666.1, 647.81, 629.36, 610.84, 592.39, 573.86, 555.24, 536.55, 517.77, 498.91, 479.97, 460.95, 441.84, 422.65, 404.16, 404.51, 404.85, 405.18, 405.52, 405.85, 406.17, 406.5, 406.82, 407.13, 407.45, 407.76, 408.26, 408.82, 409.38, 409.95, 410.51, 411.07, 411.64, 412.2, 412.76, 413.32, 413.89, 414.3, 414.55, 414.81, 415.06, 415.31, 415.56, 415.81, 416.07, 416.32, 416.57, 416.82, 416.84, 415.88, 414.91, 413.93, 412.95, 411.96, 410.96, 409.96, 408.95, 407.97, 407.12, 406.27, 406.53, 406.9, 407.26, 407.62, 407.97, 408.31, 408.65, 408.98, 3513.83, 3448.5, 3383.92, 3320.1, 3257.04, 3194.74, 3133.21, 3073.21, 3013.49, 2953.45, 2893.09, 2832.41, 2771.42, 2710.1, 2648.46, 2586.5, 2524.22, 2461.61, 2403.1, 2351.34, 2299.52, 2247.67, 2195.76, 2143.81, 2091.82, 2039.78, 1987.69, 1935.56, 1883.38, 1831.06, 1778.37, 1726.11, 1674.29, 1622.92, 1571.98, 1521.48, 1471.43, 1421.82, 1372.66, 1323.94, 1275.67, 1263.47, 1257.15, 1251.06, 1245.2, 1239.59, 1234.21, 1229.07, 1224.17, 1219.51, 1215.09, 1210.9, 1220.11, 1238.44, 1257.03, 1275.87, 1294.97, 1314.31, 1333.92, 1353.78, 1373.89, 1394.26, 1414.89, 1420.54, 1394.87, 1368.97, 1342.84, 1316.48, 1289.87, 1263.04, 1235.96, 1208.65, 1181.11, 1153.32, 1126.24, 1114.69, 1103.11, 1091.51, 1079.88, 1068.23, 1056.55, 1044.84, 1033.11, 1021.35, 1009.56, 997.75, 984.44, 970.58, 956.58, 942.44, 928.15, 913.71, 899.12, 884.39, 869.52, 854.49, 839.32, 822.74, 804.78, 786.68, 768.43, 750.04, 731.49, 712.8, 693.95, 674.95, 655.81, 636.51, 617.22, 598.31, 579.32, 560.25, 541.09, 521.86, 502.54, 483.14, 463.66, 444.09, 424.44, 404.71, 404.03, 404.36, 404.68, 404.99, 405.3, 405.61, 405.92, 406.22, 406.52, 406.81, 407.1, 407.57, 408.14, 408.7, 409.26, 409.83, 410.39, 410.95, 411.51, 412.08, 412.64, 413.2, 413.64, 413.9, 414.15, 414.4, 414.65, 414.91, 415.16, 415.41, 415.66, 415.91, 416.24, 416.45, 415.56, 414.67, 413.8, 412.93, 412.08, 411.24, 410.41, 409.58, 408.77, 407.97, 407.18, 407.31, 407.64, 407.97, 408.29, 408.6, 408.91, 409.21, 409.5, 3491.76, 3431.26, 3371.52, 3312.54, 3254.31, 3196.85, 3140.14, 3080.93, 3019.2, 2957.15, 2894.79, 2832.1, 2769.1, 2705.77, 2642.12, 2578.15, 2513.86, 2449.24, 2388.37, 2336.33, 2284.25, 2232.12, 2179.94, 2127.72, 2075.46, 2023.14, 1970.78, 1918.38, 1865.93, 1813.54, 1763.54, 1713.99, 1664.88, 1616.2, 1567.97, 1520.18, 1472.83, 1425.92, 1379.46, 1333.44, 1287.87, 1273.99, 1269.08, 1264.42, 1259.99, 1255.79, 1251.84, 1248.12, 1244.64, 1241.4, 1238.4, 1235.64, 1244.48, 1264.3, 1284.38, 1304.72, 1325.31, 1346.15, 1367.24, 1388.6, 1410.21, 1432.07, 1454.19, 1464.53, 1437.47, 1410.17, 1382.64, 1354.87, 1326.87, 1298.63, 1270.16, 1241.45, 1212.51, 1183.32, 1153.9, 1141.68, 1129.93, 1118.17, 1106.37, 1094.55, 1082.71, 1070.83, 1058.94, 1047.01, 1035.06, 1023.09, 1009.29, 994.59, 979.75, 964.75, 949.62, 934.33, 918.9, 903.33, 887.6, 871.73, 855.71, 838.5, 819.7, 800.76, 781.66, 762.42, 743.03, 723.49, 703.8, 683.95, 663.96, 643.82, 623.68, 604.31, 584.86, 565.33, 545.72, 526.03, 506.25, 486.39, 466.45, 446.43, 426.32, 406.13, 403.57, 403.87, 404.17, 404.47, 404.76, 405.05, 405.34, 405.62, 405.9, 406.18, 406.45, 406.89, 407.45, 408.01, 408.58, 409.14, 409.7, 410.27, 410.83, 411.39, 411.95, 412.64, 413.26, 413.64, 414.01, 414.37, 414.72, 415.05, 415.37, 415.67, 415.96, 416.24, 416.5, 416.75, 415.91, 415.08, 414.26, 413.45, 412.65, 411.86, 411.09, 410.32, 409.57, 408.82, 408.09, 408.09, 408.39, 408.68, 408.96, 409.24, 409.51, 409.77, 410.03, 3468.94, 3413.27, 3358.36, 3304.21, 3250.82, 3198.19, 3146.32, 3088.98, 3025.24, 2961.18, 2896.8, 2832.11, 2767.09, 2701.76, 2636.1, 2570.12, 2503.82, 2437.2, 2373.68, 2321.37, 2269.02, 2216.62, 2164.17, 2111.68, 2059.14, 2006.55, 1953.92, 1901.24, 1848.52, 1795.75, 1748.28, 1701.43, 1655.02, 1609.05, 1563.52, 1518.42, 1473.78, 1429.57, 1385.81, 1342.49, 1299.62, 1284.27, 1280.79, 1277.54, 1274.53, 1271.76, 1269.23, 1266.93, 1264.88, 1263.06, 1261.48, 1260.14, 1268.59, 1289.91, 1311.49, 1333.31, 1355.39, 1377.73, 1400.32, 1423.16, 1446.26, 1469.62, 1493.23, 1508.76, 1480.3, 1451.6, 1422.67, 1393.51, 1364.11, 1334.47, 1304.6, 1274.49, 1244.15, 1213.56, 1182.75, 1168.69, 1156.78, 1144.85, 1132.89, 1120.91, 1108.9, 1096.86, 1084.8, 1072.71, 1060.59, 1048.45, 1034.3, 1018.75, 1003.06, 987.22, 971.24, 955.11, 938.83, 922.41, 905.84, 889.12, 872.25, 854.41, 834.77, 814.98, 795.04, 774.95, 754.71, 734.33, 713.79, 693.11, 672.27, 651.28, 630.22, 610.4, 590.49, 570.5, 550.43, 530.28, 510.05, 489.73, 469.33, 448.85, 428.28, 407.64, 403.1, 403.39, 403.67, 403.95, 404.22, 404.49, 404.76, 405.03, 405.29, 405.55, 405.98, 406.59, 407.36, 408.11, 408.86, 409.59, 410.3, 411.01, 411.7, 412.37, 413.04, 413.69, 414.26, 414.57, 414.87, 415.16, 415.43, 415.69, 415.93, 416.16, 416.38, 416.58, 416.77, 416.94, 416.25, 415.47, 414.71, 413.96, 413.21, 412.48, 411.76, 411.05, 410.35, 409.66, 408.98, 408.89, 409.15, 409.4, 409.65, 409.89, 410.12, 410.35, 410.57, 3445.35, 3394.52, 3344.45, 3295.13, 3246.57, 3198.77, 3151.73, 3097.34, 3031.59, 2965.52, 2899.14, 2832.44, 2765.41, 2698.07, 2630.4, 2562.42, 2494.11, 2425.48, 2359.04, 2306.46, 2253.83, 2201.16, 2148.44, 2095.67, 2042.86, 1990.0, 1937.1, 1884.15, 1831.15, 1778.11, 1732.58, 1688.43, 1644.72, 1601.45, 1558.62, 1516.23, 1474.28, 1432.78, 1391.72, 1351.1, 1310.93, 1294.31, 1292.25, 1290.43, 1288.84, 1287.5, 1286.38, 1285.51, 1284.88, 1284.48, 1284.32, 1284.4, 1292.46, 1315.27, 1338.34, 1361.66, 1385.23, 1409.06, 1433.14, 1457.48, 1482.07, 1506.92, 1532.03, 1553.23, 1523.37, 1493.27, 1462.94, 1432.38, 1401.58, 1370.54, 1339.27, 1307.77, 1276.02, 1244.04, 1211.83, 1195.73, 1183.66, 1171.56, 1159.44, 1147.29, 1135.11, 1122.91, 1110.68, 1098.43, 1086.15, 1073.85, 1059.45, 1043.05, 1026.52, 1009.83, 993.0, 976.03, 958.9, 941.63, 924.22, 906.65, 888.94, 870.48, 849.99, 829.35, 808.57, 787.63, 766.55, 745.32, 723.94, 702.41, 680.73, 658.9, 636.91, 616.57, 596.2, 575.75, 555.23, 534.62, 513.92, 493.15, 472.29, 451.35, 430.33, 409.75, 403.85, 404.69, 405.43, 406.08, 406.65, 407.12, 407.51, 407.8, 408.0, 408.12, 408.14, 408.39, 409.08, 409.77, 410.44, 411.09, 411.74, 412.37, 412.98, 413.58, 414.17, 414.75, 415.28, 415.52, 415.74, 415.95, 416.15, 416.34, 416.51, 416.66, 416.81, 416.94, 417.05, 417.15, 416.58, 415.86, 415.15, 414.45, 413.76, 413.09, 412.42, 411.76, 411.12, 410.48, 409.86, 409.69, 409.91, 410.13, 410.34, 410.54, 410.74, 410.93, 411.11, 3421.01, 3375.01, 3329.77, 3285.28, 3241.55, 3198.58, 3156.37, 3106.02, 3038.26, 2970.19, 2901.79, 2833.08, 2764.05, 2694.7, 2625.02, 2555.03, 2484.71, 2414.07, 2344.44, 2291.59, 2238.69, 2185.74, 2132.75, 2079.71, 2026.63, 1973.5, 1920.32, 1867.1, 1813.84, 1760.52, 1716.44, 1674.99, 1633.98, 1593.41, 1553.28, 1513.59, 1474.35, 1435.54, 1397.18, 1359.27, 1321.79, 1304.12, 1303.48, 1303.08, 1302.92, 1302.99, 1303.3, 1303.85, 1304.64, 1305.66, 1306.93, 1308.43, 1316.08, 1340.38, 1364.94, 1389.75, 1414.82, 1440.14, 1465.71, 1491.54, 1517.63, 1543.97, 1570.56, 1597.42, 1566.68, 1535.18, 1503.45, 1471.49, 1439.29, 1406.86, 1374.19, 1341.28, 1308.14, 1274.77, 1241.15, 1222.8, 1210.56, 1198.3, 1186.01, 1173.7, 1161.36, 1148.99, 1136.6, 1124.18, 1111.74, 1099.27, 1084.75, 1067.51, 1050.12, 1032.59, 1014.92, 997.1, 979.13, 961.01, 942.75, 924.34, 905.78, 886.69, 865.36, 843.87, 822.24, 800.47, 778.54, 756.46, 734.24, 711.86, 689.33, 666.66, 643.56, 622.42, 601.65, 580.98, 560.4, 539.91, 519.52, 499.22, 479.02, 458.9, 438.89, 418.96, 411.07, 411.41, 411.66, 411.81, 411.88, 411.86, 411.75, 411.54, 411.25, 410.87, 410.39, 410.2, 410.82, 411.43, 412.03, 412.61, 413.18, 413.74, 414.28, 414.81, 415.32, 415.83, 416.31, 416.48, 416.63, 416.77, 416.89, 417.0, 417.1, 417.18, 417.25, 417.31, 417.35, 417.38, 416.89, 416.23, 415.58, 414.94, 414.3, 413.68, 413.07, 412.47, 411.88, 411.3, 410.73, 410.49, 410.68, 410.86, 411.03, 411.2, 411.36, 411.51, 411.66, 3395.92, 3354.75, 3314.33, 3274.67, 3235.78, 3197.64, 3160.26, 3115.02, 3045.25, 2975.17, 2904.77, 2834.05, 2763.01, 2691.65, 2619.97, 2547.96, 2475.64, 2402.99, 2330.02, 2276.76, 2223.59, 2170.37, 2117.11, 2063.8, 2010.44, 1957.04, 1903.59, 1850.1, 1796.56, 1742.98, 1699.86, 1661.11, 1622.8, 1584.93, 1547.5, 1510.51, 1473.97, 1437.86, 1402.2, 1366.98, 1332.21, 1313.7, 1314.48, 1315.5, 1316.76, 1318.25, 1319.98, 1321.95, 1324.16, 1326.61, 1329.29, 1332.22, 1339.44, 1365.24, 1391.29, 1417.6, 1444.15, 1470.96, 1498.03, 1525.35, 1552.93, 1580.76, 1608.85, 1637.19, 1610.22, 1577.33, 1544.2, 1510.84, 1477.24, 1443.41, 1409.34, 1375.04, 1340.5, 1305.73, 1270.71, 1249.9, 1237.5, 1225.07, 1212.62, 1200.14, 1187.63, 1175.1, 1162.55, 1149.96, 1137.35, 1124.72, 1110.19, 1092.11, 1073.88, 1055.5, 1036.98, 1018.31, 999.5, 980.54, 961.43, 942.18, 922.73, 902.95, 880.66, 858.18, 835.52, 812.67, 789.63, 766.42, 743.01, 719.42, 695.65, 671.69, 647.54, 626.37, 606.12, 585.96, 565.9, 545.93, 526.06, 506.27, 486.59, 466.99, 447.49, 428.09, 418.38, 418.22, 417.97, 417.63, 417.2, 416.69, 416.08, 415.38, 414.59, 413.71, 412.73, 412.02, 412.57, 413.11, 413.63, 414.14, 414.64, 415.12, 415.59, 416.05, 416.49, 416.92, 417.33, 417.45, 417.53, 417.59, 417.64, 417.68, 417.7, 417.71, 417.71, 417.69, 417.66, 417.61, 417.2, 416.6, 416.0, 415.41, 414.83, 414.27, 413.71, 413.16, 412.63, 412.11, 411.59, 411.31, 411.46, 411.6, 411.74, 411.87, 411.99, 412.11, 412.21, 3370.06, 3333.72, 3298.14, 3263.31, 3229.24, 3195.93, 3163.38, 3124.33, 3052.56, 2980.47, 2908.06, 2835.33, 2762.28, 2688.92, 2615.23, 2541.22, 2466.89, 2392.23, 2317.26, 2261.97, 2208.53, 2155.04, 2101.51, 2047.92, 1994.3, 1940.63, 1886.91, 1833.14, 1779.33, 1725.47, 1682.84, 1646.79, 1611.18, 1576.01, 1541.28, 1506.99, 1473.15, 1439.74, 1406.78, 1374.26, 1342.19, 1323.04, 1325.24, 1327.68, 1330.36, 1333.28, 1336.43, 1339.82, 1343.45, 1347.31, 1351.42, 1355.76, 1362.56, 1389.85, 1417.39, 1445.19, 1473.24, 1501.54, 1530.1, 1558.91, 1587.98, 1617.3, 1646.88, 1676.72, 1654.0, 1619.71, 1585.19, 1550.43, 1515.43, 1480.2, 1444.74, 1409.04, 1373.1, 1336.93, 1300.52, 1277.02, 1264.46, 1251.87, 1239.25, 1226.61, 1213.94, 1201.24, 1188.52, 1175.77, 1163.0, 1149.61, 1134.49, 1115.01, 1095.5, 1075.97, 1056.4, 1036.8, 1017.17, 997.51, 977.82, 958.1, 938.35, 918.57, 895.29, 871.76, 848.05, 824.16, 800.08, 775.81, 751.36, 726.73, 701.9, 676.89, 651.7, 630.22, 610.49, 590.85, 571.31, 551.86, 532.5, 513.23, 494.06, 474.99, 456.0, 437.12, 425.78, 425.13, 424.38, 423.54, 422.62, 421.6, 420.5, 419.3, 418.01, 416.64, 415.17, 413.86, 414.34, 414.8, 415.25, 415.69, 416.11, 416.52, 416.92, 417.3, 417.67, 418.02, 418.36, 418.43, 418.44, 418.43, 418.41, 418.37, 418.32, 418.26, 418.18, 418.09, 417.98, 417.86, 417.5, 416.95, 416.41, 415.87, 415.35, 414.84, 414.34, 413.85, 413.37, 412.9, 412.44, 412.13, 412.24, 412.35, 412.45, 412.54, 412.63, 412.71, 412.78, 3343.45, 3311.94, 3281.18, 3251.18, 3221.94, 3193.46, 3165.74, 3133.97, 3060.19, 2986.09, 2911.67, 2836.93, 2761.88, 2686.51, 2610.81, 2534.8, 2458.46, 2381.8, 2304.81, 2247.23, 2193.52, 2139.76, 2085.95, 2032.1, 1978.2, 1924.25, 1870.26, 1816.23, 1762.14, 1708.02, 1665.38, 1632.03, 1599.12, 1566.65, 1534.62, 1503.03, 1471.88, 1441.17, 1410.91, 1381.09, 1351.71, 1332.14, 1335.77, 1339.63, 1343.73, 1348.07, 1352.64, 1357.45, 1362.5, 1367.79, 1373.31, 1379.08, 1385.43, 1414.21, 1443.24, 1472.53, 1502.07, 1531.87, 1561.91, 1592.22, 1622.78, 1653.59, 1684.66, 1715.99, 1698.02, 1662.33, 1626.41, 1590.25, 1553.86, 1517.23, 1480.37, 1443.27, 1405.94, 1368.03, 1329.77, 1303.04, 1289.76, 1276.42, 1263.0, 1249.5, 1235.93, 1222.29, 1208.57, 1194.78, 1180.91, 1166.97, 1151.96, 1132.3, 1112.61, 1092.89, 1073.14, 1053.36, 1033.55, 1013.71, 993.84, 973.94, 954.0, 934.04, 910.11, 885.53, 860.78, 835.83, 810.71, 785.39, 759.9, 734.21, 708.34, 682.29, 656.05, 633.99, 614.77, 595.65, 576.62, 557.69, 538.85, 520.1, 501.45, 482.89, 464.42, 446.05, 433.27, 432.12, 430.88, 429.54, 428.12, 426.61, 425.01, 423.32, 421.53, 419.66, 417.7, 415.72, 416.12, 416.51, 416.88, 417.25, 417.6, 417.93, 418.25, 418.56, 418.86, 419.14, 419.41, 419.43, 419.36, 419.28, 419.18, 419.07, 418.95, 418.81, 418.66, 418.5, 418.32, 418.13, 417.79, 417.29, 416.8, 416.33, 415.86, 415.4, 414.96, 414.52, 414.1, 413.68, 413.28, 412.96, 413.03, 413.1, 413.17, 413.22, 413.27, 413.31, 413.35, 3316.08, 3289.4, 3263.47, 3238.3, 3213.89, 3190.23, 3167.34, 3143.92, 3068.13, 2992.02, 2915.6, 2838.86, 2761.8, 2684.42, 2606.71, 2528.69, 2450.35, 2371.68, 2292.69, 2232.54, 2178.55, 2124.52, 2070.44, 2016.31, 1962.14, 1907.93, 1853.67, 1799.36, 1745.0, 1690.6, 1647.48, 1616.83, 1586.62, 1556.85, 1527.52, 1498.62, 1470.17, 1442.16, 1414.6, 1387.48, 1360.8, 1341.01, 1346.06, 1351.34, 1356.86, 1362.62, 1368.61, 1374.84, 1381.31, 1388.02, 1394.97, 1402.15, 1409.58, 1438.32, 1468.84, 1499.62, 1530.65, 1561.94, 1593.48, 1625.27, 1657.32, 1689.23, 1720.08, 1750.43, 1735.62, 1697.17, 1658.77, 1620.43, 1582.15, 1543.92, 1505.76, 1467.66, 1429.61, 1391.63, 1353.71, 1324.88, 1311.17, 1297.38, 1283.52, 1269.59, 1255.58, 1241.49, 1227.34, 1213.1, 1198.79, 1184.41, 1169.47, 1149.62, 1129.75, 1109.85, 1089.92, 1069.96, 1049.96, 1029.94, 1009.89, 989.8, 969.69, 949.54, 925.11, 899.49, 873.68, 847.69, 821.52, 795.16, 768.62, 741.89, 714.97, 687.87, 660.58, 637.66, 618.96, 600.36, 581.85, 563.43, 545.11, 526.87, 508.74, 490.7, 472.75, 454.89, 440.85, 439.2, 437.46, 435.64, 433.72, 431.71, 429.61, 427.42, 425.14, 422.77, 420.31, 417.76, 417.91, 418.23, 418.53, 418.82, 419.1, 419.36, 419.61, 419.84, 420.06, 420.27, 420.46, 420.45, 420.3, 420.15, 419.98, 419.79, 419.6, 419.39, 419.16, 418.92, 418.67, 418.41, 418.07, 417.62, 417.19, 416.77, 416.36, 415.96, 415.57, 415.19, 414.82, 414.46, 414.11, 413.79, 413.83, 413.86, 413.89, 413.91, 413.92, 413.93, 413.93, 3287.96, 3266.1, 3245.0, 3224.66, 3205.07, 3186.24, 3168.17, 3150.86, 3076.39, 2998.28, 2919.85, 2841.1, 2762.03, 2682.64, 2602.94, 2522.91, 2442.56, 2361.89, 2280.89, 2217.88, 2163.62, 2109.32, 2054.97, 2000.57, 1946.13, 1891.64, 1837.11, 1782.53, 1727.9, 1673.23, 1629.14, 1601.19, 1573.68, 1546.6, 1519.97, 1493.78, 1468.02, 1442.71, 1417.84, 1393.42, 1369.44, 1349.65, 1356.12, 1362.82, 1369.76, 1376.94, 1384.35, 1392.0, 1399.89, 1408.67, 1419.1, 1429.85, 1440.92, 1470.05, 1501.91, 1533.27, 1564.14, 1594.51, 1624.38, 1653.76, 1682.64, 1711.02, 1738.9, 1766.28, 1756.15, 1718.02, 1679.96, 1641.96, 1604.01, 1566.12, 1528.3, 1490.53, 1452.82, 1415.17, 1377.58, 1346.81, 1332.65, 1318.42, 1304.12, 1289.75, 1275.3, 1260.77, 1246.17, 1231.5, 1216.75, 1201.93, 1187.0, 1166.98, 1146.93, 1126.84, 1106.73, 1086.58, 1066.41, 1046.2, 1025.97, 1005.7, 985.4, 965.07, 940.3, 913.63, 886.78, 859.74, 832.52, 805.12, 777.53, 749.75, 721.79, 693.64, 665.3, 641.24, 623.06, 604.97, 586.98, 569.08, 551.27, 533.56, 515.94, 498.41, 480.98, 463.64, 448.52, 446.38, 444.14, 441.82, 439.4, 436.9, 434.3, 431.62, 428.84, 425.98, 423.02, 419.97, 419.72, 419.96, 420.19, 420.41, 420.61, 420.8, 420.97, 421.13, 421.28, 421.41, 421.54, 421.47, 421.26, 421.03, 420.78, 420.53, 420.26, 419.97, 419.67, 419.36, 419.04, 418.7, 418.34, 417.95, 417.57, 417.2, 416.85, 416.5, 416.17, 415.84, 415.53, 415.22, 414.93, 414.65, 414.64, 414.63, 414.62, 414.61, 414.58, 414.55, 414.51, 3259.08, 3242.05, 3225.78, 3210.26, 3195.49, 3181.49, 3168.24, 3155.76, 3084.97, 3004.85, 2924.41, 2843.66, 2762.58, 2681.19, 2599.48, 2517.45, 2435.09, 2352.41, 2269.41, 2203.27, 2148.74, 2094.17, 2039.55, 1984.88, 1930.16, 1875.41, 1820.6, 1765.75, 1710.85, 1655.91, 1610.36, 1585.11, 1560.3, 1535.92, 1511.98, 1488.49, 1465.43, 1444.32, 1424.62, 1404.71, 1384.57, 1365.54, 1375.32, 1385.41, 1395.82, 1406.54, 1417.59, 1428.95, 1440.63, 1452.64, 1464.96, 1477.61, 1490.58, 1516.12, 1545.0, 1573.39, 1601.29, 1628.69, 1655.59, 1682.0, 1707.9, 1733.31, 1758.22, 1782.63, 1776.61, 1738.83, 1701.1, 1663.43, 1625.82, 1588.27, 1550.78, 1513.34, 1475.97, 1438.66, 1401.4, 1368.8, 1354.21, 1339.54, 1324.8, 1309.98, 1295.09, 1280.13, 1265.09, 1249.97, 1234.78, 1219.52, 1204.18, 1184.37, 1164.13, 1143.86, 1123.57, 1103.24, 1082.88, 1062.5, 1042.08, 1021.63, 1001.15, 980.64, 955.67, 927.95, 900.06, 871.97, 843.71, 815.26, 786.62, 757.8, 728.79, 699.59, 670.21, 644.72, 627.06, 609.49, 592.01, 574.63, 557.34, 540.14, 523.04, 506.03, 489.11, 472.29, 456.28, 453.64, 450.91, 448.09, 445.18, 442.18, 439.09, 435.9, 432.63, 429.27, 425.82, 422.28, 421.54, 421.71, 421.86, 422.01, 422.13, 422.25, 422.35, 422.44, 422.51, 422.57, 422.62, 422.51, 422.22, 421.92, 421.6, 421.27, 420.93, 420.57, 420.2, 419.81, 419.42, 419.0, 418.59, 418.26, 417.94, 417.63, 417.32, 417.03, 416.75, 416.48, 416.22, 415.97, 415.74, 415.51, 415.45, 415.41, 415.36, 415.31, 415.25, 415.18, 415.1, 3229.45, 3217.24, 3205.79, 3195.1, 3185.16, 3175.98, 3167.56, 3159.9, 3093.86, 3011.74, 2929.3, 2846.54, 2763.46, 2680.06, 2596.34, 2512.3, 2427.94, 2343.26, 2258.26, 2188.71, 2133.91, 2079.06, 2024.16, 1969.23, 1914.24, 1863.44, 1813.41, 1763.54, 1713.84, 1664.3, 1622.09, 1602.57, 1582.82, 1562.86, 1542.68, 1522.28, 1501.66, 1480.82, 1459.76, 1438.48, 1416.98, 1395.25, 1405.67, 1417.66, 1429.97, 1442.6, 1455.54, 1468.8, 1482.38, 1496.29, 1510.51, 1525.06, 1539.92, 1562.68, 1588.59, 1614.01, 1638.94, 1663.36, 1687.3, 1710.73, 1733.67, 1756.11, 1778.04, 1799.48, 1797.02, 1759.57, 1722.18, 1684.84, 1647.57, 1610.35, 1573.2, 1536.1, 1499.06, 1462.08, 1425.17, 1390.88, 1375.84, 1360.73, 1345.55, 1330.29, 1314.96, 1299.56, 1284.08, 1268.52, 1252.89, 1237.19, 1221.41, 1201.79, 1181.37, 1160.92, 1140.44, 1119.93, 1099.39, 1078.82, 1058.22, 1037.59, 1016.93, 996.24, 971.22, 942.46, 913.52, 884.39, 855.08, 825.58, 795.9, 766.03, 735.98, 705.74, 675.31, 648.12, 630.97, 613.92, 596.96, 580.09, 563.32, 546.64, 530.05, 513.56, 497.16, 480.85, 464.64, 461.0, 457.77, 454.45, 451.04, 447.55, 443.96, 440.28, 436.52, 432.66, 428.71, 424.67, 423.37, 423.47, 423.55, 423.62, 423.67, 423.71, 423.74, 423.76, 423.76, 423.75, 423.72, 423.57, 423.2, 422.83, 422.44, 422.03, 421.62, 421.18, 420.74, 420.28, 419.81, 419.32, 418.84, 418.56, 418.29, 418.04, 417.79, 417.56, 417.33, 417.12, 416.91, 416.72, 416.53, 416.36, 416.27, 416.19, 416.11, 416.02, 415.92, 415.82, 415.7, 3199.06, 3191.68, 3185.05, 3179.18, 3174.06, 3169.71, 3166.11, 3163.27, 3103.07, 3018.95, 2934.5, 2849.73, 2765.63, 2682.89, 2600.4, 2518.17, 2436.19, 2354.47, 2273.0, 2205.29, 2155.27, 2105.4, 2055.71, 2006.17, 1956.8, 1907.58, 1858.54, 1809.65, 1760.93, 1712.37, 1668.36, 1647.47, 1626.36, 1605.04, 1583.49, 1561.73, 1539.74, 1517.54, 1495.12, 1472.47, 1449.61, 1426.52, 1435.71, 1449.6, 1463.81, 1478.33, 1493.18, 1508.34, 1523.82, 1539.62, 1555.74, 1572.18, 1588.95, 1609.74, 1632.68, 1655.13, 1677.08, 1698.54, 1719.5, 1739.97, 1759.93, 1779.4, 1798.37, 1816.84, 1817.37, 1780.25, 1743.2, 1706.2, 1669.26, 1632.38, 1595.56, 1558.8, 1522.1, 1485.45, 1448.87, 1413.03, 1397.55, 1382.0, 1366.38, 1350.68, 1334.91, 1319.06, 1303.14, 1287.15, 1271.08, 1254.93, 1238.71, 1219.24, 1198.64, 1178.01, 1157.34, 1136.65, 1115.93, 1095.18, 1074.4, 1053.59, 1032.74, 1011.87, 986.97, 957.16, 927.17, 897.0, 866.64, 836.09, 805.36, 774.45, 743.35, 712.07, 680.59, 651.42, 634.79, 618.25, 601.81, 585.46, 569.2, 553.04, 536.97, 520.99, 505.11, 489.32, 473.63, 468.44, 464.72, 460.9, 457.0, 453.01, 448.92, 444.75, 440.49, 436.13, 431.69, 427.16, 425.22, 425.24, 425.25, 425.25, 425.23, 425.19, 425.15, 425.09, 425.02, 424.93, 424.83, 424.63, 424.2, 423.75, 423.28, 422.81, 422.32, 421.81, 421.29, 420.76, 420.22, 419.66, 419.08, 418.85, 418.64, 418.44, 418.25, 418.07, 417.9, 417.74, 417.59, 417.45, 417.32, 417.2, 417.09, 416.98, 416.86, 416.73, 416.6, 416.46, 416.31, 3173.46, 3174.41, 3175.11, 3175.55, 3175.74, 3175.68, 3175.37, 3174.8, 3122.53, 3040.36, 2958.44, 2876.77, 2795.35, 2714.19, 2633.28, 2552.63, 2472.23, 2392.08, 2312.19, 2243.37, 2194.33, 2145.46, 2096.74, 2048.19, 1999.8, 1951.57, 1903.5, 1855.6, 1807.86, 1760.29, 1714.85, 1692.59, 1670.12, 1647.43, 1624.53, 1601.4, 1578.05, 1554.49, 1530.7, 1506.69, 1482.47, 1458.02, 1465.44, 1481.23, 1497.33, 1513.76, 1530.5, 1547.56, 1564.93, 1582.63, 1600.65, 1618.99, 1637.66, 1657.3, 1677.27, 1696.75, 1715.73, 1734.22, 1752.21, 1769.7, 1786.7, 1803.2, 1819.2, 1834.7, 1837.66, 1800.88, 1764.16, 1727.5, 1690.9, 1654.35, 1617.87, 1581.44, 1545.07, 1508.76, 1472.52, 1436.33, 1419.34, 1403.35, 1387.28, 1371.15, 1354.93, 1338.65, 1322.29, 1305.85, 1289.34, 1272.75, 1256.09, 1236.72, 1215.94, 1195.12, 1174.28, 1153.41, 1132.5, 1111.57, 1090.61, 1069.61, 1048.59, 1027.53, 1002.89, 972.04, 941.0, 909.78, 878.38, 846.79, 815.02, 783.06, 750.91, 718.58, 686.06, 654.63, 638.51, 622.49, 606.57, 590.73, 574.99, 559.35, 543.79, 528.33, 512.97, 497.69, 482.52, 475.97, 471.75, 467.44, 463.05, 458.56, 453.98, 449.31, 444.55, 439.7, 434.76, 429.73, 427.08, 427.03, 426.96, 426.89, 426.79, 426.69, 426.57, 426.44, 426.29, 426.13, 425.96, 425.71, 425.2, 424.68, 424.14, 423.59, 423.03, 422.45, 421.86, 421.26, 420.64, 420.0, 419.36, 419.13, 418.98, 418.83, 418.69, 418.57, 418.45, 418.35, 418.25, 418.17, 418.1, 418.03, 417.93, 417.78, 417.62, 417.46, 417.29, 417.11, 416.92, 3201.44, 3200.77, 3199.84, 3198.66, 3197.23, 3195.54, 3193.6, 3191.4, 3145.69, 3065.1, 2984.75, 2904.66, 2824.83, 2745.24, 2665.91, 2586.83, 2508.01, 2429.44, 2351.13, 2281.3, 2233.24, 2185.35, 2137.62, 2090.05, 2042.64, 1995.4, 1948.31, 1901.4, 1854.64, 1808.05, 1761.62, 1737.94, 1714.11, 1690.05, 1665.78, 1641.29, 1616.58, 1591.66, 1566.51, 1541.14, 1515.55, 1489.73, 1494.86, 1512.54, 1530.54, 1548.86, 1567.5, 1586.46, 1605.74, 1625.33, 1645.25, 1665.49, 1686.05, 1705.36, 1722.36, 1738.86, 1754.88, 1770.39, 1785.42, 1799.94, 1813.97, 1827.5, 1840.53, 1853.06, 1857.9, 1821.46, 1785.07, 1748.74, 1712.47, 1676.26, 1640.11, 1604.02, 1567.99, 1532.02, 1496.11, 1460.25, 1441.2, 1424.77, 1408.27, 1391.69, 1375.03, 1358.31, 1341.51, 1324.63, 1307.68, 1290.65, 1273.55, 1254.23, 1233.27, 1212.27, 1191.25, 1170.19, 1149.11, 1127.99, 1106.85, 1085.67, 1064.46, 1043.22, 1019.0, 987.11, 955.02, 922.76, 890.31, 857.67, 824.85, 791.85, 758.66, 725.28, 691.72, 657.97, 642.15, 626.64, 611.23, 595.92, 580.69, 565.56, 550.52, 535.58, 520.73, 505.97, 491.31, 483.6, 478.88, 474.08, 469.18, 464.2, 459.12, 453.96, 448.7, 443.36, 437.92, 432.4, 428.96, 428.83, 428.69, 428.54, 428.37, 428.19, 428.0, 427.8, 427.58, 427.34, 427.1, 426.81, 426.23, 425.63, 425.02, 424.4, 423.76, 423.11, 422.44, 421.76, 421.07, 420.36, 419.64, 419.4, 419.3, 419.21, 419.13, 419.06, 419.0, 418.95, 418.91, 418.88, 418.86, 418.86, 418.77, 418.58, 418.39, 418.19, 417.98, 417.76, 417.54, 3229.68, 3227.38, 3224.83, 3222.02, 3218.96, 3215.65, 3212.08, 3208.26, 3168.6, 3089.59, 3010.82, 2932.31, 2854.05, 2776.04, 2698.29, 2620.79, 2543.54, 2466.55, 2389.82, 2319.07, 2272.0, 2225.09, 2178.34, 2131.75, 2085.33, 2039.06, 1992.96, 1947.03, 1901.25, 1855.65, 1810.2, 1783.51, 1758.31, 1732.9, 1707.26, 1681.41, 1655.34, 1629.05, 1602.54, 1575.8, 1548.85, 1521.68, 1523.96, 1543.54, 1563.44, 1583.66, 1604.19, 1625.05, 1646.22, 1667.71, 1689.53, 1711.66, 1734.12, 1753.91, 1767.94, 1781.48, 1794.52, 1807.07, 1819.12, 1830.68, 1841.74, 1852.3, 1862.36, 1871.92, 1878.08, 1841.97, 1805.92, 1769.93, 1734.0, 1698.12, 1662.31, 1626.55, 1590.85, 1555.22, 1519.64, 1484.12, 1463.14, 1446.27, 1429.32, 1412.3, 1395.21, 1378.04, 1360.8, 1343.48, 1326.09, 1308.63, 1291.09, 1271.78, 1250.63, 1229.45, 1208.25, 1187.01, 1165.74, 1144.45, 1123.12, 1101.76, 1080.37, 1058.95, 1035.3, 1002.36, 969.23, 935.92, 902.42, 868.74, 834.88, 800.83, 766.59, 732.17, 697.56, 662.77, 645.69, 630.7, 615.81, 601.0, 586.3, 571.68, 557.16, 542.73, 528.4, 514.16, 500.01, 491.31, 486.1, 480.8, 475.41, 469.93, 464.36, 458.7, 452.95, 447.11, 441.18, 435.16, 430.84, 430.65, 430.43, 430.21, 429.97, 429.72, 429.45, 429.17, 428.88, 428.57, 428.25, 427.92, 427.26, 426.59, 425.91, 425.21, 424.5, 423.77, 423.04, 422.28, 421.52, 420.74, 419.95, 419.66, 419.62, 419.58, 419.55, 419.54, 419.53, 419.54, 419.56, 419.58, 419.62, 419.67, 419.61, 419.39, 419.16, 418.92, 418.75, 418.62, 418.48, 3258.17, 3254.25, 3250.07, 3245.64, 3240.95, 3236.01, 3230.82, 3225.38, 3191.26, 3113.82, 3036.64, 2959.7, 2883.02, 2806.59, 2730.41, 2654.49, 2578.82, 2503.41, 2428.25, 2356.68, 2310.59, 2264.66, 2218.9, 2173.29, 2127.85, 2082.57, 2037.45, 1992.5, 1947.71, 1903.08, 1858.62, 1829.3, 1802.74, 1775.96, 1748.97, 1721.75, 1694.32, 1666.66, 1638.79, 1610.7, 1582.38, 1553.85, 1552.74, 1574.22, 1596.02, 1618.14, 1640.57, 1663.32, 1686.39, 1709.78, 1733.49, 1757.52, 1781.87, 1802.96, 1814.02, 1824.59, 1834.66, 1844.24, 1853.33, 1861.92, 1870.01, 1877.6, 1884.7, 1891.29, 1897.39, 1862.43, 1826.71, 1791.06, 1755.46, 1719.92, 1684.44, 1649.02, 1613.66, 1578.35, 1543.11, 1507.93, 1485.15, 1467.84, 1450.45, 1432.99, 1415.46, 1397.85, 1380.17, 1362.42, 1344.58, 1326.68, 1308.7, 1289.35, 1268.03, 1246.67, 1225.28, 1203.86, 1182.41, 1160.93, 1139.42, 1117.88, 1096.31, 1074.71, 1051.78, 1017.79, 983.62, 949.26, 914.72, 880.0, 845.09, 809.99, 774.71, 739.25, 703.6, 667.76, 649.13, 634.66, 620.29, 606.0, 591.81, 577.71, 563.71, 549.79, 535.98, 522.25, 508.62, 499.12, 493.41, 487.61, 481.73, 475.75, 469.68, 463.53, 457.28, 450.95, 444.52, 438.01, 432.75, 432.47, 432.19, 431.89, 431.58, 431.25, 430.91, 430.56, 430.19, 429.81, 429.42, 429.01, 428.31, 427.57, 426.81, 426.04, 425.25, 424.46, 423.64, 422.82, 421.98, 421.13, 420.26, 419.91, 419.92, 419.94, 419.97, 420.06, 420.31, 420.54, 420.76, 420.95, 421.12, 421.27, 421.32, 421.18, 421.03, 420.88, 420.72, 420.55, 420.37, 3286.92, 3281.37, 3275.56, 3269.51, 3263.2, 3256.64, 3249.82, 3242.75, 3213.68, 3137.81, 3062.2, 2986.84, 2911.74, 2836.88, 2762.28, 2687.94, 2613.85, 2540.01, 2466.43, 2394.13, 2349.03, 2304.08, 2259.3, 2214.68, 2170.22, 2125.92, 2081.79, 2037.81, 1994.01, 1950.36, 1906.88, 1875.31, 1847.39, 1819.25, 1790.89, 1762.31, 1733.52, 1704.5, 1675.27, 1645.81, 1616.13, 1586.24, 1581.22, 1604.6, 1628.29, 1652.3, 1676.63, 1701.28, 1726.24, 1751.53, 1777.14, 1803.06, 1829.31, 1852.5, 1860.6, 1868.2, 1875.3, 1881.92, 1888.03, 1893.66, 1898.78, 1903.41, 1907.54, 1911.16, 1914.29, 1882.83, 1847.45, 1812.13, 1776.87, 1741.66, 1706.52, 1671.43, 1636.4, 1601.44, 1566.53, 1531.68, 1507.24, 1489.49, 1471.66, 1453.76, 1435.79, 1417.74, 1399.62, 1381.42, 1363.15, 1344.81, 1326.39, 1306.96, 1285.45, 1263.91, 1242.34, 1220.74, 1199.11, 1177.45, 1155.76, 1134.03, 1112.28, 1090.5, 1068.44, 1033.41, 998.19, 962.79, 927.21, 891.44, 855.48, 819.34, 783.02, 746.51, 709.81, 672.93, 652.49, 638.54, 624.67, 610.9, 597.23, 583.65, 570.16, 556.76, 543.46, 530.25, 517.14, 507.01, 500.81, 494.51, 488.13, 481.66, 475.1, 468.45, 461.71, 454.88, 447.96, 440.95, 434.66, 434.31, 433.95, 433.58, 433.2, 432.8, 432.38, 431.96, 431.52, 431.07, 430.6, 430.12, 429.37, 428.56, 427.72, 426.88, 426.02, 425.5, 424.95, 424.37, 423.74, 423.07, 422.36, 422.13, 422.35, 422.56, 422.74, 422.9, 423.05, 423.17, 423.27, 423.35, 423.42, 423.46, 423.44, 423.26, 423.08, 422.88, 422.69, 422.48, 422.27, 3315.92, 3308.74, 3301.31, 3293.63, 3285.7, 3277.51, 3269.07, 3260.38, 3235.84, 3161.55, 3087.52, 3013.74, 2940.21, 2866.93, 2793.91, 2721.14, 2648.62, 2576.36, 2504.35, 2432.6, 2387.31, 2343.34, 2299.54, 2255.9, 2212.43, 2169.11, 2125.96, 2082.97, 2040.14, 1997.48, 1954.98, 1921.55, 1892.26, 1862.76, 1833.04, 1803.1, 1772.94, 1742.56, 1711.97, 1681.15, 1650.11, 1618.85, 1609.38, 1634.65, 1660.24, 1686.15, 1712.38, 1738.92, 1765.78, 1792.97, 1820.47, 1848.29, 1876.43, 1902.55, 1907.67, 1912.3, 1916.44, 1920.09, 1923.24, 1925.89, 1928.05, 1929.71, 1930.87, 1931.54, 1931.7, 1903.18, 1868.13, 1833.14, 1798.21, 1763.35, 1728.54, 1693.78, 1659.09, 1624.46, 1589.89, 1555.37, 1529.4, 1511.21, 1492.95, 1474.61, 1456.19, 1437.71, 1419.14, 1400.51, 1381.8, 1363.01, 1344.15, 1324.6, 1302.91, 1281.19, 1259.44, 1237.65, 1215.84, 1194.0, 1172.12, 1150.22, 1128.28, 1106.32, 1084.32, 1049.22, 1012.95, 976.51, 939.88, 903.06, 866.06, 828.88, 791.51, 753.96, 716.22, 678.29, 655.75, 642.31, 628.97, 615.72, 602.56, 589.49, 576.52, 563.64, 550.85, 538.16, 525.56, 514.99, 508.29, 501.51, 494.63, 487.66, 480.61, 473.46, 466.23, 458.9, 451.48, 443.98, 436.59, 436.17, 435.74, 435.29, 434.83, 434.36, 434.21, 434.14, 434.06, 433.98, 433.89, 433.8, 433.39, 432.83, 432.22, 431.58, 430.9, 430.17, 429.41, 428.6, 427.76, 426.87, 425.94, 425.43, 425.55, 425.64, 425.72, 425.77, 425.81, 425.82, 425.81, 425.78, 425.73, 425.66, 425.56, 425.35, 425.13, 424.9, 424.66, 424.42, 424.17, 3345.17, 3336.37, 3327.32, 3318.01, 3308.46, 3298.65, 3288.58, 3278.27, 3257.75, 3185.04, 3112.59, 3040.38, 2968.43, 2896.73, 2825.28, 2754.08, 2683.14, 2612.46, 2542.03, 2471.85, 2425.43, 2382.45, 2339.63, 2296.97, 2254.47, 2212.14, 2169.97, 2127.96, 2086.12, 2044.44, 2002.92, 1968.0, 1937.36, 1906.49, 1875.41, 1844.11, 1812.59, 1780.85, 1748.89, 1716.71, 1684.31, 1651.69, 1637.23, 1664.4, 1691.88, 1719.69, 1747.81, 1776.25, 1805.01, 1834.08, 1863.48, 1893.2, 1923.24, 1953.08, 1955.24, 1956.91, 1958.08, 1958.76, 1958.94, 1958.63, 1957.82, 1956.52, 1954.72, 1952.41, 1949.61, 1923.46, 1888.75, 1854.1, 1819.51, 1784.97, 1750.5, 1716.08, 1681.72, 1647.43, 1613.19, 1579.01, 1551.64, 1533.01, 1514.31, 1495.53, 1476.67, 1457.75, 1438.74, 1419.67, 1400.52, 1381.29, 1361.99, 1342.27, 1320.4, 1298.5, 1276.56, 1254.6, 1232.6, 1210.58, 1188.52, 1166.44, 1144.32, 1122.17, 1100.0, 1065.2, 1027.9, 990.41, 952.73, 914.87, 876.83, 838.6, 800.19, 761.59, 722.81, 683.84, 658.92, 646.0, 633.17, 620.43, 607.79, 595.24, 582.78, 570.42, 558.15, 545.97, 533.89, 523.07, 515.87, 508.59, 501.22, 493.76, 486.2, 479.0, 471.77, 464.39, 456.86, 449.17, 441.32, 440.77, 440.69, 440.61, 440.52, 440.43, 440.33, 440.22, 440.11, 439.99, 439.86, 439.73, 439.21, 438.42, 437.6, 436.74, 435.84, 434.89, 433.91, 432.88, 431.81, 430.71, 429.56, 428.76, 428.77, 428.75, 428.72, 428.66, 428.59, 428.49, 428.37, 428.23, 428.07, 427.89, 427.69, 427.44, 427.18, 426.92, 426.65, 426.37, 426.08, 3374.68, 3364.25, 3353.58, 3342.65, 3331.47, 3320.04, 3308.35, 3296.41, 3279.42, 3208.28, 3137.4, 3066.77, 2996.39, 2926.27, 2856.4, 2786.78, 2717.41, 2648.3, 2579.45, 2510.85, 2463.39, 2421.39, 2379.55, 2337.88, 2296.37, 2255.01, 2213.83, 2172.8, 2131.94, 2091.24, 2050.7, 2014.68, 1982.67, 1950.45, 1918.0, 1885.34, 1852.46, 1819.36, 1786.04, 1752.5, 1718.74, 1684.76, 1664.76, 1693.83, 1723.21, 1752.91, 1782.92, 1813.26, 1843.91, 1874.89, 1906.18, 1937.79, 1969.73, 2001.98, 2003.31, 2002.01, 2000.22, 1997.93, 1995.15, 1991.87, 1988.1, 1983.83, 1979.06, 1973.79, 1968.02, 1943.69, 1909.32, 1875.0, 1840.74, 1806.54, 1772.4, 1738.32, 1704.3, 1670.33, 1636.43, 1602.59, 1573.96, 1554.89, 1535.74, 1516.52, 1497.23, 1477.86, 1458.42, 1438.9, 1419.31, 1399.65, 1379.91, 1359.98, 1337.92, 1315.84, 1293.72, 1271.57, 1249.4, 1227.19, 1204.95, 1182.69, 1160.39, 1138.06, 1115.7, 1081.38, 1043.03, 1004.49, 965.77, 926.87, 887.78, 848.51, 809.05, 769.41, 729.58, 689.57, 662.0, 649.59, 637.28, 625.06, 612.93, 600.9, 589.8, 578.81, 567.84, 556.89, 545.96, 535.64, 528.48, 521.17, 513.71, 506.09, 498.33, 490.41, 482.33, 474.11, 465.73, 457.19, 448.51, 447.07, 446.96, 446.84, 446.72, 446.59, 446.45, 446.31, 446.16, 446.0, 445.84, 445.67, 445.07, 444.06, 443.02, 441.94, 440.81, 439.65, 438.45, 437.2, 435.91, 434.59, 433.22, 432.11, 432.01, 431.88, 431.74, 431.57, 431.39, 431.18, 430.95, 430.7, 430.43, 430.14, 429.83, 429.54, 429.25, 428.95, 428.64, 428.32, 428.0],
            "height": 233 //151
        });
        this.map.addLayer(demLayer);
        this.map.currentBasemap = demLayer;

        this.gridlayer = new GridLayer({ center: center, name: "coordinate grid", appWidth: this.width, appHeight: this.height });
        this.map.addLayer(this.gridlayer);

        new LayerControl(this.map.layers, {
            collapsed: true,
            parentDiv: 'layer-control-parent-id'
        }).addTo(this.map);

        this.basemapControl = new BasemapControl('Baselayer', {
            position: 'topright'
        }).addTo(this.map);

        //slider for scaling z value
        this.slider = new SliderControl({ layers: this.map.layers }).addTo(this.map);

        //slice on x and y axes:
        // this.slicer = new SlicerControl({ parentDiv: 'slicer-control' }).addTo(this.map);

        // this.selection = new Selection(
        //     // new Vector3(-7, -14, -14),
        //     // new Vector3(14, 9, 3)
        //     new Vector3(x.min, y.min, z.min),
        //     new Vector3(x.max, y.max, z.max)
        // );        
        this.capsScene.add(this.selection.boxMesh);
        this.scene.add(this.selection.displayMeshes);
        this.scene.add(this.selection.touchMeshes);

        // domEvent.on(window, 'resize', this.onWindowResize, this);
        // domEvent.on(window, 'keydown', this.keydown, this);
        this.start();
    }

    keydown(e) {
        if (e.ctrlKey || e.altKey) return;
        let keyPressed = e.which;
        if (!e.shiftKey) {
            //if (keyPressed == 27) app.closePopup(); // ESC           
            if (keyPressed === 87) {
                this.setWireframeMode();    // W
            }
        }
    }

    setWireframeMode() {
        let wireframe = !this.wireframeMode;
        if (wireframe === this.wireframeMode) return;

        for (var key in this.map._layers) {
            let layer = this.map._layers[key];
            layer.setWireframeMode(wireframe);
        }

        this.wireframeMode = wireframe;
        this.animate();
    }

    onWindowResize() {
        if (this._fullWindow) {
            this._setCanvasSize(window.innerWidth, window.innerHeight);
        }
        else {
            this._setCanvasSize(this.container.clientWidth, this.container.clientHeight);
        }
    }

    _setCanvasSize(width, height) {
        this.width = width;
        this.height = height;
        this.camera.aspect = width / height;
        this.camera.updateProjectionMatrix();
        this.renderer.setSize(width, height);
        this.animate();
    }

    _buildDefaultLights(scene) {
        let deg2rad = Math.PI / 180;

        // ambient light
        scene.add(new AmbientLight(0x999999));
        //scene.add(new THREE.AmbientLight(0xeeeeee));

        // directional lights
        let opt = {
            azimuth: 220,   // note: default light azimuth of gdaldem hillshade is 315.
            altitude: 45    // altitude angle
        };
        //appSettings.Options.light.directional;
        let lambda = (90 - opt.azimuth) * deg2rad;
        let phi = opt.altitude * deg2rad;

        let x = Math.cos(phi) * Math.cos(lambda),
            y = Math.cos(phi) * Math.sin(lambda),
            z = Math.sin(phi);

        let light1 = new DirectionalLight(0xffffff, 0.5);
        light1.position.set(x, y, z);
        scene.add(light1);

        // thin light from the opposite direction
        let light2 = new DirectionalLight(0xffffff, 0.1);
        light2.position.set(-x, -y, -z);
        scene.add(light2);
    }

    start() {
        this.running = true;

        this.map.addListener('change', this.animate, this); // add this only if there is no animation loop (requestAnimationFrame)
        this.animate();
    }

    animate() {
        this.renderer.clear();

        let gl = this.renderer.context;
        // if (this.showCaps) {

        //     this.renderer.state.setStencilTest(true);

        //     this.renderer.state.setStencilFunc(gl.ALWAYS, 1, 0xff);
        //     this.renderer.state.setStencilOp(gl.KEEP, gl.KEEP, gl.INCR);
        //     // this.renderer.render(this.backStencil, this.camera);

        //     this.renderer.state.setStencilFunc(gl.ALWAYS, 1, 0xff);
        //     this.renderer.state.setStencilOp(gl.KEEP, gl.KEEP, gl.DECR);
        //     // this.renderer.render(this.frontStencil, this.camera);

        //     this.renderer.state.setStencilFunc(gl.EQUAL, 1, 0xff);
        //     this.renderer.state.setStencilOp(gl.KEEP, gl.KEEP, gl.KEEP);
        //     this.renderer.render(this.capsScene, this.camera);

        //     this.renderer.state.setStencilTest(false);

        // }
        this.renderer.render(this.scene, this.camera);
        this.northArrow.animate();
        this.gridlayer.animate();
    }

    addEventListeners() {

        // domEvent.on(this.mapIcon, 'click', () => {
        //     this.basemapControl.show();
        // }, this);

        domEvent.on(window, 'resize', this.onWindowResize, this);
        domEvent.on(window, 'keydown', this.keydown, this);

        // let inputNodes = [this.$topTextInput, this.$bottomTextInput, this.$imageInput];
        // inputNodes.forEach(element => domEvent.on(element, 'keyup', this.createMeme, this));
        // //if image is changed
        // inputNodes.forEach(element => domEvent.on(element, 'change', this.createMeme, this));

        domEvent.on(this.downloadButton, 'click', this.downloadMapImage, this);

        domEvent.on(this.aboutIcon, 'click', function (e) {
            e.preventDefault();
            this.dialog.show();
        }, this);

        domEvent.on(this.menuIcon, 'click', function (e) {
            e.preventDefault();
            this.navigation.classList.toggle("active");
        }, this);

        var tabButtons = [].slice.call(document.querySelectorAll('ul.tab-nav li span.button'));

        tabButtons.map(function (button) {
            button.addEventListener('click', function () {
                document.querySelector('li span.active.button').classList.remove('active');
                button.classList.add('active');

                document.querySelector('.tab-pane.active').classList.remove('active');
                document.querySelector(button.getAttribute('name')).classList.add('active');
            })
        })

    }

    downloadMapImage() {
        // if(!this.$imageInput.files[0]) {
        //   this.$imageInput.parentElement.classList.add('has-error');
        //   return;
        // }
        // if(this.$bottomTextInput.value === '') {
        //   this.$imageInput.parentElement.classList.remove('has-error');
        //   this.$bottomTextInput.parentElement.classList.add('has-error');
        //   return;
        // }
        // this.$imageInput.parentElement.classList.remove('has-error');
        // this.$bottomTextInput.parentElement.classList.remove('has-error');

        // const imageSource = this.renderer.domElement.toDataURL('image/png');
        // const att = document.createAttribute('href');
        // att.value = imageSource.replace(/^data:image\/[^;]/, 'data:application/octet-stream');
        // this.downloadButton.setAttributeNode(att);
        // this.renderer.preserveDrawingBuffer = true;
        // this.renderer.render(this.scene, this.camera);
        this.saveCanvasImage(this.renderer.domElement);
    }

    saveCanvasImage(canvas) {
        // !HTMLCanvasElement.prototype.toBlob
        // https://developer.mozilla.org/en-US/docs/Web/API/HTMLCanvasElement.toBlob
        // decode the String
        var binStr = atob(canvas.toDataURL("image/png").split(',')[1]);
        var len = binStr.length;
        var arr = new Uint8Array(len);

        for (var i = 0; i < len; i++) {
            arr[i] = binStr.charCodeAt(i);
        }

        this.saveBlob(new Blob([arr], { type: "image/png" }));


    }

    saveBlob(blob) {
        // ie
        if (window.navigator.msSaveBlob !== undefined) {
            window.navigator.msSaveBlob(blob, filename);
            //app.popup.hide();
        }
        else {
            // create object url
            if (this._canvasImageUrl) {
                URL.revokeObjectURL(this._canvasImageUrl);
            }
            this._canvasImageUrl = URL.createObjectURL(blob);
            // display a link to save the image
            var e = this.downloadButton;//document.createElement("a");           
            e.href = this._canvasImageUrl;

        }
    }

    createMeme() {
        let context = this.$canvas.getContext('2d');

        // font size of top and bottom text
        let fontSize = ((this.canvas.width + this.canvas.height) / 2) * 4 / 100;
        context.font = `${fontSize}pt sans-serif`;
        context.textAlign = 'center';
        context.textBaseline = 'top';

        /**
         * Stroke Text Style
         */
        context.lineWidth = fontSize / 5;
        context.strokeStyle = 'black';
        /**
         * Fill Text Style
         */
        context.fillStyle = 'white';
        // Fix lines over M
        context.lineJoin = 'round';

        // get he value of the top text an dbottom text from the input fields
        let topText = this.$topTextInput.value.toUpperCase();
        let bottomText = this.$bottomTextInput.value.toUpperCase();

        // Top Text: first parameter text, second and thir parameters contain location where the text should start rendering
        context.strokeText(topText, this.canvas.width / 2, this.canvas.height * (5 / 100));
        context.fillText(topText, this.canvas.width / 2, this.canvas.height * (5 / 100));

        // Bottom Text
        context.strokeText(bottomText, this.canvas.width / 2, this.canvas.height * (90 / 100));
        context.fillText(bottomText, this.canvas.width / 2, this.canvas.height * (90 / 100));
    }

}

let container = document.getElementById("webgl");
let app = new Application(container);
app.build();
